#!/usr/bin/perl
###########################################################################
# ABI Compliance Checker (ACC) 1.23.3
# A tool for checking backward binary compatibility of a C/C++ library API
# Supported Platforms: Linux, FreeBSD, Mac OS X, Haiku (BeOS), MS Windows
# Copyright (C) 2011 Institute for System Programming, RAS
# Written by Andrey Ponomarenko
#
# System Requirements:
#  Linux - GCC (3.0-4.6.0, recommended 4.4 or newer),
#                 GNU Binutils (c++filt, readelf, objdump), Perl (5.8-5.12)
#  Mac OS X - Xcode (gcc, c++filt, otool)
#  MS Windows - MinGW (gcc.exe, c++filt.exe), Active Perl,
#                 MS Visual C++ (dumpbin.exe, undname.exe, cl.exe)
#
# In order to run the tool on Windows (Xp/Vista/7) you should:
#  1. Add path to gcc.exe (C:\MinGW\bin\) to your system PATH variable
#  2. Run vsvars32.bat script (C:\Microsoft Visual Studio 9.0\Common7\Tools\)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License or the GNU Lesser
# General Public License as published by the Free Software Foundation,
# either version 2 of the Licenses, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# and the GNU Lesser General Public License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
###########################################################################
use Getopt::Long;
Getopt::Long::Configure ("posix_default", "no_ignore_case");
use File::Path qw(mkpath rmtree);
use File::Temp qw(tempdir);
use Cwd qw(abs_path cwd);
use Data::Dumper;
use Config;

my $ABI_COMPLIANCE_CHECKER_VERSION = "1.23.3";
my $ABI_DUMP_VERSION = "2.1";
my $OLDEST_SUPPORTED_VERSION = "1.18";
my $ABI_DUMP_MAJOR = majorVersion($ABI_DUMP_VERSION);

my ($Help, $ShowVersion, %Descriptor, $TargetLibraryName, $GenDescriptor,
$TestTool, $DumpInfo_DPath, $InterfacesListPath, $CheckHeadersOnly_Opt, $UseDumps,
$CheckObjectsOnly_Opt, $AppPath, $StrictCompat, $DumpVersion, $ParamNamesFilePath,
%RelativeDirectory, $TargetLibraryFullName, $TestDump, $CheckImplementation,
%TargetVersion, $InfoMsg, $UseOldDumps, %UsedDump, $CrossGcc, %OutputLogPath,
$OutputReportPath, $OutputDumpPath, $ShowRetVal, $SystemRoot, $DumpSystem,
$CmpSystems);

my $CmdName = get_filename($0);
my $OSgroup = get_OSgroup();
my %OS_LibExt=(
    "default"=>"so",
    "macos"=>"dylib",
    "windows"=>"dll"
);

my %ERROR_CODE = (
    "Compatible"=>0,
    "Incompatible"=>1,
    "Error"=>2,
    "Not_Found"=>3,
    "Access_Error"=>4,
    "Cannot_Compile"=>5,
    "Compile_Error"=>6,
    "Invalid_Dump"=>7,
    "Dump_Version"=>8
);

my $LIB_EXT = $OS_LibExt{$OSgroup}?$OS_LibExt{$OSgroup}:$OS_LibExt{"default"};
my $SHARED = $OSgroup=~/macos|windows/?"dynamic":"shared";
my $ORIG_DIR = cwd();
my $TMP_DIR = tempdir(CLEANUP=>1);

my $ShortUsage = "ABI Compliance Checker (ACC) $ABI_COMPLIANCE_CHECKER_VERSION
A tool for checking backward binary compatibility of a C/C++ library API
Copyright (C) 2011 Institute for System Programming, RAS
License: GNU LGPLv2 or GNU GPLv2

Usage: perl $CmdName [options]
Example: perl $CmdName -lib NAME -d1 OLD.xml -d2 NEW.xml

OLD.xml and NEW.xml are XML-descriptors:

    <version>
        1.0
    </version>

    <headers>
        /path/to/headers/
    </headers>

    <libs>
        /path/to/libraries/
    </libs>

More info: perl $CmdName --help\n";

if($#ARGV==-1) {
    print $ShortUsage;
    exit(0);
}

foreach (2 .. $#ARGV)
{# correct comma separated options
    if($ARGV[$_-1] eq ",") {
        $ARGV[$_-2].=",".$ARGV[$_];
        splice(@ARGV, $_-1, 2);
    }
    elsif($ARGV[$_-1]=~/,\Z/) {
        $ARGV[$_-1].=$ARGV[$_];
        splice(@ARGV, $_, 1);
    }
    elsif($ARGV[$_]=~/\A,/
    and $ARGV[$_] ne ",") {
        $ARGV[$_-1].=$ARGV[$_];
        splice(@ARGV, $_, 1);
    }
}

GetOptions("h|help!" => \$Help,
  "i|info!" => \$InfoMsg,
  "v|version!" => \$ShowVersion,
  "dumpversion!" => \$DumpVersion,
#general options
  "l|lib|library=s" => \$TargetLibraryName,
  "d1|old|o=s" => \$Descriptor{1}{"Path"},
  "d2|new|n=s" => \$Descriptor{2}{"Path"},
  "dump|dump-abi|dump_abi=s" => \$DumpInfo_DPath,
  "old-dumps!" => \$UseOldDumps,
#extra options
  "d|descriptor-template!" => \$GenDescriptor,
  "app|application=s" => \$AppPath,
  "cross-gcc=s" =>\$CrossGcc,
  "sysroot=s" =>\$SystemRoot,
  "v1|version1|vnum=s" => \$TargetVersion{1},
  "v2|version2=s" => \$TargetVersion{2},
  "s|strict!" => \$StrictCompat,
  "symbols-list=s" => \$InterfacesListPath,
  "headers-only|headers_only!" => \$CheckHeadersOnly_Opt,
  "objects-only!" => \$CheckObjectsOnly_Opt,
  "check-implementation!" => \$CheckImplementation,
  "show-retval!" => \$ShowRetVal,
  "use-dumps!" =>\$UseDumps,
  "dump-system=s" => \$DumpSystem,
  "cmp-systems!" => \$CmpSystems,
#other options
  "test!" => \$TestTool,
  "test-dump!" =>\$TestDump,
  "p|params=s" => \$ParamNamesFilePath,
  "relpath1|relpath=s" => \$RelativeDirectory{1},
  "relpath2=s" => \$RelativeDirectory{2},
  "dump-path=s" =>\$OutputDumpPath,
  "report-path=s" =>\$OutputReportPath,
  "log1-path|log-path=s" =>\$OutputLogPath{1},
  "log2-path=s" =>\$OutputLogPath{2},
  "l-full|library-full-name=s" => \$TargetLibraryFullName
) or ERR_MESSAGE();

sub ERR_MESSAGE()
{
    print $ShortUsage;
    exit($ERROR_CODE{"Error"});
}

my $HelpMessage="
NAME:
  ABI Compliance Checker ($CmdName)
  Check backward binary compatibility of a $SHARED C/C++ library API

DESCRIPTION:
  ABI Compliance Checker (ACC) is an easy-to-use cross-platform tool for
  checking backward binary compatibility of a $SHARED C/C++ library.
  It checks header files along with $SHARED libraries (*.$LIB_EXT) of old and new
  versions and analyzes changes in Application Binary Interface (ABI) that
  may cause compatibility problems: changes in calling stack, v-table changes,
  removed symbols, etc. Breakage of the binary compatibility may result in
  crashing or incorrect behavior of applications built with an old version
  of a library when they run on a new one.

  The tool is intended for library developers and OS maintainers who are
  interested in ensuring binary compatibility (i.e. allow old applications
  to run with new library versions without the need to recompile). Also it
  can be used by ISVs for checking applications portability to new library
  versions. Found issues can be taken into account when adapting the
  application to a new library version.

  This tool is free software: you can redistribute it and/or modify it
  under the terms of the GNU LGPLv2 or GNU GPLv2.

USAGE:
  perl $CmdName [options]

EXAMPLE:
  perl $CmdName -lib NAME -d1 OLD.xml -d2 NEW.xml

  OLD.xml and NEW.xml are XML-descriptors:

    <version>
        1.0
    </version>

    <headers>
        /path1/to/header(s)/
        /path2/to/header(s)/
         ...
    </headers>

    <libs>
        /path1/to/library(ies)/
        /path2/to/library(ies)/
         ...
    </libs>

INFORMATION OPTIONS:
  -h|-help
      Print this help.

  -i|-info
      Print complete info.

  -v|-version
      Print version information.

  -dumpversion
      Print the tool version ($ABI_COMPLIANCE_CHECKER_VERSION) and don't do anything else.

GENERAL OPTIONS:
  -l|-lib|-library <name>
      Library name (without version).
      It affects only on the path and the title of the report.

  -d1|-old|-o <path(s)>
      Descriptor of 1st (old) library version.
      It may be one of the following:
      
         1. XML-descriptor (VERSION.xml file):

              <version>
                  1.0
              </version>

              <headers>
                  /path1/to/header(s)/
                  /path2/to/header(s)/
                   ...
              </headers>

              <libs>
                  /path1/to/library(ies)/
                  /path2/to/library(ies)/
                   ...
              </libs>

                 ... (XML-descriptor format is defined in
                         file generated by -d option)
             
         2. Directory with headers and/or $SHARED libraries:
              perl $CmdName -lib NAME -d1 DIR-1.0/ -d2 DIR-2.0/
              
         3. RPM or DEB \"devel\" package:
              perl $CmdName -lib NAME -d1 OLD.deb -d2 NEW.deb
              
         4. Header file:
              perl $CmdName -lib NAME -d1 OLD.h -d2 NEW.h
              
         5. ".ucfirst($SHARED)." library:
              perl $CmdName -lib NAME -d1 OLD.$LIB_EXT -d2 NEW.$LIB_EXT
              
         6. ABI dump (use -dump option to generate it):
              perl $CmdName -lib NAME -d1 OLD.abi.tar.gz -d2 NEW.abi.tar.gz
              
         7. Comma separated list of headers and/or $SHARED libraries:
              perl $CmdName -lib NAME -d1 OLD.h,OLD.$LIB_EXT -d2 NEW.h,NEW.$LIB_EXT
              
         8. Comma separated list of RPM or DEB packages:
              perl $CmdName -lib NAME -d1 1.1.deb,1.2.deb -d2 2.1.deb,2.2.deb

      If you are using an 2-8 descriptor types then you should
      specify version numbers with -v1 <num> and -v2 <num> options too.

      For more information, please see:
        http://ispras.linux-foundation.org/index.php/Library_Descriptor

  -d2|-new|-n <path(s)>
      Descriptor of 2nd (new) library version.

  -dump|-dump-abi <descriptor path(s)>
      Dump library ABI to gzipped TXT format file. You can transfer it
      anywhere and pass instead of the descriptor. Also it may be used
      for debugging the tool. Compatible dump versions: $ABI_DUMP_MAJOR.0 - $ABI_DUMP_MAJOR.9

  -old-dumps
      Enable support for old-version ABI dumps ($OLDEST_SUPPORTED_VERSION<=V<$ABI_DUMP_MAJOR.0).\n";

sub HELP_MESSAGE() {
    print $HelpMessage."
MORE INFO:
     perl $CmdName --info\n\n";
}

sub INFO_MESSAGE()
{
    print "$HelpMessage
EXTRA OPTIONS:
  -d|-descriptor-template
      Create XML-descriptor template ./VERSION.xml

  -app|-application <path>
      This option allows to specify the application that should be checked
      for portability to the new library version.

  -cross-gcc <path>
      Path to the cross GCC compiler to use instead of the usual (host) GCC.

  -sysroot <dirpath>
      Specify the alternative root directory. The tool will search for include
      paths in the <dirpath>/usr/include and <dirpath>/usr/lib directories.

  -v1|-version1 <num>
      Specify 1st library version outside the descriptor. This option is needed
      if you have prefered an alternative descriptor type (see -d1 option).

      In general case you should specify it in the XML-descriptor:
          <version>
              VERSION
          </version>

  -v2|-version2 <num>
      Specify 2nd library version outside the descriptor.

  -s|-strict
      Treat all compatibility warnings as problems. Add a number of \"Low\"
      severity problems to the return value of the tool.

  -headers-only
      Check header files without $SHARED libraries. It is easy to run, but may
      provide a low quality compatibility report with false positives and
      without detecting of added/removed symbols.
      
      Alternatively you can write \"none\" word to the <libs> section
      in the XML-descriptor:
          <libs>
              none
          </libs>

  -objects-only
      Check $SHARED libraries without header files. It is easy to run, but may
      provide a low quality compatibility report with false positives and
      without analysis of changes in parameters and data types.

      Alternatively you can write \"none\" word to the <headers> section
      in the XML-descriptor:
          <headers>
              none
          </headers>

  -check-implementation
      Compare canonified disassembled binary code of $SHARED libraries to
      detect changes in the implementation. Add \'Problems with Implementation\'
      section to the report.

  -show-retval
      Show the symbol's return type in the report.

  -symbols-list <path>
      This option allows to specify a file with a list of symbols (mangled
      names in C++) that should be checked, other symbols will not be checked.

  -use-dumps
      Make dumps for two versions of a library and compare dumps. This should
      increase the performance of the tool and decrease the operative memory usage.

  -dump-system <name>
      ( This option should be used with -sysroot <dirpath> option )
      Find all the shared libraries and header files in the <dirpath> directory,
      create XML descriptors and make ABI dumps for each library. The result set
      of ABI dumps can be compared (--cmp-systems) with the other one created for
      other version of operating system in order to check them for compatibility.
      Do not forget to specify -cross-gcc option if your target system requires
      some specific version of GCC compiler (different from the host GCC).
      The system ABI dump will be generated to:
          sys_dumps/<name>/<arch>

  -cmp-systems
      (This option should be used with -d1 sys_dumps/<name1>/<arch> and
        -d2 sys_dumps/<name2>/<arch> options)
      Compare two system ABI dumps. Create compatibility reports for each library
      and the common HTML report including the summary ot test results for all
      checked libraries. Reports will be generated to:
          sys_compat_reports/<name1>_to_<name2>/<arch>

OTHER OPTIONS:
  -test
      Run internal tests. Create two binary incompatible versions of a sample
      library and run the tool to check them for compatibility. This option
      allows to check if the tool works correctly in the current environment.

  -test-dump
      Test ability to create, restore and compare ABI dumps.

  -p|-params <path>
      Path to file with the function parameter names. It can be used
      for improving report view if the library header files have no
      parameter names. File format:
      
            func1;param1;param2;param3 ...
            func2;param1;param2;param3 ...
             ...

  -relpath <path>
      Replace {RELPATH} macros to <path> in the XML-descriptor used
      for dumping the library ABI (see -dump option).
  
  -relpath1 <path>
      Replace {RELPATH} macros to <path> in the 1st XML-descriptor (see -d1 option).

  -relpath2 <path>
      Replace {RELPATH} macros to <path> in the 2nd XML-descriptor (see -d2 option).

  -dump-path <path>
      Specify the path to a file (*.abi.tar.gz) where to generate the ABI dump.
      Default dump path: 
          abi_dumps/<library>/<library>_<version>.abi.tar.gz

  -report-path <path>
      Specify the file path where to generate the compatibility report.
      Default report path: 
          compat_reports/<library name>/<v1>_to_<v2>/abi_compat_report.html

  -log-path <path>
      Specify the file path where to generate the \"dumping\" log (see -dump option).
      Default log path:
          logs/<library>/<version>/log.txt

  -log1-path <path>
      Specify the file path where to generate the log for 1st version of a library.
      Default log path:
          logs/<library name>/<v1>/log.txt

  -log2-path <path>
      Specify the file path where to generate the log for 2nd version of a library.
      Default log path:
          logs/<library name>/<v2>/log.txt
      
  -l-full|-library-full-name <name>
      Change library name in the report title to <name>. By default
      will be displayed a name specified by -l option.

REPORT:
    Compatibility report  will be generated to:
        compat_reports/<library name>/<v1>_to_<v2>/abi_compat_report.html

    Log will be generated to:
        logs/<library name>/<v1>/log.txt
        logs/<library name>/<v2>/log.txt

EXIT CODES:
    0 - Compatible. The tool has run without any errors.
    non-zero - Incompatible or the tool has run with errors.

Report bugs to <abi-compliance-checker\@linuxtesting.org>
For more information, please see:
    http://ispras.linux-foundation.org/index.php/ABI_compliance_checker\n";
}

my $Descriptor_Template = "
<?xml version=\"1.0\" encoding=\"utf-8\"?>
<descriptor>

<!-- Primary sections -->

<version>
    <!-- Version of the library -->
</version>

<headers>
    <!-- The list of paths to header files and/or
         directories with header files, one per line -->
</headers>

<libs>
    <!-- The list of paths to $SHARED libraries (*.$LIB_EXT) and/or
         directories with $SHARED libraries, one per line -->
</libs>

<!-- Optional sections -->

<include_paths>
    <!-- The list of paths to be searched for header files
         needed for compiling of library headers, one per line.
         NOTE: If you define this section then the tool
         will not automatically detect include paths -->
</include_paths>

<add_include_paths>
    <!-- The list of include paths that should be added
         to the automatically detected include paths, one per line -->
</add_include_paths>

<gcc_options>
    <!-- Additional gcc options, one per line -->
</gcc_options>

<include_preamble>
    <!-- The list of header files that should be included before other headers, one per line.
         For example, it is a tree.h for libxml2 and ft2build.h for freetype2 -->
</include_preamble>

<opaque_types>
    <!-- The list of opaque types, one per line -->
</opaque_types>

<skip_symbols>
    <!-- The list of functions (mangled/symbol names in C++)
         that should be skipped while testing, one per line -->
</skip_symbols>

<skip_constants>
    <!-- The list of constants that should not be checked, one name per line -->
</skip_constants>

<skip_headers>
    <!-- The list of header files and/or directories with header files
         that should not be processed, one name per line -->
</skip_headers>

<skip_libs>
    <!-- The list of $SHARED libraries and/or directories with $SHARED libraries
         that should not be processed, one name per line -->
</skip_libs>

<defines>
    <!-- Add defines at the headers compiling stage, one per line:
          #define A B
          #define C D -->
</defines>

</descriptor>";

my %Operator_Indication = (
    "not" => "~",
    "assign" => "=",
    "andassign" => "&=",
    "orassign" => "|=",
    "xorassign" => "^=",
    "or" => "|",
    "xor" => "^",
    "addr" => "&",
    "and" => "&",
    "lnot" => "!",
    "eq" => "==",
    "ne" => "!=",
    "lt" => "<",
    "lshift" => "<<",
    "lshiftassign" => "<<=",
    "rshiftassign" => ">>=",
    "call" => "()",
    "mod" => "%",
    "modassign" => "%=",
    "subs" => "[]",
    "land" => "&&",
    "lor" => "||",
    "rshift" => ">>",
    "ref" => "->",
    "le" => "<=",
    "deref" => "*",
    "mult" => "*",
    "preinc" => "++",
    "delete" => " delete",
    "vecnew" => " new[]",
    "vecdelete" => " delete[]",
    "predec" => "--",
    "postinc" => "++",
    "postdec" => "--",
    "plusassign" => "+=",
    "plus" => "+",
    "minus" => "-",
    "minusassign" => "-=",
    "gt" => ">",
    "ge" => ">=",
    "new" => " new",
    "multassign" => "*=",
    "divassign" => "/=",
    "div" => "/",
    "neg" => "-",
    "pos" => "+",
    "memref" => "->*",
    "compound" => "," );

my %CppKeywords = map {$_=>1} (
    # C++ 2003 keywords
    "public",
    "protected",
    "private",
    "default",
    "template",
    "new",
    "delete",
    "this",
    "asm",
    "dynamic_cast",
    "auto",
    "operator",
    "try",
    "typeid",
    "namespace",
    "typename",
    "catch",
    "using",
    "reinterpret_cast",
    "friend",
    "class",
    "virtual",
    "const_cast",
    "mutable",
    "static_cast",
    "export",
    #"throw",
    #"inline",
    #"register",
    #"bool",
    # C++0x keywords
    "noexcept",
    "alignof",
    "nullptr",
    "thread_local",
    "constexpr",
    "decltype",
    "static_assert",
    "explicit",
    # cannot be used as a macro name
    # as it is an operator in C++
    #"and",
    #"and_eq",
    #"not",
    #"not_eq",
    #"or",
    #"or_eq",
    #"bitand",
    #"bitor",
    #"xor",
    #"xor_eq",
    #"compl"
);

my %IntrinsicMangling = (
    "void" => "v",
    "bool" => "b",
    "wchar_t" => "w",
    "char" => "c",
    "signed char" => "a",
    "unsigned char" => "h",
    "short" => "s",
    "unsigned short" => "t",
    "int" => "i",
    "unsigned int" => "j",
    "long" => "l",
    "unsigned long" => "m",
    "long long" => "x",
    "__int64" => "x",
    "unsigned long long" => "y",
    "__int128" => "n",
    "unsigned __int128" => "o",
    "float" => "f",
    "double" => "d",
    "long double" => "e",
    "__float80" => "e",
    "__float128" => "g",
    "..." => "z"
);

my %StdcxxMangling = (
    "3std"=>"St",
    "3std9allocator"=>"Sa",
    "3std11basic_string"=>"Sb",
    "3std12basic_stringIcE"=>"Ss",
    "3std13basic_istreamIcE"=>"Si",
    "3std13basic_ostreamIcE"=>"So",
    "3std14basic_iostreamIcE"=>"Sd"
);

my %ConstantSuffix = (
    "unsigned int"=>"u",
    "long"=>"l",
    "unsigned long"=>"ul",
    "long long"=>"ll",
    "unsigned long long"=>"ull"
);

my %ConstantSuffixR =
reverse(%ConstantSuffix);

my %OperatorMangling = (
    "~" => "co",
    "=" => "aS",
    "|" => "or",
    "^" => "eo",
    "&" => "an",#ad (addr)
    "==" => "eq",
    "!" => "nt",
    "!=" => "ne",
    "<" => "lt",
    "<=" => "le",
    "<<" => "ls",
    "<<=" => "lS",
    ">" => "gt",
    ">=" => "ge",
    ">>" => "rs",
    ">>=" => "rS",
    "()" => "cl",
    "%" => "rm",
    "[]" => "ix",
    "&&" => "aa",
    "||" => "oo",
    "*" => "ml",#de (deref)
    "++" => "pp",#
    "--" => "mm",#
    "new" => "nw",
    "delete" => "dl",
    "new[]" => "na",
    "delete[]" => "da",
    "+=" => "pL",
    "+" => "pl",#ps (pos)
    "-" => "mi",#ng (neg)
    "-=" => "mI",
    "*=" => "mL",
    "/=" => "dV",
    "&=" => "aN",
    "|=" => "oR",
    "%=" => "rM",
    "^=" => "eO",
    "/" => "dv",
    "->*" => "pm",
    "->" => "pt",#rf (ref)
    "," => "cm",
    "?" => "qu",
    "." => "dt",
    "sizeof"=> "sz"#st
);

my %GlibcHeader = map {$_=>1} (
    "aliases.h",
    "argp.h",
    "argz.h",
    "assert.h",
    "cpio.h",
    "ctype.h",
    "dirent.h",
    "envz.h",
    "errno.h",
    "error.h",
    "execinfo.h",
    "fcntl.h",
    "fstab.h",
    "ftw.h",
    "glob.h",
    "grp.h",
    "iconv.h",
    "ifaddrs.h",
    "inttypes.h",
    "langinfo.h",
    "limits.h",
    "link.h",
    "locale.h",
    "malloc.h",
    "math.h",
    "mntent.h",
    "monetary.h",
    "nl_types.h",
    "obstack.h",
    "printf.h",
    "pwd.h",
    "regex.h",
    "sched.h",
    "search.h",
    "setjmp.h",
    "shadow.h",
    "signal.h",
    "spawn.h",
    "stdarg.h",
    "stdint.h",
    "stdio.h",
    "stdlib.h",
    "string.h",
    "tar.h",
    "termios.h",
    "time.h",
    "ulimit.h",
    "unistd.h",
    "utime.h",
    "wchar.h",
    "wctype.h",
    "wordexp.h" );

my %GlibcDir = map {$_=>1} (
    "sys",
    "linux",
    "bits",
    "gnu",
    "netinet",
    "rpc" );

my %LocalIncludes = map {$_=>1} (
    "/usr/local/include",
    "/usr/local" );

my %OS_AddPath=(
# this data needed if tool can't determine paths automatically
"macos"=>{
    "include"=>{"/Library"=>1,"/Developer/usr/include"=>1},
    "lib"=>{"/Library"=>1,"/Developer/usr/lib"=>1},
    "bin"=>{"/Developer/usr/bin"=>1}},
"beos"=>{
    "include"=>{"/boot/common"=>1,"/boot/develop"=>1},
    "lib"=>{"/boot/common/lib"=>1,"/boot/system/lib"=>1,"/boot/apps"=>1},
    "bin"=>{"/boot/common/bin"=>1,"/boot/system/bin"=>1,"/boot/develop/abi"=>1}}
    # Haiku has GCC 2.95.3 by default, try to find GCC>=3.0 in /boot/develop/abi
);

my %Slash_Type=(
    "default"=>"/",
    "windows"=>"\\"
);

my $SLASH = $Slash_Type{$OSgroup}?$Slash_Type{$OSgroup}:$Slash_Type{"default"};

my %TypeProblems_Kind=(
    "Added_Virtual_Method"=>"High",
    "Added_Pure_Virtual_Method"=>"High",
    "Added_Virtual_Method_At_End"=>"High",
    "Added_First_Virtual_Method"=>"High",
    "Removed_Virtual_Method"=>"High",
    "Removed_Last_Virtual_Method"=>"High",
    "Virtual_Replacement"=>"Medium",
    "Pure_Virtual_Replacement"=>"Medium",
    "Virtual_Table_Size"=>"High",
    "Virtual_Method_Position"=>"High",
    "Pure_Virtual_Method_Position"=>"High",
    "Overridden_Virtual_Method"=>"Low",
    "Overridden_Virtual_Method_B"=>"Low",
    "Size_Of_Allocable_Class_Increased"=>"High",
    "Size_Of_Allocable_Class_Decreased"=>"Medium",
    "Size_Of_Copying_Class"=>"High",
    "Base_Class_Position"=>"Low",
    "Base_Class_Became_Virtually_Inherited"=>"Medium",
    "Base_Class_Became_Non_Virtually_Inherited"=>"Medium",
    "Added_Base_Class_And_Shift"=>"High",
    "Added_Base_Class"=>"Low",
    "Removed_Base_Class_And_Shift"=>"High",
    "Removed_Base_Class"=>"Low",
    "Size"=>"Dynamic",
    "Added_Field_And_Size"=>"Low",
    "Added_Field"=>"Low",
    "Added_Middle_Field_And_Size"=>"Medium",
    "Added_Middle_Field"=>"Low",
    "Added_Reserved_Field"=>"Low",
    "Added_Union_Field_And_Size"=>"Medium",
    "Added_Union_Field"=>"Low",
    "Used_Reserved_Field"=>"Low",
    "Removed_Field_And_Size"=>"Medium",
    "Removed_Field"=>"Medium",
    "Removed_Middle_Field_And_Size"=>"Dynamic",
    "Removed_Middle_Field"=>"Medium",
    "Removed_Union_Field_And_Size"=>"Medium",
    "Removed_Union_Field"=>"Low",
    "Renamed_Field"=>"Low",
    "Enum_Member_Value"=>"Medium",
    "Enum_Member_Removed"=>"Low",
    "Enum_Last_Member_Value"=>"Low",
    "Enum_Member_Name"=>"Low",
    "Field_Type_And_Size"=>"Dynamic",
    "Field_Size"=>"Dynamic",
    "BitField_Size"=>"Medium",
    "Private_Field_Size"=>"Low",
    "Reserved_Field_Size"=>"Medium",
    "Field_Type"=>"Low",
    "Field_BaseType_And_Size"=>"Dynamic",
    "Field_BaseType"=>"Low",
    "Field_PointerLevel_Increased"=>"Medium",
    "Field_PointerLevel_Decreased"=>"Low",
    "Typedef_BaseType"=>"Low"  );

my %InterfaceProblems_Kind=(
    "Added_Interface"=>"Safe",
    "Removed_Interface"=>"High",
    "Method_Became_Static"=>"High",
    "Method_Became_NonStatic"=>"High",
    "Parameter_Default_Value_Changed"=>"Low",
    "Parameter_Default_Value_Removed"=>"Low",
    "Parameter_Type_And_Size"=>"High",
    "Parameter_Type_And_Stack"=>"High",
    "Parameter_Type_And_Register"=>"High",
    "Parameter_Type_And_Pass_Through_Register"=>"High",
    "Parameter_Type_And_Pass_Through_Stack"=>"High",
    "Parameter_Type"=>"Low",
    "Parameter_Became_Non_Const"=>"Medium",
    "Parameter_BaseType_And_Size"=>"Dynamic",
    "Parameter_BaseType"=>"Low",
    "Parameter_PointerLevel_Increased"=>"High",
    "Parameter_PointerLevel_Decreased"=>"Medium",
    "Return_Type_And_Size"=>"Medium",
    "Return_Type_Became_Void"=>"Medium",
    "Return_Type_And_Stack_Layout"=>"High",
    "Return_Type_And_Return_Method"=>"Medium",
    "Return_Type"=>"Low",
    "Global_Data_Became_Non_Const"=>"Medium",
    "Return_BaseType_And_Size"=>"Medium",
    "Return_BaseType"=>"Low",
    "Return_PointerLevel_Increased"=>"Low",
    "Return_PointerLevel_Decreased"=>"Medium",
    "Removed_Parameter"=>"Medium",
    "Added_Parameter"=>"Medium",
    "Added_Middle_Parameter"=>"High",
    "Removed_Middle_Parameter"=>"High",
    "Parameter_Rename"=>"Low",
    "Symbol_Changed_Static"=>"High",
    "Symbol_Changed_NonStatic"=>"High",
    "Symbol_Changed_Virtual"=>"High",
    "Symbol_Changed_NonVirtual"=>"High",
    "Symbol_Changed_Return"=>"High",
    "Symbol_Changed_Signature"=>"High"  );

#Global variables
my %COMMON_LANGUAGE=(
  1 => "C",
  2 => "C");

my $MAX_COMMAND_LINE_ARGUMENTS = 4096;
my (%POINTER_SIZE, %CPU_ARCH, %GCC_VERSION);

my $STDCXX_TESTING = 0;
my $GLIBC_TESTING = 0;

my $CheckHeadersOnly = $CheckHeadersOnly_Opt;
my $CheckObjectsOnly = $CheckObjectsOnly_Opt;

my $MAIN_CPP_DIR;
my $CHECKER_VERDICT = 0;
my $CHECKER_WARNING = 0;
my $REPORT_PATH = $OutputReportPath;
my %LOG_PATH;
my %Cache;
my %LibInfo;
my $COMPILE_ERRORS = 0;
my %CompilerOptions;
my %CheckedDyLib;
my $TargetLibraryShortName = parse_libname($TargetLibraryName, "shortest");

#Constants (Defines)
my %Constants;
my %SkipConstants;

#Types
my %TypeInfo;
my %TemplateInstance_Func;
my %TemplateInstance;
my %OpaqueTypes = (
  "1"=>{},
  "2"=>{} );
my %Tid_TDid = (
  "1"=>{},
  "2"=>{} );
my %CheckedTypes;
my %Typedef_BaseName;
my %Typedef_Translations;
my %Typedef_Equivalent;
my %StdCxxTypedef;
my %TName_Tid;
my %EnumMembName_Id;
my %NestedNameSpaces = (
  "1"=>{},
  "2"=>{} );
my %UsedType;
my %VirtualTable;
my %VirtualTable_Full;
my %ClassVTable;
my %ClassVTable_Content;
my %VTableClass;
my %AllocableClass;
my %ClassConstructors;
my %ClassMethods;
my %ClassToId;
my %Class_SubClasses;
my %OverriddenMethods;

#Symbols
my %SymbolInfo;
my %tr_name;
my %mangled_name_gcc;
my %mangled_name;
my %SkipSymbols = (
  "1"=>{},
  "2"=>{} );
my %InterfacesList;
my %InterfacesList_App;
my %CheckedInterfaces;
my $ComparedABI;
my %DepSymbols;
my %DepSymbols_Versioned;
my %MangledNames;
my %AddIntParams;
my %Interface_Implementation;

#Headers
my %Include_Preamble;
my %Target_Headers;
my %HeaderName_Paths;
my %Header_Dependency;
my %Include_Paths;
my %INC_PATH_AUTODETECT = (
  "1"=>1,
  "2"=>1 );
my %Add_Include_Paths;
my %DependencyHeaders_All_FullPath;
my %RegisteredHeaders;
my %RegisteredDirs;
my %Header_ErrorRedirect;
my %Header_Includes;
my %Header_ShouldNotBeUsed;
my %RecursiveIncludes;
my %Header_Include_Prefix;
my %Header_Prefix;
my %SkipHeaders;
my %SkipHeaders_Pattern;
my %SkipLibs;
my %SkipLibs_Pattern;
my %Include_Order;
my %TUnit_NameSpaces;
my $C99Mode = 0;

#Shared objects
my %DyLib_DefaultPath;
my %SharedObject_Paths;
my %SharedObject_Dirs;

#System shared objects
my %SystemObjects;
my %DefaultLibPaths;

#System header files
my %SystemHeaders;
my %DefaultCppPaths;
my %DefaultGccPaths;
my %DefaultIncPaths;
my %DefaultCppHeader;
my %DefaultGccHeader;
my %UserIncPath;

#Merging
my %CompleteSignature;
my %Symbol_Library;
my %Library_Symbol = (
  "1"=>{},
  "2"=>{} );
my %Language;
my %DyLibNames_All = (
  "1"=>{},
  "2"=>{} );
my $Version;
my %AddedInt;
my %RemovedInt;
my %AddedInt_Virt;
my %RemovedInt_Virt;
my %VirtualReplacement;
my %ChangedTypedef;

#Compliance Report
my %Type_MaxPriority;

#Recursion locks
my @RecurLib;
my @RecurSymlink;
my @RecurTypes;
my @RecurInclude;
my @RecurConstant;

#System
my %SystemPaths;
my %DefaultBinPaths;
my ($GCC_PATH, $CPP_FILT);

#Symbols versioning
my %SymVer;

#Problem descriptions
my %CompatProblems;
my %ProblemsInDefines;
my %ImplProblems;

#Rerorts
my $ContentID = 1;
my $ContentSpanStart = "<span class=\"section\" onclick=\"javascript:showContent(this, 'CONTENT_ID')\">\n";
my $ContentSpanStart_Affected = "<span class=\"section_affected\" onclick=\"javascript:showContent(this, 'CONTENT_ID')\">\n";
my $ContentSpanStart_Info = "<span class=\"section_info\" onclick=\"javascript:showContent(this, 'CONTENT_ID')\">\n";
my $ContentSpanEnd = "</span>\n";
my $ContentDivStart = "<div id=\"CONTENT_ID\" style=\"display:none;\">\n";
my $ContentDivEnd = "</div>\n";
my $Content_Counter = 0;

sub numToStr($)
{
    my $Number = int($_[0]);
    if($Number>3) {
        return $Number."th";
    }
    elsif($Number==1) {
        return "1st";
    }
    elsif($Number==2) {
        return "2nd";
    }
    elsif($Number==3) {
        return "3rd";
    }
    else {
        return $Number;
    }
}

sub get_CmdPath($)
{
    my $Name = $_[0];
    return "" if(not $Name);
    return $Cache{"get_CmdPath"}{$Name} if(defined $Cache{"get_CmdPath"}{$Name});
    if($Name eq "c++filt")
    {# search for c++filt in the same directory as GCC
        my $Path = get_CmdPath("gcc");
        if($Path=~s/([\/\\])gcc\Z/$1c++filt/
        and (-f $Path or -f $Path.".exe"))
        {# synchronization of gcc with c++filt
            $Cache{"get_CmdPath"}{$Name} = $Path;
            return $Cache{"get_CmdPath"}{$Name};
        }
    }
    if(my $DefaultPath = get_CmdPath_Default($Name)) {
        $Cache{"get_CmdPath"}{$Name} = $DefaultPath;
        return $Cache{"get_CmdPath"}{$Name};
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%{$SystemPaths{"bin"}}))
    {
        if(-f $Path."/".$Name or -f $Path."/".$Name.".exe")
        {
            if($Name eq "gcc") {
                next if(not check_gcc_version(joinPath($Path,$Name), "3"));
            }
            $Cache{"get_CmdPath"}{$Name} = joinPath($Path,$Name);
            return $Cache{"get_CmdPath"}{$Name};
        }
    }
    $Cache{"get_CmdPath"}{$Name} = "";
    return "";
}

sub get_CmdPath_Default($)
{# search in PATH
    my $Name = $_[0];
    return "" if(not $Name);
    return $Cache{"get_CmdPath_Default"}{$Name} if(defined $Cache{"get_CmdPath_Default"}{$Name});
    if($Name eq "find")
    {# special case: search for "find" utility
        if(`find . -maxdepth 0 2>$TMP_DIR/null`)
        {
            $Cache{"get_CmdPath_Default"}{$Name} = "find";
            return "find";
        }
    }
    elsif($Name eq "gcc") {
        return check_gcc_version($Name, "3");
    }
    if(get_version($Name)) {
        $Cache{"get_CmdPath_Default"}{$Name} = $Name;
        return $Name;
    }
    if($OSgroup eq "windows"
    and `$Name /? 2>$TMP_DIR/null`) {
        $Cache{"get_CmdPath_Default"}{$Name} = $Name;
        return $Name;
    }
    if($Name ne "which") {
        my $WhichCmd = get_CmdPath("which");
        if($WhichCmd and `$WhichCmd $Name 2>$TMP_DIR/null`)
        {
            $Cache{"get_CmdPath_Default"}{$Name} = $Name;
            return $Cache{"get_CmdPath_Default"}{$Name};
        }
    }
    foreach my $Path (sort {length($a)<=>length($b)} keys(%DefaultBinPaths))
    {
        if(-f $Path."/".$Name or -f $Path."/".$Name.".exe") {
            $Cache{"get_CmdPath_Default"}{$Name} = joinPath($Path,$Name);
            return $Cache{"get_CmdPath_Default"}{$Name};
        }
    }
    $Cache{"get_CmdPath_Default"}{$Name} = "";
    return "";
}

sub readDescriptor($$)
{
    my ($LibVersion, $Content) = @_;
    return if(not $LibVersion);
    my $DName = $DumpInfo_DPath?"descriptor":"descriptor \"d$LibVersion\"";
    if(not $Content) {
        exitStatus("Error", "$DName is empty");
    }
    if($Content!~/\</) {
        exitStatus("Error", "descriptor should be one of the following: XML-descriptor, RPM or Deb \"devel\" package, ABI dump, standalone header/$SHARED library or directory with headers and/or $SHARED libraries");
    }
    $Content=~s/\/\*(.|\n)+?\*\///g;
    $Content=~s/<\!--(.|\n)+?-->//g;
    $Descriptor{$LibVersion}{"Version"} = parseTag(\$Content, "version");
    $Descriptor{$LibVersion}{"Version"} = $TargetVersion{$LibVersion} if($TargetVersion{$LibVersion});
    if(not $Descriptor{$LibVersion}{"Version"}) {
        exitStatus("Error", "version in the $DName is not specified (<version> section)");
    }
    if($Content=~/{RELPATH}/)
    {
        if(my $RelDir = $RelativeDirectory{$LibVersion})  {
            $Content =~ s/{RELPATH}/$RelDir/g;
        }
        else
        {
            my $NeedRelpath = $DumpInfo_DPath?"-relpath":"-relpath$LibVersion";
            exitStatus("Error", "you have not specified $NeedRelpath option, but the $DName contains {RELPATH} macro");
        }
    }
    
    if(not $CheckObjectsOnly_Opt)
    {
        my $DHeaders = parseTag(\$Content, "headers");
        if(not $DHeaders) {
            exitStatus("Error", "header files in the $DName are not specified (<headers> section)");
        }
        elsif(lc($DHeaders) ne "none")
        {# append the descriptor headers list
            $Descriptor{$LibVersion}{"Headers"} .= "\n".$DHeaders;
            foreach my $HPath (split(/\n/, $DHeaders))
            {
                $HPath=~s/\A\s+|\s+\Z//g;
                if(not -e $HPath) {
                    exitStatus("Access_Error", "can't access \'$HPath\'");
                }
            }
        }
    }
    if(not $CheckHeadersOnly_Opt)
    {
        my $DObjects = parseTag(\$Content, "libs");
        if(not $DObjects) {
            exitStatus("Error", "$SHARED libraries in the $DName are not specified (<libs> section)");
        }
        elsif(lc($DObjects) ne "none")
        {# append the descriptor libraries list
            $Descriptor{$LibVersion}{"Libs"} .= "\n".$DObjects;
            foreach my $ObjPath (split(/\n/, $DObjects))
            {
                $ObjPath=~s/\A\s+|\s+\Z//g;
                if(not -e $ObjPath) {
                    exitStatus("Access_Error", "can't access \'$ObjPath\'");
                }
            }
        }
    }
    foreach my $Dest (split(/\n/, parseTag(\$Content, "include_paths")))
    {
        $Dest=~s/\A\s+|\s+\Z//g;
        next if(not $Dest);
        if(not -d $Dest) {
            exitStatus("Access_Error", "can't access directory \'$Dest\'");
        }
        $Descriptor{$LibVersion}{"IncludePaths"}{$Dest} = 1;
    }
    foreach my $Dest (split(/\n/, parseTag(\$Content, "add_include_paths")))
    {
        $Dest=~s/\A\s+|\s+\Z//g;
        next if(not $Dest);
        if(not -d $Dest) {
            exitStatus("Access_Error", "can't access directory \'$Dest\'");
        }
        $Descriptor{$LibVersion}{"AddIncludePaths"}{$Dest} = 1;
    }
    $Descriptor{$LibVersion}{"GccOptions"} = parseTag(\$Content, "gcc_options");
    foreach my $Option (split(/\n/, $Descriptor{$LibVersion}{"GccOptions"}))
    {
        $Option=~s/\A\s+|\s+\Z//g;
        next if(not $Option);
        $CompilerOptions{$LibVersion} .= " ".$Option;
    }
    $Descriptor{$LibVersion}{"SkipHeaders"} = parseTag(\$Content, "skip_headers");
    foreach my $Name (split(/\n/, $Descriptor{$LibVersion}{"SkipHeaders"}))
    {
        $Name=~s/\A\s+|\s+\Z//g;
        next if(not $Name);
        if($Name=~/\*|\/|\[|\\/)
        {# perl regexp or directory
            $Name=~s/\*/.*/g;
            $Name=~s/\\/\\\\/g;
            $SkipHeaders_Pattern{$LibVersion}{$Name} = 1;
        }
        else {
            $SkipHeaders{$LibVersion}{$Name} = 1;
        }
    }
    $Descriptor{$LibVersion}{"SkipLibs"} = parseTag(\$Content, "skip_libs");
    foreach my $Name (split(/\n/, $Descriptor{$LibVersion}{"SkipLibs"}))
    {
        $Name=~s/\A\s+|\s+\Z//g;
        next if(not $Name);
        if($Name=~/\*|\/|\\/) {
            $Name=~s/\*/.*/g;
            $Name=~s/\\/\\\\/g;
            $SkipLibs_Pattern{$LibVersion}{$Name} = 1;
        }
        else {
            $SkipLibs{$LibVersion}{$Name} = 1;
        }
    }
    if(my $DDefines = parseTag(\$Content, "defines")) {
        $Descriptor{$LibVersion}{"Defines"} .= "\n".$DDefines;
    }
    foreach my $Order (split(/\n/, parseTag(\$Content, "include_order")))
    {
        $Order=~s/\A\s+|\s+\Z//g;
        next if(not $Order);
        if($Order=~/\A(.+):(.+)\Z/) {
            $Include_Order{$LibVersion}{$1} = $2;
        }
    }
    foreach my $Type_Name (split(/\n/, parseTag(\$Content, "opaque_types")))
    {
        $Type_Name=~s/\A\s+|\s+\Z//g;
        next if(not $Type_Name);
        $OpaqueTypes{$LibVersion}{$Type_Name} = 1;
    }
    foreach my $Interface_Name (split(/\n/, parseTag(\$Content, "skip_interfaces")),
    split(/\n/, parseTag(\$Content, "skip_symbols")))
    {# skip_interfaces renamed to skip_symbols (1.22.1)
        $Interface_Name=~s/\A\s+|\s+\Z//g;
        next if(not $Interface_Name);
        $SkipSymbols{$LibVersion}{$Interface_Name} = 1;
    }
    foreach my $Constant (split(/\n/, parseTag(\$Content, "skip_constants")))
    {
        $Constant=~s/\A\s+|\s+\Z//g;
        next if(not $Constant);
        $SkipConstants{$LibVersion}{$Constant} = 1;
    }
    if(my $DIncPreamble = parseTag(\$Content, "include_preamble")) {
        $Descriptor{$LibVersion}{"IncludePreamble"} .= "\n".$DIncPreamble;
    }
}

sub parseTag($$)
{
    my ($CodeRef, $Tag) = @_;
    return "" if(not $CodeRef or not ${$CodeRef} or not $Tag);
    if(${$CodeRef}=~s/\<\Q$Tag\E\>((.|\n)+?)\<\/\Q$Tag\E\>//)
    {
        my $Content = $1;
        $Content=~s/(\A\s+|\s+\Z)//g;
        return $Content;
    }
    else {
        return "";
    }
}

my %check_node=(
  "array_type"=>1,
  "binfo"=>1,
  "boolean_type"=>1,
  "complex_type"=>1,
  "const_decl"=>1,
  "enumeral_type"=>1,
  "field_decl"=>1,
  "function_decl"=>1,
  "function_type"=>1,
  "identifier_node"=>1,
  "integer_cst"=>1,
  "integer_type"=>1,
  "method_type"=>1,
  "namespace_decl"=>1,
  "parm_decl"=>1,
  "pointer_type"=>1,
  "real_cst"=>1,
  "real_type"=>1,
  "record_type"=>1,
  "reference_type"=>1,
  "string_cst"=>1,
  "template_decl"=>1,
  "template_type_parm"=>1,
  "tree_list"=>1,
  "tree_vec"=>1,
  "type_decl"=>1,
  "union_type"=>1,
  "var_decl"=>1,
  "void_type"=>1,
  "offset_type"=>1);

sub getInfo($)
{
    my $InfoPath = $_[0];
    return if(not $InfoPath or not -f $InfoPath);
    my $Content = readFile($InfoPath);
    unlink($InfoPath);
    $Content=~s/\n[ ]+/ /g;
    my @Lines = split(/\n/, $Content);
    $Content="";# clear
    foreach (@Lines)
    {
        if(/\A@(\d+)\s+([a-z_]+)\s+(.+)\Z/oi)
        {# get a number and attributes of node
            next if(not $check_node{$2});
            $LibInfo{$Version}{"info_type"}{$1}=$2;
            $LibInfo{$Version}{"info"}{$1}=$3;
        }
    }
    # processing info
    setTemplateParams_All();
    getTypeInfo_All();
    renameTmplInst();
    getSymbolInfo_All();
    getVarDescr_All();
    
    # cleaning memory
    %LibInfo = ();
    %TemplateInstance = ();
    %TemplateInstance_Func = ();
    %MangledNames = ();
}

sub renameTmplInst()
{
    foreach my $Base (keys(%{$Typedef_Translations{$Version}}))
    {
        next if($Base!~/</);
        my @Translations = keys(%{$Typedef_Translations{$Version}{$Base}});
        if($#Translations==0 and length($Translations[0])<=length($Base)) {
            $Typedef_Equivalent{$Version}{$Base} = $Translations[0];
        }
    }
    foreach my $TDid (keys(%{$TypeInfo{$Version}}))
    {
        foreach my $Tid (keys(%{$TypeInfo{$Version}{$TDid}}))
        {
            my $TypeName = $TypeInfo{$Version}{$TDid}{$Tid}{"Name"};
            if($TypeName=~/>/)
            {# template instances only
                foreach my $Base (sort {length($b)<=>length($a)}
                keys(%{$Typedef_Equivalent{$Version}}))
                {
                    next if(not $Base);
                    next if(index($TypeName,$Base)==-1);
                    next if(length($Base)>=length($TypeName)-2);
                    my $TypedefName = $Typedef_Equivalent{$Version}{$Base};
                    $TypeName=~s/<\Q$Base\E(\W|\Z)/<$TypedefName$1/g;
                    $TypeName=~s/<\Q$Base\E(\w|\Z)/<$TypedefName $1/g;
                    last if($TypeName!~/>/);
                }
                if($TypeName ne $TypeInfo{$Version}{$TDid}{$Tid}{"Name"})
                {
                    $TypeInfo{$Version}{$TDid}{$Tid}{"Name"} = correctName($TypeName);
                    $TName_Tid{$Version}{$TypeName} = $Tid;
                }
            }
        }
    }
}

sub setTemplateParams_All()
{
    foreach (keys(%{$LibInfo{$Version}{"info"}}))
    {
        if($LibInfo{$Version}{"info_type"}{$_} eq "template_decl") {
            setTemplateParams($_);
        }
    }
}

sub setTemplateParams($)
{
    my $TypeInfoId = $_[0];
    if($LibInfo{$Version}{"info"}{$TypeInfoId}=~/(inst|spcs)[ ]*:[ ]*@(\d+) /)
    {
        my $TmplInst_InfoId = $2;
        setTemplateInstParams($TmplInst_InfoId);
        my $TmplInst_Info = $LibInfo{$Version}{"info"}{$TmplInst_InfoId};
        while($TmplInst_Info=~/(chan|chain)[ ]*:[ ]*@(\d+) /)
        {
            $TmplInst_InfoId = $2;
            $TmplInst_Info = $LibInfo{$Version}{"info"}{$TmplInst_InfoId};
            setTemplateInstParams($TmplInst_InfoId);
        }
    }
}

sub setTemplateInstParams($)
{
    my $TmplInst_Id = $_[0];
    my $Info = $LibInfo{$Version}{"info"}{$TmplInst_Id};
    my ($Params_InfoId, $ElemId) = ();
    if($Info=~/purp[ ]*:[ ]*@(\d+) /) {
        $Params_InfoId = $1;
    }
    if($Info=~/valu[ ]*:[ ]*@(\d+) /) {
        $ElemId = $1;
    }
    if($Params_InfoId and $ElemId)
    {
        my $Params_Info = $LibInfo{$Version}{"info"}{$Params_InfoId};
        while($Params_Info=~s/ (\d+)[ ]*:[ ]*\@(\d+) / /)
        {
            my ($Param_Pos, $Param_TypeId) = ($1, $2);
            return if($LibInfo{$Version}{"info_type"}{$Param_TypeId} eq "template_type_parm");
            if($LibInfo{$Version}{"info_type"}{$ElemId} eq "function_decl") {
                $TemplateInstance_Func{$Version}{$ElemId}{$Param_Pos} = $Param_TypeId;
            }
            else {
                $TemplateInstance{$Version}{getTypeDeclId($ElemId)}{$ElemId}{$Param_Pos} = $Param_TypeId;
            }
        }
    }
}

sub getTypeDeclId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/name[ ]*:[ ]*@(\d+)/) {
        return $1;
    }
    else {
        return "";
    }
}

sub isFuncPtr($)
{
    my $Ptd = pointTo($_[0]);
    return 0 if(not $Ptd);
    if($LibInfo{$Version}{"info"}{$_[0]}=~/unql[ ]*:/
    and $LibInfo{$Version}{"info"}{$_[0]}!~/qual[ ]*:/) {
        return 0;
    }
    elsif($LibInfo{$Version}{"info_type"}{$_[0]} eq "pointer_type"
    and $LibInfo{$Version}{"info_type"}{$Ptd} eq "function_type") {
        return 1;
    }
    return 0;
}

sub isMethodPtr($)
{
    my $Ptd = pointTo($_[0]);
    return 0 if(not $Ptd);
    if($LibInfo{$Version}{"info_type"}{$_[0]} eq "record_type"
    and $LibInfo{$Version}{"info_type"}{$Ptd} eq "method_type"
    and $LibInfo{$Version}{"info"}{$_[0]}=~/ ptrmem /) {
        return 1;
    }
    return 0;
}

sub isFieldPtr($)
{
    if($LibInfo{$Version}{"info_type"}{$_[0]} eq "offset_type"
    and $LibInfo{$Version}{"info"}{$_[0]}=~/ ptrmem /) {
        return 1;
    }
    return 0;
}

sub pointTo($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/ptd[ ]*:[ ]*@(\d+)/) {
        return $1;
    }
    else {
        return "";
    }
}

sub getTypeInfo_All()
{
    foreach (sort {int($a)<=>int($b)} keys(%{$LibInfo{$Version}{"info"}}))
    {
        my $IType = $LibInfo{$Version}{"info_type"}{$_};
        if($IType=~/_type\Z/ and $IType ne "function_type"
        and $IType ne "method_type") {
            getTypeInfo(getTypeDeclId("$_"), "$_");
        }
    }
    $TypeInfo{$Version}{""}{-1}{"Name"} = "...";
    $TypeInfo{$Version}{""}{-1}{"Type"} = "Intrinsic";
    $TypeInfo{$Version}{""}{-1}{"Tid"} = -1;
}

sub getTypeInfo($$)
{
    my ($TypeDeclId, $TypeId) = @_;
    $Tid_TDid{$Version}{$TypeId} = $TypeDeclId if($TypeDeclId);
    %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = getTypeAttr($TypeDeclId, $TypeId);
    my $TName = $TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"};
    if(not $TName) {
        delete($TypeInfo{$Version}{$TypeDeclId}{$TypeId});
        return;
    }
    if(not $TName_Tid{$Version}{$TName}) {
        $TName_Tid{$Version}{$TName} = $TypeId;
    }
}

sub getArraySize($$)
{
    my ($TypeId, $BaseName) = @_;
    my $SizeBytes = getSize($TypeId)/8;
    while($BaseName=~s/\s*\[(\d+)\]//) {
        $SizeBytes/=$1;
    }
    my $BasicId = $TName_Tid{$Version}{$BaseName};
    my $BasicSize = $TypeInfo{$Version}{getTypeDeclId($BasicId)}{$BasicId}{"Size"};
    $SizeBytes/=$BasicSize if($BasicSize);
    return $SizeBytes;
}

sub getTParams($$)
{
    my ($TypeDeclId, $TypeId) = @_;
    my @Template_Params = ();
    foreach my $Param_Pos (sort {int($a)<=>int($b)} keys(%{$TemplateInstance{$Version}{$TypeDeclId}{$TypeId}}))
    {
        my $Param_TypeId = $TemplateInstance{$Version}{$TypeDeclId}{$TypeId}{$Param_Pos};
        my $Param = get_TemplateParam($Param_TypeId);
        if($Param eq "") {
            return ();
        }
        elsif($Param_Pos>=1
        and $Param=~/\Astd::(allocator|less|((char|regex)_traits)|((i|o)streambuf_iterator))\</)
        { # template<typename _Tp, typename _Alloc = std::allocator<_Tp> >
            # template<typename _Key, typename _Compare = std::less<_Key>
            # template<typename _CharT, typename _Traits = std::char_traits<_CharT> >
            # template<typename _Ch_type, typename _Rx_traits = regex_traits<_Ch_type> >
            # template<typename _CharT, typename _InIter = istreambuf_iterator<_CharT> >
            # template<typename _CharT, typename _OutIter = ostreambuf_iterator<_CharT> >
            next;
        }
        elsif($Param ne "\@skip\@") {
            @Template_Params = (@Template_Params, $Param);
        }
    }
    return @Template_Params;
}

sub getTypeAttr($$)
{
    my ($TypeDeclId, $TypeId) = @_;
    my ($BaseTypeSpec, %TypeAttr) = ();
    if($TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"}) {
        return %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}};
    }
    $TypeAttr{"Tid"} = $TypeId;
    $TypeAttr{"TDid"} = $TypeDeclId;
    $TypeAttr{"Type"} = getTypeType($TypeDeclId, $TypeId);
    if($TypeAttr{"Type"} eq "Unknown") {
        return ();
    }
    elsif($TypeAttr{"Type"}=~/(Func|Method|Field)Ptr/)
    {
        %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = getMemPtrAttr(pointTo($TypeId), $TypeDeclId, $TypeId, $TypeAttr{"Type"});
        $TName_Tid{$Version}{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"}} = $TypeId;
        return %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}};
    }
    elsif($TypeAttr{"Type"} eq "Array")
    {
        ($TypeAttr{"BaseType"}{"Tid"}, $TypeAttr{"BaseType"}{"TDid"}, $BaseTypeSpec) = selectBaseType($TypeDeclId, $TypeId);
        my %BaseTypeAttr = getTypeAttr($TypeAttr{"BaseType"}{"TDid"}, $TypeAttr{"BaseType"}{"Tid"});
        if($TypeAttr{"Size"} = getArraySize($TypeId, $BaseTypeAttr{"Name"}))
        {
            if($BaseTypeAttr{"Name"}=~/\A([^\[\]]+)(\[(\d+|)\].*)\Z/) {
                $TypeAttr{"Name"} = $1."[".$TypeAttr{"Size"}."]".$2;
            }
            else {
                $TypeAttr{"Name"} = $BaseTypeAttr{"Name"}."[".$TypeAttr{"Size"}."]";
            }
        }
        else
        {
            if($BaseTypeAttr{"Name"}=~/\A([^\[\]]+)(\[(\d+|)\].*)\Z/) {
                $TypeAttr{"Name"} = $1."[]".$2;
            }
            else {
                $TypeAttr{"Name"} = $BaseTypeAttr{"Name"}."[]";
            }
        }
        $TypeAttr{"Name"} = correctName($TypeAttr{"Name"});
        if($BaseTypeAttr{"Header"})  {
            $TypeAttr{"Header"} = $BaseTypeAttr{"Header"};
        }
        %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = %TypeAttr;
        $TName_Tid{$Version}{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"}} = $TypeId;
        return %TypeAttr;
    }
    elsif($TypeAttr{"Type"}=~/\A(Intrinsic|Union|Struct|Enum|Class)\Z/)
    {
        if(defined $TemplateInstance{$Version}{$TypeDeclId}{$TypeId})
        {
            my @Template_Params = getTParams($TypeDeclId, $TypeId);
            if(not @Template_Params)
            { # template declarations with abstract params
              # vector (tree_vec) of template_type_parm nodes in the TU dump
                return ();
            }
            %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = getTrivialTypeAttr($TypeDeclId, $TypeId);
            my $InstanceName = correctName($TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"}."< ".join(", ", @Template_Params)." >");
            $TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"Name"} = $InstanceName;
            if(not $TName_Tid{$Version}{$InstanceName}) {
                $TName_Tid{$Version}{$InstanceName} = $TypeId;
            }
            return %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}};
        }
        else
        {
            %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = getTrivialTypeAttr($TypeDeclId, $TypeId);
            return %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}};
        }
    }
    else
    {
        ($TypeAttr{"BaseType"}{"Tid"},$TypeAttr{"BaseType"}{"TDid"}, $BaseTypeSpec) = selectBaseType($TypeDeclId, $TypeId);
        my %BaseTypeAttr = getTypeAttr($TypeAttr{"BaseType"}{"TDid"}, $TypeAttr{"BaseType"}{"Tid"});
        if($BaseTypeAttr{"Type"} eq "Typedef")
        {# relinking typedefs
            my %BaseBase = get_Type($BaseTypeAttr{"BaseType"}{"TDid"},$BaseTypeAttr{"BaseType"}{"Tid"}, $Version);
            if($BaseTypeAttr{"Name"} eq $BaseBase{"Name"}) {
                ($TypeAttr{"BaseType"}{"Tid"},$TypeAttr{"BaseType"}{"TDid"}) = ($BaseBase{"Tid"},$BaseBase{"TDid"});
            }
        }
        if($BaseTypeSpec and $BaseTypeAttr{"Name"})
        {
            if($TypeAttr{"Type"} eq "Pointer"
            and $BaseTypeAttr{"Name"}=~/\([\*]+\)/) {
                $TypeAttr{"Name"} = $BaseTypeAttr{"Name"};
                $TypeAttr{"Name"}=~s/\(([*]+)\)/($1*)/g;
            }
            else {
                $TypeAttr{"Name"} = $BaseTypeAttr{"Name"}." ".$BaseTypeSpec;
            }
        }
        elsif($BaseTypeAttr{"Name"}) {
            $TypeAttr{"Name"} = $BaseTypeAttr{"Name"};
        }
        if($TypeAttr{"Type"} eq "Typedef")
        {
            $TypeAttr{"Name"} = getNameByInfo($TypeDeclId);
            if(my $NS = getNameSpace($TypeDeclId))
            {
                my $TypeName = $TypeAttr{"Name"};
                if($NS=~/\A(struct |union |class |)((.+)::|)\Q$TypeName\E\Z/)
                {# "some_type" is the typedef to "struct some_type" in C++
                    if($3) {
                        $TypeAttr{"Name"} = $3."::".$TypeName;
                    }
                }
                else
                {
                    $TypeAttr{"NameSpace"} = $NS;
                    $TypeAttr{"Name"} = $TypeAttr{"NameSpace"}."::".$TypeAttr{"Name"};
                    if($TypeAttr{"NameSpace"}=~/\Astd(::|\Z)/ and $BaseTypeAttr{"NameSpace"}=~/\Astd(::|\Z)/
                    and $BaseTypeAttr{"Name"}=~/</ and $TypeAttr{"Name"}!~/>(::\w+)+\Z/)
                    {
                        $StdCxxTypedef{$Version}{$BaseTypeAttr{"Name"}} = $TypeAttr{"Name"};
                        if(length($TypeAttr{"Name"})<=length($BaseTypeAttr{"Name"})) {
                            $Typedef_Equivalent{$Version}{$BaseTypeAttr{"Name"}} = $TypeAttr{"Name"};
                        }
                    }
                }
            }
            if($TypeAttr{"Name"} ne $BaseTypeAttr{"Name"}
            and $TypeAttr{"Name"}!~/>(::\w+)+\Z/ and $BaseTypeAttr{"Name"}!~/>(::\w+)+\Z/) {
                $Typedef_BaseName{$Version}{$TypeAttr{"Name"}} = $BaseTypeAttr{"Name"};
                $Typedef_Translations{$Version}{$BaseTypeAttr{"Name"}}{$TypeAttr{"Name"}} = 1;
            }
            ($TypeAttr{"Header"}, $TypeAttr{"Line"}) = getLocation($TypeDeclId);
        }
        if(not $TypeAttr{"Size"})
        {
            if($TypeAttr{"Type"} eq "Pointer") {
                $TypeAttr{"Size"} = $POINTER_SIZE{$Version};
            }
            else {
                $TypeAttr{"Size"} = $BaseTypeAttr{"Size"};
            }
        }
        $TypeAttr{"Name"} = correctName($TypeAttr{"Name"});
        if(not $TypeAttr{"Header"} and $BaseTypeAttr{"Header"})  {
            $TypeAttr{"Header"} = $BaseTypeAttr{"Header"};
        }
        %{$TypeInfo{$Version}{$TypeDeclId}{$TypeId}} = %TypeAttr;
        if(not $TName_Tid{$Version}{$TypeAttr{"Name"}}) {
            $TName_Tid{$Version}{$TypeAttr{"Name"}} = $TypeId;
        }
        return %TypeAttr;
    }
}

sub get_TemplateParam($)
{
    my $Type_Id = $_[0];
    return "" if(not $Type_Id);
    if($Cache{"get_TemplateParam"}{$Type_Id}) {
        return $Cache{"get_TemplateParam"}{$Type_Id};
    }
    if(getNodeType($Type_Id) eq "integer_cst")
    {# int (1), unsigned (2u), char ('c' as 99), ...
        my $CstTid = getTreeAttr($Type_Id, "type");
        my %CstType = getTypeAttr(getTypeDeclId($CstTid), $CstTid);
        my $Num = getNodeIntCst($Type_Id);
        if(my $CstSuffix = $ConstantSuffix{$CstType{"Name"}}) {
            return $Num.$CstSuffix;
        }
        else {
            return "(".$CstType{"Name"}.")".$Num;
        }
    }
    elsif(getNodeType($Type_Id) eq "string_cst") {
        return getNodeStrCst($Type_Id);
    }
    elsif(getNodeType($Type_Id) eq "tree_vec") {
        return "\@skip\@";
    }
    else
    {
        my $Type_DId = getTypeDeclId($Type_Id);
        my %ParamAttr = getTypeAttr($Type_DId, $Type_Id);
        if(not $ParamAttr{"Name"}) {
            return "";
        }
        if($ParamAttr{"Name"}=~/\>/)
        {
            if(my $Cover = cover_stdcxx_typedef($ParamAttr{"Name"})) {
                return $Cover;
            }
            else {
                return $ParamAttr{"Name"};
            }
        }
        else {
            return $ParamAttr{"Name"};
        }
    }
}

sub cover_stdcxx_typedef($)
{
    my $TypeName = $_[0];
    if(my $Cover = $StdCxxTypedef{$Version}{$TypeName}) {
        return $Cover;
    }
    my $TypeName_Covered = $TypeName;
    while($TypeName=~s/>[ ]*(const|volatile|restrict| |\*|\&)\Z/>/g){};
    if(my $Cover = $StdCxxTypedef{$Version}{$TypeName})
    {
        $TypeName_Covered=~s/(\W|\A)\Q$TypeName\E(\W|\Z)/$1$Cover$2/g;
        $TypeName_Covered=~s/(\W|\A)\Q$TypeName\E(\w|\Z)/$1$Cover $2/g;
    }
    return correctName($TypeName_Covered);
}

sub getNodeType($)
{
    return $LibInfo{$Version}{"info_type"}{$_[0]};
}

sub getNodeIntCst($)
{
    my $CstId = $_[0];
    my $CstTypeId = getTreeAttr($CstId, "type");
    if($EnumMembName_Id{$Version}{$CstId}) {
        return $EnumMembName_Id{$Version}{$CstId};
    }
    elsif((my $Value = getTreeValue($CstId)) ne "")
    {
        if($Value eq "0") {
            if(getNodeType($CstTypeId) eq "boolean_type") {
                return "false";
            }
            else {
                return "0";
            }
        }
        elsif($Value eq "1") {
            if(getNodeType($CstTypeId) eq "boolean_type") {
                return "true";
            }
            else {
                return "1";
            }
        }
        else {
            return $Value;
        }
    }
    else {
        return "";
    }
}

sub getNodeStrCst($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/strg[ ]*: (.+) lngt:[ ]*(\d+)/)
    { # string length is N-1 because of the null terminator
        return substr($1, 0, $2-1);
    }
    else {
        return "";
    }
}

sub getMemPtrAttr($$$$)
{# function, method and field pointers
    my ($PtrId, $TypeDeclId, $TypeId, $Type) = @_;
    my $MemInfo = $LibInfo{$Version}{"info"}{$PtrId};
    if($Type eq "FieldPtr") {
        $MemInfo = $LibInfo{$Version}{"info"}{$TypeId};
    }
    my $MemInfo_Type = $LibInfo{$Version}{"info_type"}{$PtrId};
    my $MemPtrName = "";
    my %TypeAttr = ("Size"=>$POINTER_SIZE{$Version}, "Type"=>$Type, "TDid"=>$TypeDeclId, "Tid"=>$TypeId);
    # Return
    if($Type eq "FieldPtr") {
        my %ReturnAttr = getTypeAttr(getTypeDeclId($PtrId), $PtrId);
        $MemPtrName .= $ReturnAttr{"Name"};
        $TypeAttr{"Return"} = $PtrId;
    }
    else
    {
        if($MemInfo=~/retn[ ]*:[ ]*\@(\d+) /)
        {
            my $ReturnTypeId = $1;
            my %ReturnAttr = getTypeAttr(getTypeDeclId($ReturnTypeId), $ReturnTypeId);
            $MemPtrName .= $ReturnAttr{"Name"};
            $TypeAttr{"Return"} = $ReturnTypeId;
        }
    }
    # Class
    if($MemInfo=~/(clas|cls)[ ]*:[ ]*@(\d+) /)
    {
        $TypeAttr{"Class"} = $2;
        my %Class = getTypeAttr(getTypeDeclId($TypeAttr{"Class"}), $TypeAttr{"Class"});
        if($Class{"Name"}) {
            $MemPtrName .= " (".$Class{"Name"}."\:\:*)";
        }
        else {
            $MemPtrName .= " (*)";
        }
    }
    else {
        $MemPtrName .= " (*)";
    }
    # Parameters
    if($Type eq "FuncPtr"
    or $Type eq "MethodPtr")
    {
        my @ParamTypeName = ();
        if($MemInfo=~/prms[ ]*:[ ]*@(\d+) /)
        {
            my $ParamTypeInfoId = $1;
            my $Position = 0;
            while($ParamTypeInfoId)
            {
                my $ParamTypeInfo = $LibInfo{$Version}{"info"}{$ParamTypeInfoId};
                last if($ParamTypeInfo!~/valu[ ]*:[ ]*@(\d+) /);
                my $ParamTypeId = $1;
                my %ParamAttr = getTypeAttr(getTypeDeclId($ParamTypeId), $ParamTypeId);
                last if($ParamAttr{"Name"} eq "void");
                if($Position!=0 or $Type ne "MethodPtr")
                {
                    $TypeAttr{"Param"}{$Position}{"type"} = $ParamTypeId;
                    push(@ParamTypeName, $ParamAttr{"Name"});
                }
                last if($ParamTypeInfo!~/(chan|chain)[ ]*:[ ]*@(\d+) /);
                $ParamTypeInfoId = $2;
                $Position+=1;
            }
        }
        $MemPtrName .= " (".join(", ", @ParamTypeName).")";
    }
    $TypeAttr{"Name"} = correctName($MemPtrName);
    return %TypeAttr;
}

sub getTreeTypeName($)
{
    my $Info = $LibInfo{$Version}{"info"}{$_[0]};
    if($Info=~/name[ ]*:[ ]*@(\d+) /) {
        return getNameByInfo($1);
    }
    else
    {
        if($LibInfo{$Version}{"info_type"}{$_[0]} eq "integer_type")
        {
            if($LibInfo{$Version}{"info"}{$_[0]}=~/unsigned/) {
                return "unsigned int";
            }
            else {
                return "int";
            }
        }
        else {
            return "";
        }
    }
}

sub getQual($)
{
    my $TypeId = $_[0];
    if($LibInfo{$Version}{"info"}{$TypeId}=~ /qual[ ]*:[ ]*(r|c|v)/)
    {
        my $Qual = $1;
        if($Qual eq "r") {
            return "restrict";
        }
        elsif($Qual eq "v") {
            return "volatile";
        }
        elsif($Qual eq "c") {
            return "const";
        }
    }
    return "";
}

sub selectBaseType($$)
{
    my ($TypeDeclId, $TypeId) = @_;
    if($LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:/
    and $LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:/
    and $LibInfo{$Version}{"info"}{$TypeId}=~/name[ ]*:[ ]*\@(\d+) /
    and (getTypeId($1) ne $TypeId)) {
        return (getTypeId($1), $1, getQual($TypeId));
    }
    elsif($LibInfo{$Version}{"info"}{$TypeId}!~/qual[ ]*:/
    and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@(\d+) /
    and getNameByInfo($TypeDeclId))
    {# typedefs
        return ($1, getTypeDeclId($1), "");
    }
    elsif($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*c /
    and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@(\d+) /)
    {
        return ($1, getTypeDeclId($1), "const");
    }
    elsif($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*r /
    and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@(\d+) /)
    {
        return ($1, getTypeDeclId($1), "restrict");
    }
    elsif($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*v /
    and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@(\d+) /)
    {
        return ($1, getTypeDeclId($1), "volatile");
    }
    elsif($LibInfo{$Version}{"info_type"}{$TypeId} eq "reference_type")
    {
        if($LibInfo{$Version}{"info"}{$TypeId}=~/refd[ ]*:[ ]*@(\d+) /) {
            return ($1, getTypeDeclId($1), "&");
        }
        else {
            return (0, 0, "");
        }
    }
    elsif($LibInfo{$Version}{"info_type"}{$TypeId} eq "array_type")
    {
        if($LibInfo{$Version}{"info"}{$TypeId}=~/elts[ ]*:[ ]*@(\d+) /) {
            return ($1, getTypeDeclId($1), "");
        }
        else {
            return (0, 0, "");
        }
    }
    elsif($LibInfo{$Version}{"info_type"}{$TypeId} eq "pointer_type")
    {
        if($LibInfo{$Version}{"info"}{$TypeId}=~/ptd[ ]*:[ ]*@(\d+) /) {
            return ($1, getTypeDeclId($1), "*");
        }
        else {
            return (0, 0, "");
        }
    }
    else {
        return (0, 0, "");
    }
}

sub getSymbolInfo_All()
{
    foreach (sort {int($b)<=>int($a)} keys(%{$LibInfo{$Version}{"info"}}))
    {
        if($LibInfo{$Version}{"info_type"}{$_} eq "function_decl") {
            getSymbolInfo("$_");
        }
    }
}

sub getVarDescr_All()
{
    foreach (sort {int($b)<=>int($a)} keys(%{$LibInfo{$Version}{"info"}}))
    {
        if($LibInfo{$Version}{"info_type"}{$_} eq "var_decl") {
            getVarDescr("$_");
        }
    }
}

sub getVarDescr($)
{
    my $InfoId = $_[0];
    if($LibInfo{$Version}{"info_type"}{getNameSpaceId($InfoId)} eq "function_decl") {
        return;
    }
    ($SymbolInfo{$Version}{$InfoId}{"Header"}, $SymbolInfo{$Version}{$InfoId}{"Line"}) = getLocation($InfoId);
    if(not $SymbolInfo{$Version}{$InfoId}{"Header"}
    or $SymbolInfo{$Version}{$InfoId}{"Header"}=~/\<built\-in\>|\<internal\>|\A\./) {
        delete($SymbolInfo{$Version}{$InfoId});
        return;
    }
    my $ShortName = $SymbolInfo{$Version}{$InfoId}{"ShortName"} = getVarShortName($InfoId);
    if($SymbolInfo{$Version}{$InfoId}{"ShortName"}=~/\Atmp_add_class_\d+\Z/) {
        delete($SymbolInfo{$Version}{$InfoId});
        return;
    }
    $SymbolInfo{$Version}{$InfoId}{"MnglName"} = getFuncMnglName($InfoId);
    if($SymbolInfo{$Version}{$InfoId}{"MnglName"}
    and $SymbolInfo{$Version}{$InfoId}{"MnglName"}!~/\A_Z/)
    {# validate mangled name
        delete($SymbolInfo{$Version}{$InfoId});
        return;
    }
    $SymbolInfo{$Version}{$InfoId}{"Data"} = 1;
    set_Class_And_Namespace($InfoId);
    if($COMMON_LANGUAGE{$Version} eq "C++" or $CheckHeadersOnly)
    {
        if(not $SymbolInfo{$Version}{$InfoId}{"MnglName"})
        {# for some symbols (_ZTI) the short name is the mangled name
            if($ShortName=~/\A_Z/) {
                $SymbolInfo{$Version}{$InfoId}{"MnglName"} = $ShortName;
            }
        }
        if(not $SymbolInfo{$Version}{$InfoId}{"MnglName"})
        {# try to mangle symbol (link with libraries)
            $SymbolInfo{$Version}{$InfoId}{"MnglName"} = linkSymbol($InfoId);
        }
    }
    if(not $SymbolInfo{$Version}{$InfoId}{"MnglName"}) {
        $SymbolInfo{$Version}{$InfoId}{"MnglName"} = $ShortName;
    }
    if(not is_in_library($SymbolInfo{$Version}{$InfoId}{"MnglName"}, $Version)
    and not $CheckHeadersOnly)
    {
        if(is_in_library($ShortName, $Version)
        and $SymbolInfo{$Version}{$InfoId}{"MnglName"}=~/_ZL\d+$ShortName\Z/)
        {# const int global_data is mangled as _ZL11global_data in the tree
            $SymbolInfo{$Version}{$InfoId}{"MnglName"} = $ShortName;
        }
        else {
            delete($SymbolInfo{$Version}{$InfoId});
            return;
        }
    }
    if(not $SymbolInfo{$Version}{$InfoId}{"MnglName"}) {
        delete($SymbolInfo{$Version}{$InfoId});
        return;
    }
    $SymbolInfo{$Version}{$InfoId}{"Return"} = getTypeId($InfoId);
    delete($SymbolInfo{$Version}{$InfoId}{"Return"}) if(not $SymbolInfo{$Version}{$InfoId}{"Return"});
    setFuncAccess($InfoId);
    if($SymbolInfo{$Version}{$InfoId}{"MnglName"}=~/\A_ZTV/) {
        delete($SymbolInfo{$Version}{$InfoId}{"Return"});
    }
    if($ShortName=~/\A(_Z|\?)/) {
        delete($SymbolInfo{$Version}{$InfoId}{"ShortName"});
    }
}

sub getTrivialTypeAttr($$)
{
    my ($TypeInfoId, $TypeId) = @_;
    my %TypeAttr = ();
    return () if(getTypeTypeByTypeId($TypeId)!~/\A(Intrinsic|Union|Struct|Enum)\Z/);
    setTypeAccess($TypeId, \%TypeAttr);
    ($TypeAttr{"Header"}, $TypeAttr{"Line"}) = getLocation($TypeInfoId);
    if($TypeAttr{"Header"}=~/\<built\-in\>|\<internal\>|\A\./)
    {
        delete($TypeAttr{"Header"});
        delete($TypeAttr{"Line"});
    }
    $TypeAttr{"Name"} = getNameByInfo($TypeInfoId);
    if(not $TypeAttr{"Name"}) {
        $TypeAttr{"Name"} = getTreeTypeName($TypeId);
    }
    $TypeAttr{"Name"}=~s/<(.+)\Z//g; # GCC 3.4.4 add template params to the name
    if(my $NameSpaceId = getNameSpaceId($TypeInfoId)) {
        if($NameSpaceId ne $TypeId) {
            $TypeAttr{"NameSpace"} = getNameSpace($TypeInfoId);
        }
    }
    if($TypeAttr{"NameSpace"} and isNotAnon($TypeAttr{"Name"})) {
        $TypeAttr{"Name"} = $TypeAttr{"NameSpace"}."::".$TypeAttr{"Name"};
    }
    $TypeAttr{"Name"} = correctName($TypeAttr{"Name"});
    $TypeAttr{"Type"} = getTypeType($TypeInfoId, $TypeId);
    if(isAnon($TypeAttr{"Name"}))
    {
        $TypeAttr{"Name"} = "anon-".lc($TypeAttr{"Type"})."-";
        $TypeAttr{"Name"} .= $TypeAttr{"Header"}."-".$TypeAttr{"Line"};
    }
    $TypeAttr{"Size"} = getSize($TypeId)/8;
    if($TypeAttr{"Type"} eq "Struct" and detect_lang($TypeId)) {
        $TypeAttr{"Type"} = "Class";
    }
    if($TypeAttr{"Type"} eq "Struct"
    or $TypeAttr{"Type"} eq "Class") {
        setBaseClasses($TypeInfoId, $TypeId, \%TypeAttr);
    }
    setTypeMemb($TypeId, \%TypeAttr);
    $TypeAttr{"Tid"} = $TypeId;
    $TypeAttr{"TDid"} = $TypeInfoId;
    $Tid_TDid{$Version}{$TypeId} = $TypeInfoId if($TypeInfoId);
    if(not $TName_Tid{$Version}{$TypeAttr{"Name"}}) {
        $TName_Tid{$Version}{$TypeAttr{"Name"}} = $TypeId;
    }
    if(my $VTable = $ClassVTable_Content{$Version}{$TypeAttr{"Name"}})
    {
        my @Entries = split(/\n/, $VTable);
        foreach (1 .. $#Entries)
        {
            my $Entry = $Entries[$_];
            if($Entry=~/\A(\d+)\s+(.+)\Z/) {
                $TypeAttr{"VTable"}{$1} = $2;
            }
        }
    }
    return %TypeAttr;
}

sub detect_lang($)
{
    my $TypeId = $_[0];
    my $Info = $LibInfo{$Version}{"info"}{$TypeId};
    if(check_gcc_version($GCC_PATH, "4"))
    {# GCC 4 fncs-node points to only non-artificial methods
        return ($Info=~/(fncs)[ ]*:[ ]*@(\d+) /);
    }
    else
    {# GCC 3
        my $Fncs = getTreeAttr($TypeId, "fncs");
        while($Fncs)
        {
            my $Info = $LibInfo{$Version}{"info"}{$Fncs};
            if($Info!~/artificial/) {
                return 1;
            }
            $Fncs = getTreeAttr($Fncs, "chan");
        }
    }
    return 0;
}

sub setBaseClasses($$$)
{
    my ($TypeInfoId, $TypeId, $TypeAttr) = @_;
    my $Info = $LibInfo{$Version}{"info"}{$TypeId};
    if($Info=~/binf[ ]*:[ ]*@(\d+) /)
    {
        $Info = $LibInfo{$Version}{"info"}{$1};
        my $Pos = 0;
        while($Info=~s/(pub|public|prot|protected|priv|private|)[ ]+binf[ ]*:[ ]*@(\d+) //)
        {
            my ($Access, $BInfoId) = ($1, $2);
            my $ClassId = getBinfClassId($BInfoId);
            my $BaseInfo = $LibInfo{$Version}{"info"}{$BInfoId};
            if($Access=~/prot/)
            {
                $TypeAttr->{"Base"}{$ClassId}{"access"} = "protected";
            }
            elsif($Access=~/priv/)
            {
                $TypeAttr->{"Base"}{$ClassId}{"access"} = "private";
            }
            $TypeAttr->{"Base"}{$ClassId}{"pos"} = $Pos++;
            if($BaseInfo=~/virt/)
            {# virtual base
                $TypeAttr->{"Base"}{$ClassId}{"virtual"} = 1;
            }
            $Class_SubClasses{$Version}{$ClassId}{$TypeId}=1;
        }
    }
}

sub getBinfClassId($)
{
    my $Info = $LibInfo{$Version}{"info"}{$_[0]};
    $Info=~/type[ ]*:[ ]*@(\d+) /;
    return $1;
}

sub construct_signature($)
{
    my $InfoId = $_[0];
    if($Cache{"construct_signature"}{$Version}{$InfoId}) {
        return $Cache{"construct_signature"}{$Version}{$InfoId};
    }
    my $PureSignature = $SymbolInfo{$Version}{$InfoId}{"ShortName"};
    if($SymbolInfo{$Version}{$InfoId}{"Destructor"}) {
        $PureSignature = "~".$PureSignature;
    }
    if(not $SymbolInfo{$Version}{$InfoId}{"Data"})
    {
        my @ParamTypes = ();
        foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$SymbolInfo{$Version}{$InfoId}{"Param"}}))
        {# checking parameters
            my $ParamType_Id = $SymbolInfo{$Version}{$InfoId}{"Param"}{$ParamPos}{"type"};
            my $ParamType_Name = uncover_typedefs(get_TypeName($ParamType_Id, $Version));
            if($OSgroup eq "windows") {
                $ParamType_Name=~s/(\W|\A)long long(\W|\Z)/$1__int64$2/;
            }
            @ParamTypes = (@ParamTypes, $ParamType_Name);
        }
        $PureSignature .= "(".join(", ", @ParamTypes).")";
        $PureSignature = delete_keywords($PureSignature);
    }
    if(my $ClassId = $SymbolInfo{$Version}{$InfoId}{"Class"}) {
        $PureSignature = get_TypeName($ClassId, $Version)."::".$PureSignature;
    }
    elsif(my $NS = $SymbolInfo{$Version}{$InfoId}{"NameSpace"}) {
        $PureSignature = $NS."::".$PureSignature;
    }
    if($SymbolInfo{$Version}{$InfoId}{"Const"}) {
        $PureSignature .= " const";
    }
    if(defined $TemplateInstance_Func{$Version}{$InfoId}
    and keys(%{$TemplateInstance_Func{$Version}{$InfoId}}))
    { # mangled names for template function specializations include return value
        if(my $Return = $SymbolInfo{$Version}{$InfoId}{"Return"}) {
            $PureSignature = get_TypeName($Return, $Version)." ".$PureSignature;
        }
    }
    return ($Cache{"construct_signature"}{$Version}{$InfoId} = correctName($PureSignature));
}

sub mangle_symbol($)
{ # mangling for simple methods
  # see gcc-4.6.0/gcc/cp/mangle.c
    my $InfoId = $_[0];
    if($Cache{"mangle_symbol"}{$Version}{$InfoId}) {
        return $Cache{"mangle_symbol"}{$Version}{$InfoId};
    }
    my ($Mangled, $ClassId, $NameSpace) = ("_Z", 0, "");
    my %Repl = ();# SN_ replacements
    if($ClassId = $SymbolInfo{$Version}{$InfoId}{"Class"})
    {
        my $MangledClass = mangle_param($ClassId, \%Repl);
        if($MangledClass!~/\AN/) {
            $MangledClass = "N".$MangledClass;
        }
        else {
            $MangledClass=~s/E\Z//;
        }
        if($SymbolInfo{$Version}{$InfoId}{"Const"}) {
            $MangledClass=~s/\AN/NK/;
        }
        
        $Mangled .= $MangledClass;
    }
    elsif($NameSpace = $SymbolInfo{$Version}{$InfoId}{"NameSpace"})
    {
        my $MangledNS = "";
        my @Parts = split(/::/, $NameSpace);
        foreach my $Num (0 .. $#Parts)
        {
            $MangledNS .= length($Parts[$Num]).$Parts[$Num];
            add_substitution($MangledNS, \%Repl);
        }
        $Mangled .= "N".$MangledNS;
    }
    my @TParams = getTParams_Func($InfoId);
    if($SymbolInfo{$Version}{$InfoId}{"Constructor"}) {
        $Mangled .= "C1";
    }
    elsif($SymbolInfo{$Version}{$InfoId}{"Destructor"}) {
        $Mangled .= "D0";
    }
    elsif(my $ShortName = $SymbolInfo{$Version}{$InfoId}{"ShortName"})
    {
        if($NameSpace eq "__gnu_cxx"
        and not $ClassId)
        { # _ZN9__gnu_cxxL25__exchange_and_add_singleEPii
          # _ZN9__gnu_cxxL19__atomic_add_singleEPii
            $Mangled .= "L";
        }
        $ShortName=~s/ <.+>\Z//g;#templates
        if($ShortName=~/\Aoperator(\W.*)\Z/)
        {
            my $Op = $1;
            $Op=~s/\A[ ]+//g;
            if(my $OpMngl = $OperatorMangling{$Op}) {
                $Mangled .= $OpMngl;
            }
            else {# conversion operator
                $Mangled .= "cv".mangle_param(getTypeIdByName($Op, $Version), \%Repl);
            }
        }
        else {
            $Mangled .= length($ShortName).$ShortName;
        }
        if(@TParams)
        {# templates
            $Mangled .= "I";
            foreach my $TParam (@TParams) {
                $Mangled .= mangle_template_param($TParam, \%Repl);
            }
            $Mangled .= "E";
        }
        if(not $ClassId) {
            add_substitution($ShortName, \%Repl);
        }
    }
    if($ClassId or $NameSpace) {
        $Mangled .= "E";
    }
    if(@TParams) {
        if(my $Return = $SymbolInfo{$Version}{$InfoId}{"Return"}) {
            $Mangled .= mangle_param($Return, \%Repl);
        }
    }
    if(not $SymbolInfo{$Version}{$InfoId}{"Data"})
    {
        my @Params = keys(%{$SymbolInfo{$Version}{$InfoId}{"Param"}});
        foreach my $ParamPos (sort {int($a) <=> int($b)} @Params)
        {# checking parameters
            my $ParamType_Id = $SymbolInfo{$Version}{$InfoId}{"Param"}{$ParamPos}{"type"};
            $Mangled .= mangle_param($ParamType_Id, \%Repl);
        }
        if(not @Params) {
            $Mangled .= "v";
        }
    }
    # optimization: one of SN_ is the superposition of others
    my %Dups = ();
    foreach my $RType1 (sort {$Repl{$b}<=>$Repl{$a}} keys(%Repl))
    {
        my $Num1 = $Repl{$RType1};
        my $Converted = $RType1;
        foreach my $RType2 (keys(%Repl))
        {
            next if($RType1 eq $RType2);
            my $Replace = macro_mangle($Repl{$RType2});
            $Converted=~s/\Q$RType2\E/$Replace/g;
        }
        foreach my $RType2 (sort {$Repl{$a}<=>$Repl{$b}} keys(%Repl))
        {
            my $Num2 = $Repl{$RType2};
            if($Num1!=$Num2)
            {
                if($Converted eq $RType2
                or $Converted eq macro_mangle($Num2)) {
                    $Dups{$RType2}=macro_mangle($Num1);
                }
            }
        }
    }
    foreach my $Dup (keys(%Dups))
    {
        my $Replace = $Dups{$Dup};
        $Mangled=~s/\Q$Dup\E/$Replace/g;
    }
    $Mangled = correct_incharge($InfoId, $Mangled);
    $Mangled = write_stdcxx_substitution($Mangled);
    $Cache{"mangle_symbol"}{$Version}{$InfoId} = $Mangled;
    return $Cache{"mangle_symbol"}{$Version}{$InfoId};
}

sub correct_incharge($$)
{
    my ($InfoId, $Mangled) = @_;
    if($SymbolInfo{$Version}{$InfoId}{"Constructor"})
    {
        if($MangledNames{$Version}{$Mangled}) {
            $Mangled=~s/C1E/C2E/;
        }
    }
    elsif($SymbolInfo{$Version}{$InfoId}{"Destructor"})
    {
        if($MangledNames{$Version}{$Mangled}) {
            $Mangled=~s/D0E/D1E/;
        }
        if($MangledNames{$Version}{$Mangled}) {
            $Mangled=~s/D1E/D2E/;
        }
    }
    return $Mangled;
}

sub mangle_param($$)
{
    my ($PTid, $Repl) = @_;
    my ($MPrefix, $Mangled) = ("", "");
    my %ReplCopy = %{$Repl};
    my %BaseType = get_BaseType($Tid_TDid{$Version}{$PTid}, $PTid, $Version);
    my $BaseType_Name = $BaseType{"Name"};
    my $ShortName = $BaseType_Name;
    $ShortName=~s/<(.*)\Z//g;#templates
    my $Suffix = get_BaseTypeQual($Tid_TDid{$Version}{$PTid}, $PTid, $Version);
    $Suffix=~s/(const|volatile)\Z//g;
    while($Suffix=~/(&|\*|const)\Z/)
    {
        if($Suffix=~s/[ ]*&\Z//) {
            $MPrefix .= "R";
        }
        if($Suffix=~s/[ ]*\*\Z//) {
            $MPrefix .= "P";
        }
        if($Suffix=~s/[ ]*const\Z//)
        {
            if($MPrefix=~/R|P/
            or $Suffix=~/&|\*/) {
                $MPrefix .= "K";
            }
        }
        if($Suffix=~s/[ ]*volatile\Z//) {
            $MPrefix .= "V";
        }
        if($Suffix=~s/[ ]*restrict\Z//) {
            $MPrefix .= "r";
        }
    }
    if(my $Token = $IntrinsicMangling{$BaseType_Name}) {
        $Mangled .= $Token;
    }
    elsif($BaseType{"Type"}=~/(Class|Struct|Union|Enum)/)
    {
        $Mangled .= "N";
        my @TParams = getTParams($BaseType{"TDid"}, $BaseType{"Tid"});
        my $MangledNS = "";
        my @Parts = split(/::/, $ShortName);
        foreach my $Num (0 .. $#Parts)
        {
            $MangledNS .= length($Parts[$Num]).$Parts[$Num];
            if($Num!=$#Parts or @TParams) {
                add_substitution($MangledNS, $Repl);
            }
        }
        $Mangled .= $MangledNS;
        if(@TParams)
        {# templates
            $Mangled .= "I";
            foreach my $TParam (@TParams) {
                $Mangled .= mangle_template_param($TParam, $Repl);
            }
            $Mangled .= "E";
        }
        $Mangled .= "E";
    }
    elsif($BaseType{"Type"}=~/(FuncPtr|MethodPtr)/)
    {
        if($BaseType{"Type"} eq "MethodPtr") {
            $Mangled .= "M".mangle_param($BaseType{"Class"}, $Repl)."F";
        }
        else {
            $Mangled .= "PF";
        }
        $Mangled .= mangle_param($BaseType{"Return"}, $Repl);
        foreach my $Num (sort {int($a)<=>int($b)} keys(%{$BaseType{"Param"}})) {
            $Mangled .= mangle_param($BaseType{"Param"}{$Num}{"type"}, $Repl);
        }
        $Mangled .= "E";
    }
    elsif($BaseType{"Type"} eq "FieldPtr")
    {
        $Mangled .= "M".mangle_param($BaseType{"Class"}, $Repl);
        $Mangled .= mangle_param($BaseType{"Return"}, $Repl);
    }
    if(my $Optimized = write_substitution($Mangled, \%ReplCopy))
    {
        if($Mangled eq $Optimized)
        {
            if($ShortName!~/::/) {
                $Mangled=~s/\AN(.+)E\Z/$1/g;
            }
        }
        else {
            $Mangled = $Optimized;
        }
    }
    $Mangled = $MPrefix.$Mangled;# add prefix (RPK)
    add_substitution($Mangled, $Repl);
    return $Mangled;
}

sub mangle_template_param($$)
{# types + literals
    my ($TParam, $Repl) = @_;
    if(my $TPTid = $TName_Tid{$Version}{$TParam}) {
        return mangle_param($TPTid, $Repl);
    }
    elsif($TParam=~/\A(\d+)(\w+)\Z/)
    {# class_name<1u>::method(...)
        return "L".$IntrinsicMangling{$ConstantSuffixR{$2}}.$1."E";
    }
    elsif($TParam=~/\A\(([\w ]+)\)(\d+)\Z/)
    { # class_name<(signed char)1>::method(...)
        return "L".$IntrinsicMangling{$1}.$2."E";
    }
    elsif($TParam eq "true")
    { # class_name<true>::method(...)
        return "Lb1E";
    }
    elsif($TParam eq "false")
    { # class_name<true>::method(...)
        return "Lb0E";
    }
    else { # internal error
        return length($TParam).$TParam;
    }
}

sub add_substitution($$)
{
    my ($Value, $Repl) = @_;
    return if($Value=~/\AS(\d*)_\Z/);
    return if(defined $Repl->{$Value});
    return if(length($Value)<=1);
    return if($StdcxxMangling{$Value});
    $Value=~s/\AN(.+)E\Z/$1/g;
    my @Repls = sort {$b<=>$a} values(%{$Repl});
    if(@Repls) {
        $Repl->{$Value} = $Repls[0]+1;
    }
    else {
        $Repl->{$Value} = -1;
    }
}

sub macro_mangle($)
{
    my $Num = $_[0];
    if($Num==-1) {
        return "S_";
    }
    else {
        return "S".$Num."_";
    }
}

sub write_stdcxx_substitution($)
{
    my $Mangled = $_[0];
    if($StdcxxMangling{$Mangled}) {
        return $StdcxxMangling{$Mangled};
    }
    else
    {
        my @Repls = keys(%StdcxxMangling);
        @Repls = sort {length($b)<=>length($a)} @Repls;
        foreach my $MangledType (@Repls)
        {
            my $Replace = $StdcxxMangling{$MangledType};
            if($Mangled!~/$Replace/) {
                $Mangled=~s/N\Q$MangledType\EE/$Replace/g;
                $Mangled=~s/\Q$MangledType\E/$Replace/g;
            }
        }
    }
    return $Mangled;
}

sub write_substitution($$)
{
    my ($Mangled, $Repl) = @_;
    if(defined $Repl->{$Mangled}
    and my $MnglNum = $Repl->{$Mangled}) {
        $Mangled = macro_mangle($MnglNum);
    }
    else
    {
        my @Repls = keys(%{$Repl});
        #@Repls = sort {$Repl->{$a}<=>$Repl->{$b}} @Repls;
        # FIXME: how to apply replacements? by num or by pos
        @Repls = sort {length($b)<=>length($a)} @Repls;
        foreach my $MangledType (@Repls)
        {
            my $Replace = macro_mangle($Repl->{$MangledType});
            if($Mangled!~/$Replace/) {
                $Mangled=~s/N\Q$MangledType\EE/$Replace/g;
                $Mangled=~s/\Q$MangledType\E/$Replace/g;
            }
        }
    }
    return $Mangled;
}

sub delete_keywords($)
{
    my $TypeName = $_[0];
    $TypeName=~s/(\W|\A)(enum |struct |union |class )/$1/g;
    return $TypeName;
}

my %Intrinsic_Keywords = (
    "true"=>1,
    "false"=>1,
    "_Bool"=>1,
    "_Complex"=>1,
    "const"=>1,
    "int"=>1,
    "long"=>1,
    "void"=>1,
    "short"=>1,
    "float"=>1,
    "volatile"=>1,
    "restrict"=>1,
    "unsigned"=>1,
    "signed"=>1,
    "char"=>1,
    "double"=>1,
    "class"=>1,
    "struct"=>1,
    "union"=>1,
    "enum"=>1
);

sub uncover_typedefs($)
{
    my $TypeName = $_[0];
    return "" if(not $TypeName);
    return $Cache{"uncover_typedefs"}{$Version}{$TypeName} if(defined $Cache{"uncover_typedefs"}{$Version}{$TypeName});
    my ($TypeName_New, $TypeName_Pre) = (correctName($TypeName), "");
    while($TypeName_New ne $TypeName_Pre)
    {
        $TypeName_Pre = $TypeName_New;
        my $TypeName_Copy = $TypeName_New;
        my %Words = ();
        while($TypeName_Copy=~s/(\W|\A)([a-z_][\w:]*)(\W|\Z)//io)
        {
            my $Word = $2;
            next if(not $Word or $Intrinsic_Keywords{$Word});
            $Words{$Word} = 1;
        }
        foreach my $Word (keys(%Words))
        {
            my $BaseType_Name = $Typedef_BaseName{$Version}{$Word};
            next if(not $BaseType_Name);
            next if($TypeName_New=~/(\A|\W)(struct|union|enum)\s\Q$Word\E(\W|\Z)/);
            if($BaseType_Name=~/\([\*]+\)/)
            {# FuncPtr
                if($TypeName_New=~/\Q$Word\E(.*)\Z/)
                {
                    my $Type_Suffix = $1;
                    $TypeName_New = $BaseType_Name;
                    if($TypeName_New=~s/\(([\*]+)\)/($1 $Type_Suffix)/) {
                        $TypeName_New = correctName($TypeName_New);
                    }
                }
            }
            else
            {
                if($TypeName_New=~s/(\W|\A)\Q$Word\E(\W|\Z)/$1$BaseType_Name$2/g) {
                    $TypeName_New = correctName($TypeName_New);
                }
            }
        }
    }
    $Cache{"uncover_typedefs"}{$Version}{$TypeName} = $TypeName_New;
    return $TypeName_New;
}

sub isInternal($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    return 0 if($FuncInfo!~/mngl[ ]*:[ ]*@(\d+) /);
    my $FuncMnglNameInfoId = $1;
    return ($LibInfo{$Version}{"info"}{$FuncMnglNameInfoId}=~/\*[ ]*INTERNAL[ ]*\*/);
}

sub set_Class_And_Namespace($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/scpe[ ]*:[ ]*@(\d+) /)
    {
        my $NameSpaceInfoId = $1;
        if($LibInfo{$Version}{"info_type"}{$NameSpaceInfoId} eq "namespace_decl") {
            $SymbolInfo{$Version}{$FuncInfoId}{"NameSpace"} = getNameSpace($FuncInfoId);
        }
        elsif($LibInfo{$Version}{"info_type"}{$NameSpaceInfoId} eq "record_type") {
            $SymbolInfo{$Version}{$FuncInfoId}{"Class"} = $NameSpaceInfoId;
        }
    }
}

sub linkSymbol($)
{ # link symbols from shared libraries
  # with the symbols from header files
    my $InfoId = $_[0];
    # try to mangle symbol
    if(not $SymbolInfo{$Version}{$InfoId}{"Class"}
    or not check_gcc_version($GCC_PATH, "4") or $CheckHeadersOnly)
    {
        if(my $Mangled = $mangled_name_gcc{construct_signature($InfoId)})
        { # for C++ functions that mangled in library, but are not mangled in TU dump (using GCC 4.x)
            $Mangled = correct_incharge($InfoId, $Mangled);
            return $Mangled;
        }
        elsif(my $Mangled = $mangled_name{$Version}{synchTr(construct_signature($InfoId))})
        { # for linking MS C++ symbols from library with GCC symbols from headers
            return $Mangled;
        }
        elsif(my $Mangled = mangle_symbol($InfoId))
        { # 1. for GCC 3.4.4 that doesn't mangle names in the TU dump (only functions and global data)
          # 2. for --headers-only option
            return $Mangled;
        }
        else
        { # internal error
            return "";
        }
    }
}

sub getTParams_Func($)
{
    my $FuncInfoId = $_[0];
    my @TmplParams = ();
    foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$TemplateInstance_Func{$Version}{$FuncInfoId}}))
    {
        my $Param = get_TemplateParam($TemplateInstance_Func{$Version}{$FuncInfoId}{$ParamPos});
        if($Param eq "") {
            delete($SymbolInfo{$Version}{$FuncInfoId});
            return;
        }
        elsif($Param ne "\@skip\@") {
            push(@TmplParams, $Param);
        }
    }
    return @TmplParams;
}

sub getSymbolInfo($)
{
    my $FuncInfoId = $_[0];
    return if(isInternal($FuncInfoId));
    ($SymbolInfo{$Version}{$FuncInfoId}{"Header"}, $SymbolInfo{$Version}{$FuncInfoId}{"Line"}) = getLocation($FuncInfoId);
    if(not $SymbolInfo{$Version}{$FuncInfoId}{"Header"}
    or $SymbolInfo{$Version}{$FuncInfoId}{"Header"}=~/\<built\-in\>|\<internal\>|\A\./) {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    setFuncAccess($FuncInfoId);
    setFuncKind($FuncInfoId);
    if($SymbolInfo{$Version}{$FuncInfoId}{"PseudoTemplate"}) {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    $SymbolInfo{$Version}{$FuncInfoId}{"Type"} = getFuncType($FuncInfoId);
    $SymbolInfo{$Version}{$FuncInfoId}{"Return"} = getFuncReturn($FuncInfoId);
    delete($SymbolInfo{$Version}{$FuncInfoId}{"Return"}) if(not $SymbolInfo{$Version}{$FuncInfoId}{"Return"});
    $SymbolInfo{$Version}{$FuncInfoId}{"ShortName"} = getFuncShortName(getFuncOrig($FuncInfoId));
    if($SymbolInfo{$Version}{$FuncInfoId}{"ShortName"}=~/\._/) {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    if(defined $TemplateInstance_Func{$Version}{$FuncInfoId})
    {
        my $PrmsInLine = join(", ", getTParams_Func($FuncInfoId));
        if($PrmsInLine=~/>\Z/) {
            $PrmsInLine = " ".$PrmsInLine." ";
        }
        $SymbolInfo{$Version}{$FuncInfoId}{"ShortName"} .= " <".$PrmsInLine.">";
    }
    $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"} = getFuncMnglName($FuncInfoId);
    if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}
    and $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}!~/\A_Z/)
    {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    if($SymbolInfo{$FuncInfoId}{"MnglName"} and not $STDCXX_TESTING)
    {# stdc++ interfaces
        if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}=~/\A(_ZS|_ZNS|_ZNKS)/) {
            delete($SymbolInfo{$Version}{$FuncInfoId});
            return;
        }
    }
    setFuncParams($FuncInfoId);
    set_Class_And_Namespace($FuncInfoId);
    if((is_in_library($SymbolInfo{$Version}{$FuncInfoId}{"ShortName"}, $Version) or $CheckHeadersOnly)
    and $SymbolInfo{$Version}{$FuncInfoId}{"Type"} eq "Function") {
        # functions (C++): not mangled in library, but are mangled in TU dump
        if(not $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}
          or not is_in_library($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}, $Version)) {
            $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"} = $SymbolInfo{$Version}{$FuncInfoId}{"ShortName"};
        }
    }
    if($COMMON_LANGUAGE{$Version} eq "C++" or $CheckHeadersOnly)
    {
        if($SymbolInfo{$Version}{$FuncInfoId}{"ShortName"}=~/\A__(comp|base|deleting)_(c|d)tor\Z/)
        { # support for old GCC versions: reconstruct real names for constructors and destructors
            $SymbolInfo{$Version}{$FuncInfoId}{"ShortName"} = getNameByInfo(getTypeDeclId($SymbolInfo{$Version}{$FuncInfoId}{"Class"}));
        }
        if(not $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"})
        {# try to mangle symbol (link with libraries)
            $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"} = linkSymbol($FuncInfoId);
        }
    }
    if(not $SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}) {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    if(getFuncSpec($FuncInfoId) eq "Virt")
    {# virtual methods
        $SymbolInfo{$Version}{$FuncInfoId}{"Virt"} = 1;
    }
    if(getFuncSpec($FuncInfoId) eq "PureVirt")
    {# pure virtual methods
        $SymbolInfo{$Version}{$FuncInfoId}{"PureVirt"} = 1;
    }
    if(not is_in_library($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}, $Version)
    and not $CheckHeadersOnly and $OSgroup ne "windows"
    and not $SymbolInfo{$Version}{$FuncInfoId}{"Virt"}
    and not $SymbolInfo{$Version}{$FuncInfoId}{"PureVirt"})
    {# removing src only and external non-virtual functions
     # non-virtual template instances going here
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"})
    {
        if($MangledNames{$Version}{$SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}})
        {# one instance for one mangled name only
            delete($SymbolInfo{$Version}{$FuncInfoId});
            return;
        }
        else {
            $MangledNames{$Version}{$SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}} = 1;
        }
    }
    if($SymbolInfo{$Version}{$FuncInfoId}{"Constructor"}
    or $SymbolInfo{$Version}{$FuncInfoId}{"Destructor"}) {
        delete($SymbolInfo{$Version}{$FuncInfoId}{"Return"});
    }
    if(isInline($FuncInfoId)) {
        $SymbolInfo{$Version}{$FuncInfoId}{"InLine"} = 1;
    }
    if($SymbolInfo{$Version}{$FuncInfoId}{"Type"} eq "Method"
    or $SymbolInfo{$Version}{$FuncInfoId}{"Constructor"}
    or $SymbolInfo{$Version}{$FuncInfoId}{"Destructor"}) {
        if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}!~/\A_Z/) {
            delete($SymbolInfo{$Version}{$FuncInfoId});
            return;
        }
    }
    if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}=~/\A_Z/
    and $SymbolInfo{$Version}{$FuncInfoId}{"Class"})
    {
        if($SymbolInfo{$Version}{$FuncInfoId}{"Type"} eq "Function")
        {# static methods
            $SymbolInfo{$Version}{$FuncInfoId}{"Static"} = 1;
        }
    }
    if(getFuncLink($FuncInfoId) eq "Static") {
        $SymbolInfo{$Version}{$FuncInfoId}{"Static"} = 1;
    }
    if($SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}=~/\A(_Z|\?)/
    and $tr_name{$SymbolInfo{$Version}{$FuncInfoId}{"MnglName"}}=~/\.\_\d/) {
        delete($SymbolInfo{$Version}{$FuncInfoId});
        return;
    }
    delete($SymbolInfo{$Version}{$FuncInfoId}{"Type"});
}

sub isInline($)
{# "body: undefined" in the tree
 # -fkeep-inline-functions GCC option should be specified
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    if($FuncInfo=~/ undefined /i) {
        return 0;
    }
    return 1;
}

sub getTypeId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/type[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub setTypeMemb($$)
{
    my ($TypeId, $TypeAttr) = @_;
    my $TypeType = $TypeAttr->{"Type"};
    my ($Position, $UnnamedPos) = (0, 0);
    if($TypeType eq "Enum")
    {
        my $TypeMembInfoId = getEnumMembInfoId($TypeId);
        while($TypeMembInfoId)
        {
            $TypeAttr->{"Memb"}{$Position}{"value"} = getEnumMembVal($TypeMembInfoId);
            my $MembName = getEnumMembName($TypeMembInfoId);
            $TypeAttr->{"Memb"}{$Position}{"name"} = getEnumMembName($TypeMembInfoId);
            $EnumMembName_Id{$Version}{getTreeAttr($TypeMembInfoId, "valu")} = ($TypeAttr->{"NameSpace"})?$TypeAttr->{"NameSpace"}."::".$MembName:$MembName;
            $TypeMembInfoId = getNextMembInfoId($TypeMembInfoId);
            $Position += 1;
        }
    }
    elsif($TypeType=~/\A(Struct|Class|Union)\Z/)
    {
        my $TypeMembInfoId = getStructMembInfoId($TypeId);
        while($TypeMembInfoId)
        {
            if($LibInfo{$Version}{"info_type"}{$TypeMembInfoId} ne "field_decl") {
                $TypeMembInfoId = getNextStructMembInfoId($TypeMembInfoId);
                next;
            }
            my $StructMembName = getStructMembName($TypeMembInfoId);
            if($StructMembName=~/_vptr\./)
            {# virtual tables
                $TypeMembInfoId = getNextStructMembInfoId($TypeMembInfoId);
                next;
            }
            if(not $StructMembName)
            { # unnamed fields
                if($TypeAttr->{"Name"}!~/_type_info_pseudo/)
                {
                    my $UnnamedTid = getTreeAttr($TypeMembInfoId, "type");
                    my $UnnamedTName = getNameByInfo(getTypeDeclId($UnnamedTid));
                    if(isAnon($UnnamedTName))
                    { # rename unnamed fields to unnamed0, unnamed1, ...
                        $StructMembName = "unnamed".($UnnamedPos++);
                    }
                }
            }
            if(not $StructMembName)
            { # unnamed fields and base classes
                $TypeMembInfoId = getNextStructMembInfoId($TypeMembInfoId);
                next;
            }
            $TypeAttr->{"Memb"}{$Position}{"type"} = getTreeAttr($TypeMembInfoId, "type");
            $TypeAttr->{"Memb"}{$Position}{"name"} = $StructMembName;
            if((my $Access = getTreeAccess($TypeMembInfoId)) ne "public")
            {# marked only protected and private, public by default
                $TypeAttr->{"Memb"}{$Position}{"access"} = $Access;
            }
            if(my $BFSize = getStructMembBitFieldSize($TypeMembInfoId)) {
                $TypeAttr->{"Memb"}{$Position}{"bitfield"} = $BFSize;
            }
            else
            { # set alignment for non-bit fields
              # alignment for bitfields is always equal to 1 bit
                $TypeAttr->{"Memb"}{$Position}{"algn"} = getAlignment($TypeMembInfoId)/8;
            }
            $TypeMembInfoId = getNextStructMembInfoId($TypeMembInfoId);
            $Position += 1;
        }
    }
}

sub setFuncParams($)
{
    my $FuncInfoId = $_[0];
    my $ParamInfoId = getFuncParamInfoId($FuncInfoId);
    my $FunctionType = getFuncType($FuncInfoId);
    if($FunctionType eq "Method")
    {
        my $ObjectTypeId = getFuncParamType($ParamInfoId);
        if(get_TypeName($ObjectTypeId, $Version)=~/ const\*const\Z/) {
            $SymbolInfo{$Version}{$FuncInfoId}{"Const"} = 1;
        }
        $ParamInfoId = getNextElem($ParamInfoId);
    }
    my ($Position, $Vtt_Pos) = (0, -1);
    while($ParamInfoId)
    {
        my $ParamTypeId = getFuncParamType($ParamInfoId);
        my $ParamName = getFuncParamName($ParamInfoId);
        last if($TypeInfo{$Version}{getTypeDeclId($ParamTypeId)}{$ParamTypeId}{"Name"} eq "void");
        if($TypeInfo{$Version}{getTypeDeclId($ParamTypeId)}{$ParamTypeId}{"Type"} eq "Restrict")
        {# delete restrict spec
            $ParamTypeId = getRestrictBase($ParamTypeId);
        }
        if($ParamName eq "__vtt_parm"
        and get_TypeName($ParamTypeId, $Version) eq "void const**")
        {
            $Vtt_Pos = $Position;
            $ParamInfoId = getNextElem($ParamInfoId);
            next;
        }
        $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"type"} = $ParamTypeId;
        $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"name"} = $ParamName;
        $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"algn"} = getAlignment($ParamInfoId)/8;
        if(not $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"name"}) {
            $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"name"} = "p".($Position+1);
        }
        $ParamInfoId = getNextElem($ParamInfoId);
        $Position += 1;
    }
    if(detect_nolimit_args($FuncInfoId, $Vtt_Pos)) {
        $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"type"} = -1;
    }
}

sub detect_nolimit_args($$)
{
    my ($FuncInfoId, $Vtt_Pos) = @_;
    my $FuncTypeId = getFuncTypeId($FuncInfoId);
    my $ParamListElemId = getTreeAttr($FuncTypeId, "prms");
    if(getFuncType($FuncInfoId) eq "Method") {
        $ParamListElemId = getNextElem($ParamListElemId);
    }
    return 1 if(not $ParamListElemId);# foo(...)
    my $HaveVoid = 0;
    my $Position = 0;
    while($ParamListElemId)
    {
        if($Vtt_Pos!=-1 and $Position==$Vtt_Pos)
        {
            $Vtt_Pos=-1;
            $ParamListElemId = getNextElem($ParamListElemId);
            next;
        }
        my $ParamTypeId = getTreeAttr($ParamListElemId, "valu");
        if(my $PurpId = getTreeAttr($ParamListElemId, "purp"))
        {
            if($LibInfo{$Version}{"info_type"}{$PurpId} eq "integer_cst") {
                $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"default"} = getTreeValue($PurpId);
            }
            elsif($LibInfo{$Version}{"info_type"}{$PurpId} eq "string_cst") {
                $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"default"} = getNodeStrCst($PurpId);
            }
        }
        if($TypeInfo{$Version}{getTypeDeclId($ParamTypeId)}{$ParamTypeId}{"Name"} eq "void") {
            $HaveVoid = 1;
            last;
        }
        elsif(not defined $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"type"})
        {
            $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"type"} = $ParamTypeId;
            if(not $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"name"}) {
                $SymbolInfo{$Version}{$FuncInfoId}{"Param"}{$Position}{"name"} = "p".($Position+1);
            }
        }
        $ParamListElemId = getNextElem($ParamListElemId);
        $Position += 1;
    }
    return ($Position>=1 and not $HaveVoid);
}

sub getTreeAttr($$)
{
    my ($Id, $Attr) = @_;
    if($LibInfo{$Version}{"info"}{$Id}=~/\Q$Attr\E[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    return "";
}

sub getTreeValue($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/low[ ]*:[ ]*([^ ]+) /) {
        return $1;
    }
    return "";
}

sub getRestrictBase($)
{
    my $TypeId = $_[0];
    my $TypeDeclId = getTypeDeclId($TypeId);
    return $TypeInfo{$Version}{$TypeDeclId}{$TypeId}{"BaseType"}{"Tid"};
}

sub getTreeAccess($)
{
    my $InfoId = $_[0];
    if($LibInfo{$Version}{"info"}{$InfoId}=~/accs[ ]*:[ ]*([a-zA-Z]+) /)
    {
        my $Access = $1;
        if($Access eq "prot") {
            return "protected";
        }
        elsif($Access eq "priv") {
            return "private";
        }
    }
    elsif($LibInfo{$Version}{"info"}{$InfoId}=~/ protected /)
    {# support for old GCC versions
        return "protected";
    }
    elsif($LibInfo{$Version}{"info"}{$InfoId}=~/ private /)
    {# support for old GCC versions
        return "private";
    }
    return "public";
}

sub setFuncAccess($)
{
    my $FuncInfoId = $_[0];
    my $Access = getTreeAccess($FuncInfoId);
    if($Access eq "protected") {
        $SymbolInfo{$Version}{$FuncInfoId}{"Protected"} = 1;
    }
    elsif($Access eq "private") {
        $SymbolInfo{$Version}{$FuncInfoId}{"Private"} = 1;
    }
}

sub setTypeAccess($$)
{
    my ($TypeId, $TypeAttr) = @_;
    my $Access = getTreeAccess($TypeId);
    if($Access eq "protected") {
        $TypeAttr->{"Protected"} = 1;
    }
    elsif($Access eq "private") {
        $TypeAttr->{"Private"} = 1;
    }
}

sub setFuncKind($)
{
    my $FuncInfoId = $_[0];
    if($LibInfo{$Version}{"info"}{$FuncInfoId}=~/pseudo tmpl/) {
        $SymbolInfo{$Version}{$FuncInfoId}{"PseudoTemplate"} = 1;
    }
    elsif($LibInfo{$Version}{"info"}{$FuncInfoId}=~/ constructor /) {
        $SymbolInfo{$Version}{$FuncInfoId}{"Constructor"} = 1;
    }
    elsif($LibInfo{$Version}{"info"}{$FuncInfoId}=~/ destructor /) {
        $SymbolInfo{$Version}{$FuncInfoId}{"Destructor"} = 1;
    }
}

sub getFuncSpec($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/spec[ ]*:[ ]*pure /) {
        return "PureVirt";
    }
    elsif($FuncInfo=~/spec[ ]*:[ ]*virt /) {
        return "Virt";
    }
    elsif($FuncInfo=~/ pure\s+virtual /)
    {# support for old GCC versions
        return "PureVirt";
    }
    elsif($FuncInfo=~/ virtual /)
    {# support for old GCC versions
        return "Virt";
    }
    return "";
}

sub getFuncClass($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/scpe[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub getFuncLink($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/link[ ]*:[ ]*static /) {
        return "Static";
    }
    else {
        if($FuncInfo=~/link[ ]*:[ ]*([a-zA-Z]+) /) {
            return $1;
        }
        else {
            return "";
        }
    }
}

sub getNextElem($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/(chan|chain)[ ]*:[ ]*@(\d+) /) {
        return $2;
    }
    else {
        return "";
    }
}

sub getFuncParamInfoId($)
{
    my $FuncInfoId = $_[0];
    my $FuncInfo = $LibInfo{$Version}{"info"}{$FuncInfoId};
    if($FuncInfo=~/args[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub getFuncParamType($)
{
    my $ParamInfoId = $_[0];
    my $ParamInfo = $LibInfo{$Version}{"info"}{$ParamInfoId};
    if($ParamInfo=~/type[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub getFuncParamName($)
{
    my $ParamInfoId = $_[0];
    my $ParamInfo = $LibInfo{$Version}{"info"}{$ParamInfoId};
    if($ParamInfo=~/name[ ]*:[ ]*@(\d+) /) {
        return getTreeStr($1);
    }
    return "";
}

sub getEnumMembInfoId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/csts[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub getStructMembInfoId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/flds[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub get_IntNameSpace($$)
{
    my ($Interface, $LibVersion) = @_;
    return "" if(not $Interface or not $LibVersion);
    return $Cache{"get_IntNameSpace"}{$Interface}{$LibVersion} if(defined $Cache{"get_IntNameSpace"}{$Interface}{$LibVersion});
    my $Signature = get_Signature($Interface, $LibVersion);
    if($Signature=~/\:\:/)
    {
        my $FounNameSpace = 0;
        foreach my $NameSpace (sort {get_depth($b)<=>get_depth($a)} keys(%{$NestedNameSpaces{$LibVersion}}))
        {
            if($Signature=~/\A\Q$NameSpace\E\:\:/
            or $Signature=~/\s+for\s+\Q$NameSpace\E\:\:/) {
                $Cache{"get_IntNameSpace"}{$Interface}{$LibVersion} = $NameSpace;
                return $NameSpace;
            }
        }
    }
    else {
        $Cache{"get_IntNameSpace"}{$Interface}{$LibVersion} = "";
        return "";
    }
}

sub get_TypeNameSpace($$)
{
    my ($TypeName, $LibVersion) = @_;
    return "" if(not $TypeName or not $LibVersion);
    if(defined $Cache{"get_TypeNameSpace"}{$TypeName}{$LibVersion}) {
        return $Cache{"get_TypeNameSpace"}{$TypeName}{$LibVersion};
    }
    if($TypeName=~/\:\:/)
    {
        my $FounNameSpace = 0;
        foreach my $NameSpace (sort {get_depth($b)<=>get_depth($a)} keys(%{$NestedNameSpaces{$LibVersion}}))
        {
            if($TypeName=~/\A\Q$NameSpace\E\:\:/) {
                $Cache{"get_TypeNameSpace"}{$TypeName}{$LibVersion} = $NameSpace;
                return $NameSpace;
            }
        }
    }
    else {
        $Cache{"get_TypeNameSpace"}{$TypeName}{$LibVersion} = "";
        return "";
    }
}

sub getNameSpace($)
{
    my $TypeInfoId = $_[0];
    my $NameSpaceInfoId = getTreeAttr($TypeInfoId, "scpe");
    return "" if(not $NameSpaceInfoId);
    if($LibInfo{$Version}{"info_type"}{$NameSpaceInfoId} eq "namespace_decl")
    {
        my $NameSpaceInfo = $LibInfo{$Version}{"info"}{$NameSpaceInfoId};
        if($NameSpaceInfo=~/name[ ]*:[ ]*@(\d+) /)
        {
            my $NameSpace = getTreeStr($1);
            return "" if($NameSpace eq "::");
            if(my $BaseNameSpace = getNameSpace($NameSpaceInfoId)) {
                $NameSpace = $BaseNameSpace."::".$NameSpace;
            }
            $NestedNameSpaces{$Version}{$NameSpace} = 1;
            return $NameSpace;
        }
        else {
            return "";
        }
    }
    elsif($LibInfo{$Version}{"info_type"}{$NameSpaceInfoId} eq "record_type")
    {# inside data type
        my %NameSpaceAttr = getTypeAttr(getTypeDeclId($NameSpaceInfoId), $NameSpaceInfoId);
        return $NameSpaceAttr{"Name"};
    }
    else {
        return "";
    }
}

sub getNameSpaceId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/scpe[ ]*:[ ]*\@(\d+)/) {
        return $1;
    }
    else {
        return "";
    }
}

sub getEnumMembName($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/purp[ ]*:[ ]*\@(\d+)/) {
        return getTreeStr($1);
    }
    else {
        return "";
    }
}

sub getStructMembName($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/name[ ]*:[ ]*\@(\d+)/) {
        return getTreeStr($1);
    }
    else {
        return "";
    }
}

sub getEnumMembVal($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/valu[ ]*:[ ]*\@(\d+)/)
    {
        if($LibInfo{$Version}{"info"}{$1}=~/cnst[ ]*:[ ]*\@(\d+)/)
        {# in newer versions of GCC the value is in the "const_decl->cnst" node
            return getTreeValue($1);
        }
        else
        {# some old versions of GCC (3.3) have the value in the "integer_cst" node
            return getTreeValue($1);
        }
    }
    return "";
}

sub getSize($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/size[ ]*:[ ]*\@(\d+)/) {
        return getTreeValue($1);
    }
    else {
        return 0;
    }
}

sub getAlignment($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/algn[ ]*:[ ]*(\d+) /) {
        return $1;
    }
    else {
        return "";
    }
}

sub getStructMembBitFieldSize($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/ bitfield /) {
        return getSize($_[0]);
    }
    else {
        return 0;
    }
}

sub getNextMembInfoId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/(chan|chain)[ ]*:[ ]*@(\d+) /) {
        return $2;
    }
    else {
        return "";
    }
}

sub getNextStructMembInfoId($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/(chan|chain)[ ]*:[ ]*@(\d+) /) {
        return $2;
    }
    else {
        return "";
    }
}

sub fieldHasName($)
{
    my $TypeMembInfoId = $_[0];
    if($LibInfo{$Version}{"info_type"}{$TypeMembInfoId} eq "field_decl")
    {
        if($LibInfo{$Version}{"info"}{$TypeMembInfoId}=~/name[ ]*:[ ]*@(\d+) /) {
            return $1;
        }
        else {
            return "";
        }
    }
    else {
        return 0;
    }
}

sub register_header($$)
{# input: header absolute path, relative path or name
    my ($Header, $LibVersion) = @_;
    return "" if(not $Header);
    if($Header=~/\A(\/|\w+:[\/\\])/ and not -f $Header) {
        print STDERR "\nERROR: can't access \'$Header\'\n";
        return "";
    }
    return "" if(skip_header($Header, $LibVersion));
    my $Header_Path = identify_header($Header, $LibVersion);
    return "" if(not $Header_Path);
    if(my $RHeader_Path = $Header_ErrorRedirect{$LibVersion}{$Header_Path})
    {
        return "" if(skip_header($RHeader_Path, $LibVersion));
        $Header_Path = $RHeader_Path;
        return "" if($RegisteredHeaders{$LibVersion}{$Header_Path});
    }
    elsif($Header_ShouldNotBeUsed{$LibVersion}{$Header_Path}) {
        return "";
    }
    $Target_Headers{$LibVersion}{$Header_Path}{"Identity"} = get_filename($Header_Path);
    $HeaderName_Paths{$LibVersion}{get_filename($Header_Path)}{$Header_Path} = 1;
    $RegisteredHeaders{$LibVersion}{$Header_Path} = 1;
    return $Header_Path;
}

sub register_directory($$)
{
    my ($Dir, $LibVersion) = @_;
    $Dir=~s/[\/\\]+\Z//g;
    return 0 if(not $LibVersion or not $Dir or not -d $Dir);
    return 0 if(skip_header($Dir, $LibVersion));
    $Dir = abs_path($Dir) if($Dir!~/\A(\/|\w+:[\/\\])/);
    return 1 if($RegisteredDirs{$LibVersion}{$Dir});
    $Header_Dependency{$LibVersion}{$Dir} = 1;
    foreach my $Path (sort {length($b)<=>length($a)} cmd_find($Dir,"f","",""))
    {
        next if(not is_header($Path, 0, $LibVersion));
        next if(ignore_path($Path, $Dir));
        next if(skip_header($Path, $LibVersion));
        $DependencyHeaders_All_FullPath{$LibVersion}{get_filename($Path)} = $Path;
        my $Prefix = get_filename(get_dirname($Path));
        $DependencyHeaders_All_FullPath{$LibVersion}{joinPath($Prefix, get_filename($Path))} = $Path;
    }
    $RegisteredDirs{$LibVersion}{$Dir} = 1;
    if(get_filename($Dir) eq "include")
    {# search for "lib/include/" directory
        my $LibDir = $Dir;
        if($LibDir=~s/([\/\\])include\Z/$1lib/g and -d $LibDir) {
            register_directory($LibDir, $LibVersion);
        }
    }
    return 1;
}

sub parse_redirect($$$)
{
    my ($Content, $Path, $LibVersion) = @_;
    return $Cache{"parse_redirect"}{$LibVersion}{$Path} if(defined $Cache{"parse_redirect"}{$LibVersion}{$Path});
    return "" if(not $Content);
    my @ErrorMacros = ();
    while($Content=~s/#\s*error\s+([^\n]+?)\s*(\n|\Z)//) {
        push(@ErrorMacros, $1);
    }
    my $Redirect = "";
    foreach my $ErrorMacro (@ErrorMacros)
    {
        if($ErrorMacro=~/(only|must\s+include|update\s+to\s+include|replaced\s+with|replaced\s+by|renamed\s+to|is\s+in)\s+(<[^<>]+>|[a-z0-9-_\/\\]+\.(h|hh|hp|hxx|hpp|h\+\+|tcc))/i)
        {
            $Redirect = $2;
            last;
        }
        elsif($ErrorMacro=~/(include|use|is\s+in)\s+(<[^<>]+>|[a-z0-9-_\/\\]+\.(h|hh|hp|hxx|hpp|h\+\+|tcc))\s+instead/i)
        {
            $Redirect = $2;
            last;
        }
        elsif($ErrorMacro=~/this\s+header\s+should\s+not\s+be\s+used/i
        or $ErrorMacro=~/programs\s+should\s+not\s+directly\s+include/i
        or $ErrorMacro=~/you\s+should\s+not\s+include/i
        or $ErrorMacro=~/you\s+should\s+not\s+be\s+including\s+this\s+file/i
        or $ErrorMacro=~/you\s+should\s+not\s+be\s+using\s+this\s+header/i
        or $ErrorMacro=~/is\s+not\s+supported\s+API\s+for\s+general\s+use/i
        or $ErrorMacro=~/cannot\sbe\sincluded\sdirectly/i) {
            $Header_ShouldNotBeUsed{$LibVersion}{$Path} = 1;
        }
    }
    $Redirect=~s/\A<//g;
    $Redirect=~s/>\Z//g;
    $Cache{"parse_redirect"}{$LibVersion}{$Path} = $Redirect;
    return $Redirect;
}

sub parse_includes($$)
{
    my ($Content, $Path) = @_;
    my %Includes = ();
    while($Content=~s/#([ \t]*)(include|import)([ \t]*)(<|")([^<>"]+)(>|")//)
    {# C/C++: include, Objective C/C++: import directive
        my ($Header, $Method) = ($5, $4);
        if(($Method eq "\"" and -e joinPath(get_dirname($Path),$Header))
        or $Header=~/\A(\/|\w+:[\/\\])/) {
        # include "path/header.h" that doesn't exist is equal to include <path/header.h>
            $Includes{$Header} = -1;
        }
        else {
            $Includes{$Header} = 1;
        }
    }
    return \%Includes;
}

sub ignore_path($$)
{
    my ($Path, $Prefix) = @_;
    return 1 if(not $Path or not -e $Path
    or not $Prefix or not -e $Prefix);
    return 1 if($Path=~/\~\Z/);# skipping system backup files
    # skipping hidden .svn, .git, .bzr, .hg and CVS directories
    return 1 if(cut_path_prefix($Path, $Prefix)=~/(\A|[\/\\]+)(\.(svn|git|bzr|hg)|CVS)([\/\\]+|\Z)/);
    return 0;
}

sub sort_by_word($$)
{
    my ($ArrRef, $W) = @_;
    return if(length($W)<2);
    @{$ArrRef} = sort {get_filename($b)=~/\Q$W\E/i<=>get_filename($a)=~/\Q$W\E/i} @{$ArrRef};
}

sub natural_sorting($$)
{
    my ($H1, $H2) = @_;
    $H1=~s/\.[a-z]+\Z//ig;
    $H2=~s/\.[a-z]+\Z//ig;
    my ($HDir1, $Hname1) = separate_path($H1);
    my ($HDir2, $Hname2) = separate_path($H2);
    my $Dirname1 = get_filename($HDir1);
    my $Dirname2 = get_filename($HDir2);
    if($H1 eq $H2) {
        return 0;
    }
    elsif($H1=~/\A\Q$H2\E/) {
        return 1;
    }
    elsif($H2=~/\A\Q$H1\E/) {
        return -1;
    }
    elsif($HDir1=~/\Q$Hname1\E/i
    and $HDir2!~/\Q$Hname2\E/i)
    {# include/glib-2.0/glib.h
        return -1;
    }
    elsif($HDir2=~/\Q$Hname2\E/i
    and $HDir1!~/\Q$Hname1\E/i)
    {# include/glib-2.0/glib.h
        return 1;
    }
    elsif($Hname1=~/\Q$Dirname1\E/i
    and $Hname2!~/\Q$Dirname2\E/i)
    {# include/hildon-thumbnail/hildon-thumbnail-factory.h
        return -1;
    }
    elsif($Hname2=~/\Q$Dirname2\E/i
    and $Hname1!~/\Q$Dirname1\E/i)
    {# include/hildon-thumbnail/hildon-thumbnail-factory.h
        return 1;
    }
    elsif($Hname1=~/(config|lib)/i
    and $Hname2!~/(config|lib)/i)
    {# include/alsa/asoundlib.h
        return -1;
    }
    elsif($Hname2=~/(config|lib)/i
    and $Hname1!~/(config|lib)/i)
    {# include/alsa/asoundlib.h
        return 1;
    }
    elsif(checkRelevance($H1)
    and not checkRelevance($H2))
    {# libebook/e-book.h
        return -1;
    }
    elsif(checkRelevance($H2)
    and not checkRelevance($H1))
    {# libebook/e-book.h
        return 1;
    }
    else {
        return (lc($H1) cmp lc($H2));
    }
}

sub register_dependency($$)
{
    my ($Path, $LibVersion) = @_;
    foreach my $Dir (cmd_find($Path,"d","","")) {
        $Dir=~s/[\/\\]+\Z//g;
        $Header_Dependency{$LibVersion}{$Dir} = 1;
    }
}

sub searchForHeaders($)
{
    my $LibVersion = $_[0];
    # gcc standard include paths
    find_gcc_cxx_headers($LibVersion);
    # processing header paths
    foreach my $Dest (keys(%{$Descriptor{$LibVersion}{"IncludePaths"}}),
    keys(%{$Descriptor{$LibVersion}{"AddIncludePaths"}}))
    {
        my $IDest = $Dest;
        if(not -e $Dest) {
            print STDERR "\nERROR: can't access \'$Dest\'\n";
        }
        elsif(-f $Dest) {
            print STDERR "\nERROR: \'$Dest\' - not a directory\n";
        }
        elsif(-d $Dest)
        {
            $Dest = abs_path($Dest) if($Dest!~/\A(\/|\w+:[\/\\])/);
            register_directory($Dest, $LibVersion);
            if($Descriptor{$LibVersion}{"AddIncludePaths"}{$IDest}) {
                $Add_Include_Paths{$LibVersion}{$Dest} = 1;
            }
            else {
                $Include_Paths{$LibVersion}{$Dest} = 1;
            }
        }
    }
    if(keys(%{$Include_Paths{$LibVersion}})) {
        $INC_PATH_AUTODETECT{$LibVersion} = 0;
    }
    # registering directories
    foreach my $Dest (split(/\n/, $Descriptor{$LibVersion}{"Headers"}))
    {
        $Dest=~s/\A\s+|\s+\Z//g;
        next if(not $Dest or not -e $Dest);
        if($Dest!~/\A(\/|\w+:[\/\\])/) {
            $Dest = abs_path($Dest);
        }
        if(-d $Dest)
        {
            if(register_directory($Dest, $LibVersion))
            {# register all sub directories
                register_dependency($Dest, $LibVersion);
            }
        }
        elsif(-f $Dest)
        {
            my $Dir = get_dirname($Dest);
            if(not $SystemPaths{"include"}{$Dir}
            and not $LocalIncludes{$Dir})
            {
                if(register_directory($Dir, $LibVersion))
                {# register all sub directories
                    register_dependency($Dir, $LibVersion);
                }
                if(my $OutDir = get_dirname($Dir))
                {# registering the outer directory
                    if(not $SystemPaths{"include"}{$OutDir}
                    and not $LocalIncludes{$OutDir}) {
                        register_directory($OutDir, $LibVersion);
                    }
                }
            }
        }
    }
    # detecting library header includes
    my @DepPaths = values(%{$DependencyHeaders_All_FullPath{$LibVersion}});
    detect_headers_array_includes($LibVersion, \@DepPaths);
    foreach my $Dir (keys(%DefaultIncPaths))
    {# searching for bits directory
        if(-d $Dir."/bits") {
            my @BitsPaths = cmd_find($Dir."/bits","f","","");
            detect_headers_array_includes($LibVersion, \@BitsPaths);
            last;
        }
    }
    # recursive redirects
    foreach my $Path (keys(%{$Header_ErrorRedirect{$LibVersion}}))
    {
        my $Redirect = $Header_ErrorRedirect{$LibVersion}{$Path};
        if($Path eq $Redirect) {
            delete($Header_ErrorRedirect{$LibVersion}{$Path});
        }
        if(my $RecurRedirect = $Header_ErrorRedirect{$LibVersion}{$Redirect}) {
            if($Path ne $RecurRedirect) {
                $Header_ErrorRedirect{$LibVersion}{$Path} = $RecurRedirect;
            }
        }
    }
    # registering headers
    my $Position = 0;
    foreach my $Dest (split(/\n/, $Descriptor{$LibVersion}{"Headers"}))
    {
        $Dest=~s/\A\s+|\s+\Z//g;
        next if(not $Dest);
        if($Dest=~/\A(\/|\w+:[\/\\])/ and not -e $Dest) {
            print STDERR "ERROR: can't access \'$Dest\'\n";
            next;
        }
        if(is_header($Dest, 1, $LibVersion)) {
            if(my $HPath = register_header($Dest, $LibVersion))
            {
                $Target_Headers{$LibVersion}{$HPath}{"Position"} = $Position;
                $Position += 1;
            }
        }
        elsif(-d $Dest)
        {
            my @Registered = ();
            foreach my $Path (cmd_find($Dest,"f","",""))
            {
                next if(ignore_path($Path, $Dest));
                next if(not is_header($Path, 0, $LibVersion));
                if(my $HPath = register_header($Path, $LibVersion)) {
                    push(@Registered, $HPath);
                }
            }
            @Registered = sort {natural_sorting($a, $b)} @Registered;
            sort_by_word(\@Registered, $TargetLibraryShortName);
            foreach my $Path (@Registered)
            {
                $Target_Headers{$LibVersion}{$Path}{"Position"} = $Position;
                $Position += 1;
            }
        }
        else {
            print STDERR "ERROR: can't identify \'$Dest\' as a header file\n";
        }
    }
    # preparing preamble headers
    my $Preamble_Position=0;
    foreach my $Header (split(/\n/, $Descriptor{$LibVersion}{"IncludePreamble"}))
    {
        $Header=~s/\A\s+|\s+\Z//g;
        next if(not $Header);
        if($Header=~/\A(\/|\w+:[\/\\])/ and not -f $Header) {
            print STDERR "ERROR: can't access file \'$Header\'\n";
            next;
        }
        if(my $Header_Path = is_header($Header, 1, $LibVersion)) {
            $Include_Preamble{$LibVersion}{$Header_Path}{"Position"} = $Preamble_Position++;
        }
        else {
            print STDERR "ERROR: can't identify \'$Header\' as a header file\n";
        }
    }
    if($INC_PATH_AUTODETECT{$LibVersion})
    {
        my @RegPaths = (keys(%{$RegisteredHeaders{$LibVersion}}), keys(%{$Include_Preamble{$LibVersion}}));
        detect_headers_array_includes($LibVersion, \@RegPaths);
        foreach my $AbsPath (sort keys(%{$Header_Includes{$LibVersion}})) {
            detect_recursive_includes($AbsPath, $LibVersion);
        }
    }
    foreach my $Header_Name (keys(%{$HeaderName_Paths{$LibVersion}}))
    {# set relative paths (for duplicates)
        if(keys(%{$HeaderName_Paths{$LibVersion}{$Header_Name}})>=2)
        {# search for duplicates
            my $FirstPath = (keys(%{$HeaderName_Paths{$LibVersion}{$Header_Name}}))[0];
            my $Prefix = get_dirname($FirstPath);
            while($Prefix=~/\A(.+)[\/\\]+[^\/\\]+\Z/)
            {# detect a shortest distinguishing prefix
                my $NewPrefix = $1;
                my %Identity = ();
                foreach my $Path (keys(%{$HeaderName_Paths{$LibVersion}{$Header_Name}}))
                {
                    if($Path=~/\A\Q$Prefix\E[\/\\]+(.*)\Z/) {
                        $Identity{$Path} = $1;
                    }
                }
                if(keys(%Identity)==keys(%{$HeaderName_Paths{$LibVersion}{$Header_Name}}))
                {# all names are differend with current prefix
                    foreach my $Path (keys(%{$HeaderName_Paths{$LibVersion}{$Header_Name}})) {
                        $Target_Headers{$LibVersion}{$Path}{"Identity"} = $Identity{$Path};
                    }
                    last;
                }
                $Prefix = $NewPrefix; # increase prefix
            }
        }
    }
    foreach my $HeaderName (keys(%{$Include_Order{$LibVersion}}))
    {# ordering headers according to descriptor
        my $PairName=$Include_Order{$LibVersion}{$HeaderName};
        my ($Pos, $PairPos)=(-1, -1);
        my ($Path, $PairPath)=();
        foreach my $Header_Path (sort {int($Target_Headers{$LibVersion}{$a}{"Position"})<=>int($Target_Headers{$LibVersion}{$b}{"Position"})}
        keys(%{$Target_Headers{$LibVersion}}))  {
            if(get_filename($Header_Path) eq $PairName)
            {
                $PairPos = $Target_Headers{$LibVersion}{$Header_Path}{"Position"};
                $PairPath = $Header_Path;
            }
            if(get_filename($Header_Path) eq $HeaderName)
            {
                $Pos = $Target_Headers{$LibVersion}{$Header_Path}{"Position"};
                $Path = $Header_Path;
            }
        }
        if($PairPos!=-1 and $Pos!=-1
        and int($PairPos)<int($Pos))
        {
            my %Tmp = %{$Target_Headers{$LibVersion}{$Path}};
            %{$Target_Headers{$LibVersion}{$Path}} = %{$Target_Headers{$LibVersion}{$PairPath}};
            %{$Target_Headers{$LibVersion}{$PairPath}} = %Tmp;
        }
    }
    if(not keys(%{$Target_Headers{$LibVersion}}))
    {
        if($LibVersion==1) {
            $CheckObjectsOnly = 1;
        }
        else {
            exitStatus("Error", "header files are not found in the ".$Descriptor{$LibVersion}{"Version"});
        }
    }
}

sub detect_headers_array_includes($$)
{
    my ($LibVersion, $HeadersArrayRef) = @_;
    return if(not $LibVersion or not $HeadersArrayRef);
    foreach my $Path (@{$HeadersArrayRef}) {
        detect_header_includes($Path, $LibVersion);
    }
}

sub detect_real_includes($$)
{
    my ($AbsPath, $LibVersion) = @_;
    return () if(not $LibVersion or not $AbsPath or not -e $AbsPath);
    if($Cache{"detect_real_includes"}{$LibVersion}{$AbsPath}
    or keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}})) {
        return keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}});
    }
    my $Path = callPreprocessor($AbsPath, "", $LibVersion);
    return () if(not $Path);
    open(PREPROC, $Path);
    while(<PREPROC>)
    {
        if(/#\s+\d+\s+"([^"]+)"[\s\d]*\n/)
        {
            my $Include = $1;
            if($Include=~/\<(built\-in|internal|command\-line)\>|\A\./
            or $Include eq $AbsPath) {
                next;
            }
            $RecursiveIncludes{$LibVersion}{$AbsPath}{$Include} = 1;
        }
    }
    close(PREPROC);
    $Cache{"detect_real_includes"}{$LibVersion}{$AbsPath}=1;
    return keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}});
}

sub detect_header_includes($$)
{
    my ($Path, $LibVersion) = @_;
    return if(not $LibVersion or not $Path or not -e $Path);
    return if($Cache{"detect_header_includes"}{$LibVersion}{$Path});
    my $Content = readFile($Path);
    if($Content=~/#[ \t]*error[ \t]+/ and (my $Redirect = parse_redirect($Content, $Path, $LibVersion)))
    {# detecting error directive in the headers
        if(my $RedirectPath = identify_header($Redirect, $LibVersion)) {
            if($RedirectPath=~/\/usr\/include\// and $Path!~/\/usr\/include\//) {
                $RedirectPath = identify_header(get_filename($Redirect), $LibVersion);
            }
            $Header_ErrorRedirect{$LibVersion}{$Path} = $RedirectPath;
        }
    }
    my $Inc = parse_includes($Content, $Path);
    foreach my $Include (keys(%{$Inc}))
    {# detecting includes
        $Header_Includes{$LibVersion}{$Path}{$Include} = $Inc->{$Include};
        if(my $Prefix = get_dirname($Include)) {
            $Header_Prefix{$LibVersion}{get_filename($Include)} = $Prefix;
        }
    }
    $Cache{"detect_header_includes"}{$LibVersion}{$Path} = 1;
}

sub detect_recursive_includes($)
{
    my ($AbsPath, $LibVersion) = @_;
    return () if(not $AbsPath);
    if(isCyclical(\@RecurInclude, $AbsPath)) {
        return keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}});
    }
    return () if($SystemPaths{"include"}{get_dirname($AbsPath)} and $GlibcHeader{get_filename($AbsPath)});
    return () if($SystemPaths{"include"}{get_dirname(get_dirname($AbsPath))}
    and (get_dirname($AbsPath)=~/[\/\\](asm-.+)\Z/ or $GlibcDir{get_filename(get_dirname($AbsPath))}));
    if(keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}})) {
        return keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}});
    }
    return () if($OSgroup ne "windows" and get_filename($AbsPath)=~/windows|win32|win64/i);
    return () if(get_filename($AbsPath)=~/atomic/i);
    return () if(skip_header($AbsPath, $LibVersion));
    return () if($MAIN_CPP_DIR and $AbsPath=~/\A\Q$MAIN_CPP_DIR\E/ and not $STDCXX_TESTING);
    if($DefaultGccPaths{get_dirname($AbsPath)}) {
        return detect_real_includes($AbsPath, $LibVersion);
    }
    push(@RecurInclude, $AbsPath);
    if(not keys(%{$Header_Includes{$LibVersion}{$AbsPath}})) {
        detect_header_includes($AbsPath, $LibVersion);
    }
    foreach my $Include (keys(%{$Header_Includes{$LibVersion}{$AbsPath}}))
    {
        my $HPath = "";
        if($Header_Includes{$LibVersion}{$AbsPath}{$Include}==-1
        and -f get_dirname($AbsPath)."/".$Include) {
            $HPath = get_dirname($AbsPath)."/".$Include;
        }
        if(not $HPath) {
            $HPath = identify_header($Include, $LibVersion);
        }
        next if(not $HPath);
        $RecursiveIncludes{$LibVersion}{$AbsPath}{$HPath} = 1;
        if($Header_Includes{$LibVersion}{$AbsPath}{$Include}==1)
        { # only include < ... >, skip include " ... " prefixes
            $Header_Include_Prefix{$LibVersion}{$AbsPath}{$HPath}{get_dirname($Include)} = 1;
        }
        foreach my $IncPath (detect_recursive_includes($HPath, $LibVersion))
        {
            $RecursiveIncludes{$LibVersion}{$AbsPath}{$IncPath} = 1;
            foreach my $Prefix (keys(%{$Header_Include_Prefix{$LibVersion}{$HPath}{$IncPath}})) {
                $Header_Include_Prefix{$LibVersion}{$AbsPath}{$IncPath}{$Prefix} = 1;
            }
        }
        foreach my $Dep (keys(%{$Header_Include_Prefix{$LibVersion}{$AbsPath}}))
        {
            if($GlibcHeader{get_filename($Dep)} and keys(%{$Header_Include_Prefix{$LibVersion}{$AbsPath}{$Dep}})>=2
            and defined $Header_Include_Prefix{$LibVersion}{$AbsPath}{$Dep}{""})
            {# distinguish math.h from glibc and math.h from the tested library
                delete($Header_Include_Prefix{$LibVersion}{$AbsPath}{$Dep}{""});
                last;
            }
        }
    }
    pop(@RecurInclude);
    return keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}});
}

sub find_in_dependencies($$)
{
    my ($Header, $LibVersion) = @_;
    return "" if(not $Header or not $LibVersion);
    return $Cache{"find_in_dependencies"}{$LibVersion}{$Header} if(defined $Cache{"find_in_dependencies"}{$LibVersion}{$Header});
    foreach my $Dependency (sort {get_depth($a)<=>get_depth($b)} keys(%{$Header_Dependency{$LibVersion}}))
    {
        next if(not $Dependency);
        next if(skip_header($Dependency."/".$Header, $LibVersion));
        if(-e $Dependency."/".$Header)
        {
            $Cache{"find_in_dependencies"}{$LibVersion}{$Header} = $Dependency;
            return $Cache{"find_in_dependencies"}{$LibVersion}{$Header};
        }
        if($OSgroup eq "macos" and $Header=~/([^\/]+\.framework)\/Headers\/([^\/]+)/)
        {# searching in frameworks
            my ($HFramework, $HName) = ($1, $2);
            if(get_filename($Dependency) eq $HFramework
            and -e get_dirname($Dependency)."/".$Header)
            {
                $Cache{"find_in_dependencies"}{$LibVersion}{$Header} = get_dirname($Dependency);
                return $Cache{"find_in_dependencies"}{$LibVersion}{$Header};
            }
        }
    }
    return "";
}

sub find_in_defaults($)
{
    my $Header = $_[0];
    return "" if(not $Header);
    foreach my $DefaultPath (sort {get_depth($a)<=>get_depth($b)}
    (keys(%DefaultIncPaths),keys(%DefaultGccPaths),keys(%DefaultCppPaths),keys(%UserIncPath)))
    {
        next if(not $DefaultPath);
        if(-f $DefaultPath."/".$Header)
        {
            return $DefaultPath;
        }
    }
    return "";
}

sub cmp_paths($$)
{
    my ($Path1, $Path2) = @_;
    my @Parts1 = split(/[\/\\]/, $Path1);
    my @Parts2 = split(/[\/\\]/, $Path2);
    foreach my $Num (0 .. $#Parts1)
    {
        my $Part1 = $Parts1[$Num];
        my $Part2 = $Parts2[$Num];
        if($GlibcDir{$Part1}
        and not $GlibcDir{$Part2}) {
            return 1;
        }
        elsif($GlibcDir{$Part2}
        and not $GlibcDir{$Part1}) {
            return -1;
        }
        elsif($Part1=~/glib/
        and $Part2!~/glib/) {
            return 1;
        }
        elsif($Part1!~/glib/
        and $Part2=~/glib/) {
            return -1;
        }
        elsif(my $CmpRes = ($Part1 cmp $Part2)) {
            return $CmpRes;
        }
    }
    return 0;
}

sub checkRelevance($)
{
    my ($Path) = @_;
    return 0 if(not $Path);
    if($SystemRoot) {
        $Path=~s/\A\Q$SystemRoot\E//g;
    }
    my ($Dir, $Name) = separate_path($Path);
    $Name=~s/\.\w+\Z//g;# remove extension (.h)
    my @Tokens = split(/[_\d\W]+/, $Name);
    foreach (@Tokens)
    {
        next if(not $_);
        if($Dir=~/(\A|lib|[_\d\W])\Q$_\E([_\d\W]|lib|\Z)/i
        or length($_)>=4 and $Dir=~/\Q$_\E/i)
        { # include/gupnp-1.0/libgupnp/gupnp-context.h
          # include/evolution-data-server-1.4/libebook/e-book.h
            return 1;
        }
    }
    return 0;
}

sub selectSystemHeader($$)
{
    my ($Header, $LibVersion) = @_;
    return $Header if(-f $Header);
    return "" if($Header=~/\A(\/|\w+:[\/\\])/ and not -f $Header);
    return "" if($Header=~/\A(atomic|config|configure|build|conf|setup)\.h\Z/i);
    return "" if($OSgroup ne "windows" and get_filename($Header)=~/windows|win32|win64|\A(dos|process|winsock)\.h\Z/i);
    my $Arch = getArch($LibVersion);
    return $Cache{"selectSystemHeader"}{$Arch}{$Header} if(defined $Cache{"selectSystemHeader"}{$Arch}{$Header});
    if($Header!~/[\/\\]/)
    {
        foreach my $Path (keys(%{$SystemPaths{"include"}}))
        {# search in default paths
            if(-f $Path."/".$Header)
            {
                $Cache{"selectSystemHeader"}{$Arch}{$Header} = joinPath($Path,$Header);
                return $Cache{"selectSystemHeader"}{$Arch}{$Header};
            }
        }
    }
    detectSystemHeaders() if(not keys(%SystemHeaders));
    my $HName = get_filename($Header);
    my $SystemDepth = $SystemRoot?get_depth($SystemRoot):0;
    foreach my $Path (sort {get_depth($a)<=>get_depth($b)} sort {cmp_paths($b, $a)}
    keys(%{$SystemHeaders{$HName}}))
    {
        if($Path=~/[\/\\]\Q$Header\E\Z/)
        {
            if(keys(%{$SystemHeaders{$HName}})>=2 # multiple candidates
            and $Header!~/[\/\\]/ # name without prefix
            and (get_depth($Path)-$SystemDepth)>=5 # too deep directories
            and $Header!~/config/ and $Path!~/\/lib\d*\// # not a config file
            and not checkRelevance($Path))
            { # Disallow to search for "abstract" headers in too deep directories
              # Disallow to search for sstream.h or typeinfo.h in /usr/include/wx-2.9/wx/
              # Allow to search for glibconfig.h in /usr/lib/glib-2.0/include/
              # Allow to search for atk.h in /usr/include/atk-1.0/atk/
                next;
            }
            if($Arch ne "unknown" and $Path=~/[\/\\]asm-/
            and $Path!~/[\/\\]asm-\Q$Arch\E/) {
                # skip ../asm-arm/ if using x86 architecture
                next;
            }
            if($Path=~/c\+\+[\/\\]\d+/ and $MAIN_CPP_DIR
            and $Path!~/\A\Q$MAIN_CPP_DIR\E/) {
                # skip ../c++/3.3.3/ if using ../c++/4.4.0/
                next;
            }
            if($HName eq "parser.h" and $Path!~/\/libxml2\//) {
                # select parser.h from xml2 library
                next;
            }
            return ($Cache{"selectSystemHeader"}{$Arch}{$Header} = $Path);
        }
    }
    return ($Cache{"selectSystemHeader"}{$Arch}{$Header} = ""); # error
}

sub cut_path_prefix($$)
{
    my ($Path, $Prefix) = @_;
    return $Path if(not $Prefix);
    $Prefix=~s/[\/\\]+\Z//;
    $Path=~s/\A\Q$Prefix\E([\/\\]+|\Z)//;
    return $Path;
}

sub is_default_include_dir($)
{
    my $Dir = $_[0];
    $Dir=~s/[\/\\]+\Z//;
    return ($DefaultGccPaths{$Dir} or $DefaultCppPaths{$Dir} or $DefaultIncPaths{$Dir});
}

sub identify_header($$)
{
    my ($Header, $LibVersion) = @_;
    return $Cache{"identify_header"}{$Header}{$LibVersion} if(defined $Cache{"identify_header"}{$Header}{$LibVersion});
    my $Path = identify_header_m($Header, $LibVersion);
    if(not $Path and $OSgroup eq "macos" and $Header=~/\A([^\/]+)\/([^\/]+)/)
    {# search in frameworks: "OpenGL/gl.h" is "OpenGL.framework/Headers/gl.h"
        my ($HFramework, $HName) = ($1, $2);
        my $RelPath = "$HFramework\.framework\/Headers\/$HName";
        if(my $HeaderDir = find_in_dependencies($RelPath, $LibVersion)) {
            $Path = joinPath($HeaderDir, $RelPath);
        }
    }
    $Cache{"identify_header"}{$Header}{$LibVersion} = $Path;
    return $Cache{"identify_header"}{$Header}{$LibVersion};
}

sub identify_header_m($$)
{# search for header by absolute path, relative path or name
    my ($Header, $LibVersion) = @_;
    $Header=~s/[\/\\]/$SLASH/g if($OSgroup eq "windows");
    if(not $Header
    or $Header=~/\.tcc\Z/) {
        return "";
    }
    elsif(-f $Header) {
        if($Header!~/\A(\/|\w+:[\/\\])/) {
            $Header = abs_path($Header);
        }
        return $Header;
    }
    elsif($GlibcHeader{$Header} and not $GLIBC_TESTING
    and my $HeaderDir = find_in_defaults($Header)) {
        return joinPath($HeaderDir,$Header);
    }
    elsif(my $HeaderDir = find_in_dependencies($Header, $LibVersion)) {
        return joinPath($HeaderDir,$Header);
    }
    elsif(my $Path = $DependencyHeaders_All_FullPath{$LibVersion}{$Header}) {
        return $Path;
    }
    elsif($DefaultGccHeader{$Header}) {
        return $DefaultGccHeader{$Header};
    }
    elsif(my $Path = search_in_public_dirs($Header)) {
        return $Path;
    }
    elsif(my $HeaderDir = find_in_defaults($Header)) {
        return joinPath($HeaderDir,$Header);
    }
    elsif($DefaultCppHeader{$Header}) {
        return $DefaultCppHeader{$Header};
    }
    elsif($Header_Prefix{$LibVersion}{$Header}
    and my $AnyPath = selectSystemHeader(joinPath($Header_Prefix{$LibVersion}{$Header},$Header), $LibVersion))
    {# gfile.h in glib-2.0/gio/ and poppler/goo/
        return $AnyPath;
    }
    elsif(my $AnyPath = selectSystemHeader($Header, $LibVersion)) {
        return $AnyPath;
    }
    else {
        return "";
    }
}

sub search_in_public_dirs($)
{
    my $Header = $_[0];
    return "" if(not $Header);
    my @DefaultDirs = ("sys", "netinet");
    foreach my $Dir (@DefaultDirs)
    {
        if(my $HeaderDir = find_in_defaults(joinPath($Dir,$Header))) {
            return joinPath(joinPath($HeaderDir,$Dir),$Header);
        }
    }
    return "";
}

sub get_filename($)
{# much faster than basename() from File::Basename module
    my $Path = $_[0];
    $Path=~s/[\/\\]+\Z//;
    return $Cache{"get_filename"}{$Path} if($Cache{"get_filename"}{$Path});
    if($Path=~/([^\/\\]+)\Z/) {
        return ($Cache{"get_filename"}{$Path} = $1);
    }
    return "";
}

sub get_dirname($)
{# much faster than dirname() from File::Basename module
    my $Path = $_[0];
    $Path=~s/[\/\\]+\Z//;
    if($Path=~/\A(.*)[\/\\]+([^\/\\]*)\Z/) {
        return $1;
    }
    return "";
}

sub separate_path($) {
    return (get_dirname($_[0]), get_filename($_[0]));
}

sub esc($)
{
    my $Str = $_[0];
    $Str=~s/([()\[\]{}$ &'"`;,<>\+])/\\$1/g;
    return $Str;
}

sub getLocation($)
{
    if($LibInfo{$Version}{"info"}{$_[0]}=~/srcp[ ]*:[ ]*([\w\-\<\>\.\+\/\\]+):(\d+) /) {
        return ($1, $2);
    }
    else {
        return ();
    }
}

sub getTypeType($$)
{
    my ($TypeDeclId, $TypeId) = @_;
    if($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*c /
    and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@/) {
        return "Const";
    }
    return "Typedef" if($LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:/ and $LibInfo{$Version}{"info"}{$TypeId}!~/qual[ ]*:/ and getNameByInfo($TypeDeclId));
    return "Volatile" if($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*v / and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@/);
    return "Restrict" if($LibInfo{$Version}{"info"}{$TypeId}=~/qual[ ]*:[ ]*r / and $LibInfo{$Version}{"info"}{$TypeId}=~/unql[ ]*:[ ]*\@/);
    my $TypeType = getTypeTypeByTypeId($TypeId);
    if($TypeType eq "Struct")
    {
        if($TypeDeclId
        and $LibInfo{$Version}{"info_type"}{$TypeDeclId} eq "template_decl") {
            return "Template";
        }
        else {
            return "Struct";
        }
    }
    else {
        return $TypeType;
    }
    
}

sub getTypeTypeByTypeId($)
{
    my $TypeId = $_[0];
    my $TypeType = $LibInfo{$Version}{"info_type"}{$TypeId};
    if($TypeType=~/integer_type|real_type|boolean_type|void_type|complex_type/)
    {
        return "Intrinsic";
    }
    elsif(isFuncPtr($TypeId)) {
        return "FuncPtr";
    }
    elsif(isMethodPtr($TypeId)) {
        return "MethodPtr";
    }
    elsif(isFieldPtr($TypeId)) {
        return "FieldPtr";
    }
    elsif($TypeType eq "pointer_type") {
        return "Pointer";
    }
    elsif($TypeType eq "reference_type") {
        return "Ref";
    }
    elsif($TypeType eq "union_type") {
        return "Union";
    }
    elsif($TypeType eq "enumeral_type") {
        return "Enum";
    }
    elsif($TypeType eq "record_type") {
        return "Struct";
    }
    elsif($TypeType eq "array_type") {
        return "Array";
    }
    elsif($TypeType eq "complex_type") {
        return "Intrinsic";
    }
    elsif($TypeType eq "function_type") {
        return "FunctionType";
    }
    elsif($TypeType eq "method_type") {
        return "MethodType";
    }
    else {
        return "Unknown";
    }
}

sub getNameByInfo($)
{
    return "" if($LibInfo{$Version}{"info"}{$_[0]}!~/name[ ]*:[ ]*@(\d+) /);
    if($LibInfo{$Version}{"info"}{$1}=~/strg[ ]*:[ ]*(.*?)[ ]+lngt/)
    {# short unsigned int (may include spaces)
        return $1;
    }
    else {
        return "";
    }
}

sub getTreeStr($)
{
    my $Info = $LibInfo{$Version}{"info"}{$_[0]};
    if($Info=~/strg[ ]*:[ ]*([^ ]*)/)
    {
        my $Str = $1;
        if($C99Mode and $Str=~/\Ac99_(.+)\Z/) {
            if($CppKeywords{$1}) {
                $Str=$1;
            }
        }
        return $Str;
    }
    else {
        return "";
    }
}

sub getVarShortName($)
{
    my $VarInfo = $LibInfo{$Version}{"info"}{$_[0]};
    return "" if($VarInfo!~/name[ ]*:[ ]*@(\d+) /);
    return getTreeStr($1);
}

sub getFuncShortName($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    if($FuncInfo=~/ operator /)
    {
        if($FuncInfo=~/ conversion /) {
            return "operator ".get_TypeName($SymbolInfo{$Version}{$_[0]}{"Return"}, $Version);
        }
        else
        {
            return "" if($FuncInfo!~/ operator[ ]+([a-zA-Z]+) /);
            return "operator".$Operator_Indication{$1};
        }
    }
    else
    {
        return "" if($FuncInfo!~/name[ ]*:[ ]*@(\d+) /);
        return getTreeStr($1);
    }
}

sub getFuncMnglName($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    return "" if($FuncInfo!~/mngl[ ]*:[ ]*@(\d+) /);
    return getTreeStr($1);
}

sub getFuncReturn($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    return "" if($FuncInfo!~/type[ ]*:[ ]*@(\d+) /);
    my $FuncTypeInfoId = $1;
    return "" if($LibInfo{$Version}{"info"}{$FuncTypeInfoId}!~/retn[ ]*:[ ]*@(\d+) /);
    my $FuncReturnTypeId = $1;
    if($TypeInfo{$Version}{getTypeDeclId($FuncReturnTypeId)}{$FuncReturnTypeId}{"Type"} eq "Restrict")
    {# delete restrict spec
        $FuncReturnTypeId = getRestrictBase($FuncReturnTypeId);
    }
    return $FuncReturnTypeId;
}

sub getFuncOrig($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    if($FuncInfo=~/orig[ ]*:[ ]*@(\d+) /) {
        return $1;
    }
    else {
        return $_[0];
    }
}

sub unmangleArray(@)
{
    if(@_[0]=~/\A\?/)
    {# "cl" mangling
        my $UndNameCmd = get_CmdPath("undname");
        if(not $UndNameCmd) {
            exitStatus("Not_Found", "can't find \"undname\"");
        }
        writeFile("$TMP_DIR/unmangle", join("\n", @_));
        return split(/\n/, `$UndNameCmd 0x8386 $TMP_DIR/unmangle`);
    }
    else
    {# "gcc" mangling
        my $CppFiltInfo = `$CPP_FILT -h 2>&1`;
        if($CppFiltInfo=~/\@<file>/)
        {# new version of c++filt can take a file
            my $Und = ($OSgroup eq "macos" or $OSgroup eq "windows")?"_":"";
            writeFile("$TMP_DIR/unmangle", $Und.join("\n".$Und, @_));
            return split(/\n/, `$CPP_FILT \@\"$TMP_DIR/unmangle\"`);
        }
        else {
            return unmangleArray_Old(@_);
        }
    }
}

sub unmangleArray_Old(@)
{
    if($#_>$MAX_COMMAND_LINE_ARGUMENTS) {
        my @Half = splice(@_, 0, ($#_+1)/2);
        return (unmangleArray(@Half), unmangleArray(@_))
    }
    else {
        my $Und = ($OSgroup eq "macos" or $OSgroup eq "windows")?"_":"";
        my $UnmangleCommand = $CPP_FILT." $Und".join(" $Und", @_);
        return split(/\n/, `$UnmangleCommand`);
    }
}

sub get_SignatureNoInfo($$)
{
    my ($Interface, $LibVersion) = @_;
    return $Cache{"get_SignatureNoInfo"}{$LibVersion}{$Interface} if($Cache{"get_SignatureNoInfo"}{$LibVersion}{$Interface});
    my ($MnglName, $VersionSpec, $SymbolVersion) = get_symbol_name_version($Interface);
    my $Signature = $tr_name{$MnglName}?$tr_name{$MnglName}:$MnglName;
    if($Interface=~/\A(_Z|\?)/)
    {# C++
        $Signature=~s/\Qstd::basic_string<char, std::char_traits<char>, std::allocator<char> >\E/std::string/g;
        $Signature=~s/\Qstd::map<std::string, std::string, std::less<std::string >, std::allocator<std::pair<std::string const, std::string > > >\E/std::map<std::string, std::string>/g;
    }
    if(not $CheckObjectsOnly or $OSgroup=~/linux|bsd|beos/)
    {# ELF format marks data as OBJECT
        if($CompleteSignature{$LibVersion}{$Interface}{"Object"}) {
            $Signature .= " [data]";
        }
        elsif($Interface!~/\A(_Z|\?)/) {
            $Signature .= "(...)";
        }
    }
    if(my $ChargeLevel = get_ChargeLevel($Interface, $LibVersion))
    {
        my $ShortName = substr($Signature, 0, get_signature_center($Signature));
        $Signature=~s/\A\Q$ShortName\E/$ShortName $ChargeLevel/g;
    }
    $Signature .= $VersionSpec.$SymbolVersion if($SymbolVersion);
    $Cache{"get_SignatureNoInfo"}{$LibVersion}{$Interface} = $Signature;
    return $Signature;
}

sub get_ChargeLevel($$)
{
    my ($Interface, $LibVersion) = @_;
    return "" if($Interface!~/\A(_Z|\?)/);
    if(defined $CompleteSignature{$LibVersion}{$Interface}
    and $CompleteSignature{$LibVersion}{$Interface}{"Header"})
    {
        if($CompleteSignature{$LibVersion}{$Interface}{"Constructor"})
        {
            if($Interface=~/C1E/) {
                return "[in-charge]";
            }
            elsif($Interface=~/C2E/) {
                return "[not-in-charge]";
            }
        }
        elsif($CompleteSignature{$LibVersion}{$Interface}{"Destructor"})
        {
            if($Interface=~/D1E/) {
                return "[in-charge]";
            }
            elsif($Interface=~/D2E/) {
                return "[not-in-charge]";
            }
            elsif($Interface=~/D0E/) {
                return "[in-charge-deleting]";
            }
        }
    }
    else
    {
        if($Interface=~/C1E/) {
            return "[in-charge]";
        }
        elsif($Interface=~/C2E/) {
            return "[not-in-charge]";
        }
        elsif($Interface=~/D1E/) {
            return "[in-charge]";
        }
        elsif($Interface=~/D2E/) {
            return "[not-in-charge]";
        }
        elsif($Interface=~/D0E/) {
            return "[in-charge-deleting]";
        }
    }
    return "";
}

sub get_Signature($$)
{
    my ($Interface, $LibVersion) = @_;
    return $Cache{"get_Signature"}{$LibVersion}{$Interface} if($Cache{"get_Signature"}{$LibVersion}{$Interface});
    my ($MnglName, $VersionSpec, $SymbolVersion) = get_symbol_name_version($Interface);
    if(skipGlobalData($MnglName) or not $CompleteSignature{$LibVersion}{$Interface}{"Header"}) {
        return get_SignatureNoInfo($Interface, $LibVersion);
    }
    my ($Func_Signature, @Param_Types_FromUnmangledName) = ();
    my $ShortName = $CompleteSignature{$LibVersion}{$Interface}{"ShortName"};
    if($Interface=~/\A(_Z|\?)/)
    {
        if(my $ClassId = $CompleteSignature{$LibVersion}{$Interface}{"Class"}) {
            $Func_Signature = get_TypeName($ClassId, $LibVersion)."::".(($CompleteSignature{$LibVersion}{$Interface}{"Destructor"})?"~":"").$ShortName;
        }
        elsif(my $NameSpace = $CompleteSignature{$LibVersion}{$Interface}{"NameSpace"}) {
            $Func_Signature = $NameSpace."::".$ShortName;
        }
        else {
            $Func_Signature = $ShortName;
        }
        @Param_Types_FromUnmangledName = get_signature_parts($tr_name{$MnglName}, 0);
    }
    else {
        $Func_Signature = $MnglName;
    }
    my @ParamArray = ();
    foreach my $Pos (sort {int($a) <=> int($b)} keys(%{$CompleteSignature{$LibVersion}{$Interface}{"Param"}}))
    {
        next if($Pos eq "");
        my $ParamTypeId = $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$Pos}{"type"};
        next if(not $ParamTypeId);
        my $ParamTypeName = get_TypeName($ParamTypeId, $LibVersion);
        $ParamTypeName = $Param_Types_FromUnmangledName[$Pos] if(not $ParamTypeName);
        foreach my $Typedef (keys(%ChangedTypedef)) {
            my $Base = $Typedef_BaseName{$LibVersion}{$Typedef};
            $ParamTypeName=~s/(\A|\W)\Q$Typedef\E(\W|\Z)/$1$Base$2/g;
        }
        if(my $ParamName = $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$Pos}{"name"}) {
            push(@ParamArray, create_member_decl($ParamTypeName, $ParamName));
        }
        else {
            push(@ParamArray, $ParamTypeName);
        }
    }
    if($CompleteSignature{$LibVersion}{$Interface}{"Data"}
    or $CompleteSignature{$LibVersion}{$Interface}{"Object"}) {
        $Func_Signature .= " [data]";
    }
    else
    {
        if(my $ChargeLevel = get_ChargeLevel($Interface, $LibVersion))
        {# add [in-charge]
            $Func_Signature .= " ".$ChargeLevel;
        }
        $Func_Signature .= " (".join(", ", @ParamArray).")";
        if($Interface=~/\A_ZNK/) {
            $Func_Signature .= " const";
        }
        if($CompleteSignature{$LibVersion}{$Interface}{"Static"}
        and $Interface=~/\A(_Z|\?)/)
        {# for static methods
            $Func_Signature .= " [static]";
        }
    }
    if(defined $ShowRetVal
    and my $ReturnTId = $CompleteSignature{$LibVersion}{$Interface}{"Return"}) {
        $Func_Signature .= ":".get_TypeName($ReturnTId, $LibVersion);
    }
    if($SymbolVersion) {
        $Func_Signature .= $VersionSpec.$SymbolVersion;
    }
    $Cache{"get_Signature"}{$LibVersion}{$Interface} = $Func_Signature;
    return $Func_Signature;
}

sub create_member_decl($$)
{
    my ($TName, $Member) = @_;
    if($TName=~/\([\*]+\)/) {
        $TName=~s/\(([\*]+)\)/\($1$Member\)/;
        return $TName;
    }
    else
    {
        my @ArraySizes = ();
        while($TName=~s/(\[[^\[\]]*\])\Z//) {
            push(@ArraySizes, $1);
        }
        return $TName." ".$Member.join("", @ArraySizes);
    }
}

sub getFuncType($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    return "" if($FuncInfo!~/type[ ]*:[ ]*@(\d+) /);
    my $FuncTypeInfoId = $1;
    my $FunctionType = $LibInfo{$Version}{"info_type"}{$FuncTypeInfoId};
    if($FunctionType eq "method_type")
    {
        return "Method";
    }
    elsif($FunctionType eq "function_type")
    {
        return "Function";
    }
    else
    {
        return $FunctionType;
    }
}

sub getFuncTypeId($)
{
    my $FuncInfo = $LibInfo{$Version}{"info"}{$_[0]};
    if($FuncInfo=~/type[ ]*:[ ]*@(\d+)( |\Z)/)
    {
        return $1;
    }
    else
    {
        return 0;
    }
}

sub isNotAnon($)
{
    return (not isAnon($_[0]));
}

sub isAnon($)
{# "._N" or "$_N" in older GCC versions
    return ($_[0]=~/(\.|\$)\_\d+|anon\-/);
}

sub unmangled_Compact($)
{ # Removes all non-essential (for C++ language) whitespace from a string.  If 
  # the whitespace is essential it will be replaced with exactly one ' ' 
  # character. Works correctly only for unmangled names.
    return $Cache{"unmangled_Compact"}{$_[0]} if(defined $Cache{"unmangled_Compact"}{$_[0]});
    my $result=$_[0];
    # First, we reduce all spaces that we can
    my $coms='[-()<>:*&~!|+=%@~"?.,/[^'."']";
    my $coms_nobr='[-()<:*&~!|+=%@~"?.,'."']";
    my $clos='[),;:\]]';
    $result=~s/^\s+//gm;
    $result=~s/\s+$//gm;
    $result=~s/((?!\n)\s)+/ /g;
    $result=~s/(\w+)\s+($coms+)/$1$2/gm;
    $result=~s/($coms+)\s+(\w+)/$1$2/gm;
    $result=~s/(\w)\s+($clos)/$1$2/gm;
    $result=~s/($coms+)\s+($coms+)/$1 $2/gm;
    $result=~s/($coms_nobr+)\s+($coms+)/$1$2/gm;
    $result=~s/($coms+)\s+($coms_nobr+)/$1$2/gm;
    # don't forget about >> and <:.  In unmangled names global-scope modifier 
    # is not used, so <: will always be a digraph and requires no special treatment.
    # We also try to remove other parts that are better to be removed here than in other places
    # double-cv
    $result=~s/\bconst\s+const\b/const/gm;
    $result=~s/\bvolatile\s+volatile\b/volatile/gm;
    $result=~s/\bconst\s+volatile\b\s+const\b/const volatile/gm;
    $result=~s/\bvolatile\s+const\b\s+volatile\b/const volatile/gm;
    # Place cv in proper order
    $result=~s/\bvolatile\s+const\b/const volatile/gm;
    $Cache{"unmangled_Compact"}{$_[0]} = $result;
    return $result;
}

sub unmangled_PostProcess($)
{
    my $result = $_[0];
    #s/\bunsigned int\b/unsigned/g;
    $result=~s/\bshort unsigned int\b/unsigned short/g;
    $result=~s/\bshort int\b/short/g;
    $result=~s/\blong long unsigned int\b/unsigned long long/g;
    $result=~s/\blong unsigned int\b/unsigned long/g;
    $result=~s/\blong long int\b/long long/g;
    $result=~s/\blong int\b/long/g;
    $result=~s/\)const\b/\) const/g;
    $result=~s/\blong long unsigned\b/unsigned long long/g;
    $result=~s/\blong unsigned\b/unsigned long/g;
    return $result;
}

sub correctName($)
{# type name correction
    my $CorrectName = $_[0];
    $CorrectName = unmangled_Compact($CorrectName);
    $CorrectName = unmangled_PostProcess($CorrectName);
    return $CorrectName;
}

sub get_HeaderDeps($$)
{
    my ($AbsPath, $LibVersion) = @_;
    return () if(not $AbsPath or not $LibVersion);
    return @{$Cache{"get_HeaderDeps"}{$LibVersion}{$AbsPath}} if(defined $Cache{"get_HeaderDeps"}{$LibVersion}{$AbsPath});
    my %IncDir = ();
    foreach my $HeaderPath (keys(%{$RecursiveIncludes{$LibVersion}{$AbsPath}}))
    {
        next if(not $HeaderPath);
        next if($MAIN_CPP_DIR and $HeaderPath=~/\A\Q$MAIN_CPP_DIR\E([\/\\]|\Z)/);
        my $Dir = get_dirname($HeaderPath);
        foreach my $Prefix (keys(%{$Header_Include_Prefix{$LibVersion}{$AbsPath}{$HeaderPath}}))
        {
            my $Dep = $Dir;
            if($Prefix)
            {
                if(not $Dep=~s/[\/\\]+\Q$Prefix\E\Z//g)
                {
                    if($OSgroup eq "macos" and $HeaderPath=~/(.+\.framework)\/Headers\/([^\/]+)/)
                    {# frameworks
                        my ($HFramework, $HName) = ($1, $2);
                        $Dep = $HFramework;
                    }
                    else
                    {# mismatch
                        next;
                    }
                }
            }
            next if(not $Dep or is_default_include_dir($Dep) or get_depth($Dep)==1);
            next if($DefaultIncPaths{get_dirname($Dep)} and $GlibcDir{get_filename($Dep)});
            $IncDir{$Dep}=1;
        }
    }
    $Cache{"get_HeaderDeps"}{$LibVersion}{$AbsPath} = sortIncPaths([keys(%IncDir)], $LibVersion);
    return @{$Cache{"get_HeaderDeps"}{$LibVersion}{$AbsPath}};
}

sub sortIncPaths($$)
{
    my ($ArrRef, $LibVersion) = @_;
    @{$ArrRef} =  sort {$Header_Dependency{$LibVersion}{$b}<=>$Header_Dependency{$LibVersion}{$a}}
      sort {get_depth($a)<=>get_depth($b)} sort {$b cmp $a} @{$ArrRef};
    return $ArrRef;
}

sub joinPath($$)
{
    return join($SLASH, @_);
}

sub get_namespace_additions($)
{
    my $NameSpaces = $_[0];
    my ($Additions, $AddNameSpaceId) = ("", 1);
    foreach my $NS (sort {$a=~/_/ <=> $b=~/_/} sort {lc($a) cmp lc($b)} keys(%{$NameSpaces}))
    {
        next if(not $NS or $NameSpaces->{$NS}==-1);
        next if($NS=~/(\A|::)iterator(::|\Z)/i);
        next if($NS=~/\A__/i);
        next if(($NS=~/\Astd::/ or $NS=~/\A(std|tr1|rel_ops|fcntl)\Z/) and not $STDCXX_TESTING);
        $NestedNameSpaces{$Version}{$NS} = 1;# for future use in reports
        my ($TypeDecl_Prefix, $TypeDecl_Suffix) = ();
        my @NS_Parts = split(/::/, $NS);
        next if($#NS_Parts==-1);
        next if($NS_Parts[0]=~/\A(random|or)\Z/);
        foreach my $NS_Part (@NS_Parts)
        {
            $TypeDecl_Prefix .= "namespace $NS_Part\{";
            $TypeDecl_Suffix .= "}";
        }
        my $TypeDecl = $TypeDecl_Prefix."typedef int tmp_add_type_$AddNameSpaceId;".$TypeDecl_Suffix;
        my $FuncDecl = "$NS\:\:tmp_add_type_$AddNameSpaceId tmp_add_func_$AddNameSpaceId(){return 0;};";
        $Additions.="  $TypeDecl\n  $FuncDecl\n";
        $AddNameSpaceId+=1;
    }
    return $Additions;
}

sub forward_slash($)
{# forward slash to pass into MinGW GCC
    my $Path = $_[0];
    if($OSgroup eq "windows") {
        $Path=~s/\\/\//g;
    }
    return $Path;
}

sub inc_opt($$)
{
    my ($Path, $Style) = @_;
    if($Style eq "GCC")
    {# GCC options
        if($OSgroup eq "windows")
        {# to MinGW GCC
            return "-I\"".forward_slash($Path)."\"";
        }
        elsif($OSgroup eq "macos"
        and $Path=~/\.framework\Z/)
        {# to Apple's GCC
            return "-F".esc(get_dirname($Path));
        }
        else {
            return "-I".esc($Path);
        }
    }
    elsif($Style eq "CL") {
        return "/I \"$Path\"";
    }
    return "";
}

sub getPlatformOpts($)
{
    my $LibVersion = $_[0];
    my $Arch = getArch($LibVersion);
    my $Options = "";    
    if($OSgroup eq "windows")
    {# emulate the CL compiler
        if($Arch eq "x86") {
            $Options .= " -D_M_IX86=300";
        }
        elsif($Arch eq "x86_64") {
            $Options .= " -D_M_AMD64=300";
        }
        elsif($Arch eq "ia64") {
            $Options .= " -D_M_IA64=300";
        }
        $Options .= " -D_STDCALL_SUPPORTED";
        $Options .= " -D__int64=\"long long\"";
        $Options .= " -D__int32=long";
        $Options .= " -D__int16=short";
        $Options .= " -D__int8=char";
        $Options .= " -D__possibly_notnullterminated=\" \"";
        $Options .= " -D__nullterminated=\" \"";
        $Options .= " -D__nullnullterminated=\" \"";
        $Options .= " -D__w64=\" \" -D__ptr32=\" \" -D__ptr64=\" \"";
        $Options .= " -D__forceinline=inline -D__inline=inline";
        $Options .= " -D_MSC_VER=1500";
    }
    return $Options;
}

my %C_Structure = (
# FIXME: Can't separate union and struct data types before dumping,
# so it sometimes cause compilation errors for unknown reason
# when trying to declare TYPE* tmp_add_class_N
# This is a list of such structures + list of other C structures
    "sigval"=>1,
    "sigevent"=>1,
    "sigaction"=>1,
    "sigvec"=>1,
    "sigstack"=>1,
    "timeval"=>1,
    "timezone"=>1,
    "rusage"=>1,
    "rlimit"=>1,
    "wait"=>1,
    "flock"=>1,
    "stat"=>1,
    "stat64"=>1,
    "if_nameindex"=>1,
    "usb_device"=>1,
    "sigaltstack"=>1,
    "sysinfo"=>1,
# Other C structures appearing in every dump
    "timespec"=>1,
    "random_data"=>1,
    "drand48_data"=>1,
    "_IO_marker"=>1,
    "_IO_FILE"=>1,
    "lconv"=>1,
    "sched_param"=>1,
    "tm"=>1,
    "itimerspec"=>1,
    "_pthread_cleanup_buffer"=>1
);

sub getCompileCmd($$$)
{
    my ($Path, $Opt, $Inc) = @_;
    my $GccCall = $GCC_PATH;
    if($Opt) {
        $GccCall .= " ".$Opt;
    }
    $GccCall .= " -x ";
    if($OSgroup eq "macos") {
        $GccCall .= "objective-";
    }
    if(check_gcc_version($GCC_PATH, "4"))
    { # compile as "C++" header
      # to obtain complete dump using GCC 4.0
        $GccCall .= "c++-header";
    }
    else
    { # compile as "C++" source
      # GCC 3.3 cannot compile headers
        $GccCall .= "c++";
    }
    $GccCall .= " ".esc($Path);
    if($CompilerOptions{$Version})
    {# user-defined options
        $GccCall .= " ".$CompilerOptions{$Version};
    }
    if($Inc)
    {# include paths
        $GccCall .= " ".$Inc;
    }
    if(my $PlatformOpts = getPlatformOpts($Version)) {
        $GccCall .= " ".$PlatformOpts;
    }
    return $GccCall;
}

sub getDump()
{
    my $TmpHeaderPath = "$TMP_DIR/dump$Version.h";
    my $MHeaderPath = $TmpHeaderPath;
    open(LIB_HEADER, ">$TmpHeaderPath") || die ("can't open file \'$TmpHeaderPath\': $!\n");
    if(my $AddDefines = $Descriptor{$Version}{"Defines"})
    {
        $AddDefines=~s/\n\s+/\n/g;
        print LIB_HEADER "\n  // add defines\n".$AddDefines."\n";
    }
    print LIB_HEADER "\n  // add includes\n";
    my @PreambleHeaders = keys(%{$Include_Preamble{$Version}});
    @PreambleHeaders = sort {int($Include_Preamble{$Version}{$a}{"Position"})<=>int($Include_Preamble{$Version}{$b}{"Position"})} @PreambleHeaders;
    foreach my $Header_Path (@PreambleHeaders) {
        print LIB_HEADER "  #include \"".forward_slash($Header_Path)."\"\n";
    }
    my @Headers = keys(%{$Target_Headers{$Version}});
    @Headers = sort {int($Target_Headers{$Version}{$a}{"Position"})<=>int($Target_Headers{$Version}{$b}{"Position"})} @Headers;
    foreach my $Header_Path (@Headers) {
        next if($Include_Preamble{$Version}{$Header_Path});
        print LIB_HEADER "  #include \"".forward_slash($Header_Path)."\"\n";
    }
    close(LIB_HEADER);
    my $IncludeString = getIncString(getIncPaths(@PreambleHeaders, @Headers), "GCC");
    # preprocessing stage
    checkPreprocessedUnit(callPreprocessor($TmpHeaderPath, $IncludeString, $Version));
    if($COMMON_LANGUAGE{$Version} eq "C"
    and not $CheckHeadersOnly and $C99Mode!=-1)
    { # rename "C++" keywords in "C" code
        my $RegExp = join("|", keys(%CppKeywords));
        my $PreprocessCmd = getCompileCmd($TmpHeaderPath, "-E", $IncludeString);
        my $Content = `$PreprocessCmd 2>$TMP_DIR/null`;
        if($Content=~s/(\*\s*|\s+|\@|\()($RegExp)(\s*(,|\(|\)|;|->|\.|\:\s*\d))/$1c99_$2$3/g)
        {# functions:
         #   int operator(...);
         # parameters:
         #   int foo(int new, int class, int (*this)(int));
         # bitfields:
         #   unsigned private: 8;
            $C99Mode = 1;
        }
        if($C99Mode==1)
        {# compile the corrected preprocessor output
            $MHeaderPath = "$TMP_DIR/dump$Version.i";
            writeFile($MHeaderPath, $Content);
            print "Using c99 compatibility mode\n";
        }
    }
    if($COMMON_LANGUAGE{$Version} eq "C++" or $CheckHeadersOnly)
    {# add classes and namespaces to the dump
        my $CHdump = "-fdump-class-hierarchy -c";
        if($C99Mode==1) {
            $CHdump .= " -fpreprocessed";
        }
        my $ClassHierarchyCmd = getCompileCmd($MHeaderPath, $CHdump, $IncludeString);
        chdir($TMP_DIR);
        system("$ClassHierarchyCmd >null 2>&1");
        chdir($ORIG_DIR);
        if(my $ClassDump = (cmd_find($TMP_DIR,"f","*.class",1))[0])
        {
            my %AddClasses = ();
            my $Content = readFile($ClassDump);
            foreach my $ClassInfo (split(/\n\n/, $Content))
            {
                if($ClassInfo=~/\AClass\s+(.+)\s*/i)
                {
                    my $CName = $1;
                    next if($CName=~/\A(__|_objc_|_opaque_)/);
                    $TUnit_NameSpaces{$Version}{$CName} = -1;
                    if($CName=~/\A[\w:]+\Z/)
                    {# classes
                        $AddClasses{$CName} = 1;
                    }
                    if($CName=~/(\w[\w:]*)::/)
                    {# namespaces
                        my $NS = $1;
                        if(not defined $TUnit_NameSpaces{$Version}{$NS}) {
                            $TUnit_NameSpaces{$Version}{$NS} = 1;
                        }
                    }
                }
                elsif($ClassInfo=~/\AVtable\s+for\s+(.+)\n((.|\n)+)\Z/i)
                {# read v-tables (advanced approach)
                    my ($CName, $VTable) = ($1, $2);
                    $ClassVTable_Content{$Version}{$CName} = $VTable;
                }
            }
            unlink($ClassDump);
            if(my $NS_Add = get_namespace_additions($TUnit_NameSpaces{$Version}))
            {# GCC on all supported platforms does not include namespaces to the dump by default
                appendFile($MHeaderPath, "\n  // add namespaces\n".$NS_Add);
            }
            # on some OS (MacOS, FreeBSD, CentOS, ...) GCC does not include class methods to the dump by default
            my ($AddClass, $ClassNum) = ("", 0);
            foreach my $CName (sort keys(%AddClasses))
            {
                next if($C_Structure{$CName});
                next if(not $STDCXX_TESTING and $CName=~/\Astd::/);
                next if(($CName=~tr![:]!!)>2);
                if($CName=~/\A(.+)::[^:]+\Z/
                and $AddClasses{$1}) {
                    next;
                }
                $AddClass .= "  $CName* tmp_add_class_".($ClassNum++).";\n";
            }
            appendFile($MHeaderPath, "\n  // add classes\n".$AddClass);
        }
    }
    appendFile($LOG_PATH{$Version}, "Temporary header file \'$TmpHeaderPath\' with the following content will be compiled to create GCC translation unit dump:\n".readFile($TmpHeaderPath)."\n");
    # create TU dump
    my $TUdump = "-fdump-translation-unit -fkeep-inline-functions -c";
    if($C99Mode==1) {
        $TUdump .= " -fpreprocessed";
    }
    my $SyntaxTreeCmd = getCompileCmd($MHeaderPath, $TUdump, $IncludeString);
    appendFile($LOG_PATH{$Version}, "The GCC parameters:\n  $SyntaxTreeCmd\n\n");
    chdir($TMP_DIR);
    system($SyntaxTreeCmd." >$TMP_DIR/tu_errors 2>&1");
    if($?)
    {# failed to compile, but the TU dump still can be created
        my $Errors = readFile("$TMP_DIR/tu_errors");
        if($Errors=~/c99_/)
        {# disable c99 mode
            $C99Mode=-1;
            unlink($LOG_PATH{$Version});
            return getDump();
        }
        # FIXME: handle other errors and try to recompile
        appendFile($LOG_PATH{$Version}, $Errors);# new line
        print STDERR "\nERROR: some errors occurred, see error log for details:\n  ".$LOG_PATH{$Version}."\n\n";
        $COMPILE_ERRORS = 6;
        appendFile($LOG_PATH{$Version}, "\n");# new line
    }
    chdir($ORIG_DIR);
    unlink($TmpHeaderPath, $MHeaderPath);
    return (cmd_find($TMP_DIR,"f","*.tu",1))[0];
}

sub cmd_file($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -e $Path);
    if(my $CmdPath = get_CmdPath("file")) {
        my $Cmd = $CmdPath." -b ".esc($Path);
        my $Cmd_Out = `$Cmd`;
        return $Cmd_Out;
    }
    return "";
}

sub getIncString($$)
{
    my ($ArrRef, $Style) = @_;
    return if(not $ArrRef or $#{$ArrRef}<0);
    my $String = "";
    foreach (@{$ArrRef}) {
        $String .= " ".inc_opt($_, $Style);
    }
    return $String;
}

sub getIncPaths(@)
{
    my @HeaderPaths = @_;
    my @IncPaths = ();
    if($INC_PATH_AUTODETECT{$Version})
    {# autodetecting dependencies
        my %Includes = ();
        foreach my $HPath (@HeaderPaths) {
            foreach my $Dir (get_HeaderDeps($HPath, $Version)) {
                $Includes{$Dir}=1;
            }
        }
        foreach my $Dir (keys(%{$Add_Include_Paths{$Version}}))
        {
            next if($Includes{$Version}{$Dir});
            push(@IncPaths, $Dir);
        }
        foreach my $Dir (@{sortIncPaths([keys(%Includes)], $Version)}) {
            push(@IncPaths, $Dir);
        }
    }
    else
    {# user defined paths
        foreach my $Dir (sort {get_depth($a)<=>get_depth($b)}
        sort {$b cmp $a} keys(%{$Include_Paths{$Version}})) {
            push(@IncPaths, $Dir);
        }
    }
    return \@IncPaths;
}

sub callPreprocessor($$$)
{
    my ($Path, $Inc, $LibVersion) = @_;
    return "" if(not $Path or not -f $Path);
    my $IncludeString=$Inc;
    if(not $Inc) {
        $IncludeString = getIncString(getIncPaths($Path), "GCC");
    }
    my $Cmd = getCompileCmd($Path, "-dD -E", $IncludeString)." 2>$TMP_DIR/null";
    my $Path = "$TMP_DIR/preprocessed";
    system("$Cmd >$Path");
    return $Path;
}

sub cmd_find($$$$)
{ # native "find" is much faster than File::Find (~6x)
  # also the File::Find doesn't support --maxdepth N option
  # so using the cross-platform wrapper for the native one
    my ($Path, $Type, $Name, $MaxDepth) = @_;
    return () if(not $Path or not -e $Path);
    if($OSgroup eq "windows")
    {
        my $DirCmd = get_CmdPath("dir");
        if(not $DirCmd) {
            exitStatus("Not_Found", "can't find \"dir\" command");
        }
        $Path = abs_path($Path);
        $Path=~s/\//\\/g;
        my $Cmd = $DirCmd." \"$Path\" /B /O";
        if($MaxDepth!=1) {
            $Cmd .= " /S";
        }
        if($Type eq "d") {
            $Cmd .= " /AD";
        }
        my @Files = ();
        if($Name)
        { # FIXME: how to search file names in MS shell?
            $Name=~s/\*/.*/g if($Name!~/\]/);
            foreach my $File (split(/\n/, `$Cmd`))
            {
                if($File=~/$Name\Z/i) {
                    push(@Files, $File);
                }
            }
        }
        else {
            @Files = split(/\n/, `$Cmd 2>$TMP_DIR/null`);
        }
        my @AbsPaths = ();
        foreach my $File (@Files) {
            if($File!~/\A(\/|\w+:[\/\\])/) {
                $File = joinPath($Path, $File);
            }
            push(@AbsPaths, $File);
        }
        if($Type eq "d") {
            push(@AbsPaths, $Path);
        }
        return @AbsPaths;
    }
    else
    {
        my $FindCmd = get_CmdPath("find");
        if(not $FindCmd) {
            exitStatus("Not_Found", "can't find a \"find\" command");
        }
        if($Path!~/\A(\/|\w+:[\/\\])/) {
            $Path = abs_path($Path);
        }
        if(-d $Path and -l $Path
        and $Path!~/\/\Z/)
        { # for directories that are symlinks
            $Path.="/";
        }
        my $Cmd = $FindCmd." ".esc($Path);
        if($MaxDepth) {
            $Cmd .= " -maxdepth $MaxDepth";
        }
        if($Type) {
            $Cmd .= " -type $Type";
        }
        if($Name) {
            if($Name=~/\]/) {
                $Cmd .= " -regex \"$Name\"";
            }
            else {
                $Cmd .= " -name \"$Name\"";
            }
        }
        return split(/\n/, `$Cmd 2>$TMP_DIR/null`);
    }
}

sub unpackDump($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -e $Path);
    my $TarCmd = get_CmdPath("tar");
    if(not $TarCmd) {
        exitStatus("Not_Found", "can't find \"tar\" command");
    }
    if($OSgroup eq "windows")
    {# -xvzf option is not implemented in tar.exe (2003)
     # use "gzip.exe -k -d -f" + "tar.exe -xvf" instead
        my $GzipCmd = get_CmdPath("gzip");
        if(not $GzipCmd) {
            exitStatus("Not_Found", "can't find \"gzip\" command");
        }
        system("$GzipCmd -k -d -f $Path");# keep input files (-k)
        if($?) {
            exitStatus("Error", "can't extract \'$Path\'");
        }
        $Path=~s/\.gz\Z//g;
        my $Cmd = "$TarCmd -xvf ".get_filename($Path);
        chdir(get_dirname($Path));
        my $Cmd_Out = `$Cmd`;
        chdir($ORIG_DIR);
        chomp($Cmd_Out);
        unlink($Path);
        return get_dirname($Path)."/".$Cmd_Out;
    }
    else
    {# Linux, MacOS
        my $Cmd = "$TarCmd -xvzf ".esc($Path)." --directory $TMP_DIR";
        my $Cmd_Out = `$Cmd`;
        chomp($Cmd_Out);
        return $TMP_DIR."/".$Cmd_Out;
    }
}

sub is_header_file($)
{# allowed extensions for header files
    if($_[0]=~/\.(h|hh|hp|hxx|hpp|h\+\+|tcc)\Z/i) {
        return $_[0];
    }
    return 0;
}

sub is_header($$$)
{
    my ($Header, $UserDefined, $LibVersion) = @_;
    return 0 if(-d $Header);
    if(-f $Header) {
        $Header = abs_path($Header);
    }
    else {
        if($Header=~/\A(\/|\w+:[\/\\])/) {
            # incorrect absolute path
            return 0;
        }
        if(my $HPath = identify_header($Header, $LibVersion)) {
            $Header = $HPath;
        }
        else {
            # can't find header
            return 0;
        }
    }
    if($Header=~/\.\w+\Z/) {
        # have an extension
        return is_header_file($Header);
    }
    else {
        if($UserDefined) {
            # user defined file without extension
            # treated as header file by default
            return $Header;
        }
        else {
            # directory content
            if(cmd_file($Header)=~/C[\+]*\s+program/i) {
                # !~/HTML|XML|shared|dynamic/i
                return $Header;
            }
        }
    }
    return 0;
}

sub readHeaders($)
{
    $Version = $_[0];
    print "checking header(s) ".$Descriptor{$Version}{"Version"}." ...\n";
    my $DumpPath = getDump();
    if(not $DumpPath) {
        exitStatus("Cannot_Compile", "can't compile header(s)");
    }
    getInfo($DumpPath);
}

sub is_in_library($$)
{
    my ($MnglName, $LibVersion) = @_;
    if(my $Lib = $Symbol_Library{$LibVersion}{$MnglName}) {
        return $Lib;
    }
    elsif(my $SVer = $SymVer{$LibVersion}{$MnglName}) {
        if(my $Lib = $Symbol_Library{$LibVersion}{$SVer}) {
            return $Lib;
        }
    }
    return "";
}

sub prepareInterfaces($)
{
    my $LibVersion = $_[0];
    foreach my $FuncInfoId (keys(%{$SymbolInfo{$LibVersion}}))
    {
        my $MnglName = $SymbolInfo{$LibVersion}{$FuncInfoId}{"MnglName"};
        if(not $MnglName)
        { # ABI dumps don't contain mangled names for C-functions
            $MnglName = $SymbolInfo{$LibVersion}{$FuncInfoId}{"ShortName"};
            $SymbolInfo{$LibVersion}{$FuncInfoId}{"MnglName"} = $MnglName;
        }
        %{$CompleteSignature{$LibVersion}{$MnglName}} = %{$SymbolInfo{$LibVersion}{$FuncInfoId}};
        # symbol and its symlink have same signatures
        if($SymVer{$LibVersion}{$MnglName}) {
            %{$CompleteSignature{$LibVersion}{$SymVer{$LibVersion}{$MnglName}}} = %{$SymbolInfo{$LibVersion}{$FuncInfoId}};
        }
        delete($SymbolInfo{$LibVersion}{$FuncInfoId});
    }
    if($COMMON_LANGUAGE{$LibVersion} eq "C++"
    or $CheckHeadersOnly or $OSgroup eq "windows") {
        translateSymbols(keys(%{$CompleteSignature{$LibVersion}}), $LibVersion);
        translateSymbols(keys(%{$DepSymbols{$LibVersion}}), $LibVersion);
    }
    if($OSgroup eq "windows")
    {# translate symbols from CL mangling to GCC
        foreach my $MnglName (keys(%{$CompleteSignature{$LibVersion}}))
        {
            if($MnglName=~/\A_Z/)
            {
                if(my $TargetMangled = $mangled_name{$LibVersion}{synchTr($tr_name{$MnglName})}) {
                    %{$CompleteSignature{$LibVersion}{$TargetMangled}} = %{$CompleteSignature{$LibVersion}{$MnglName}};
                    delete($CompleteSignature{$LibVersion}{$MnglName});
                }
            }
        }
    }
    if(not keys(%{$CompleteSignature{$LibVersion}}))
    {
        if($CheckHeadersOnly or $CheckObjectsOnly) {
            exitStatus("Error", "the set of public symbols is empty (".$Descriptor{$Version}{"Version"}.")");
        }
        else {
            exitStatus("Error", "the sets of public symbols in headers and $SHARED libraries have empty intersection (".$Descriptor{$Version}{"Version"}.")");
        }
        
    }
    foreach my $MnglName (keys(%{$CompleteSignature{$LibVersion}}))
    {# detect allocable classes with public exported constructors
      # or classes with auto-generated or inline-only constructors
        if(my $ClassId = $CompleteSignature{$LibVersion}{$MnglName}{"Class"})
        {
            my $ClassName = get_TypeName($ClassId, $LibVersion);
            if($CompleteSignature{$LibVersion}{$MnglName}{"Constructor"}
            and not $CompleteSignature{$LibVersion}{$MnglName}{"InLine"})
            {# Class() { ... } will not be exported
                if(not $CompleteSignature{$LibVersion}{$MnglName}{"Private"}) {
                    $AllocableClass{$LibVersion}{$ClassName}{$MnglName} = 1;
                }
                $ClassConstructors{$LibVersion}{$ClassName}{$MnglName} = 1;
            }
            if(not $CompleteSignature{$LibVersion}{$MnglName}{"Private"})
            {# all imported class methods
                if($CheckHeadersOnly) {
                    if(not $CompleteSignature{$LibVersion}{$MnglName}{"InLine"}
                    or $CompleteSignature{$LibVersion}{$MnglName}{"Virt"})
                    {# all symbols except non-virtual inline
                        $ClassMethods{$LibVersion}{$ClassName}{$MnglName} = 1;
                    }
                }
                elsif(is_in_library($MnglName, $LibVersion))
                {# all symbols
                    $ClassMethods{$LibVersion}{$ClassName}{$MnglName} = 1;
                }
            }
            $ClassToId{$LibVersion}{$ClassName} = $ClassId;
        }
    }
    foreach my $MnglName (keys(%VTableClass))
    {# reconstruct header name for v-tables
        if($MnglName=~/\A_ZTV/)
        {
            if(my $ClassName = $VTableClass{$MnglName})
            {
                if(my $ClassId = $TName_Tid{$LibVersion}{$ClassName}) {
                    $CompleteSignature{$LibVersion}{$MnglName}{"Header"} = get_TypeHeader($ClassId, $LibVersion);
                }
            }
        }
    }
}

sub synchTr($)
{
    my $Translation = $_[0];
    if($OSgroup eq "windows") {
        $Translation=~s/\(\)(|\s*const)\Z/\(void\)$1/g;
    }
    return $Translation;
}

sub cleanData($)
{# remove unnecessary data from the ABI dump
    my $LibVersion = $_[0];
    foreach my $FuncInfoId (keys(%{$SymbolInfo{$LibVersion}}))
    {
        my $MnglName = $SymbolInfo{$LibVersion}{$FuncInfoId}{"MnglName"};
        if(not $MnglName) {
            delete($SymbolInfo{$LibVersion}{$FuncInfoId});
            next;
        }
        if($MnglName eq $SymbolInfo{$LibVersion}{$FuncInfoId}{"ShortName"}) {
            delete($SymbolInfo{$LibVersion}{$FuncInfoId}{"MnglName"});
        }
        if(not is_in_library($MnglName, $LibVersion) and not $CheckHeadersOnly
        and not $SymbolInfo{$LibVersion}{$FuncInfoId}{"Virt"}
        and not $SymbolInfo{$LibVersion}{$FuncInfoId}{"PureVirt"})
        {# removing src only and all external functions
            delete($SymbolInfo{$LibVersion}{$FuncInfoId});
            next;
        }
        if(not interfaceFilter($MnglName, $LibVersion, "Public")) {
            delete($SymbolInfo{$LibVersion}{$FuncInfoId});
            next;
        }
        my %FuncInfo = %{$SymbolInfo{$LibVersion}{$FuncInfoId}};
        register_TypeUsing($Tid_TDid{$LibVersion}{$FuncInfo{"Return"}}, $FuncInfo{"Return"}, $LibVersion);
        register_TypeUsing($Tid_TDid{$LibVersion}{$FuncInfo{"Class"}}, $FuncInfo{"Class"}, $LibVersion);
        foreach my $Param_Pos (keys(%{$FuncInfo{"Param"}}))
        {
            my $Param_TypeId = $FuncInfo{"Param"}{$Param_Pos}{"type"};
            register_TypeUsing($Tid_TDid{$LibVersion}{$Param_TypeId}, $Param_TypeId, $LibVersion);
        }
        if(not keys(%{$SymbolInfo{$LibVersion}{$FuncInfoId}{"Param"}})) {
            delete($SymbolInfo{$LibVersion}{$FuncInfoId}{"Param"});
        }
    }
    foreach my $TDid (keys(%{$TypeInfo{$LibVersion}}))
    {
        if(not keys(%{$TypeInfo{$LibVersion}{$TDid}})) {
            delete($TypeInfo{$LibVersion}{$TDid});
        }
        else
        {
            foreach my $Tid (keys(%{$TypeInfo{$LibVersion}{$TDid}}))
            {
                if(not $UsedType{$LibVersion}{$TDid}{$Tid})
                {
                    delete($TypeInfo{$LibVersion}{$TDid}{$Tid});
                    if(not keys(%{$TypeInfo{$LibVersion}{$TDid}})) {
                        delete($TypeInfo{$LibVersion}{$TDid});
                    }
                    delete($Tid_TDid{$LibVersion}{$Tid}) if($Tid_TDid{$LibVersion}{$Tid} eq $TDid);
                }
                else
                {# clean attributes
                    if(not $TypeInfo{$LibVersion}{$TDid}{$Tid}{"TDid"}) {
                        delete($TypeInfo{$LibVersion}{$TDid}{$Tid}{"TDid"});
                    }
                    if(not $TypeInfo{$LibVersion}{$TDid}{$Tid}{"NameSpace"}) {
                        delete($TypeInfo{$LibVersion}{$TDid}{$Tid}{"NameSpace"});
                    }
                    if(defined $TypeInfo{$LibVersion}{$TDid}{$Tid}{"BaseType"}
                    and not $TypeInfo{$LibVersion}{$TDid}{$Tid}{"BaseType"}{"TDid"}) {
                        delete($TypeInfo{$LibVersion}{$TDid}{$Tid}{"BaseType"}{"TDid"});
                    }
                    if(not $TypeInfo{$LibVersion}{$TDid}{$Tid}{"Header"}) {
                        delete($TypeInfo{$LibVersion}{$TDid}{$Tid}{"Header"});
                    }
                    if(not $TypeInfo{$LibVersion}{$TDid}{$Tid}{"Line"}) {
                        delete($TypeInfo{$LibVersion}{$TDid}{$Tid}{"Line"});
                    }
                }
            }
        }
    }
}

sub register_TypeUsing($$$)
{
    my ($TypeDeclId, $TypeId, $LibVersion) = @_;
    return if($UsedType{$LibVersion}{$TypeDeclId}{$TypeId});
    my %Type = get_Type($TypeDeclId, $TypeId, $LibVersion);
    if($Type{"Type"}=~/\A(Struct|Union|Class|FuncPtr|MethodPtr|FieldPtr|Enum)\Z/)
    {
        $UsedType{$LibVersion}{$TypeDeclId}{$TypeId} = 1;
        if($Type{"Type"}=~/\A(Struct|Class)\Z/)
        {
            if(my $ThisPtrId = getTypeIdByName(get_TypeName($TypeId, $LibVersion)."*const", $LibVersion))
            {# register "this" pointer
                my $ThisPtrDId = $Tid_TDid{$LibVersion}{$ThisPtrId};
                my %ThisPtrType = get_Type($ThisPtrDId, $ThisPtrId, $LibVersion);
                $UsedType{$LibVersion}{$ThisPtrDId}{$ThisPtrId} = 1;
                register_TypeUsing($ThisPtrType{"BaseType"}{"TDid"}, $ThisPtrType{"BaseType"}{"Tid"}, $LibVersion);
            }
            foreach my $BaseId (keys(%{$Type{"Base"}}))
            {# register base classes
                register_TypeUsing($Tid_TDid{$LibVersion}{$BaseId}, $BaseId, $LibVersion);
            }
        }
        foreach my $Memb_Pos (keys(%{$Type{"Memb"}}))
        {
            my $Member_TypeId = $Type{"Memb"}{$Memb_Pos}{"type"};
            register_TypeUsing($Tid_TDid{$LibVersion}{$Member_TypeId}, $Member_TypeId, $LibVersion);
        }
        if($Type{"Type"} eq "FuncPtr"
        or $Type{"Type"} eq "MethodPtr") {
            my $ReturnType = $Type{"Return"};
            register_TypeUsing($Tid_TDid{$LibVersion}{$ReturnType}, $ReturnType, $LibVersion);
            foreach my $Memb_Pos (keys(%{$Type{"Param"}}))
            {
                my $Member_TypeId = $Type{"Param"}{$Memb_Pos}{"type"};
                register_TypeUsing($Tid_TDid{$LibVersion}{$Member_TypeId}, $Member_TypeId, $LibVersion);
            }
        }
    }
    elsif($Type{"Type"}=~/\A(Const|Pointer|Ref|Volatile|Restrict|Array|Typedef)\Z/)
    {
        $UsedType{$LibVersion}{$TypeDeclId}{$TypeId} = 1;
        register_TypeUsing($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
    }
    elsif($Type{"Type"} eq "Intrinsic") {
        $UsedType{$LibVersion}{$TypeDeclId}{$TypeId} = 1;
    }
    else
    {
        delete($TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId});
        if(not keys(%{$TypeInfo{$LibVersion}{$TypeDeclId}})) {
            delete($TypeInfo{$LibVersion}{$TypeDeclId});
        }
        if($Tid_TDid{$LibVersion}{$TypeId} eq $TypeDeclId) {
            delete($Tid_TDid{$LibVersion}{$TypeId});
        }
    }
}

sub findMethod($$$)
{
    my ($VirtFunc, $ClassId, $LibVersion) = @_;
    foreach my $BaseClass_Id (keys(%{$TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$ClassId}}{$ClassId}{"Base"}}))
    {
        if(my $VirtMethodInClass = findMethod_Class($VirtFunc, $BaseClass_Id, $LibVersion)) {
            return $VirtMethodInClass;
        }
        elsif(my $VirtMethodInBaseClasses = findMethod($VirtFunc, $BaseClass_Id, $LibVersion)) {
            return $VirtMethodInBaseClasses;
        }
    }
    return "";
}

sub findMethod_Class($$$)
{
    my ($VirtFunc, $ClassId, $LibVersion) = @_;
    my $ClassName = get_TypeName($ClassId, $LibVersion);
    return "" if(not defined $VirtualTable{$LibVersion}{$ClassName});
    my $TargetSuffix = get_func_suffix($VirtFunc);
    my $TargetShortName = $CompleteSignature{$LibVersion}{$VirtFunc}{"ShortName"};
    foreach my $Candidate (keys(%{$VirtualTable{$LibVersion}{$ClassName}}))
    {# search for interface with the same parameters suffix (overridden)
        if($TargetSuffix eq get_func_suffix($Candidate))
        {
            if($CompleteSignature{$LibVersion}{$VirtFunc}{"Destructor"}) {
                if($CompleteSignature{$LibVersion}{$Candidate}{"Destructor"}) {
                    if(($VirtFunc=~/D0E/ and $Candidate=~/D0E/)
                    or ($VirtFunc=~/D1E/ and $Candidate=~/D1E/)
                    or ($VirtFunc=~/D2E/ and $Candidate=~/D2E/)) {
                        return $Candidate;
                    }
                }
            }
            else {
                if($TargetShortName eq $CompleteSignature{$LibVersion}{$Candidate}{"ShortName"}) {
                    return $Candidate;
                }
            }
        }
    }
    return "";
}

sub registerVirtualTable($)
{
    my $LibVersion = $_[0];
    foreach my $Interface (keys(%{$CompleteSignature{$LibVersion}}))
    {
        if($CompleteSignature{$LibVersion}{$Interface}{"Virt"}
        or $CompleteSignature{$LibVersion}{$Interface}{"PureVirt"})
        {
            my $ClassName = get_TypeName($CompleteSignature{$LibVersion}{$Interface}{"Class"}, $LibVersion);
            next if(not $STDCXX_TESTING and $ClassName=~/\A(std::|__cxxabi)/);
            if($CompleteSignature{$LibVersion}{$Interface}{"Destructor"}
            and $Interface=~/D2E/)
            { # pure virtual D2-destructors are marked as "virt" in the dump
              # virtual D2-destructors are NOT marked as "virt" in the dump
              # both destructors are not presented in the v-table
                next;
            }
            $CheckedTypes{$ClassName} = 1;
            my ($MnglName, $VersionSpec, $SymbolVersion) = get_symbol_name_version($Interface);
            $VirtualTable{$LibVersion}{$ClassName}{$MnglName} = 1;
        }
        if($CheckHeadersOnly
        and $CompleteSignature{$LibVersion}{$Interface}{"Virt"})
        { # Register added and removed virtual symbols
          # This is necessary for --headers-only mode
          # Virtual function cannot be inline, so:
          # presence in headers <=> presence in shared libs
            if($LibVersion==2 and not $CompleteSignature{1}{$Interface}{"Header"})
            { # not presented in old-version headers
                $AddedInt{$Interface} = 1;
            }
            if($LibVersion==1 and not $CompleteSignature{2}{$Interface}{"Header"})
            { # not presented in new-version headers
                $RemovedInt{$Interface} = 1;
            }
        }
    }
}

sub registerOverriding($)
{
    my $LibVersion = $_[0];
    foreach my $ClassName (keys(%{$VirtualTable{$LibVersion}}))
    {
        $CheckedTypes{$ClassName} = 1;
        foreach my $VirtFunc (keys(%{$VirtualTable{$LibVersion}{$ClassName}}))
        {
            next if($CompleteSignature{$LibVersion}{$VirtFunc}{"PureVirt"});
            if(my $OverriddenMethod = findMethod($VirtFunc, $TName_Tid{$LibVersion}{$ClassName}, $LibVersion))
            {# both overridden virtual and implemented pure virtual functions
                $CompleteSignature{$LibVersion}{$VirtFunc}{"Override"} = $OverriddenMethod;
                $OverriddenMethods{$LibVersion}{$OverriddenMethod}{$VirtFunc} = 1;
                delete($VirtualTable{$LibVersion}{$ClassName}{$VirtFunc});
            }
        }
        if(not keys(%{$VirtualTable{$LibVersion}{$ClassName}})) {
            delete($VirtualTable{$LibVersion}{$ClassName});
        }
    }
}

sub setVirtFuncPositions($)
{
    my $LibVersion = $_[0];
    foreach my $ClassName (keys(%{$VirtualTable{$LibVersion}}))
    {
        $CheckedTypes{$ClassName} = 1;
        my ($Num, $RelPos, $AbsNum) = (1, 0, 1);
        foreach my $VirtFunc (sort {int($CompleteSignature{$LibVersion}{$a}{"Line"}) <=> int($CompleteSignature{$LibVersion}{$b}{"Line"})}
        sort keys(%{$VirtualTable{$LibVersion}{$ClassName}}))
        {
            if(not $CompleteSignature{1}{$VirtFunc}{"Override"}
            and not $CompleteSignature{2}{$VirtFunc}{"Override"})
            {
                if(defined $VirtualTable{1}{$ClassName} and defined $VirtualTable{1}{$ClassName}{$VirtFunc}
                and defined $VirtualTable{2}{$ClassName} and defined $VirtualTable{2}{$ClassName}{$VirtFunc})
                {# relative position excluding added and removed virtual functions
                    $CompleteSignature{$LibVersion}{$VirtFunc}{"RelPos"} = $RelPos++;
                }
                $VirtualTable{$LibVersion}{$ClassName}{$VirtFunc}=$Num++;
            }
            
        }
    }
    foreach my $ClassName (keys(%{$ClassToId{$LibVersion}}))
    {
        $CheckedTypes{$ClassName} = 1;
        my $AbsNum = 1;
        foreach my $VirtFunc (getVTable($ClassToId{$LibVersion}{$ClassName}, $LibVersion)) {
            $VirtualTable_Full{$LibVersion}{$ClassName}{$VirtFunc}=$AbsNum++;
        }
    }
}

sub get_class_bases($$$)
{
    my ($ClassId, $LibVersion, $Recursive) = @_;
    my %ClassType = get_Type($Tid_TDid{$LibVersion}{$ClassId}, $ClassId, $LibVersion);
    return () if(not defined $ClassType{"Base"});
    my (@Bases, %Added) = ();
    foreach my $BaseId (sort {int($ClassType{"Base"}{$a}{"pos"})<=>int($ClassType{"Base"}{$b}{"pos"})}
    keys(%{$ClassType{"Base"}}))
    {
        if($Recursive) {
            foreach my $SubBaseId (get_class_bases($BaseId, $LibVersion, 1)) {
                push(@Bases, $SubBaseId);
            }
        }
        push(@Bases, $BaseId);
    }
    return @Bases;
}

sub getVTable($$)
{# return list of v-table elements
    my ($ClassId, $LibVersion) = @_;
    my @Bases = get_class_bases($ClassId, $LibVersion, 1);
    my @Elements = ();
    foreach my $BaseId (@Bases, $ClassId)
    {
        my $BName = get_TypeName($BaseId, $LibVersion);
        my @VFunctions = keys(%{$VirtualTable{$LibVersion}{$BName}});
        @VFunctions = sort {int($CompleteSignature{$LibVersion}{$a}{"Line"}) <=> int($CompleteSignature{$LibVersion}{$b}{"Line"})} @VFunctions;
        foreach my $VFunc (@VFunctions)
        {
            push(@Elements, $VFunc);
        }
    }
    return @Elements;
}

sub getVShift($$)
{
    my ($ClassId, $LibVersion) = @_;
    my @Bases = get_class_bases($ClassId, $LibVersion, 1);
    my $VShift = 0;
    foreach my $BaseId (@Bases)
    {
        my $BName = get_TypeName($BaseId, $LibVersion);
        if(defined $VirtualTable{$LibVersion}{$BName}) {
            $VShift+=keys(%{$VirtualTable{$LibVersion}{$BName}});
        }
    }
    return $VShift;
}

sub getShift($$)
{
    my ($ClassId, $LibVersion) = @_;
    my @Bases = get_class_bases($ClassId, $LibVersion, 0);
    my $Shift = 0;
    foreach my $BaseId (@Bases)
    {
        my $Size = get_TypeSize($BaseId, $LibVersion);
        if($Size!=1) {
            # empty base class
            $Shift+=get_TypeSize($BaseId, $LibVersion);
        }
    }
    return $Shift;
}

sub getVSize($$)
{
    my ($ClassName, $LibVersion) = @_;
    if(defined $VirtualTable{$LibVersion}{$ClassName})  {
        return keys(%{$VirtualTable{$LibVersion}{$ClassName}});
    }
    else {
        return 0;
    }
}

sub isAllocableClass($$)
{
    my ($ClassName, $LibVersion) = @_;
    return keys(%{$AllocableClass{$LibVersion}{$ClassName}});
}

sub isCopyingClass($$)
{
    my ($ClassName, $LibVersion) = @_;
    return (not keys(%{$ClassConstructors{$LibVersion}{$ClassName}}));
}

sub isLeafClass($$)
{
    my ($ClassId, $LibVersion) = @_;
    return (not keys(%{$Class_SubClasses{$LibVersion}{$ClassId}}));
}

sub havePubFields($)
{# check structured type for public fields
    return isAccessible($_[0], 0, -1);
}

sub isAccessible($$$)
{# check interval in structured type for public fields
    my ($TypePtr, $Start, $End) = @_;
    return if(not $TypePtr);
    if(defined $Cache{"isAccessible"}{$TypePtr->{"Name"}}{$End}{$Start}) {
        return $Cache{"isAccessible"}{$TypePtr->{"Name"}}{$End}{$Start};
    }
    if($End==-1) {
        $End = keys(%{$TypePtr->{"Memb"}})-1;
    }
    foreach my $MembPos (keys(%{$TypePtr->{"Memb"}}))
    {
        if(int($MembPos)>=$Start and int($MembPos)<=$End) {
            if(isPublic($TypePtr, $MembPos)) {
                $Cache{"isAccessible"}{$TypePtr->{"Name"}}{$End}{$Start} = 1;
                return 1;
            }
        }
    }
    $Cache{"isAccessible"}{$TypePtr->{"Name"}}{$End}{$Start} = 0;
    return 0;
}

sub havePadding($)
{# check structured type for padding fields
    my $TypePtr = $_[0];
    return 0 if(not $TypePtr or not keys(%{$TypePtr->{"Memb"}}));
    return isReserved($TypePtr->{"Memb"}{keys(%{$TypePtr->{"Memb"}})-1}{"name"});
}

sub isReserved($)
{
    my $MName = $_[0];
    return $MName=~/reserved|padding|f_spare/i;
}

sub isPublic($$)
{
    my ($TypePtr, $MembPos) = @_;
    return 0 if(not $TypePtr);
    if(not $TypePtr->{"Memb"}{$MembPos}{"access"}) {
        # by name in C language
        # FIXME: add other methods to detect private members
        my $MName = $TypePtr->{"Memb"}{$MembPos}{"name"};
        if($MName!~/priv|abidata/i and lc($MName) ne "abi"
        and not isReserved($MName)) {
            return 1;
        }
    }
    elsif($TypePtr->{"Memb"}{$MembPos}{"access"} ne "private") {
        # by access in C++ language
        return 1;
    }
    return 0;
}

sub mergeBases()
{
    foreach my $ClassName (keys(%{$ClassToId{1}}))
    {# detect added and removed virtual functions
        my $ClassId = $ClassToId{1}{$ClassName};
        next if(not $ClassId);
        foreach my $VirtFunc (keys(%{$VirtualTable{2}{$ClassName}}))
        {
            if($ClassToId{1}{$ClassName} and not defined $VirtualTable{1}{$ClassName}{$VirtFunc})
            {# added to virtual table
                if($AddedInt{$VirtFunc}
                # added pure virtual method (without a symbol in shared library)
                or (not $CompleteSignature{1}{$VirtFunc}{"Header"} and $CompleteSignature{2}{$VirtFunc}{"PureVirt"})
                # presented as non-virtual in old-version headers
                or ($CompleteSignature{1}{$VirtFunc}{"Header"} and not $CompleteSignature{1}{$VirtFunc}{"Virt"}))
                {# reasons: added, became virtual, added pure virtual, added to header
                    if(not $CompleteSignature{2}{$VirtFunc}{"Override"}) {
                        $AddedInt_Virt{$ClassName}{$VirtFunc} = 1;
                    }
                }
            }
        }
        foreach my $VirtFunc (keys(%{$VirtualTable{1}{$ClassName}}))
        {
            if($ClassToId{2}{$ClassName} and not defined $VirtualTable{2}{$ClassName}{$VirtFunc})
            {# removed from virtual table
                if($RemovedInt{$VirtFunc}
                # removed pure virtual method (without a symbol in shared library)
                or ($CompleteSignature{1}{$VirtFunc}{"PureVirt"} and not $CompleteSignature{2}{$VirtFunc}{"Header"})
                # presented as non-virtual in new-version headers
                or ($CompleteSignature{2}{$VirtFunc}{"Header"} and not $CompleteSignature{2}{$VirtFunc}{"Virt"}))
                {# reasons: removed, became non-virtual, removed pure virtual, removed from header
                    if(not $CompleteSignature{1}{$VirtFunc}{"Override"}) {
                        $RemovedInt_Virt{$ClassName}{$VirtFunc} = 1;
                    }
                }
            }
        }
        my %Class_Type = get_Type($Tid_TDid{1}{$ClassId}, $ClassId, 1);
        foreach my $AddedVFunc (keys(%{$AddedInt_Virt{$ClassName}}))
        {# check replacements, including pure virtual methods
            my $AddedPos = $VirtualTable{2}{$ClassName}{$AddedVFunc};
            foreach my $RemovedVFunc (keys(%{$RemovedInt_Virt{$ClassName}}))
            {
                my $RemovedPos = $VirtualTable{1}{$ClassName}{$RemovedVFunc};
                if($AddedPos==$RemovedPos) {
                    $VirtualReplacement{$AddedVFunc} = $RemovedVFunc;
                    $VirtualReplacement{$RemovedVFunc} = $AddedVFunc;
                    last;# other methods will be reported as "added" or "removed"
                }
            }
            if(my $RemovedVFunc = $VirtualReplacement{$AddedVFunc}) {
                my $ProblemType = "Virtual_Replacement";
                my @Affected = ($RemovedVFunc);
                if($CompleteSignature{1}{$RemovedVFunc}{"PureVirt"}) {
                    $ProblemType = "Pure_Virtual_Replacement";
                    @Affected = (keys(%{$ClassMethods{1}{$ClassName}}))
                }
                foreach my $AffectedInt (@Affected) {
                    if(not $CompleteSignature{1}{$AffectedInt}{"PureVirt"})
                    {# affected all exported methods
                        %{$CompatProblems{$AffectedInt}{$ProblemType}{$tr_name{$AddedVFunc}}}=(
                            "Type_Name"=>$Class_Type{"Name"},
                            "Type_Type"=>"Class",
                            "Target"=>get_Signature($AddedVFunc, 2),
                            "Old_Value"=>get_Signature($RemovedVFunc, 1));
                    }
                    # FIXME: check usage of the abstract class as the parameter of some exported method
                    # such methods will be affected
                }
            }
        }
    }
    if($UsedDump{1}
    and cmpVersions($UsedDump{1}, "2.0")<0)
    {# support for old dumps
        # "Base" attribute introduced in ACC 1.22 (dump 2.0 format)
        return;
    }
    if($UsedDump{2}
    and cmpVersions($UsedDump{2}, "2.0")<0)
    {# support for old dumps
        # "Base" attribute introduced in ACC 1.22 (dump 2.0 format)
        return;
    }
    foreach my $ClassName (keys(%{$ClassToId{1}}))
    {
        my $ClassId_Old = $ClassToId{1}{$ClassName};
        next if(not $ClassId_Old);
        if(not isAllocableClass($ClassName, 1) and not isCopyingClass($ClassName, 1)) {
            # skip classes without public constructors (including auto-generated)
            # example: class has only a private exported or private inline constructor
            next;
        }
        if($ClassName=~/>/) {
            # skip affected template instances
            next;
        }
        my %Class_Old = get_Type($Tid_TDid{1}{$ClassId_Old}, $ClassId_Old, 1);
        my $ClassId_New = $ClassToId{2}{$ClassName};
        next if(not $ClassId_New);
        my %Class_New = get_Type($Tid_TDid{2}{$ClassId_New}, $ClassId_New, 2);
        my @Bases_Old = sort {$Class_Old{"Base"}{$a}{"pos"}<=>$Class_Old{"Base"}{$b}{"pos"}} keys(%{$Class_Old{"Base"}});
        my @Bases_New = sort {$Class_New{"Base"}{$a}{"pos"}<=>$Class_New{"Base"}{$b}{"pos"}} keys(%{$Class_New{"Base"}});
        my ($BNum1, $BNum2) = (1, 1);
        my %BasePos_Old = map {get_TypeName($_, 1) => $BNum1++} @Bases_Old;
        my %BasePos_New = map {get_TypeName($_, 2) => $BNum2++} @Bases_New;
        my %ShortBase_Old = map {get_ShortName($_, 1) => 1} @Bases_Old;
        my %ShortBase_New = map {get_ShortName($_, 2) => 1} @Bases_New;
        my $Shift_Old = getShift($ClassId_Old, 1);
        my $Shift_New = getShift($ClassId_New, 2);
        my %BaseId_New = map {get_TypeName($_, 2) => $_} @Bases_New;
        my ($Added, $Removed) = (0, 0);
        my @StableBases_Old = ();
        foreach my $BaseId (@Bases_Old)
        {
            my $BaseName = get_TypeName($BaseId, 1);
            if($BasePos_New{$BaseName}) {
                push(@StableBases_Old, $BaseId);
            }
            elsif(not $ShortBase_New{$BaseName}
            and not $ShortBase_New{get_ShortName($BaseId, 1)})
            {# removed base
             # excluding namespace::SomeClass to SomeClass renaming
                my $ProblemKind = "Removed_Base_Class";
                if($Shift_Old ne $Shift_New
                and (havePubFields(\%Class_Old) or $Class_Old{"Size"} ne $Class_New{"Size"})) {
                    $ProblemKind = "Removed_Base_Class_And_Shift";
                }
                foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                {
                    %{$CompatProblems{$Interface}{$ProblemKind}{"this"}}=(
                        "Type_Name"=>$ClassName,
                        "Type_Type"=>"Class",
                        "Target"=>$BaseName,
                        "Old_Size"=>$Class_Old{"Size"},
                        "New_Size"=>$Class_New{"Size"},
                        "Memory_Shift"=>abs($Shift_New-$Shift_Old)  );
                }
                $Removed+=1;
            }
        }
        my @StableBases_New = ();
        foreach my $BaseId (@Bases_New)
        {
            my $BaseName = get_TypeName($BaseId, 2);
            if($BasePos_Old{$BaseName}) {
                push(@StableBases_New, $BaseId);
            }
            elsif(not $ShortBase_Old{$BaseName}
            and not $ShortBase_Old{get_ShortName($BaseId, 2)})
            {# added base
             # excluding namespace::SomeClass to SomeClass renaming
                my $ProblemKind = "Added_Base_Class";
                if($Shift_Old ne $Shift_New
                and (havePubFields(\%Class_Old) or $Class_Old{"Size"} ne $Class_New{"Size"})) {
                    $ProblemKind = "Added_Base_Class_And_Shift";
                }
                foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                {
                    %{$CompatProblems{$Interface}{$ProblemKind}{"this"}}=(
                        "Type_Name"=>$ClassName,
                        "Type_Type"=>"Class",
                        "Target"=>$BaseName,
                        "Old_Size"=>$Class_Old{"Size"},
                        "New_Size"=>$Class_New{"Size"},
                        "Memory_Shift"=>abs($Shift_New-$Shift_Old)  );
                }
                $Added+=1;
            }
        }
        ($BNum1, $BNum2) = (1, 1);
        my %BaseRelPos_Old = map {get_TypeName($_, 1) => $BNum1++} @StableBases_Old;
        my %BaseRelPos_New = map {get_TypeName($_, 2) => $BNum2++} @StableBases_New;
        foreach my $BaseId (@Bases_Old)
        {
            my $BaseName = get_TypeName($BaseId, 1);
            if(my $NewPos = $BaseRelPos_New{$BaseName})
            {
                my $BaseNewId = $BaseId_New{$BaseName};
                my $OldPos = $BaseRelPos_Old{$BaseName};
                if($NewPos!=$OldPos)
                {# changed position of the base class
                    foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                    {
                        %{$CompatProblems{$Interface}{"Base_Class_Position"}{"this"}}=(
                            "Type_Name"=>$ClassName,
                            "Type_Type"=>"Class",
                            "Target"=>$BaseName,
                            "Old_Value"=>$OldPos-1,
                            "New_Value"=>$NewPos-1  );
                    }
                }
                if($Class_Old{"Base"}{$BaseId}{"virtual"}
                and not $Class_New{"Base"}{$BaseNewId}{"virtual"})
                {# became non-virtual base
                    foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                    {
                        %{$CompatProblems{$Interface}{"Base_Class_Became_Non_Virtually_Inherited"}{"this->".$BaseName}}=(
                            "Type_Name"=>$ClassName,
                            "Type_Type"=>"Class",
                            "Target"=>$BaseName  );
                    }
                }
                elsif(not $Class_Old{"Base"}{$BaseId}{"virtual"}
                and $Class_New{"Base"}{$BaseNewId}{"virtual"})
                {# became virtual base
                    foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                    {
                        %{$CompatProblems{$Interface}{"Base_Class_Became_Virtually_Inherited"}{"this->".$BaseName}}=(
                            "Type_Name"=>$ClassName,
                            "Type_Type"=>"Class",
                            "Target"=>$BaseName  );
                    }
                }
            }
        }
        # detect size changes in base classes
        if($Shift_Old!=$Shift_New)
        {# size of allocable class
            foreach my $BaseId (@StableBases_Old)
            {# search for changed base
                my %BaseType = get_Type($Tid_TDid{1}{$BaseId}, $BaseId, 1);
                my $Size_Old = get_TypeSize($BaseId, 1);
                my $Size_New = get_TypeSize($BaseId_New{$BaseType{"Name"}}, 2);
                if($Size_Old ne $Size_New
                and $Size_Old and $Size_New)
                {
                    my $ProblemType = "";
                    if(isAllocableClass($BaseType{"Name"}, 1))
                    {
                        if($Size_New>$Size_Old) {
                            # increased size
                            $ProblemType = "Size_Of_Allocable_Class_Increased";
                        }
                        else {
                            # decreased size
                            $ProblemType = "Size_Of_Allocable_Class_Decreased";
                            if(not havePubFields(\%Class_Old)) {
                                # affected class has no public members
                                next;
                            }
                        }
                    }
                    else {
                        $ProblemType = "Size_Of_Copying_Class";
                    }
                    foreach my $Interface (keys(%{$ClassMethods{1}{$ClassName}}))
                    {# base class size changes affecting current class
                        %{$CompatProblems{$Interface}{$ProblemType}{"this->".$BaseType{"Name"}}}=(
                            "Type_Name"=>$BaseType{"Name"},
                            "Type_Type"=>"Class",
                            "Target"=>$BaseType{"Name"},
                            "Old_Value"=>$Size_Old,
                            "New_Value"=>$Size_New  );
                    }
                }
            }
        }
        if(my @VFunctions = keys(%{$VirtualTable{1}{$ClassName}}))
        {# compare virtual tables size in base classes
            my $VShift_Old = getVShift($ClassId_Old, 1);
            my $VShift_New = getVShift($ClassId_New, 2);
            if($VShift_Old ne $VShift_New)
            {# changes in the base class or changes in the list of base classes
                my @AllBases_Old = get_class_bases($ClassId_Old, 1, 1);
                my @AllBases_New = get_class_bases($ClassId_New, 2, 1);
                ($BNum1, $BNum2) = (1, 1);
                my %StableBase = map {get_TypeName($_, 2) => $_} @AllBases_New;
                foreach my $BaseId (@AllBases_Old)
                {
                    my %BaseType = get_Type($Tid_TDid{1}{$BaseId}, $BaseId, 1);
                    if(not $StableBase{$BaseType{"Name"}}) {
                        # lost base
                        next;
                    }
                    my $VSize_Old = getVSize($BaseType{"Name"}, 1);
                    my $VSize_New = getVSize($BaseType{"Name"}, 2);
                    if($VSize_Old!=$VSize_New)
                    {
                        my $VRealSize_Old = get_VTableSymbolSize($BaseType{"Name"}, 1);
                        my $VRealSize_New = get_VTableSymbolSize($BaseType{"Name"}, 2);
                        if(not $VRealSize_Old or not $VRealSize_New) {
                            # try to compute a model v-table size
                            $VRealSize_Old = ($VSize_Old+2+getVShift($BaseId, 1))*$POINTER_SIZE{1};
                            $VRealSize_New = ($VSize_New+2+getVShift($StableBase{$BaseType{"Name"}}, 2))*$POINTER_SIZE{2};
                        }
                        foreach my $Interface (@VFunctions)
                        {
                            if(not defined $VirtualTable{2}{$ClassName}{$Interface})
                            {# Removed_Virtual_Method, will be registered in mergeVirtualTables()
                                next;
                            }
                            if($VirtualTable{2}{$ClassName}{$Interface}-$VirtualTable{1}{$ClassName}{$Interface}+$VSize_New-$VSize_Old==0)
                            {# skip interfaces that have not changed the absolute virtual position
                                next;
                            }
                            if(not is_in_library($Interface, 1) and not $CheckHeadersOnly)
                            {# affected symbols in shared library
                                next;
                            }
                            %{$CompatProblems{$Interface}{"Virtual_Table_Size"}{$BaseType{"Name"}}}=(
                                "Type_Name"=>$BaseType{"Name"},
                                "Type_Type"=>"Class",
                                "Target"=>get_Signature($Interface, 1),
                                "Old_Value"=>$VRealSize_Old,
                                "New_Value"=>$VRealSize_New  );
                            foreach my $VirtFunc (keys(%{$AddedInt_Virt{$BaseType{"Name"}}}))
                            {# the reason of the layout change: added virtual functions
                                next if($VirtualReplacement{$VirtFunc});
                                my $ProblemType = "Added_Virtual_Method";
                                if($CompleteSignature{2}{$VirtFunc}{"PureVirt"}) {
                                    $ProblemType = "Added_Pure_Virtual_Method";
                                }
                                %{$CompatProblems{$Interface}{$ProblemType}{get_Signature($VirtFunc, 2)}}=(
                                    "Type_Name"=>$BaseType{"Name"},
                                    "Type_Type"=>"Class",
                                    "Target"=>get_Signature($VirtFunc, 2)  );
                            }
                            foreach my $VirtFunc (keys(%{$RemovedInt_Virt{$BaseType{"Name"}}}))
                            {# the reason of the layout change: removed virtual functions
                                next if($VirtualReplacement{$VirtFunc});
                                %{$CompatProblems{$Interface}{"Removed_Virtual_Method"}{get_Signature($VirtFunc, 1)}}=(
                                    "Type_Name"=>$BaseType{"Name"},
                                    "Type_Type"=>"Class",
                                    "Target"=>get_Signature($VirtFunc, 1)  );
                            }
                        }
                    }
                }
            }
        }
    }
}

sub mergeVirtualTables($)
{# check for changes in the virtual table
    my $Interface = $_[0];# affected method
    my $Class_Id = $CompleteSignature{1}{$Interface}{"Class"};
    my %Class_Type = get_Type($Tid_TDid{1}{$Class_Id}, $Class_Id, 1);
    $CheckedTypes{$Class_Type{"Name"}} = 1;
    # check virtual table structure
    foreach my $AddedVFunc (keys(%{$AddedInt_Virt{$Class_Type{"Name"}}}))
    {
        next if($Interface eq $AddedVFunc);
        next if($VirtualReplacement{$AddedVFunc});
        my $VPos_Added = $VirtualTable{2}{$Class_Type{"Name"}}{$AddedVFunc};
        if($CompleteSignature{2}{$AddedVFunc}{"PureVirt"})
        { # pure virtual methods affect all others (virtual and non-virtual)
            %{$CompatProblems{$Interface}{"Added_Pure_Virtual_Method"}{$tr_name{$AddedVFunc}}}=(
                "Type_Name"=>$Class_Type{"Name"},
                "Type_Type"=>"Class",
                "Target"=>get_Signature($AddedVFunc, 2)  );
            # FIXME: check usage of the abstract class as the parameter of some exported method
            # pure virtual methods should not be registered as affected
        }
        elsif($VPos_Added>keys(%{$VirtualTable{1}{$Class_Type{"Name"}}}))
        {# added virtual function at the end of v-table
            if(not keys(%{$VirtualTable_Full{1}{$Class_Type{"Name"}}}))
            {# became polymorphous class, added v-table pointer
                %{$CompatProblems{$Interface}{"Added_First_Virtual_Method"}{$tr_name{$AddedVFunc}}}=(
                    "Type_Name"=>$Class_Type{"Name"},
                    "Type_Type"=>"Class",
                    "Target"=>get_Signature($AddedVFunc, 2)  );
            }
            elsif(isCopyingClass($Class_Type{"Name"}, 1))
            {# class has no constructors and v-table will be copied by applications, it may affect all methods
                my $VSize_Old = getVSize($Class_Type{"Name"}, 1);
                my $VSize_New = getVSize($Class_Type{"Name"}, 2);
                next if($VSize_Old==$VSize_New);# exception: register as removed and added virtual method
                my $ProblemType = "Added_Virtual_Method";
                if(isLeafClass($Class_Id, 1)) {
                    $ProblemType = "Added_Virtual_Method_At_End";
                }
                %{$CompatProblems{$Interface}{$ProblemType}{$tr_name{$AddedVFunc}}}=(
                    "Type_Name"=>$Class_Type{"Name"},
                    "Type_Type"=>"Class",
                    "Target"=>get_Signature($AddedVFunc, 2)  );
            }
            else {
                # safe
            }
        }
        elsif($CompleteSignature{1}{$Interface}{"Virt"}
        or $CompleteSignature{1}{$Interface}{"PureVirt"})
        {
            my $VPos_Old = $VirtualTable{1}{$Class_Type{"Name"}}{$Interface};
            my $VPos_New = $VirtualTable{2}{$Class_Type{"Name"}}{$Interface};
            if($VPos_Added<=$VPos_Old and $VPos_Old!=$VPos_New) {
                my @Affected = ($Interface, keys(%{$OverriddenMethods{1}{$Interface}}));
                foreach my $AffectedInterface (@Affected) {
                    %{$CompatProblems{$AffectedInterface}{"Added_Virtual_Method"}{$tr_name{$AddedVFunc}}}=(
                        "Type_Name"=>$Class_Type{"Name"},
                        "Type_Type"=>"Class",
                        "Target"=>get_Signature($AddedVFunc, 2)  );
                }
            }
        }
        else {
            # safe
        }
    }
    foreach my $RemovedVFunc (keys(%{$RemovedInt_Virt{$Class_Type{"Name"}}}))
    {
        next if($VirtualReplacement{$RemovedVFunc});
        if($RemovedVFunc eq $Interface
        and $CompleteSignature{1}{$RemovedVFunc}{"PureVirt"})
        { # This case is for removed virtual methods
          # implemented in both versions of a library
            next;
        }
        if(not keys(%{$VirtualTable_Full{2}{$Class_Type{"Name"}}}))
        {# became non-polymorphous class, removed v-table pointer
            %{$CompatProblems{$Interface}{"Removed_Last_Virtual_Method"}{$tr_name{$RemovedVFunc}}}=(
                "Type_Name"=>$Class_Type{"Name"},
                "Type_Type"=>"Class",
                "Target"=>get_Signature($RemovedVFunc, 1)  );
        }
        elsif($CompleteSignature{1}{$Interface}{"Virt"}
        or $CompleteSignature{1}{$Interface}{"PureVirt"})
        {
            my $VPos_Removed = $VirtualTable{1}{$Class_Type{"Name"}}{$RemovedVFunc};
            my $VPos_Old = $VirtualTable{1}{$Class_Type{"Name"}}{$Interface};
            my $VPos_New = $VirtualTable{2}{$Class_Type{"Name"}}{$Interface};
            if($VPos_Removed<=$VPos_Old and $VPos_Old!=$VPos_New) {
                my @Affected = ($Interface, keys(%{$OverriddenMethods{1}{$Interface}}));
                foreach my $AffectedInterface (@Affected) {
                    %{$CompatProblems{$AffectedInterface}{"Removed_Virtual_Method"}{$tr_name{$RemovedVFunc}}}=(
                        "Type_Name"=>$Class_Type{"Name"},
                        "Type_Type"=>"Class",
                        "Target"=>get_Signature($RemovedVFunc, 1)  );
                }
            }
        }
    }
}

sub get_func_suffix($)
{
    my $Interface = $_[0];
    my $Signature = $tr_name{$Interface};
    my $Suffix = substr($Signature, get_signature_center($Signature));
    return $Suffix;
}

sub find_MemberPair_Pos_byName($$)
{
    my ($Member_Name, $Pair_Type) = @_;
    $Member_Name=~s/\A[_]+|[_]+\Z//g;
    foreach my $MemberPair_Pos (sort {int($a)<=>int($b)} keys(%{$Pair_Type->{"Memb"}}))
    {
        if(defined $Pair_Type->{"Memb"}{$MemberPair_Pos})
        {
            my $Name = $Pair_Type->{"Memb"}{$MemberPair_Pos}{"name"};
            $Name=~s/\A[_]+|[_]+\Z//g;
            if($Name eq $Member_Name) {
                return $MemberPair_Pos;
            }
        }
    }
    return "lost";
}

sub getBitfieldSum($$)
{
    my ($Member_Pos, $Pair_Type) = @_;
    my $BitfieldSum = 0;
    while($Member_Pos>-1)
    {
        if($Pair_Type->{"Memb"}{$Member_Pos}{"bitfield"})
        {
            $BitfieldSum += $Pair_Type->{"Memb"}{$Member_Pos}{"bitfield"};
            $Member_Pos -= 1;
        }
        else {
            last;
        }
    }
    return $BitfieldSum;
}

sub find_MemberPair_Pos_byVal($$)
{
    my ($Member_Value, $Pair_Type) = @_;
    foreach my $MemberPair_Pos (sort {int($a)<=>int($b)} keys(%{$Pair_Type->{"Memb"}}))
    {
        if(defined $Pair_Type->{"Memb"}{$MemberPair_Pos}
        and $Pair_Type->{"Memb"}{$MemberPair_Pos}{"value"} eq $Member_Value) {
            return $MemberPair_Pos;
        }
    }
    return "lost";
}

my %Priority_Value=(
    "High"=>3,
    "Medium"=>2,
    "Low"=>1);

sub max_priority($$)
{
    my ($Priority1, $Priority2) = @_;
    if(cmp_priority($Priority1, $Priority2)) {
        return $Priority1;
    }
    else {
        return $Priority2;
    }
}

sub cmp_priority($$)
{
    my ($Priority1, $Priority2) = @_;
    return ($Priority_Value{$Priority1}>$Priority_Value{$Priority2});
}

sub set_Problems_Priority()
{
    foreach my $InterfaceName (sort keys(%CompatProblems))
    {
        foreach my $Kind (sort keys(%{$CompatProblems{$InterfaceName}}))
        {
            foreach my $Location (sort keys(%{$CompatProblems{$InterfaceName}{$Kind}}))
            {
                if($TypeProblems_Kind{$Kind}
                and $TypeProblems_Kind{$Kind} ne "Dynamic") {
                    $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = $TypeProblems_Kind{$Kind};
                }
                elsif($InterfaceProblems_Kind{$Kind}
                and $InterfaceProblems_Kind{$Kind} ne "Dynamic") {
                    $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = $InterfaceProblems_Kind{$Kind};
                }
                else {
                    my $IsInTypeInternals = $CompatProblems{$InterfaceName}{$Kind}{$Location}{"IsInTypeInternals"};
                    my $InitialType_Type = $CompatProblems{$InterfaceName}{$Kind}{$Location}{"InitialType_Type"};
                    if($Kind eq "Parameter_BaseType_And_Size")
                    {
                        if($InitialType_Type=~/\A(Pointer|Ref)\Z/) {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Low";
                        }
                        else {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "High";
                        }
                    }
                    elsif($Kind eq "Size")
                    {
                        if($InitialType_Type=~/\A(Pointer|Ref)\Z/) {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Low";
                        }
                        else {
                            if($IsInTypeInternals) {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Medium";
                            }
                            else {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "High";
                            }
                        }
                    }
                    elsif($Kind eq "Removed_Middle_Field_And_Size")
                    {
                        if($IsInTypeInternals) {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Medium";
                        }
                        else {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "High";
                        }
                    }
                    elsif($Kind eq "Field_Type_And_Size" or $Kind eq "Field_Size")
                    {
                        if($IsInTypeInternals) {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Low";
                        }
                        else {
                            if($InitialType_Type=~/\A(Pointer|Ref)\Z/) {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Medium";
                            }
                            else {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "High";
                            }
                        }
                    }
                    elsif($Kind eq "Field_BaseType_And_Size")
                    {
                        if($InitialType_Type=~/\A(Pointer|Ref)\Z/) {
                            $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Low";
                        }
                        else {
                            if($IsInTypeInternals) {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Low";
                            }
                            else {
                                $CompatProblems{$InterfaceName}{$Kind}{$Location}{"Priority"} = "Medium";
                            }
                        }
                    }
                }
            }
        }
    }
}

sub isRecurType($$$$)
{
    foreach (@RecurTypes)
    {
        if($_->{"Tid1"} eq $_[0]
        and $_->{"TDid1"} eq $_[1]
        and $_->{"Tid2"} eq $_[2]
        and $_->{"TDid2"} eq $_[3])
        {
            return 1;
        }
    }
    return 0;
}

sub pushType($$$$)
{
    my %TypeIDs=(
        "Tid1"  => $_[0],
        "TDid1" => $_[1],
        "Tid2"  => $_[2],
        "TDid2" => $_[3]  );
    push(@RecurTypes, \%TypeIDs);
}

sub mergeTypes($$$$)
{
    my ($Type1_Id, $Type1_DId, $Type2_Id, $Type2_DId) = @_;
    return () if((not $Type1_Id and not $Type1_DId) or (not $Type2_Id and not $Type2_DId));
    my (%Sub_SubProblems, %SubProblems) = ();
    if($Cache{"mergeTypes"}{$Type1_Id}{$Type1_DId}{$Type2_Id}{$Type2_DId})
    {# already merged
        return %{$Cache{"mergeTypes"}{$Type1_Id}{$Type1_DId}{$Type2_Id}{$Type2_DId}};
    }
    my %Type1 = get_Type($Type1_DId, $Type1_Id, 1);
    my %Type2 = get_Type($Type2_DId, $Type2_Id, 2);
    my %Type1_Pure = get_PureType($Type1_DId, $Type1_Id, 1);
    my %Type2_Pure = get_PureType($Type2_DId, $Type2_Id, 2);
    return () if(not $Type1_Pure{"Size"} or not $Type2_Pure{"Size"});
    if(isRecurType($Type1_Pure{"Tid"}, $Type1_Pure{"TDid"}, $Type2_Pure{"Tid"}, $Type2_Pure{"TDid"}))
    {# skip recursive declarations
        return ();
    }
    return () if(not $Type1_Pure{"Name"} or not $Type2_Pure{"Name"});
    return () if($OpaqueTypes{1}{$Type1_Pure{"Name"}});
    return () if($OpaqueTypes{1}{$Type1{"Name"}});
    
    my %Typedef_1 = goToFirst($Type1{"TDid"}, $Type1{"Tid"}, 1, "Typedef");
    my %Typedef_2 = goToFirst($Type2{"TDid"}, $Type2{"Tid"}, 2, "Typedef");
    if(not $UseOldDumps
    and $Typedef_1{"Type"} eq "Typedef"
    and $Typedef_2{"Type"} eq "Typedef"
    and $Typedef_1{"Name"} eq $Typedef_2{"Name"})
    {
        my %Base_1 = get_OneStep_BaseType($Typedef_1{"TDid"}, $Typedef_1{"Tid"}, 1);
        my %Base_2 = get_OneStep_BaseType($Typedef_2{"TDid"}, $Typedef_2{"Tid"}, 2);
        if($Base_1{"Name"}!~/anon\-/ and $Base_2{"Name"}!~/anon\-/
        and $Base_1{"Name"} ne $Base_2{"Name"})
        {
            %{$SubProblems{"Typedef_BaseType"}{$Typedef_1{"Name"}}}=(
                "Target"=>$Typedef_1{"Name"},
                "Type_Name"=>$Typedef_1{"Name"},
                "Type_Type"=>"Typedef",
                "Old_Value"=>$Base_1{"Name"},
                "New_Value"=>$Base_2{"Name"}  );
        }
    }
    if($Type1_Pure{"Name"} ne $Type2_Pure{"Name"}
    and ($Type1_Pure{"Type"} ne "Pointer" or $Type2_Pure{"Type"} ne "Pointer"
    or $Type1_Pure{"Name"}!~/\Avoid\s*\*/ and $Type2_Pure{"Name"}=~/\Avoid\s*\*/)
    and $Type1_Pure{"Name"}!~/anon\-/ and $Type2_Pure{"Name"}!~/anon\-/)
    {# different types
        %{$Cache{"mergeTypes"}{$Type1_Id}{$Type1_DId}{$Type2_Id}{$Type2_DId}} = %SubProblems;
        return %SubProblems;
    }
    pushType($Type1_Pure{"Tid"}, $Type1_Pure{"TDid"},
             $Type2_Pure{"Tid"}, $Type2_Pure{"TDid"});
    if($Type1_Pure{"Name"} eq $Type2_Pure{"Name"}
    and $Type1_Pure{"Type"}=~/\A(Struct|Class|Union)\Z/)
    {# checking size
        if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
        {
            my $ProblemKind = "Size";
            if($Type1_Pure{"Type"} eq "Class"
            and keys(%{$ClassMethods{1}{$Type1_Pure{"Name"}}}))
            {
                if(isAllocableClass($Type1_Pure{"Name"}, 1))
                {
                    if(int($Type2_Pure{"Size"})>int($Type1_Pure{"Size"})) {
                        $ProblemKind = "Size_Of_Allocable_Class_Increased";
                    }
                    else {
                        # descreased size of allocable class
                        # it has no special effects
                    }
                }
                elsif(isCopyingClass($Type1_Pure{"Name"}, 1)) {
                    $ProblemKind = "Size_Of_Copying_Class";
                }
            }
            %{$SubProblems{$ProblemKind}{$Type1_Pure{"Name"}}}=(
                "Target"=>$Type1_Pure{"Name"},
                "Type_Name"=>$Type1_Pure{"Name"},
                "Type_Type"=>$Type1_Pure{"Type"},
                "Old_Value"=>$Type1_Pure{"Size"},
                "New_Value"=>$Type2_Pure{"Size"},
                "InitialType_Type"=>$Type1_Pure{"Type"});
        }
    }
    if($Type1_Pure{"BaseType"}{"Tid"} and $Type2_Pure{"BaseType"}{"Tid"})
    {# checking base types
        %Sub_SubProblems = mergeTypes($Type1_Pure{"BaseType"}{"Tid"}, $Type1_Pure{"BaseType"}{"TDid"},
                                       $Type2_Pure{"BaseType"}{"Tid"}, $Type2_Pure{"BaseType"}{"TDid"});
        foreach my $Sub_SubProblemType (keys(%Sub_SubProblems))
        {
            foreach my $Sub_SubLocation (keys(%{$Sub_SubProblems{$Sub_SubProblemType}}))
            {
                foreach my $Attr (keys(%{$Sub_SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}})) {
                    $SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}{$Attr} = $Sub_SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}{$Attr};
                }
                $SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}{"InitialType_Type"} = $Type1_Pure{"Type"};
            }
        }
    }
    foreach my $Member_Pos (sort {int($a) <=> int($b)} keys(%{$Type1_Pure{"Memb"}}))
    {# checking older members
        next if($Type1_Pure{"Memb"}{$Member_Pos}{"access"} eq "private");
        my $Member_Name = $Type1_Pure{"Memb"}{$Member_Pos}{"name"};
        next if(not $Member_Name);
        my $Member_Location = $Member_Name;
        my $MemberPair_Pos = (defined $Type2_Pure{"Memb"}{$Member_Pos} and $Type2_Pure{"Memb"}{$Member_Pos}{"name"} eq $Member_Name)?$Member_Pos:find_MemberPair_Pos_byName($Member_Name, \%Type2_Pure);
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"}=~/\A(Struct|Class|Union)\Z/) {
            if($Member_Name=~/\Aunnamed\d+\Z/)
            { # support for old-version dumps
              # unnamed fields have been introduced in the ACC 1.23 (dump 2.1 format)
                if($UsedDump{2}
                and cmpVersions($UsedDump{2}, "2.1")<0) {
                    next;
                }
            }
        }
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"}=~/\A(Struct|Class)\Z/)
        {# removed members
            if($Member_Pos > keys(%{$Type2_Pure{"Memb"}}) - 1)
            {# removed from the end of structure
                if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
                {
                    %{$SubProblems{"Removed_Field_And_Size"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"},
                        "Old_Size"=>$Type1_Pure{"Size"},
                        "New_Size"=>$Type2_Pure{"Size"}  );
                }
                elsif(isReserved($Member_Name))
                {# skip removed reserved members
                    next;
                }
                else
                {
                    %{$SubProblems{"Removed_Field"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"}  );
                }
                next;
            }
            elsif($Member_Pos < keys(%{$Type1_Pure{"Memb"}}) - 1)
            {
                my $MemberType_Id = $Type1_Pure{"Memb"}{$Member_Pos}{"type"};
                my %MemberType_Pure = get_PureType($Tid_TDid{1}{$MemberType_Id}, $MemberType_Id, 1);
                
                my $MemberStraightPairType_Id = $Type2_Pure{"Memb"}{$Member_Pos}{"type"};
                my %MemberStraightPairType_Pure = get_PureType($Tid_TDid{2}{$MemberStraightPairType_Id}, $MemberStraightPairType_Id, 2);
                
                my $MemberStraightPair_Name = $Type2_Pure{"Memb"}{$Member_Pos}{"name"};
                my $MemberPair_Pos_Rev = ($Member_Name eq $MemberStraightPair_Name)?$Member_Pos:find_MemberPair_Pos_byName($MemberStraightPair_Name, \%Type1_Pure);
                if($MemberPair_Pos_Rev eq "lost" and ($MemberType_Pure{"Name"} eq $MemberStraightPairType_Pure{"Name"}
                or get_TypeName($MemberType_Id, 1) eq get_TypeName($MemberStraightPairType_Id, 2)
                or $MemberType_Pure{"Size"} eq $MemberStraightPairType_Pure{"Size"}))
                {
                    if(isReserved($Member_Name))
                    {
                        %{$SubProblems{"Used_Reserved_Field"}{$Member_Name}}=(
                            "Target"=>$MemberStraightPair_Name,
                            "Reserved"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"} );
                    }
                    else
                    {
                        %{$SubProblems{"Renamed_Field"}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"},
                            "Old_Value"=>$Member_Name,
                            "New_Value"=>$MemberStraightPair_Name  );
                    }
                    $MemberPair_Pos = $Member_Pos;
                }
                else
                {
                    if($Type1_Pure{"Memb"}{$Member_Pos}{"bitfield"})
                    {
                        my $BitfieldSum = getBitfieldSum($Member_Pos-1, \%Type1_Pure)%($POINTER_SIZE{1}*8);
                        if($BitfieldSum and $BitfieldSum-$Type2_Pure{"Memb"}{$Member_Pos}{"bitfield"}>=0)
                        {# bitfields of any type are "word"-aligned
                            %{$SubProblems{"Removed_Middle_Field"}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"}  );
                            next;
                        }
                    }
                    if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
                    {
                        %{$SubProblems{"Removed_Middle_Field_And_Size"}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"}  );
                    }
                    elsif(isReserved($Member_Name))
                    {# skip removed reserved members
                        next;
                    }
                    else
                    {
                        %{$SubProblems{"Removed_Middle_Field"}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"}  );
                    }
                    next;
                }
            }
        }
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"} eq "Union")
        {
            if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
            {
                %{$SubProblems{"Removed_Union_Field_And_Size"}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"}  );
            }
            else
            {
                %{$SubProblems{"Removed_Union_Field"}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"}  );
            }
        }
        if($Type1_Pure{"Type"} eq "Enum")
        {
            my $Member_Value1 = $Type1_Pure{"Memb"}{$Member_Pos}{"value"};
            next if(not $Member_Value1);
            my $Member_Value2 = $Type2_Pure{"Memb"}{$MemberPair_Pos}{"value"};
            if($MemberPair_Pos eq "lost")
            {
                $MemberPair_Pos = find_MemberPair_Pos_byVal($Member_Value1, \%Type2_Pure);
                if($MemberPair_Pos eq "lost")
                {
                    %{$SubProblems{"Enum_Member_Removed"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"},
                        "Old_Value"=>$Member_Name  );
                }
                else
                {
                    %{$SubProblems{"Enum_Member_Name"}{$Type1_Pure{"Memb"}{$Member_Pos}{"value"}}}=(
                        "Target"=>$Type1_Pure{"Memb"}{$Member_Pos}{"value"},
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"},
                        "Old_Value"=>$Member_Name,
                        "New_Value"=>$Type2_Pure{"Memb"}{$MemberPair_Pos}{"name"}  );
                }
            }
            else
            {
                if($Member_Value1 ne "" and $Member_Value2 ne "")
                {
                    if($Member_Value1 ne $Member_Value2)
                    {
                        my $ProblemType = "Enum_Member_Value";
                        if($Member_Name=~/last|count/i) {
                            # GST_LEVEL_COUNT, GST_RTSP_ELAST
                            $ProblemType = "Enum_Last_Member_Value";
                        }
                        %{$SubProblems{$ProblemType}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"},
                            "Old_Value"=>$Type1_Pure{"Memb"}{$Member_Pos}{"value"},
                            "New_Value"=>$Type2_Pure{"Memb"}{$MemberPair_Pos}{"value"}  );
                    }
                }
            }
            next;
        }
        my $MemberType1_Id = $Type1_Pure{"Memb"}{$Member_Pos}{"type"};
        my $MemberType2_Id = $Type2_Pure{"Memb"}{$MemberPair_Pos}{"type"};
        my $MemberType1_Name = get_TypeName($MemberType1_Id, 1);
        my $MemberType2_Name = get_TypeName($MemberType2_Id, 2);
        if(not $Type1_Pure{"Memb"}{$Member_Pos}{"bitfield"}
        and not $Type2_Pure{"Memb"}{$MemberPair_Pos}{"bitfield"})
        { # GCC 4.2.1 treats guint bitfield as the 1-bit "unsigned int" field in the TU dump
          # but GCC 3.4.4 treats it as the 4-byte typedef to "unsigned int" (with "unql" attribute)
            %Sub_SubProblems = detectTypeChange($MemberType1_Id, $MemberType2_Id, "Field");
            foreach my $Sub_SubProblemType (keys(%Sub_SubProblems))
            {
                if(isReserved($Member_Name))
                { # skip changed reserved members
                  # protobuf-c 0.13 to 0.14
                    if($Sub_SubProblemType eq "Field_Type_And_Size")
                    {# size of reserved field changed
                        if($Type1_Pure{"Size"} eq $Type2_Pure{"Size"}
                        and $Member_Pos==keys(%{$Type1_Pure{"Memb"}})-1)
                        {# reserved field at end of structure
                            next;
                        }
                        # skip changed reserved unions
                        next if($Type1_Pure{"Name"}=~/\Aanon\-union/);
                    }
                    elsif($Sub_SubProblemType eq "Field_Type"
                    or $Sub_SubProblemType=~/Field_PointerLevel_(Increased|Decreased)/)
                    {# type name of reserved field changed
                        next;
                    }
                }
                %{$SubProblems{$Sub_SubProblemType}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Member_Position"=>$Member_Pos,
                    "Max_Position"=>keys(%{$Type1_Pure{"Memb"}})-1,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"}  );
                foreach my $Attr (keys(%{$Sub_SubProblems{$Sub_SubProblemType}})) {
                    $SubProblems{$Sub_SubProblemType}{$Member_Name}{$Attr} = $Sub_SubProblems{$Sub_SubProblemType}{$Attr};
                }
            }
        }
        if($MemberType1_Id and $MemberType2_Id)
        {# checking member type changes (replace)
            %Sub_SubProblems = mergeTypes($MemberType1_Id, $Tid_TDid{1}{$MemberType1_Id},
                                          $MemberType2_Id, $Tid_TDid{2}{$MemberType2_Id});
            foreach my $Sub_SubProblemType (keys(%Sub_SubProblems))
            {
                foreach my $Sub_SubLocation (keys(%{$Sub_SubProblems{$Sub_SubProblemType}}))
                {
                    my $NewLocation = ($Sub_SubLocation)?$Member_Location."->".$Sub_SubLocation:$Member_Location;
                    $SubProblems{$Sub_SubProblemType}{$NewLocation}{"IsInTypeInternals"}=1;
                    foreach my $Attr (keys(%{$Sub_SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}})) {
                        $SubProblems{$Sub_SubProblemType}{$NewLocation}{$Attr} = $Sub_SubProblems{$Sub_SubProblemType}{$Sub_SubLocation}{$Attr};
                    }
                    if($Sub_SubLocation!~/\-\>/)
                    {
                        $SubProblems{$Sub_SubProblemType}{$NewLocation}{"Member_Type_Name"} = $MemberType1_Name;
                        $SubProblems{$Sub_SubProblemType}{$NewLocation}{"Start_Type_Name"} = $MemberType1_Name;
                    }
                }
            }
        }
        my $SizeV1 = get_TypeSize($MemberType1_Id, 1);
        if(my $BSize1 = $Type1_Pure{"Memb"}{$Member_Pos}{"bitfield"})
        { # support for GCC 3.4.4
          # bitfields have 4-byte "unsigned int" type
            $SizeV1 = $BSize1;
        }
        my $SizeV2 = get_TypeSize($MemberType2_Id, 2);
        if(my $BSize2 = $Type2_Pure{"Memb"}{$MemberPair_Pos}{"bitfield"})
        { # support for GCC 3.4.4
          # bitfields have 4-byte "unsigned int" type
            $SizeV2 = $BSize2;
        }
        if($SizeV1 ne $SizeV2)
        {
            if(($MemberType1_Name=~/\Aanon\-/ and $MemberType2_Name=~/\Aanon\-/) or $MemberType1_Name eq $MemberType2_Name
            or ($Type1_Pure{"Memb"}{$Member_Pos}{"bitfield"} and $Type2_Pure{"Memb"}{$MemberPair_Pos}{"bitfield"}))
            { # member type size change (including anon-structures and unions)
              # - same types
              # - unnamed types
              # - bitfields
                my $ProblemType = "Field_Size";
                if(not isAccessible(\%Type1_Pure, $Member_Pos+1, -1))
                {# structure has no public members after the current
                    if(not isPublic(\%Type1_Pure, $Member_Pos)) {
                        # should not be accessed by applications, goes to "Low Severity"
                        # example: "abidata" members in GStreamer types
                        $ProblemType = "Private_Field_Size";
                    }
                    elsif($Type1_Pure{"Size"} eq $Type2_Pure{"Size"}
                    and havePadding(\%Type1_Pure)) {
                        # public member before reserved member
                        $ProblemType = "Reserved_Field_Size";
                    }
                }
                if($Type1_Pure{"Memb"}{$Member_Pos}{"bitfield"}) {
                    $ProblemType = "BitField_Size";
                }
                %{$SubProblems{$ProblemType}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Member_Position"=>$Member_Pos,
                    "Max_Position"=>keys(%{$Type1_Pure{"Memb"}})-1,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"},
                    "Old_Size"=>$SizeV1,
                    "New_Size"=>$SizeV2);
            }
        }
    }
    foreach my $Member_Pos (sort {int($a) <=> int($b)} keys(%{$Type2_Pure{"Memb"}}))
    {# checking added members, public and private
        my $Member_Name = $Type2_Pure{"Memb"}{$Member_Pos}{"name"};
        next if(not $Member_Name);
        my $MemberPair_Pos = (defined $Type1_Pure{"Memb"}{$Member_Pos} and $Type1_Pure{"Memb"}{$Member_Pos}{"name"} eq $Member_Name)?$Member_Pos:find_MemberPair_Pos_byName($Member_Name, \%Type1_Pure);
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"}=~/\A(Struct|Class|Union)\Z/) {
            if($Member_Name=~/\Aunnamed\d+\Z/)
            { # support for old-version dumps
              # unnamed fields have been introduced in the ACC 1.23 (dump 2.1 format)
                if($UsedDump{1}
                and cmpVersions($UsedDump{1}, "2.1")<0) {
                    next;
                }
            }
        }
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"}=~/\A(Struct|Class)\Z/)
        {# added members
            if($Member_Pos > keys(%{$Type1_Pure{"Memb"}}) - 1)
            {# added at the end of structure
                if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
                {
                    if($Type2_Pure{"Memb"}{$Member_Pos}{"bitfield"})
                    {# bitfields are "word"-aligned
                        my $BitfieldSum = getBitfieldSum($Member_Pos-1, \%Type2_Pure)%($POINTER_SIZE{2}*8);
                        next if($BitfieldSum and $BitfieldSum<=$POINTER_SIZE{2}*8-$Type2_Pure{"Memb"}{$Member_Pos}{"bitfield"});
                    }
                    %{$SubProblems{"Added_Field_And_Size"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"}  );
                }
                elsif(havePadding(\%Type1_Pure))
                {# added member before the reserved member
                    %{$SubProblems{"Added_Reserved_Field"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Reserved"=>$Type1_Pure{"Memb"}{keys(%{$Type1_Pure{"Memb"}})-1}{"name"},
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"}  );
                }
                else
                {
                    %{$SubProblems{"Added_Field"}{$Member_Name}}=(
                        "Target"=>$Member_Name,
                        "Type_Name"=>$Type1_Pure{"Name"},
                        "Type_Type"=>$Type1_Pure{"Type"}  );
                    # FIXME: mysql 1.7.40 to 2.0.0 change in the data type mysqlpp::Query
                    # add rules to check for compatibility
                }
            }
            else
            {
                my $MemberType_Id = $Type2_Pure{"Memb"}{$Member_Pos}{"type"};
                my %MemberType_Pure = get_PureType($Tid_TDid{2}{$MemberType_Id}, $MemberType_Id, 2);
                
                my $MemberStraightPairType_Id = $Type1_Pure{"Memb"}{$Member_Pos}{"type"};
                my %MemberStraightPairType_Pure = get_PureType($Tid_TDid{1}{$MemberStraightPairType_Id}, $MemberStraightPairType_Id, 1);
                
                my $MemberStraightPair_Name = $Type1_Pure{"Memb"}{$Member_Pos}{"name"};
                my $MemberPair_Pos_Rev = ($Member_Name eq $MemberStraightPair_Name)?$Member_Pos:find_MemberPair_Pos_byName($MemberStraightPair_Name, \%Type2_Pure);
                if($MemberPair_Pos_Rev eq "lost" and ($MemberType_Pure{"Name"} eq $MemberStraightPairType_Pure{"Name"}
                or get_TypeName($MemberType_Id, 2) eq get_TypeName($MemberStraightPairType_Id, 1)
                or $MemberType_Pure{"Size"} eq $MemberStraightPairType_Pure{"Size"}))
                {
                    next if($Type1_Pure{"Memb"}{$Member_Pos}{"access"} eq "private");
                    if(isReserved($MemberStraightPair_Name))
                    {
                        %{$SubProblems{"Used_Reserved_Field"}{$Member_Name}}=(
                            "Target"=>$Member_Name,
                            "Reserved"=>$MemberStraightPair_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"}  );
                    }
                    else
                    {
                        %{$SubProblems{"Renamed_Field"}{$Member_Name}}=(
                            "Target"=>$MemberStraightPair_Name,
                            "Type_Name"=>$Type1_Pure{"Name"},
                            "Type_Type"=>$Type1_Pure{"Type"},
                            "Old_Value"=>$MemberStraightPair_Name,
                            "New_Value"=>$Member_Name  );
                    }
                }
                else
                {
                    if($Type2_Pure{"Memb"}{$Member_Pos}{"bitfield"})
                    {# bitfields are "word"-aligned
                        my $BitfieldSum = getBitfieldSum($Member_Pos-1, \%Type2_Pure)%($POINTER_SIZE{2}*8);
                        next if($BitfieldSum and $BitfieldSum<=$POINTER_SIZE{2}*8-$Type2_Pure{"Memb"}{$Member_Pos}{"bitfield"});
                    }
                    if($Type1_Pure{"Size"} eq $Type2_Pure{"Size"})
                    {
                        if(keys(%{$SubProblems{"Removed_Field"}})
                        or keys(%{$SubProblems{"Removed_Middle_Field"}}))
                        {# FIXME: check this case accurately, check the base classes structure
                            %{$SubProblems{"Added_Middle_Field_And_Size"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                        }
                        elsif(defined $Type1_Pure{"Memb"}{$Member_Pos}
                        and $MemberPair_Pos_Rev eq "lost"
                        and isReserved($MemberStraightPair_Name))
                        {# added member instead of reserved member ( example: void (*_g_reservedN) (void) )
                            %{$SubProblems{"Used_Reserved_Field"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Reserved"=>$MemberStraightPair_Name,
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                        }
                        elsif(havePadding(\%Type1_Pure))
                        {# added member before the reserved member with flexible type size ( example: gpointer padding[N] )
                            %{$SubProblems{"Added_Reserved_Field"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Reserved"=>$Type1_Pure{"Memb"}{keys(%{$Type1_Pure{"Memb"}})-1}{"name"},
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                        }
                        else
                        {
                            %{$SubProblems{"Added_Middle_Field"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                            # FIXME: mysql 1.7.40 to 2.0.0 change in the data type mysqlpp::Query
                            # add rules to check for compatibility
                        }
                    }
                    else
                    {
                        if($MemberPair_Pos_Rev eq "lost"
                        and isReserved($MemberStraightPair_Name))
                        {# added member instead of reserved member
                            # example in gst-plugins-base: union abidata instead of _gst_reserved[GST_PADDING]
                            %{$SubProblems{"Used_Reserved_Field"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Reserved"=>$MemberStraightPair_Name,
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                        }
                        else
                        {
                            %{$SubProblems{"Added_Middle_Field_And_Size"}{$Member_Name}}=(
                                "Target"=>$Member_Name,
                                "Type_Name"=>$Type1_Pure{"Name"},
                                "Type_Type"=>$Type1_Pure{"Type"}  );
                        }
                    }
                }
            }
        }
        if($MemberPair_Pos eq "lost"
        and $Type2_Pure{"Type"} eq "Union")
        {
            if($Type1_Pure{"Size"} ne $Type2_Pure{"Size"})
            {
                %{$SubProblems{"Added_Union_Field_And_Size"}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"}  );
            }
            else
            {
                %{$SubProblems{"Added_Union_Field"}{$Member_Name}}=(
                    "Target"=>$Member_Name,
                    "Type_Name"=>$Type1_Pure{"Name"},
                    "Type_Type"=>$Type1_Pure{"Type"}  );
            }
        }
    }
    %{$Cache{"mergeTypes"}{$Type1_Id}{$Type1_DId}{$Type2_Id}{$Type2_DId}} = %SubProblems;
    pop(@RecurTypes);
    return %SubProblems;
}

sub get_TypeName($$)
{
    my ($TypeId, $LibVersion) = @_;
    return $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"Name"};
}

sub get_ClassVTable($$)
{
    my ($ClassId, $LibVersion) = @_;
    return $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$ClassId}}{$ClassId}{"VTable"};
}

sub get_ShortName($$)
{
    my ($TypeId, $LibVersion) = @_;
    my $TypeName = $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"Name"};
    my $NameSpace = $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"NameSpace"};
    $TypeName=~s/\A$NameSpace\:\://g;
    return $TypeName;
}

sub get_TypeType($$)
{
    my ($TypeId, $LibVersion) = @_;
    return $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"Type"};
}

sub get_TypeSize($$)
{
    my ($TypeId, $LibVersion) = @_;
    return $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"Size"};
}

sub get_TypeHeader($$)
{
    my ($TypeId, $LibVersion) = @_;
    return $TypeInfo{$LibVersion}{$Tid_TDid{$LibVersion}{$TypeId}}{$TypeId}{"Header"};
}

sub goToFirst($$$$)
{
    my ($TypeDId, $TypeId, $LibVersion, $Type_Type) = @_;
    if(defined $Cache{"goToFirst"}{$TypeDId}{$TypeId}{$LibVersion}{$Type_Type}) {
        return %{$Cache{"goToFirst"}{$TypeDId}{$TypeId}{$LibVersion}{$Type_Type}};
    }
    return () if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    return () if(not $Type{"Type"});
    if($Type{"Type"} ne $Type_Type)
    {
        return () if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
        %Type = goToFirst($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion, $Type_Type);
    }
    $Cache{"goToFirst"}{$TypeDId}{$TypeId}{$LibVersion}{$Type_Type} = \%Type;
    return %Type;
}

my %TypeSpecAttributes = (
    "Const" => 1,
    "Volatile" => 1,
    "Restrict" => 1,
    "Typedef" => 1
);

sub get_PureType($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return "" if(not $TypeId);
    if(defined $Cache{"get_PureType"}{$TypeDId}{$TypeId}{$LibVersion}) {
        return %{$Cache{"get_PureType"}{$TypeDId}{$TypeId}{$LibVersion}};
    }
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    if($Type{"Type"}=~/\A(Struct|Union|Typedef|Class|Enum)\Z/) {
        $CheckedTypes{$Type{"Name"}} = 1;
    }
    return %Type if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
    if($TypeSpecAttributes{$Type{"Type"}}) {
        %Type = get_PureType($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
    }
    $Cache{"get_PureType"}{$TypeDId}{$TypeId}{$LibVersion} = \%Type;
    return %Type;
}

sub get_PointerLevel($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return 0 if(not $TypeId);
    if(defined $Cache{"get_PointerLevel"}{$TypeDId}{$TypeId}{$LibVersion}) {
        return $Cache{"get_PointerLevel"}{$TypeDId}{$TypeId}{$LibVersion};
    }
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    return 1 if($Type{"Type"}=~/FuncPtr|MethodPtr|FieldPtr/);
    return 0 if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
    my $PointerLevel = 0;
    if($Type{"Type"} =~/Pointer|Ref|FuncPtr|MethodPtr|FieldPtr/) {
        $PointerLevel += 1;
    }
    $PointerLevel += get_PointerLevel($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
    $Cache{"get_PointerLevel"}{$TypeDId}{$TypeId}{$LibVersion} = $PointerLevel;
    return $PointerLevel;
}

sub get_BaseType($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return "" if(not $TypeId);
    if(defined $Cache{"get_BaseType"}{$TypeDId}{$TypeId}{$LibVersion}) {
        return %{$Cache{"get_BaseType"}{$TypeDId}{$TypeId}{$LibVersion}};
    }
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    if($Type{"Type"}=~/\A(Struct|Union|Typedef|Class|Enum)\Z/) {
        $CheckedTypes{$Type{"Name"}} = 1;
    }
    return %Type if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
    %Type = get_BaseType($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
    $Cache{"get_BaseType"}{$TypeDId}{$TypeId}{$LibVersion} = \%Type;
    return %Type;
}

sub get_BaseTypeQual($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return "" if(not $TypeId);
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    return  if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
    my $Qual = "";
    if($Type{"Type"} eq "Const") {
        $Qual .= "const";
    }
    elsif($Type{"Type"} eq "Pointer") {
        $Qual .= "*";
    }
    elsif($Type{"Type"} eq "Ref") {
        $Qual .= "&";
    }
    elsif($Type{"Type"} eq "Volatile") {
        $Qual .= "volatile";
    }
    elsif($Type{"Type"} eq "Restrict") {
        $Qual .= "restrict";
    }
    my $BQual = get_BaseTypeQual($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
    return $BQual.$Qual;
}

sub get_OneStep_BaseType($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return "" if(not $TypeId);
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    my %Type = %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
    return %Type if(not $Type{"BaseType"}{"TDid"} and not $Type{"BaseType"}{"Tid"});
    return get_Type($Type{"BaseType"}{"TDid"}, $Type{"BaseType"}{"Tid"}, $LibVersion);
}

sub get_Type($$$)
{
    my ($TypeDId, $TypeId, $LibVersion) = @_;
    return "" if(not $TypeId);
    return "" if(not $TypeInfo{$LibVersion}{$TypeDId}{$TypeId});
    return %{$TypeInfo{$LibVersion}{$TypeDId}{$TypeId}};
}

sub is_template_instance($$)
{
    my ($Interface, $LibVersion) = @_;
    return 0 if($Interface!~/\A(_Z|\?)/);
    my $Signature = get_SignatureNoInfo($Interface, $LibVersion);
    return 0 if($Signature!~/>/);
    my $ShortName = substr($Signature, 0, get_signature_center($Signature));
    $ShortName=~s/::operator .*//;# class::operator template<instance>
    return ($ShortName=~/>/);
}

sub skipGlobalData($)
{
    my $Symbol = $_[0];
    return ($Symbol=~/\A(_ZGV|_ZTI|_ZTS|_ZTT|_ZTV|_ZTC|_ZThn|_ZTv0_n)/);
}

sub interfaceFilter($$$)
{# some special cases when the symbol cannot be imported
    my ($Interface, $LibVersion, $Type) = @_;
    if(skipGlobalData($Interface))
    {# non-public global data
        return 0;
    }
    if($CheckObjectsOnly) {
        return 0 if($Interface=~/\A(_init|_fini)\Z/);
    }
    if($Type eq "Imported")
    {
        if(not $STDCXX_TESTING and $Interface=~/\A(_ZS|_ZNS|_ZNKS)/)
        {# stdc++ interfaces
            return 0;
        }
        if($CompleteSignature{$LibVersion}{$Interface}{"InLine"}
        and not $CompleteSignature{$LibVersion}{$Interface}{"Virt"})
        {# inline non-virtual methods
            return 0;
        }
        if(is_template_instance($Interface, $LibVersion))
        {
            if(defined $CompleteSignature{$LibVersion}{$Interface}
            and ($CompleteSignature{$LibVersion}{$Interface}{"Virt"}
            or $CompleteSignature{$LibVersion}{$Interface}{"PureVirt"}))
            {
                # virtual (pure) template instances can be imported
            }
            else {
              # cannot be imported (will be copied)
              return 0;
            }
        }
        if($SkipSymbols{$LibVersion}{$Interface})
        {# user defined symbols to ignore
            return 0;
        }
        if($InterfacesListPath and not $InterfacesList{$Interface})
        {# user defined symbols to accept
            return 0;
        }
        if($AppPath and not $InterfacesList_App{$Interface})
        {# user defined application
            return 0;
        }
    }
    return 1;
}

sub mergeImplementations()
{
    my $DiffCmd = get_CmdPath("diff");
    if(not $DiffCmd) {
        exitStatus("Not_Found", "can't find \"diff\"");
    }
    foreach my $Interface (sort keys(%{$Symbol_Library{1}}))
    {# implementation changes
        next if($CompleteSignature{1}{$Interface}{"Private"});
        next if(not $CompleteSignature{1}{$Interface}{"Header"} and not $CheckObjectsOnly);
        next if(not $Symbol_Library{2}{$Interface} and not $Symbol_Library{2}{$SymVer{2}{$Interface}});
        next if(not interfaceFilter($Interface, 1, "Imported"));
        my $Impl1 = canonify_implementation($Interface_Implementation{1}{$Interface});
        next if(not $Impl1);
        my $Impl2 = canonify_implementation($Interface_Implementation{2}{$Interface});
        next if(not $Impl2);
        if($Impl1 ne $Impl2)
        {
            writeFile("$TMP_DIR/impl1", $Impl1);
            writeFile("$TMP_DIR/impl2", $Impl2);
            my $Diff = `$DiffCmd -rNau $TMP_DIR/impl1 $TMP_DIR/impl2`;
            $Diff=~s/(---|\+\+\+).+\n//g;
            $Diff=~s/[ ]{3,}/ /g;
            $Diff=~s/\n\@\@/\n \n\@\@/g;
            unlink("$TMP_DIR/impl1", "$TMP_DIR/impl2");
            %{$ImplProblems{$Interface}}=(
                "Diff" => get_CodeView($Diff)  );
        }
    }
}

sub canonify_implementation($)
{
    my $FuncBody=  $_[0];
    return "" if(not $FuncBody);
    $FuncBody=~s/0x[a-f\d]+/0x?/g;# addr
    $FuncBody=~s/((\A|\n)[a-z]+[\t ]+)[a-f\d]+([^x]|\Z)/$1?$3/g;# call, jump
    $FuncBody=~s/# [a-f\d]+ /# ? /g;# call, jump
    $FuncBody=~s/%([a-z]+[a-f\d]*)/\%reg/g;# registers
    while($FuncBody=~s/\nnop[ \t]*(\n|\Z)/$1/g){};# empty op
    $FuncBody=~s/<.+?\.cpp.+?>/<name.cpp>/g;
    $FuncBody=~s/(\A|\n)[a-f\d]+ </$1? </g;# 5e74 <_ZN...
    $FuncBody=~s/\.L\d+/.L/g;
    $FuncBody=~s/#(-?)\d+/#$1?/g;# r3, [r3, #120]
    $FuncBody=~s/[\n]{2,}/\n/g;
    return $FuncBody;
}

sub get_CodeView($)
{
    my $Code = $_[0];
    my $View = "";
    foreach my $Line (split(/\n/, $Code))
    {
        if($Line=~s/\A(\+|-)/$1 /g)
        {# bold line
            $View .= "<tr><td><b>".htmlSpecChars($Line)."</b></td></tr>\n";
        }
        else {
            $View .= "<tr><td>".htmlSpecChars($Line)."</td></tr>\n";
        }
    }
    return "<table class='code_view'>$View</table>\n";
}

sub getImplementations($$)
{
    my ($LibVersion, $Path) = @_;
    return if(not $LibVersion or not -e $Path);
    if($OSgroup eq "macos")
    {
        my $OtoolCmd = get_CmdPath("otool");
        if(not $OtoolCmd) {
            exitStatus("Not_Found", "can't find \"otool\"");
        }
        my $CurInterface = "";
        foreach my $Line (split(/\n/, `$OtoolCmd -tv $Path 2>$TMP_DIR/null`))
        {
            if($Line=~/\A\s*_(\w+)\s*:/i) {
                $CurInterface = $1;
            }
            elsif($Line=~/\A\s*[\da-z]+\s+(.+?)\Z/i) {
                $Interface_Implementation{$LibVersion}{$CurInterface} .= "$1\n";
            }
        }
    }
    else
    {
        my $ObjdumpCmd = get_CmdPath("objdump");
        if(not $ObjdumpCmd) {
            exitStatus("Not_Found", "can't find \"objdump\"");
        }
        my $CurInterface = "";
        foreach my $Line (split(/\n/, `$ObjdumpCmd -d $Path 2>$TMP_DIR/null`))
        {
            if($Line=~/\A[\da-z]+\s+<(\w+)>/i) {
                $CurInterface = $1;
            }
            else
            { # x86:    51fa:(\t)89 e5               (\t)mov    %esp,%ebp
              # arm:    5020:(\t)e24cb004(\t)sub(\t)fp, ip, #4(\t); 0x4
                if($Line=~/\A\s*[a-f\d]+:\s+([a-f\d]+\s+)+([a-z]+\s+.*?)\s*(;.*|)\Z/i) {
                    $Interface_Implementation{$LibVersion}{$CurInterface} .= "$2\n";
                }
            }
        }
    }
}

sub mergeLibs()
{
    foreach my $Interface (sort keys(%AddedInt))
    {# checking added symbols
        next if($CompleteSignature{2}{$Interface}{"Private"});
        next if(not $CompleteSignature{2}{$Interface}{"Header"} and not $CheckObjectsOnly);
        next if(not interfaceFilter($Interface, 2, "Imported"));
        %{$CompatProblems{$Interface}{"Added_Interface"}{""}}=();
    }
    foreach my $Interface (sort keys(%RemovedInt))
    {# checking removed symbols
        next if($CompleteSignature{1}{$Interface}{"Private"});
        next if(not $CompleteSignature{1}{$Interface}{"Header"} and not $CheckObjectsOnly);
        if($Interface=~/\A_ZTV/) {
            # skip v-tables for templates, that should not be imported by applications
            next if($tr_name{$Interface}=~/</);
            if(not keys(%{$ClassMethods{1}{$VTableClass{$Interface}}}))
            { # vtables for "private" classes
              # use case: vtable for QDragManager (Qt 4.5.3 to 4.6.0) became HIDDEN symbol
                next;
            }
        }
        else {
            next if(not interfaceFilter($Interface, 1, "Imported"));
        }
        if($CompleteSignature{1}{$Interface}{"PureVirt"})
        { # symbols for pure virtual methods cannot be called by clients
            next;
        }
        %{$CompatProblems{$Interface}{"Removed_Interface"}{""}}=();
    }
}

sub add_absent_param_names($)
{
    my $LibraryVersion = $_[0];
    return if(not keys(%AddIntParams));
    my $SecondVersion = $LibraryVersion==1?2:1;
    foreach my $Interface (sort keys(%{$CompleteSignature{$LibraryVersion}}))
    {
        next if(not keys(%{$AddIntParams{$Interface}}));
        foreach my $ParamPos (sort {int($a)<=>int($b)} keys(%{$CompleteSignature{$LibraryVersion}{$Interface}{"Param"}}))
        {# add absent parameter names
            my $ParamName = $CompleteSignature{$LibraryVersion}{$Interface}{"Param"}{$ParamPos}{"name"};
            if($ParamName=~/\Ap\d+\Z/ and my $NewParamName = $AddIntParams{$Interface}{$ParamPos})
            {# names from the external file
                if(defined $CompleteSignature{$SecondVersion}{$Interface}
                and defined $CompleteSignature{$SecondVersion}{$Interface}{"Param"}{$ParamPos})
                {
                    if($CompleteSignature{$SecondVersion}{$Interface}{"Param"}{$ParamPos}{"name"}=~/\Ap\d+\Z/) {
                        $CompleteSignature{$LibraryVersion}{$Interface}{"Param"}{$ParamPos}{"name"} = $NewParamName;
                    }
                }
                else {
                    $CompleteSignature{$LibraryVersion}{$Interface}{"Param"}{$ParamPos}{"name"} = $NewParamName;
                }
            }
        }
    }
}

sub detectChangedTypedefs()
{# detect changed typedefs to create correct function signatures
    foreach my $Typedef (keys(%{$Typedef_BaseName{1}}))
    {
        next if(not $Typedef);
        next if(isAnon($Typedef_BaseName{1}{$Typedef}));
        next if(isAnon($Typedef_BaseName{2}{$Typedef}));
        next if(not $Typedef_BaseName{1}{$Typedef});
        next if(not $Typedef_BaseName{2}{$Typedef});# exclude added/removed
        if($Typedef_BaseName{1}{$Typedef} ne $Typedef_BaseName{2}{$Typedef}) {
            $ChangedTypedef{$Typedef} = 1;
        }
    }
}

sub get_func_prefix($$)
{
    my ($Symbol, $LibVersion) = @_;
    my $ShortName = $CompleteSignature{$LibVersion}{$Symbol}{"ShortName"};
    if(my $ClassId = $CompleteSignature{$LibVersion}{$Symbol}{"Class"})
    {# methods
        $ShortName = get_TypeName($ClassId, $LibVersion)."::".$ShortName;
    }
    return $ShortName;
}

sub mergeSignatures()
{
    my %SubProblems = ();
    
    prepareInterfaces(1);
    prepareInterfaces(2);
    %SymbolInfo=();
    
    registerVirtualTable(1);
    registerVirtualTable(2);
    
    registerOverriding(1);
    registerOverriding(2);
    
    setVirtFuncPositions(1);
    setVirtFuncPositions(2);
    
    add_absent_param_names(1);
    add_absent_param_names(2);

    detectChangedTypedefs();

    mergeBases();
    
    my %AddedOverloads = ();
    foreach my $Interface (sort keys(%AddedInt))
    {# check all added exported symbols
        next if(not $CompleteSignature{2}{$Interface}{"Header"});
        if(defined $CompleteSignature{1}{$Interface}
        and $CompleteSignature{1}{$Interface}{"Header"}) {
            # double-check added symbol
            next;
        }
        next if(not interfaceFilter($Interface, 2, "Imported"));
        $CheckedInterfaces{$Interface} = 1;
        $AddedOverloads{get_func_prefix($Interface, 2)}{get_func_suffix($Interface)} = $Interface;
        if(my $OverriddenMethod = $CompleteSignature{2}{$Interface}{"Override"})
        {# register virtual redefinition
            my $AffectedClass_Name = get_TypeName($CompleteSignature{2}{$Interface}{"Class"}, 2);
            if(defined $CompleteSignature{1}{$OverriddenMethod}
            and $CompleteSignature{1}{$OverriddenMethod}{"Virt"} and $ClassToId{1}{$AffectedClass_Name}
            and not $CompleteSignature{1}{$OverriddenMethod}{"Private"})
            { # public virtual methods, virtual destructors: class should exist in previous version
                if(isCopyingClass($AffectedClass_Name, 1)) {
                    # old v-table (copied) will be used by applications
                    next;
                }
                if(defined $CompleteSignature{1}{$Interface}
                and $CompleteSignature{1}{$Interface}{"InLine"}) {
                    # auto-generated virtual destructors stay in the header (and v-table), added to library
                    # use case: Ice 3.3.1 -> 3.4.0
                    next;
                }
                %{$CompatProblems{$OverriddenMethod}{"Overridden_Virtual_Method"}{$tr_name{$Interface}}}=(
                    "Type_Name"=>$AffectedClass_Name,
                    "Type_Type"=>"Class",
                    "Target"=>get_Signature($Interface, 2),
                    "Old_Value"=>get_Signature($OverriddenMethod, 2),
                    "New_Value"=>get_Signature($Interface, 2)  );
            }
        }
    }
    foreach my $Interface (sort keys(%RemovedInt))
    {
        next if(not $CompleteSignature{1}{$Interface}{"Header"});
        if(defined $CompleteSignature{2}{$Interface}
        and $CompleteSignature{2}{$Interface}{"Header"}) {
            # double-check removed symbol
            next;
        }
        next if($CompleteSignature{1}{$Interface}{"Private"}); # skip private methods
        next if(not interfaceFilter($Interface, 1, "Imported"));
        $CheckedInterfaces{$Interface} = 1;
        if(my $OverriddenMethod = $CompleteSignature{1}{$Interface}{"Override"})
        {# register virtual redefinition
            my $AffectedClass_Name = get_TypeName($CompleteSignature{1}{$Interface}{"Class"}, 1);
            if(defined $CompleteSignature{2}{$OverriddenMethod}
            and $CompleteSignature{2}{$OverriddenMethod}{"Virt"} and $ClassToId{2}{$AffectedClass_Name})
            {# virtual methods, virtual destructors: class should exist in newer version
                if(isCopyingClass($AffectedClass_Name, 1)) {
                    # old v-table (copied) will be used by applications
                    next;
                }
                if(defined $CompleteSignature{2}{$Interface}
                and $CompleteSignature{2}{$Interface}{"InLine"}) {
                    # auto-generated virtual destructors stay in the header (and v-table), removed from library
                    # use case: Ice 3.3.1 -> 3.4.0
                    next;
                }
                %{$CompatProblems{$Interface}{"Overridden_Virtual_Method_B"}{$tr_name{$OverriddenMethod}}}=(
                    "Type_Name"=>$AffectedClass_Name,
                    "Type_Type"=>"Class",
                    "Target"=>get_Signature($OverriddenMethod, 1),
                    "Old_Value"=>get_Signature($Interface, 1),
                    "New_Value"=>get_Signature($OverriddenMethod, 1)  );
            }
        }
        if($OSgroup eq "windows")
        {# register the reason of symbol name change
            if(my $NewSymbol = $mangled_name{2}{$tr_name{$Interface}})
            {
                if($AddedInt{$NewSymbol})
                {
                    if($CompleteSignature{1}{$Interface}{"Static"} ne $CompleteSignature{2}{$NewSymbol}{"Static"})
                    {
                        if($CompleteSignature{2}{$NewSymbol}{"Static"}) {
                            %{$CompatProblems{$Interface}{"Symbol_Changed_Static"}{$tr_name{$Interface}}}=(
                                "Target"=>$tr_name{$Interface},
                                "Old_Value"=>$Interface,
                                "New_Value"=>$NewSymbol  );
                        }
                        else {
                            %{$CompatProblems{$Interface}{"Symbol_Changed_NonStatic"}{$tr_name{$Interface}}}=(
                                "Target"=>$tr_name{$Interface},
                                "Old_Value"=>$Interface,
                                "New_Value"=>$NewSymbol  );
                        }
                    }
                    if($CompleteSignature{1}{$Interface}{"Virt"} ne $CompleteSignature{2}{$NewSymbol}{"Virt"})
                    {
                        if($CompleteSignature{2}{$NewSymbol}{"Virt"}) {
                            %{$CompatProblems{$Interface}{"Symbol_Changed_Virtual"}{$tr_name{$Interface}}}=(
                                "Target"=>$tr_name{$Interface},
                                "Old_Value"=>$Interface,
                                "New_Value"=>$NewSymbol  );
                        }
                        else {
                            %{$CompatProblems{$Interface}{"Symbol_Changed_NonVirtual"}{$tr_name{$Interface}}}=(
                                "Target"=>$tr_name{$Interface},
                                "Old_Value"=>$Interface,
                                "New_Value"=>$NewSymbol  );
                        }
                    }
                    my $ReturnTypeName1 = get_TypeName($CompleteSignature{1}{$Interface}{"Return"}, 1);
                    my $ReturnTypeName2 = get_TypeName($CompleteSignature{2}{$NewSymbol}{"Return"}, 2);
                    if($ReturnTypeName1 ne $ReturnTypeName2)
                    {
                        %{$CompatProblems{$Interface}{"Symbol_Changed_Return"}{$tr_name{$Interface}}}=(
                            "Target"=>$tr_name{$Interface},
                            "Old_Type"=>$ReturnTypeName1,
                            "New_Type"=>$ReturnTypeName2,
                            "Old_Value"=>$Interface,
                            "New_Value"=>$NewSymbol  );
                    }
                }
            }
        }
        my $Prefix = get_func_prefix($Interface, 1);
        if(my @Overloads = sort keys(%{$AddedOverloads{$Prefix}})
        and not $AddedOverloads{$Prefix}{get_func_suffix($Interface)})
        {# changed signature
            my $NewSymbol = $AddedOverloads{$Prefix}{$Overloads[0]};
            if($CompleteSignature{1}{$Interface}{"Constructor"}) {
                if($Interface=~/(C1E|C2E)/) {
                    my $CtorType = $1;
                    $NewSymbol=~s/(C1E|C2E)/$CtorType/g;
                }
            }
            elsif($CompleteSignature{1}{$Interface}{"Destructor"}) {
                if($Interface=~/(D0E|D1E|D2E)/) {
                    my $DtorType = $1;
                    $NewSymbol=~s/(D0E|D1E|D2E)/$DtorType/g;
                }
            }
            %{$CompatProblems{$Interface}{"Symbol_Changed_Signature"}{$tr_name{$Interface}}}=(
                "Target"=>$tr_name{$Interface},
                "New_Signature"=>get_Signature($NewSymbol, 2),
                "Old_Value"=>$Interface,
                "New_Value"=>$NewSymbol  );
        }
    }
    foreach my $Interface (sort keys(%{$CompleteSignature{1}}))
    {# checking interfaces
        if($Interface!~/[\@\$\?]/
        and $SymVer{1}{$Interface}=~/\A([^\@\$\?]+)[\@\$]+/) {
            next if($1 eq $Interface);
        }
        next if(not $CompleteSignature{1}{$Interface}{"MnglName"} or not $CompleteSignature{2}{$Interface}{"MnglName"});
        next if($CompleteSignature{1}{$Interface}{"Private"});
        next if(not $CompleteSignature{1}{$Interface}{"Header"} or not $CompleteSignature{2}{$Interface}{"Header"});
        if($CheckHeadersOnly)
        {# skip added and removed pure virtual methods
            next if(not $CompleteSignature{1}{$Interface}{"PureVirt"} and $CompleteSignature{2}{$Interface}{"PureVirt"});
            next if($CompleteSignature{1}{$Interface}{"PureVirt"} and not $CompleteSignature{2}{$Interface}{"PureVirt"});
        }
        else
        {# skip external, added and removed functions except pure virtual methods
            if(not is_in_library($Interface, 1) or not is_in_library($Interface, 2)) {
                if(not $CompleteSignature{1}{$Interface}{"PureVirt"}
                or not $CompleteSignature{2}{$Interface}{"PureVirt"}) {
                    next;
                }
            }
        }
        if(not interfaceFilter($Interface, 1, "Imported"))
        {# symbols that cannot be imported
            next;
        }
        # checking virtual table
        mergeVirtualTables($Interface);
        if($COMPILE_ERRORS)
        {# if some errors occurred at the compiling stage
         # then some false positives can be skipped here
            if(not $CompleteSignature{1}{$Interface}{"Data"} and $CompleteSignature{2}{$Interface}{"Data"}
            and not $CompleteSignature{2}{$Interface}{"Object"})
            {# missed information about parameters in newer version
                next;
            }
            if($CompleteSignature{1}{$Interface}{"Data"} and not $CompleteSignature{1}{$Interface}{"Object"}
            and not $CompleteSignature{2}{$Interface}{"Data"})
            {# missed information about parameters in older version
                next;
            }
        }
        my ($MnglName, $VersionSpec, $SymbolVersion) = get_symbol_name_version($Interface);
        $CheckedInterfaces{$Interface}=1;
        $ComparedABI+=1;
        # checking attributes
        if($CompleteSignature{2}{$Interface}{"Static"}
        and not $CompleteSignature{1}{$Interface}{"Static"} and $Interface=~/\A(_Z|\?)/) {
            %{$CompatProblems{$Interface}{"Method_Became_Static"}{""}}=();
        }
        elsif(not $CompleteSignature{2}{$Interface}{"Static"}
        and $CompleteSignature{1}{$Interface}{"Static"} and $Interface=~/\A(_Z|\?)/) {
            %{$CompatProblems{$Interface}{"Method_Became_NonStatic"}{""}}=();
        }
        if(($CompleteSignature{1}{$Interface}{"Virt"} and $CompleteSignature{2}{$Interface}{"Virt"})
        or ($CompleteSignature{1}{$Interface}{"PureVirt"} and $CompleteSignature{2}{$Interface}{"PureVirt"}))
        {# relative position of virtual and pure virtual methods
            if($CompleteSignature{1}{$Interface}{"RelPos"}!=$CompleteSignature{2}{$Interface}{"RelPos"})
            {
                my $Class_Id = $CompleteSignature{1}{$Interface}{"Class"};
                if($VirtualTable{1}{get_TypeName($Class_Id, 1)}{$Interface}!=$VirtualTable{2}{get_TypeName($Class_Id, 1)}{$Interface})
                {# check the absolute position of virtual method (including added and removed methods)
                    my %Class_Type = get_Type($Tid_TDid{1}{$Class_Id}, $Class_Id, 1);
                    my $ProblemType = "Virtual_Method_Position";
                    if($CompleteSignature{1}{$Interface}{"PureVirt"}) {
                        $ProblemType = "Pure_Virtual_Method_Position";
                    }
                    my @Affected = ($Interface, keys(%{$OverriddenMethods{1}{$Interface}}));
                    foreach my $AffectedInterface (@Affected) {
                        %{$CompatProblems{$AffectedInterface}{$ProblemType}{$tr_name{$MnglName}}}=(
                            "Type_Name"=>$Class_Type{"Name"},
                            "Type_Type"=>"Class",
                            "Old_Value"=>$CompleteSignature{1}{$Interface}{"RelPos"},
                            "New_Value"=>$CompleteSignature{2}{$Interface}{"RelPos"},
                            "Target"=>get_Signature($Interface, 1)  );
                    }
                }
            }
        }
        if(keys(%{$CompleteSignature{1}{$Interface}{"Param"}})==keys(%{$CompleteSignature{2}{$Interface}{"Param"}}))
        {
            foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$CompleteSignature{1}{$Interface}{"Param"}}))
            {# checking parameters
                mergeParameters($Interface, $ParamPos, $ParamPos);
            }
        }
        elsif($Interface!~/\A(_Z|\?)/)
        {# signature changes in C functions
            foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$CompleteSignature{2}{$Interface}{"Param"}}))
            {# checking added parameters
                my $ParamType2_Id = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"type"};
                last if(get_TypeName($ParamType2_Id, 2) eq "...");
                my $Parameter_Name = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"name"};
                my $Parameter_OldName = (defined $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos})?$CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"name"}:"";
                my $ParamPos_Prev = "-1";
                if($Parameter_Name=~/\Ap\d+\Z/i)
                {# added unnamed parameter ( pN )
                    my @Positions1 = find_ParamPair_Pos_byTypeAndPos(get_TypeName($ParamType2_Id, 2), $ParamPos, "backward", $Interface, 1);
                    my @Positions2 = find_ParamPair_Pos_byTypeAndPos(get_TypeName($ParamType2_Id, 2), $ParamPos, "backward", $Interface, 2);
                    if($#Positions1==-1 or $#Positions2>$#Positions1) {
                        $ParamPos_Prev = "lost";
                    }
                }
                else {
                    $ParamPos_Prev = find_ParamPair_Pos_byName($Parameter_Name, $Interface, 1);
                }
                if($ParamPos_Prev eq "lost")
                {
                    if($ParamPos>keys(%{$CompleteSignature{1}{$Interface}{"Param"}})-1)
                    {
                        %{$CompatProblems{$Interface}{"Added_Parameter"}{numToStr($ParamPos+1)." Parameter"}}=(
                            "Target"=>$Parameter_Name,
                            "Parameter_Position"=>$ParamPos,
                            "Parameter_TypeName"=>get_TypeName($ParamType2_Id, 2),
                            "New_Signature"=>get_Signature($Interface, 2)  );
                    }
                    else
                    {
                        my %ParamType_Pure = get_PureType($Tid_TDid{2}{$ParamType2_Id}, $ParamType2_Id, 2);
                        my $ParamStraightPairType_Id = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"type"};
                        my %ParamStraightPairType_Pure = get_PureType($Tid_TDid{1}{$ParamStraightPairType_Id}, $ParamStraightPairType_Id, 1);
                        if(($ParamType_Pure{"Name"} eq $ParamStraightPairType_Pure{"Name"} or get_TypeName($ParamType2_Id, 2) eq get_TypeName($ParamStraightPairType_Id, 1))
                        and find_ParamPair_Pos_byName($Parameter_OldName, $Interface, 2) eq "lost")
                        {
                            if($Parameter_OldName!~/\Ap\d+\Z/ and $Parameter_Name!~/\Ap\d+\Z/)
                            {
                                %{$CompatProblems{$Interface}{"Parameter_Rename"}{numToStr($ParamPos+1)." Parameter"}}=(
                                    "Target"=>$Parameter_OldName,
                                    "Parameter_Position"=>$ParamPos,
                                    "Parameter_TypeName"=>get_TypeName($ParamType2_Id, 2),
                                    "Old_Value"=>$Parameter_OldName,
                                    "New_Value"=>$Parameter_Name,
                                    "New_Signature"=>get_Signature($Interface, 2)  );
                            }
                        }
                        else
                        {
                            %{$CompatProblems{$Interface}{"Added_Middle_Parameter"}{numToStr($ParamPos+1)." Parameter"}}=(
                                "Target"=>$Parameter_Name,
                                "Parameter_Position"=>$ParamPos,
                                "Parameter_TypeName"=>get_TypeName($ParamType2_Id, 2),
                                "New_Signature"=>get_Signature($Interface, 2)  );
                        }
                    }
                }
            }
            foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$CompleteSignature{1}{$Interface}{"Param"}}))
            {# check relevant parameters
                my $ParamType1_Id = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"type"};
                my $ParamName1 = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"name"};
                if(defined $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos})
                {
                    my $ParamType2_Id = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"type"};
                    my $ParamName2 = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"name"};
                    if(($ParamName1!~/\Ap\d+\Z/i and $ParamName1 eq $ParamName2)
                    or get_TypeName($ParamType1_Id, 1) eq get_TypeName($ParamType2_Id, 2)) {
                        mergeParameters($Interface, $ParamPos, $ParamPos);
                    }
                }
            }
            foreach my $ParamPos (sort {int($a) <=> int($b)} keys(%{$CompleteSignature{1}{$Interface}{"Param"}}))
            {# checking removed parameters
                my $ParamType1_Id = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"type"};
                last if(get_TypeName($ParamType1_Id, 1) eq "...");
                my $Parameter_Name = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos}{"name"};
                my $Parameter_NewName = (defined $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos})?$CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"name"}:"";
                my $ParamPos_New = "-1";
                if($Parameter_Name=~/\Ap\d+\Z/i)
                {# removed unnamed parameter ( pN )
                    my @Positions1 = find_ParamPair_Pos_byTypeAndPos(get_TypeName($ParamType1_Id, 1), $ParamPos, "forward", $Interface, 1);
                    my @Positions2 = find_ParamPair_Pos_byTypeAndPos(get_TypeName($ParamType1_Id, 1), $ParamPos, "forward", $Interface, 2);
                    if($#Positions2==-1 or $#Positions2<$#Positions1) {
                        $ParamPos_New = "lost";
                    }
                }
                else {
                    $ParamPos_New = find_ParamPair_Pos_byName($Parameter_Name, $Interface, 2);
                }
                if($ParamPos_New eq "lost")
                {
                    if($ParamPos>keys(%{$CompleteSignature{2}{$Interface}{"Param"}})-1)
                    {
                        %{$CompatProblems{$Interface}{"Removed_Parameter"}{numToStr($ParamPos+1)." Parameter"}}=(
                            "Target"=>$Parameter_Name,
                            "Parameter_Position"=>$ParamPos,
                            "Parameter_TypeName"=>get_TypeName($ParamType1_Id, 1),
                            "New_Signature"=>get_Signature($Interface, 2)  );
                    }
                    elsif($ParamPos<keys(%{$CompleteSignature{1}{$Interface}{"Param"}})-1)
                    {
                        my %ParamType_Pure = get_PureType($Tid_TDid{1}{$ParamType1_Id}, $ParamType1_Id, 1);
                        my $ParamStraightPairType_Id = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos}{"type"};
                        my %ParamStraightPairType_Pure = get_PureType($Tid_TDid{2}{$ParamStraightPairType_Id}, $ParamStraightPairType_Id, 2);
                        if(($ParamType_Pure{"Name"} eq $ParamStraightPairType_Pure{"Name"} or get_TypeName($ParamType1_Id, 1) eq get_TypeName($ParamStraightPairType_Id, 2))
                        and find_ParamPair_Pos_byName($Parameter_NewName, $Interface, 1) eq "lost")
                        {
                            if($Parameter_NewName!~/\Ap\d+\Z/ and $Parameter_Name!~/\Ap\d+\Z/)
                            {
                                %{$CompatProblems{$Interface}{"Parameter_Rename"}{numToStr($ParamPos+1)." Parameter"}}=(
                                    "Target"=>$Parameter_Name,
                                    "Parameter_Position"=>$ParamPos,
                                    "Parameter_TypeName"=>get_TypeName($ParamType1_Id, 1),
                                    "Old_Value"=>$Parameter_Name,
                                    "New_Value"=>$Parameter_NewName,
                                    "New_Signature"=>get_Signature($Interface, 2)  );
                            }
                        }
                        else
                        {
                            %{$CompatProblems{$Interface}{"Removed_Middle_Parameter"}{numToStr($ParamPos+1)." Parameter"}}=(
                                "Target"=>$Parameter_Name,
                                "Parameter_Position"=>$ParamPos,
                                "Parameter_TypeName"=>get_TypeName($ParamType1_Id, 1),
                                "New_Signature"=>get_Signature($Interface, 2)  );
                        }
                    }
                }
            }
        }
        # checking return type
        my $ReturnType1_Id = $CompleteSignature{1}{$Interface}{"Return"};
        my $ReturnType2_Id = $CompleteSignature{2}{$Interface}{"Return"};
        %SubProblems = detectTypeChange($ReturnType1_Id, $ReturnType2_Id, "Return");
        foreach my $SubProblemType (keys(%SubProblems))
        {
            my $New_Value = $SubProblems{$SubProblemType}{"New_Value"};
            my $Old_Value = $SubProblems{$SubProblemType}{"Old_Value"};
            if($SubProblemType eq "Return_Type_And_Stack_Layout"
            and not keys(%{$CompleteSignature{1}{$Interface}{"Param"}}))
            {# function has no parameters
                %{$SubProblems{"Return_Type_And_Return_Method"}} = %{$SubProblems{$SubProblemType}};
                delete($SubProblems{$SubProblemType});
                $SubProblemType = "Return_Type_And_Return_Method";
            }
            elsif($SubProblemType eq "Return_Type" and $CompleteSignature{1}{$Interface}{"Data"}
            and $Old_Value=~/\A\Q$New_Value\E const\Z/)
            {# const -> non-const global data
                %{$SubProblems{"Global_Data_Became_Non_Const"}} = %{$SubProblems{$SubProblemType}};
                delete($SubProblems{$SubProblemType});
                $SubProblemType = "Global_Data_Became_Non_Const";
            }
            @{$CompatProblems{$Interface}{$SubProblemType}{"RetVal"}}{keys(%{$SubProblems{$SubProblemType}})} = values %{$SubProblems{$SubProblemType}};
        }
        if($ReturnType1_Id and $ReturnType2_Id)
        {
            @RecurTypes = ();
            %SubProblems = mergeTypes($ReturnType1_Id, $Tid_TDid{1}{$ReturnType1_Id},
                                      $ReturnType2_Id, $Tid_TDid{2}{$ReturnType2_Id});
            foreach my $SubProblemType (keys(%SubProblems))
            {
                foreach my $SubLocation (keys(%{$SubProblems{$SubProblemType}}))
                {
                    my $NewLocation = ($SubLocation)?"RetVal->".$SubLocation:"RetVal";
                    %{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}=(
                        "Return_Type_Name"=>get_TypeName($ReturnType1_Id, 1) );
                    @{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}{keys(%{$SubProblems{$SubProblemType}{$SubLocation}})} = values %{$SubProblems{$SubProblemType}{$SubLocation}};
                    if($SubLocation!~/\-\>/) {
                        $CompatProblems{$Interface}{$SubProblemType}{$NewLocation}{"Start_Type_Name"} = get_TypeName($ReturnType1_Id, 1);
                    }
                }
            }
        }
        
        # checking object type
        my $ObjectType1_Id = $CompleteSignature{1}{$Interface}{"Class"};
        my $ObjectType2_Id = $CompleteSignature{2}{$Interface}{"Class"};
        if($ObjectType1_Id and $ObjectType2_Id)
        {
            my $ThisPtr1_Id = getTypeIdByName(get_TypeName($ObjectType1_Id, 1)."*const", 1);
            my $ThisPtr2_Id = getTypeIdByName(get_TypeName($ObjectType2_Id, 2)."*const", 2);
            if($ThisPtr1_Id and $ThisPtr2_Id)
            {
                @RecurTypes = ();
                %SubProblems = mergeTypes($ThisPtr1_Id, $Tid_TDid{1}{$ThisPtr1_Id},
                                          $ThisPtr2_Id, $Tid_TDid{2}{$ThisPtr2_Id});
                foreach my $SubProblemType (keys(%SubProblems))
                {
                    foreach my $SubLocation (keys(%{$SubProblems{$SubProblemType}}))
                    {
                        my $NewLocation = ($SubLocation)?"this->".$SubLocation:"this";
                        %{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}=(
                            "Object_Type_Name"=>get_TypeName($ObjectType1_Id, 1) );
                        @{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}{keys(%{$SubProblems{$SubProblemType}{$SubLocation}})} = values %{$SubProblems{$SubProblemType}{$SubLocation}};
                        if($SubLocation!~/\-\>/) {
                            $CompatProblems{$Interface}{$SubProblemType}{$NewLocation}{"Start_Type_Name"} = get_TypeName($ObjectType1_Id, 1);
                        }
                    }
                }
            }
        }
    }
    set_Problems_Priority();
}

sub mergeDefaultParams($$$)
{
    my ($Interface, $ParamPos1, $ParamPos2) = @_;
    if($UsedDump{1}
    and cmpVersions($UsedDump{1}, "2.0")<0)
    {# support for old dumps
        # "default" attribute added in ACC 1.22 (dump 2.0 format)
        return;
    }
    if($UsedDump{2}
    and cmpVersions($UsedDump{2}, "2.0")<0)
    {# support for old dumps
        # "default" attribute added in ACC 1.22 (dump 2.0 format)
        return;
    }
    my $ParamType1_Id = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos1}{"type"};
    my $Parameter_Name = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos1}{"name"};
    my $ParamType2_Id = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos2}{"type"};
    return if(not $ParamType1_Id or not $ParamType2_Id);
    my %PType1 = get_PureType($Tid_TDid{1}{$ParamType1_Id}, $ParamType1_Id, 1);
    my $Parameter_Location = ($Parameter_Name)?$Parameter_Name:numToStr($ParamPos1+1)." Parameter";
    my $DefaultValue_Old = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos1}{"default"};
    my $DefaultValue_New = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos2}{"default"};
    if($PType1{"Name"}=~/\A(char\*|char const\*)\Z/)
    {
        if($DefaultValue_Old)
        {# FIXME: how to distinguish "0" and 0 (NULL)
            $DefaultValue_Old = "\"$DefaultValue_Old\"";
        }
        if($DefaultValue_New) {
            $DefaultValue_New = "\"$DefaultValue_New\"";
        }
    }
    elsif($PType1{"Name"}=~/\A(char)\Z/)
    {
        if($DefaultValue_Old) {
            $DefaultValue_Old = "\'$DefaultValue_Old\'";
        }
        if($DefaultValue_New) {
            $DefaultValue_New = "\'$DefaultValue_New\'";
        }
    }
    return if($DefaultValue_Old eq "");
    if($DefaultValue_New ne "")
    {
        if($DefaultValue_Old ne $DefaultValue_New)
        {
            %{$CompatProblems{$Interface}{"Parameter_Default_Value_Changed"}{$Parameter_Location}}=(
                "Target"=>$Parameter_Name,
                "Parameter_Position"=>$ParamPos1,
                "Old_Value"=>$DefaultValue_Old,
                "New_Value"=>$DefaultValue_New  );
        }
    }
    else
    {
        %{$CompatProblems{$Interface}{"Parameter_Default_Value_Removed"}{$Parameter_Location}}=(
                "Target"=>$Parameter_Name,
                "Parameter_Position"=>$ParamPos1,
                "Old_Value"=>$DefaultValue_Old  );
    }
}

sub mergeParameters($$$)
{
    my ($Interface, $ParamPos1, $ParamPos2) = @_;
    return if(not $Interface);
    return if(not defined $CompleteSignature{1}{$Interface}{"Param"});
    return if(not defined $CompleteSignature{2}{$Interface}{"Param"});
    my $ParamType1_Id = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos1}{"type"};
    my $Parameter_Name = $CompleteSignature{1}{$Interface}{"Param"}{$ParamPos1}{"name"};
    my $ParamType2_Id = $CompleteSignature{2}{$Interface}{"Param"}{$ParamPos2}{"type"};
    return if(not $ParamType1_Id or not $ParamType2_Id);
    mergeDefaultParams($Interface, $ParamPos1, $ParamPos2);
    my $Parameter_Location = ($Parameter_Name)?$Parameter_Name:numToStr($ParamPos1+1)." Parameter";
    if($Interface!~/\A(_Z|\?)/)
    {
        # checking type change (replace)
        my %SubProblems = detectTypeChange($ParamType1_Id, $ParamType2_Id, "Parameter");
        foreach my $SubProblemType (keys(%SubProblems))
        {
            my $New_Value = $SubProblems{$SubProblemType}{"New_Value"};
            my $Old_Value = $SubProblems{$SubProblemType}{"Old_Value"};
            my $Old_Value_Non_Const = $Old_Value;
            if($SubProblemType eq "Parameter_Type" and $Old_Value_Non_Const=~s/\Wconst(\W|\Z)/$1$2/g) {
                $Old_Value_Non_Const = correctName($Old_Value_Non_Const);
                if($Old_Value=~/\A\Q$New_Value\E const\Z/ or $Old_Value_Non_Const eq $New_Value)
                {# const -> non-const parameter
                    %{$SubProblems{"Parameter_Became_Non_Const"}} = %{$SubProblems{$SubProblemType}};
                    delete($SubProblems{$SubProblemType});
                    $SubProblemType = "Parameter_Became_Non_Const";
                }
            }
            if($SubProblemType eq "Parameter_Type_And_Size"
            or $SubProblemType eq "Parameter_Type")
            {
                my ($Arch1, $Arch2) = (getArch(1), getArch(2));
                if($Arch1 eq "unknown" or $Arch2 eq "unknown")
                { # if one of the architectures is unknown
                  # then set other arhitecture to unknown too
                    ($Arch1, $Arch2) = ("unknown", "unknown");
                }
                my ($Method1, $Passed1, $SizeOnStack1, $RegName1) = callingConvention($Interface, $ParamPos1, 1, $Arch1);
                my ($Method2, $Passed2, $SizeOnStack2, $RegName2) = callingConvention($Interface, $ParamPos2, 2, $Arch2);
                my $NewProblemType = $SubProblemType;
                if($Method1 eq $Method2)
                {
                    if($Method1 eq "stack" and $SizeOnStack1 ne $SizeOnStack2) {
                        $NewProblemType = "Parameter_Type_And_Stack";
                    }
                    elsif($Method1 eq "register" and $RegName1 ne $RegName2) {
                        $NewProblemType = "Parameter_Type_And_Register";
                    }
                }
                else
                {
                    if($Method1 eq "stack") {
                        $NewProblemType = "Parameter_Type_And_Pass_Through_Register";
                    }
                    elsif($Method1 eq "register") {
                        $NewProblemType = "Parameter_Type_And_Pass_Through_Stack";
                    }
                }
                if($NewProblemType ne $SubProblemType)
                {
                    %{$SubProblems{$NewProblemType}} = %{$SubProblems{$SubProblemType}};
                    delete($SubProblems{$SubProblemType});
                    $SubProblemType = $NewProblemType;
                    # add attributes
                    $SubProblems{$NewProblemType}{"Old_Reg"} = $RegName1;
                    $SubProblems{$NewProblemType}{"New_Reg"} = $RegName2;
                }
            }
            %{$CompatProblems{$Interface}{$SubProblemType}{$Parameter_Location}}=(
                "Target"=>$Parameter_Name,
                "Parameter_Position"=>$ParamPos1  );
            @{$CompatProblems{$Interface}{$SubProblemType}{$Parameter_Location}}{keys(%{$SubProblems{$SubProblemType}})} = values %{$SubProblems{$SubProblemType}};
        }
    }
    @RecurTypes = ();
    # checking type definition changes
    my %SubProblems_Merge = mergeTypes($ParamType1_Id, $Tid_TDid{1}{$ParamType1_Id}, $ParamType2_Id, $Tid_TDid{2}{$ParamType2_Id});
    foreach my $SubProblemType (keys(%SubProblems_Merge))
    {
        foreach my $SubLocation (keys(%{$SubProblems_Merge{$SubProblemType}}))
        {
            my $NewLocation = ($SubLocation)?$Parameter_Location."->".$SubLocation:$Parameter_Location;
            %{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}=(
                "Parameter_Type_Name"=>get_TypeName($ParamType1_Id, 1),
                "Parameter_Position"=>$ParamPos1,
                "Parameter_Name"=>$Parameter_Name  );
            @{$CompatProblems{$Interface}{$SubProblemType}{$NewLocation}}{keys(%{$SubProblems_Merge{$SubProblemType}{$SubLocation}})} = values %{$SubProblems_Merge{$SubProblemType}{$SubLocation}};
            if($SubLocation!~/\-\>/) {
                $CompatProblems{$Interface}{$SubProblemType}{$NewLocation}{"Start_Type_Name"} = get_TypeName($ParamType1_Id, 1);
            }
        }
    }
}

sub callingConvention($$$$)
{ # calling conventions for different compilers and operating systems
    my ($Interface, $ParamPos, $LibVersion, $Arch) = @_;
    my $ParamTypeId = $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$ParamPos}{"type"};
    my %Type = get_PureType($Tid_TDid{$LibVersion}{$ParamTypeId}, $ParamTypeId, $LibVersion);
    my ($Method, $Alignment, $Passed, $Register) = ("", 0, "", "");
    if($OSgroup=~/\A(linux|macos|freebsd)\Z/)
    {# GCC
        if($Arch eq "x86")
        { # System V ABI Intel386 ("Function Calling Sequence")
          # The stack is word aligned. Although the architecture does not require any
          # alignment of the stack, software convention and the operating system
          # requires that the stack be aligned on a word boundary.

          # Argument words are pushed onto the stack in reverse order (that is, the
          # rightmost argument in C call syntax has the highest address), preserving the
          # stack’s word alignment. All incoming arguments appear on the stack, residing
          # in the stack frame of the caller.

          # An argument’s size is increased, if necessary, to make it a multiple of words.
          # This may require tail padding, depending on the size of the argument.

          # Other areas depend on the compiler and the code being compiled. The stan-
          # dard calling sequence does not define a maximum stack frame size, nor does
          # it restrict how a language system uses the ‘‘unspecified’’ area of the stan-
          # dard stack frame.
            ($Method, $Alignment) = ("stack", 4);
        }
        elsif($Arch eq "x86_64")
        { # System V AMD64 ABI ("Function Calling Sequence")
            ($Method, $Alignment) = ("stack", 8);# eightbyte aligned
        }
        elsif($Arch eq "arm")
        { # Procedure Call Standard for the ARM Architecture
          # The stack must be double-word aligned
            ($Method, $Alignment) = ("stack", 8);# double-word
        }
    }
    elsif($OSgroup eq "windows")
    {# MS C++ Compiler
        if($Arch eq "x86")
        {
            if($ParamPos==0) {
                ($Method, $Register, $Passed) = ("register", "ecx", "value");
            }
            elsif($ParamPos==1) {
                ($Method, $Register, $Passed) = ("register", "edx", "value");
            }
            else {
                ($Method, $Alignment) = ("stack", 4);
            }
        }
        elsif($Arch eq "x86_64")
        {
            if($ParamPos<=3) {
                if($Type{"Name"}=~/\A(float|double|long double)\Z/) {
                    ($Method, $Passed) = ("xmm".$ParamPos, "value");
                }
                elsif($Type{"Name"}=~/\A(unsigned |)(short|int|long|long long)\Z/
                or $Type{"Type"}=~/\A(Struct|Union|Enum|Array)\Z/
                or $Type{"Name"}=~/\A(__m64|__m128)\Z/) {
                    if($ParamPos==0) {
                        ($Method, $Register, $Passed) = ("register", "rcx", "value");
                    }
                    elsif($ParamPos==1) {
                        ($Method, $Register, $Passed) = ("register", "rdx", "value");
                    }
                    elsif($ParamPos==2) {
                        ($Method, $Register, $Passed) = ("register", "r8", "value");
                    }
                    elsif($ParamPos==3) {
                        ($Method, $Register, $Passed) = ("register", "r9", "value");
                    }
                    if($Type{"Size"}>64
                    or $Type{"Type"} eq "Array") {
                        $Passed = "pointer";
                    }
                }
            }
            else {
                ($Method, $Alignment) = ("stack", 8);# word alignment
            }
        }
    }
    if($Method eq "register") {
        return ("register", $Passed, "", $Register);
    }
    else
    {# on the stack
        if(not $Alignment)
        {# default convention
            $Alignment = $POINTER_SIZE{$LibVersion};
        }
        if(not $Passed)
        {# default convention
            $Passed = "value";
        }
        return ("stack", $Passed, (int(($Type{"Size"}+$Alignment)/$Alignment)*$Alignment), "");
    }
}

sub find_ParamPair_Pos_byName($$$)
{
    my ($Name, $Interface, $LibVersion) = @_;
    foreach my $ParamPos (sort {int($a)<=>int($b)} keys(%{$CompleteSignature{$LibVersion}{$Interface}{"Param"}}))
    {
        next if(not defined $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$ParamPos});
        if($CompleteSignature{$LibVersion}{$Interface}{"Param"}{$ParamPos}{"name"} eq $Name)
        {
            return $ParamPos;
        }
    }
    return "lost";
}

sub find_ParamPair_Pos_byTypeAndPos($$$$$)
{
    my ($TypeName, $MediumPos, $Order, $Interface, $LibVersion) = @_;
    my @Positions = ();
    foreach my $ParamPos (sort {int($a)<=>int($b)} keys(%{$CompleteSignature{$LibVersion}{$Interface}{"Param"}}))
    {
        next if($Order eq "backward" and $ParamPos>$MediumPos);
        next if($Order eq "forward" and $ParamPos<$MediumPos);
        next if(not defined $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$ParamPos});
        my $PTypeId = $CompleteSignature{$LibVersion}{$Interface}{"Param"}{$ParamPos}{"type"};
        if(get_TypeName($PTypeId, $LibVersion) eq $TypeName) {
            push(@Positions, $ParamPos);
        }
    }
    return @Positions;
}

sub getTypeIdByName($$)
{
    my ($TypeName, $Version) = @_;
    return $TName_Tid{$Version}{correctName($TypeName)};
}

sub detectTypeChange($$$)
{
    my ($Type1_Id, $Type2_Id, $Prefix) = @_;
    my %LocalProblems = ();
    my $Type1_DId = $Tid_TDid{1}{$Type1_Id};
    my $Type2_DId = $Tid_TDid{2}{$Type2_Id};
    my %Type1 = get_Type($Type1_DId, $Type1_Id, 1);
    my %Type2 = get_Type($Type2_DId, $Type2_Id, 2);
    my %Type1_Pure = get_PureType($Type1_DId, $Type1_Id, 1);
    my %Type2_Pure = get_PureType($Type2_DId, $Type2_Id, 2);
    my %Type1_Base = ($Type1_Pure{"Type"} eq "Array")?get_OneStep_BaseType($Tid_TDid{1}{$Type1_Pure{"Tid"}}, $Type1_Pure{"Tid"}, 1):get_BaseType($Type1_DId, $Type1_Id, 1);
    my %Type2_Base = ($Type2_Pure{"Type"} eq "Array")?get_OneStep_BaseType($Tid_TDid{2}{$Type2_Pure{"Tid"}}, $Type2_Pure{"Tid"}, 2):get_BaseType($Type2_DId, $Type2_Id, 2);
    my $Type1_PLevel = get_PointerLevel($Type1_DId, $Type1_Id, 1);
    my $Type2_PLevel = get_PointerLevel($Type2_DId, $Type2_Id, 2);
    return () if(not $Type1{"Name"} or not $Type2{"Name"});
    return () if(not $Type1_Base{"Name"} or not $Type2_Base{"Name"});
    return () if($Type1_PLevel eq "" or $Type2_PLevel eq "");
    if($Type1_Base{"Name"} ne $Type2_Base{"Name"}
    and ($Type1{"Name"} eq $Type2{"Name"} or ($Type1_PLevel>=1 and $Type1_PLevel==$Type2_PLevel
    and $Type1_Base{"Name"} ne "void" and $Type2_Base{"Name"} ne "void")))
    {# base type change
        if($Type1{"Type"} eq "Typedef" and $Type2{"Type"} eq "Typedef"
        and $Type1{"Name"} eq $Type2{"Name"})
        {# will be reported in mergeTypes() as typedef problem
            return ();
        }
        if($Type1_Base{"Name"}!~/anon\-/ and $Type2_Base{"Name"}!~/anon\-/)
        {
            if($Type1_Base{"Size"} ne $Type2_Base{"Size"}
            and $Type1_Base{"Size"} and $Type2_Base{"Size"})
            {
                %{$LocalProblems{$Prefix."_BaseType_And_Size"}}=(
                    "Old_Value"=>$Type1_Base{"Name"},
                    "New_Value"=>$Type2_Base{"Name"},
                    "Old_Size"=>$Type1_Base{"Size"},
                    "New_Size"=>$Type2_Base{"Size"},
                    "InitialType_Type"=>$Type1_Pure{"Type"});
            }
            elsif(not $UseOldDumps)
            {# support for old dumps: type names have been corrected in ACC 1.22, 1.23
                %{$LocalProblems{$Prefix."_BaseType"}}=(
                    "Old_Value"=>$Type1_Base{"Name"},
                    "New_Value"=>$Type2_Base{"Name"},
                    "InitialType_Type"=>$Type1_Pure{"Type"});
            }
        }
    }
    elsif($Type1{"Name"} ne $Type2{"Name"})
    {# type change
        if($Type1{"Name"}!~/anon\-/ and $Type2{"Name"}!~/anon\-/)
        {
            if($Prefix eq "Return" and $Type1{"Name"} eq "void"
            and $Type2_Pure{"Type"}=~/Intrinsic|Enum/) {
                # safe change
            }
            elsif($Prefix eq "Return" and $Type1_Pure{"Type"}=~/Intrinsic|Enum/
            and $Type2_Pure{"Type"}=~/Struct|Class|Union/)
            {# returns into hidden first parameter instead of a register

                # System V ABI Intel386 ("Function Calling Sequence")
                # A function that returns an integral or pointer value places its result in register %eax.

                # A floating-point return value appears on the top of the Intel387 register stack. The
                # caller then must remove the value from the Intel387 stack, even if it doesn’t use the
                # value.

                # If a function returns a structure or union, then the caller provides space for the
                # return value and places its address on the stack as argument word zero. In effect,
                # this address becomes a ‘‘hidden’’ first argument.

                %{$LocalProblems{"Return_Type_And_Stack_Layout"}}=(
                    "Old_Value"=>$Type1{"Name"},
                    "New_Value"=>$Type2{"Name"},
                    "Old_Size"=>$Type1{"Size"},
                    "New_Size"=>$Type2{"Size"},
                    "Old_Type"=>"Intrinsic",
                    "New_Type"=>"Struct",
                    "InitialType_Type"=>$Type1_Pure{"Type"});
            }
            elsif($Prefix eq "Return" and $Type2_Pure{"Type"}=~/Intrinsic|Enum/
            and $Type1_Pure{"Type"}=~/Struct|Class|Union/)
            {# returns in a register instead of a hidden first parameter
                %{$LocalProblems{"Return_Type_And_Stack_Layout"}}=(
                    "Old_Value"=>$Type1{"Name"},
                    "New_Value"=>$Type2{"Name"},
                    "Old_Size"=>$Type1{"Size"},
                    "New_Size"=>$Type2{"Size"},
                    "Old_Type"=>"Struct",
                    "New_Type"=>"Intrinsic",
                    "InitialType_Type"=>$Type1_Pure{"Type"});
                if($Type2{"Name"} eq "void") {
                    %{$LocalProblems{"Return_Type_Became_Void"}}=(
                        "Old_Value"=>$Type1{"Name"},
                        "New_Value"=>"void",
                        "InitialType_Type"=>$Type1_Pure{"Type"});
                }
            }
            elsif($Prefix eq "Return" and $Type2{"Name"} eq "void")
            {
                %{$LocalProblems{"Return_Type_Became_Void"}}=(
                    "Old_Value"=>$Type1{"Name"},
                    "New_Value"=>"void",
                    "InitialType_Type"=>$Type1_Pure{"Type"});
            }
            else
            {
                if($Type1{"Size"} ne $Type2{"Size"}
                and $Type1{"Size"} and $Type2{"Size"})
                {
                    %{$LocalProblems{$Prefix."_Type_And_Size"}}=(
                        "Old_Value"=>$Type1{"Name"},
                        "New_Value"=>$Type2{"Name"},
                        "Old_Size"=>($Type1{"Type"} eq "Array")?$Type1{"Size"}*$Type1_Base{"Size"}:$Type1{"Size"},
                        "New_Size"=>($Type2{"Type"} eq "Array")?$Type2{"Size"}*$Type2_Base{"Size"}:$Type2{"Size"},
                        "InitialType_Type"=>$Type1_Pure{"Type"});
                }
                elsif(not $UseOldDumps)
                { # support for old dumps: type names have been corrected in ACC 1.22, 1.23
                  # FIXME: type names may be changed in minor versions of the dump
                    %{$LocalProblems{$Prefix."_Type"}}=(
                        "Old_Value"=>$Type1{"Name"},
                        "New_Value"=>$Type2{"Name"},
                        "InitialType_Type"=>$Type1_Pure{"Type"});
                }
            }
        }
    }
    if($Type1_PLevel!=$Type2_PLevel)
    {
        if($Type1{"Name"} ne "void"
        and $Type2{"Name"} ne "void")
        {
            if($Type2_PLevel>$Type1_PLevel) {
                %{$LocalProblems{$Prefix."_PointerLevel_Increased"}}=(
                    "Old_Value"=>$Type1_PLevel,
                    "New_Value"=>$Type2_PLevel);
            }
            else {
                %{$LocalProblems{$Prefix."_PointerLevel_Decreased"}}=(
                    "Old_Value"=>$Type1_PLevel,
                    "New_Value"=>$Type2_PLevel);
            }
        }
    }
    if($Type1_Pure{"Type"} eq "Array") {
        # base_type[N] -> base_type[N]
        # base_type: older_structure -> typedef to newer_structure
        my %SubProblems = detectTypeChange($Type1_Base{"Tid"}, $Type2_Base{"Tid"}, $Prefix."_Base");
        foreach my $SubProblemType (keys(%SubProblems))
        {
            next if(defined $LocalProblems{$SubProblemType});
            foreach my $Attr (keys(%{$SubProblems{$SubProblemType}})) {
                $LocalProblems{$SubProblemType}{$Attr} = $SubProblems{$SubProblemType}{$Attr};
            }
        }
    }
    return %LocalProblems;
}

sub htmlSpecChars($)
{
    my $Str = $_[0];
    $Str=~s/\&([^#]|\Z)/&amp;$1/g;
    $Str=~s/</&lt;/g;
    $Str=~s/\-\>/&minus;&gt;/g;
    $Str=~s/>/&gt;/g;
    $Str=~s/([^ ])( )([^ ])/$1\@ALONE_SP\@$3/g;
    $Str=~s/ /&nbsp;/g;
    $Str=~s/\@ALONE_SP\@/ /g;
    $Str=~s/\n/<br\/>/g;
    return $Str;
}

sub highLight_Signature($)
{
    my $Signature = $_[0];
    return highLight_Signature_PPos_Italic($Signature, "", 0, 0, 0);
}

sub highLight_Signature_Italic_Color($)
{
    my $Signature = $_[0];
    return highLight_Signature_PPos_Italic($Signature, "", 1, 1, 1);
}

sub get_symbol_name_version($)
{
    my $fullname = $_[0];
    my ($realname, $version_spec, $version) = ($fullname, "", "");
    if($fullname=~/\A([^\@\$\?]+)([\@\$]+)([^\@\$]+)\Z/) {
        ($realname, $version_spec, $version) = ($1, $2, $3);
    }
    return ($realname, $version_spec, $version);
}

sub highLight_Signature_PPos_Italic($$$$$)
{
    my ($FullSignature, $Parameter_Position, $ItalicParams, $ColorParams, $ShowReturn) = @_;
    $ItalicParams=$ColorParams=0 if($CheckObjectsOnly);
    my ($Signature, $VersionSpec, $SymbolVersion) = get_symbol_name_version($FullSignature);
    my $Return = "";
    if($ShowRetVal and $Signature=~s/([^:]):([^:].+?)\Z/$1/g) {
        $Return = $2;
    }
    my $SCenter = get_signature_center($Signature);
    if(not $SCenter)
    {# global data
        $Signature = htmlSpecChars($Signature);
        $Signature=~s!(\[data\])!<span style='color:Black;font-weight:normal;'>$1</span>!g;
        $Signature .= (($SymbolVersion)?"<span class='symver'>&nbsp;$VersionSpec&nbsp;$SymbolVersion</span>":"");
        if($Return and $ShowReturn) {
            $Signature .= "<span class='int_p' style='white-space:nowrap'> &nbsp;<b>:</b>&nbsp;&nbsp;".htmlSpecChars($Return)."</span>";
        }
        return $Signature;
    }
    my ($Begin, $End) = (substr($Signature, 0, $SCenter), "");
    $Begin.=" " if($Begin!~/ \Z/);
    if($Signature=~/\)((| const)(| \[static\]))\Z/) {
        $End = $1;
    }
    my @Parts = ();
    my $Part_Num = 0;
    foreach my $Part (get_signature_parts($Signature, 1))
    {
        $Part=~s/\A\s+|\s+\Z//g;
        my ($Part_Styled, $ParamName) = (htmlSpecChars($Part), "");
        if($Part=~/\([\*]+(\w+)\)/i) {
            $ParamName = $1;#func-ptr
        }
        elsif($Part=~/(\w+)[\,\)]*\Z/i) {
            $ParamName = $1;
        }
        if(not $ParamName) {
            push(@Parts, $Part_Styled);
            next;
        }
        if($ItalicParams and not $TName_Tid{1}{$Part}
        and not $TName_Tid{2}{$Part})
        {
            if($Parameter_Position ne ""
            and $Part_Num==$Parameter_Position) {
                $Part_Styled =~ s!(\W)$ParamName([\,\)]|\Z)!$1<span class='focus_p'>$ParamName</span>$2!ig;
            }
            elsif($ColorParams) {
                $Part_Styled =~ s!(\W)$ParamName([\,\)]|\Z)!$1<span class='color_p'>$ParamName</span>$2!ig;
            }
            else {
                $Part_Styled =~ s!(\W)$ParamName([\,\)]|\Z)!$1<span style='font-style:italic;'>$ParamName</span>$2!ig;
            }
        }
        $Part_Styled=~s/,(\w)/, $1/g;
        if(length($Part)<=45) {
            $Part_Styled = "<span style='white-space:nowrap;'>".$Part_Styled."</span>";
        }
        push(@Parts, $Part_Styled);
        $Part_Num += 1;
    }
    $Signature = htmlSpecChars($Begin)."<span class='int_p'>"."(&nbsp;".join(" ", @Parts).(@Parts?"&nbsp;":"").")"."</span>".$End;
    if($Return and $ShowReturn) {
        $Signature .= "<span class='int_p' style='white-space:nowrap'> &nbsp;<b>:</b>&nbsp;&nbsp;".htmlSpecChars($Return)."</span>";
    }
    $Signature=~s!\[\]![<span style='padding-left:2px;'>]</span>!g;
    $Signature=~s!operator=!operator<span style='padding-left:2px'>=</span>!g;
    $Signature=~s!(\[in-charge\]|\[not-in-charge\]|\[in-charge-deleting\]|\[static\])!<span style='color:Black;font-weight:normal;'>$1</span>!g;
    return $Signature.(($SymbolVersion)?"<span class='symver'>&nbsp;$VersionSpec&nbsp;$SymbolVersion</span>":"");
}

sub get_signature_parts($$)
{
    my ($Signature, $Comma) = @_;
    my @Parts = ();
    my $Parameters = $Signature;
    my $ShortName = substr($Parameters, 0, get_signature_center($Parameters));
    $Parameters=~s/\A\Q$ShortName\E\(//g;
    $Parameters=~s/\)(| const)(| \[static\])\Z//g;
    return separate_params($Parameters, $Comma);
}

sub separate_params($$)
{
    my ($Params, $Comma) = @_;
    my @Parts = ();
    my ($Bracket_Num, $Bracket2_Num, $Part_Num) = (0, 0, 0);
    foreach my $Pos (0 .. length($Params) - 1)
    {
        my $Symbol = substr($Params, $Pos, 1);
        $Bracket_Num += 1 if($Symbol eq "(");
        $Bracket_Num -= 1 if($Symbol eq ")");
        $Bracket2_Num += 1 if($Symbol eq "<");
        $Bracket2_Num -= 1 if($Symbol eq ">");
        if($Symbol eq "," and $Bracket_Num==0 and $Bracket2_Num==0)
        {
            if($Comma) {# include comma
                $Parts[$Part_Num] .= $Symbol;
            }
            $Part_Num += 1;
        }
        else {
            $Parts[$Part_Num] .= $Symbol;
        }
    }
    return @Parts;
}

sub get_signature_center($)
{
    my $Signature = $_[0];
    my ($Bracket_Num, $Bracket2_Num) = (0, 0);
    my $Center = 0;
    if($Signature=~s/(operator([<>\-\=\*]+|\(\)))//g)
    {# operators: (),->,->*,<,<=,<<,<<=,>,>=,>>,>>=
        $Center+=length($1);
    }
    foreach my $Pos (0 .. length($Signature)-1)
    {
        my $Symbol = substr($Signature, $Pos, 1);
        $Bracket_Num += 1 if($Symbol eq "(");
        $Bracket_Num -= 1 if($Symbol eq ")");
        $Bracket2_Num += 1 if($Symbol eq "<");
        $Bracket2_Num -= 1 if($Symbol eq ">");
        if($Symbol eq "(" and
        $Bracket_Num==1 and $Bracket2_Num==0) {
            return $Center;
        }
        $Center+=1;
    }
    return 0;
}

sub testSystemCpp()
{
    print "verifying detectable C++ library changes\n";
    my ($HEADER1, $SOURCE1, $HEADER2, $SOURCE2) = ();
    my $DECL_SPEC = ($OSgroup eq "windows")?"__declspec( dllexport )":"";
    my $EXTERN = ($OSgroup eq "windows")?"extern ":"";# add "extern" for CL compiler
    
    # Begin namespace
    $HEADER1 .= "namespace TestNS {\n";
    $HEADER2 .= "namespace TestNS {\n";
    $SOURCE1 .= "namespace TestNS {\n";
    $SOURCE2 .= "namespace TestNS {\n";
    # Removed_Virtual_Method (inline)
    $HEADER1 .= "
        class $DECL_SPEC RemovedInlineVirtualFunction {
        public:
            virtual int removedMethod(int param) { return 0; }
            virtual int method(int param);
    };";
    $SOURCE1 .= "
        int RemovedInlineVirtualFunction::method(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedInlineVirtualFunction {
        public:
            virtual int method(int param);
        };";
    $SOURCE2 .= "
        int RemovedInlineVirtualFunction::method(int param) { return param; }";
    
    # MethodPtr
    $HEADER1 .= "
        class TestMethodPtr {
            public:
                typedef void (TestMethodPtr::*Method)(int*);
                Method _method;
                TestMethodPtr();
                void method();
        };";
    $SOURCE1 .= "
        TestMethodPtr::TestMethodPtr() { }
        void TestMethodPtr::method() { }";
    
    $HEADER2 .= "
        class TestMethodPtr {
            public:
                typedef void (TestMethodPtr::*Method)(int*, void*);
                Method _method;
                TestMethodPtr();
                void method();
        };";
    $SOURCE2 .= "
        TestMethodPtr::TestMethodPtr() { }
        void TestMethodPtr::method() { }";
    
    # FieldPtr
    $HEADER1 .= "
        class TestFieldPtr {
            public:
                typedef void* (TestFieldPtr::*Field);
                Field _field;
                TestFieldPtr();
                void method(void*);
        };";
    $SOURCE1 .= "
        TestFieldPtr::TestFieldPtr(){ }
        void TestFieldPtr::method(void*) { }";
    
    $HEADER2 .= "
        class TestFieldPtr {
            public:
                typedef int (TestFieldPtr::*Field);
                Field _field;
                TestFieldPtr();
                void method(void*);
        };";
    $SOURCE2 .= "
        TestFieldPtr::TestFieldPtr(){ }
        void TestFieldPtr::method(void*) { }";

    # Removed_Symbol (Template Specializations)
    $HEADER1 .= "
        template <unsigned int _TP, typename AAA>
        class Template {
            public:
                char const *field;
        };
        template <unsigned int _TP, typename AAA>
        class TestRemovedTemplate {
            public:
                char const *field;
                void method(int);
        };
        template <>
        class TestRemovedTemplate<7, char> {
            public:
                char const *field;
                void method(int);
        };";
    $SOURCE1 .= "
        void TestRemovedTemplate<7, char>::method(int){ }";

    # Removed_Symbol (Template Specializations)
    $HEADER1 .= "
        template <typename TName>
        int removedTemplateSpec(TName);

        template <> int removedTemplateSpec<char>(char);";
    $SOURCE1 .= "
        template <> int removedTemplateSpec<char>(char){return 0;}";
    
    # Global_Data_Value (int)
    $HEADER1 .= "
        $EXTERN $DECL_SPEC const int globalDataValue = 10;";
    
    $HEADER2 .= "
        $EXTERN $DECL_SPEC const int globalDataValue = 15;";
    
    # Global_Data_Became_Non_Const (int)
    $HEADER1 .= "
        $EXTERN $DECL_SPEC const int globalDataBecameNonConst = 10;";
    
    $HEADER2 .= "
        extern $DECL_SPEC int globalDataBecameNonConst;";
    $SOURCE2 .= "
        int globalDataBecameNonConst = 15;";

    # Global_Data_Became_Const (safe)
    $HEADER1 .= "
        extern $DECL_SPEC int globalDataBecameConst;";
    $SOURCE1 .= "
        int globalDataBecameConst = 10;";
    
    $HEADER2 .= "
        $EXTERN $DECL_SPEC const int globalDataBecameConst=15;";
    
    # Removed_Field (Ref)
    $HEADER1 .= "
        struct TestRefChange {
            int a, b, c;
        };
        $DECL_SPEC int paramRefChange(const TestRefChange & p1, int p2);";
    $SOURCE1 .= "
        int paramRefChange(const TestRefChange & p1, int p2) { return p2; }";
    
    $HEADER2 .= "
        struct TestRefChange {
            int a, b;
        };
        $DECL_SPEC int paramRefChange(const TestRefChange & p1, int p2);";
    $SOURCE2 .= "
        int paramRefChange(const TestRefChange & p1, int p2) { return p2; }";
    
    # Removed_Parameter
    $HEADER1 .= "
        $DECL_SPEC int removedParameter(int param, int removed_param);";
    $SOURCE1 .= "
        int removedParameter(int param, int removed_param) { return 0; }";
    
    $HEADER2 .= "
        $DECL_SPEC int removedParameter(int param);";
    $SOURCE2 .= "
        int removedParameter(int param) { return 0; }";
    
    # Added_Parameter
    $HEADER1 .= "
        $DECL_SPEC int addedParameter(int param);";
    $SOURCE1 .= "
        int addedParameter(int param) { return 0; }";
    
    $HEADER2 .= "
        $DECL_SPEC int addedParameter(int param, int added_param);";
    $SOURCE2 .= "
        int addedParameter(int param, int added_param) { return 0; }";
    
    # Added
    $HEADER2 .= "
        typedef int (*FUNCPTR_TYPE)(int a, int b);
        $DECL_SPEC int addedFunc(FUNCPTR_TYPE*const** f);";
    $SOURCE2 .= "
        int addedFunc(FUNCPTR_TYPE*const** f) { return 0; }";
    
    # Added_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC AddedVirtualMethod {
        public:
            virtual int method(int param);
        };";
    $SOURCE1 .= "
        int AddedVirtualMethod::method(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC AddedVirtualMethod {
        public:
            virtual int addedMethod(int param);
            virtual int method(int param);
        };";
    $SOURCE2 .= "
        int AddedVirtualMethod::addedMethod(int param) {
            return param;
        }
        int AddedVirtualMethod::method(int param) { return param; }";

    # Added_Virtual_Method (added "virtual" attribute)
    $HEADER1 .= "
        class $DECL_SPEC BecameVirtualMethod {
        public:
            int becameVirtual(int param);
            virtual int method(int param);
        };";
    $SOURCE1 .= "
        int BecameVirtualMethod::becameVirtual(int param) { return param; }
        int BecameVirtualMethod::method(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC BecameVirtualMethod {
        public:
            virtual int becameVirtual(int param);
            virtual int method(int param);
        };";
    $SOURCE2 .= "
        int BecameVirtualMethod::becameVirtual(int param) { return param; }
        int BecameVirtualMethod::method(int param) { return param; }";

    # Added_Pure_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC AddedPureVirtualMethod {
        public:
            virtual int method(int param);
            int otherMethod(int param);
        };";
    $SOURCE1 .= "
        int AddedPureVirtualMethod::method(int param) { return param; }
        int AddedPureVirtualMethod::otherMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC AddedPureVirtualMethod {
        public:
            virtual int addedMethod(int param)=0;
            virtual int method(int param);
            int otherMethod(int param);
        };";
    $SOURCE2 .= "
        int AddedPureVirtualMethod::method(int param) { return param; }
        int AddedPureVirtualMethod::otherMethod(int param) { return param; }";

    # Added_Virtual_Method_At_End (Safe)
    $HEADER1 .= "
        class $DECL_SPEC AddedVirtualMethodAtEnd {
        public:
            AddedVirtualMethodAtEnd();
            int method1(int param);
            virtual int method2(int param);
        };";
    $SOURCE1 .= "
        AddedVirtualMethodAtEnd::AddedVirtualMethodAtEnd() { }
        int AddedVirtualMethodAtEnd::method1(int param) { return param; }
        int AddedVirtualMethodAtEnd::method2(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC AddedVirtualMethodAtEnd {
        public:
            AddedVirtualMethodAtEnd();
            int method1(int param);
            virtual int method2(int param);
            virtual int addedMethod(int param);
        };";
    $SOURCE2 .= "
        AddedVirtualMethodAtEnd::AddedVirtualMethodAtEnd() { }
        int AddedVirtualMethodAtEnd::method1(int param) { return param; }
        int AddedVirtualMethodAtEnd::method2(int param) { return param; }
        int AddedVirtualMethodAtEnd::addedMethod(int param) { return param; }";

    # Added_Virtual_Method_At_End (With Default Constructor)
    $HEADER1 .= "
        class $DECL_SPEC AddedVirtualMethodAtEnd_DefaultConstructor {
        public:
            int method1(int param);
            virtual int method2(int param);
        };";
    $SOURCE1 .= "
        int AddedVirtualMethodAtEnd_DefaultConstructor::method1(int param) { return param; }
        int AddedVirtualMethodAtEnd_DefaultConstructor::method2(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC AddedVirtualMethodAtEnd_DefaultConstructor {
        public:
            int method1(int param);
            virtual int method2(int param);
            virtual int addedMethod(int param);
        };";
    $SOURCE2 .= "
        int AddedVirtualMethodAtEnd_DefaultConstructor::method1(int param) { return param; }
        int AddedVirtualMethodAtEnd_DefaultConstructor::method2(int param) { return param; }
        int AddedVirtualMethodAtEnd_DefaultConstructor::addedMethod(int param) { return param; }";
    
    # Added_First_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC AddedFirstVirtualMethod {
        public:
            int method(int param);
        };";
    $SOURCE1 .= "
        int AddedFirstVirtualMethod::method(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC AddedFirstVirtualMethod {
        public:
            int method(int param);
            virtual int addedMethod(int param);
        };";
    $SOURCE2 .= "
        int AddedFirstVirtualMethod::method(int param) { return param; }
        int AddedFirstVirtualMethod::addedMethod(int param) { return param; }";
    
    # Removed_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC RemovedVirtualFunction {
        public:
            virtual int removedMethod(int param);
            virtual int vMethod(int param);
    };";
    $SOURCE1 .= "
        int RemovedVirtualFunction::removedMethod(int param) { return param; }
        int RemovedVirtualFunction::vMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedVirtualFunction {
        public:
            int removedMethod(int param);
            virtual int vMethod(int param);
    };";
    $SOURCE2 .= "
        int RemovedVirtualFunction::removedMethod(int param) { return param; }
        int RemovedVirtualFunction::vMethod(int param) { return param; }";

    # Removed_Virtual_Method (Pure, From the End)
    $HEADER1 .= "
        class $DECL_SPEC RemovedPureVirtualMethodFromEnd {
        public:
            virtual int method(int param);
            virtual int removedMethod(int param)=0;
        };";
    $SOURCE1 .= "
        int RemovedPureVirtualMethodFromEnd::method(int param) { return param; }
        int RemovedPureVirtualMethodFromEnd::removedMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedPureVirtualMethodFromEnd
        {
        public:
            virtual int method(int param);
            int removedMethod(int param);
        };";
    $SOURCE2 .= "
        int RemovedPureVirtualMethodFromEnd::method(int param) { return param; }
        int RemovedPureVirtualMethodFromEnd::removedMethod(int param) { return param; }";

    # Removed_Symbol (Pure with Implementation)
    $HEADER1 .= "
        class $DECL_SPEC RemovedPureSymbol {
        public:
            virtual int method(int param);
            virtual int removedMethod(int param)=0;
        };";
    $SOURCE1 .= "
        int RemovedPureSymbol::method(int param) { return param; }
        int RemovedPureSymbol::removedMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedPureSymbol
        {
        public:
            virtual int method(int param);
        };";
    $SOURCE2 .= "
        int RemovedPureSymbol::method(int param) { return param; }";

    # Removed_Virtual_Method (From the End)
    $HEADER1 .= "
        class $DECL_SPEC RemovedVirtualMethodFromEnd {
        public:
            virtual int method(int param);
            virtual int removedMethod(int param);
        };";
    $SOURCE1 .= "
        int RemovedVirtualMethodFromEnd::method(int param) { return param; }
        int RemovedVirtualMethodFromEnd::removedMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedVirtualMethodFromEnd
        {
        public:
            virtual int method(int param);
            int removedMethod(int param);
        };";
    $SOURCE2 .= "
        int RemovedVirtualMethodFromEnd::method(int param) { return param; }
        int RemovedVirtualMethodFromEnd::removedMethod(int param) { return param; }";

    # Removed_Last_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC RemovedLastVirtualMethod
        {
        public:
            int method(int param);
            virtual int removedMethod(int param);
        };";
    $SOURCE1 .= "
        int RemovedLastVirtualMethod::method(int param) { return param; }";
    $SOURCE1 .= "
        int RemovedLastVirtualMethod::removedMethod(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC RemovedLastVirtualMethod
        {
        public:
            int method(int param);
            int removedMethod(int param);
        };";
    $SOURCE2 .= "
        int RemovedLastVirtualMethod::method(int param) { return param; }";
    $SOURCE2 .= "
        int RemovedLastVirtualMethod::removedMethod(int param) { return param; }";
    
    # Virtual_Table_Size
    $HEADER1 .= "
        class $DECL_SPEC VirtualTableSize
        {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };
        class $DECL_SPEC VirtualTableSize_SubClass: public VirtualTableSize
        {
        public:
            virtual int method3(int param);
            virtual int method4(int param);
        };";
    $SOURCE1 .= "
        int VirtualTableSize::method1(int param) { return param; }
        int VirtualTableSize::method2(int param) { return param; }
        int VirtualTableSize_SubClass::method3(int param) { return param; }
        int VirtualTableSize_SubClass::method4(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC VirtualTableSize
        {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
            virtual int addedMethod(int param);
        };
        class $DECL_SPEC VirtualTableSize_SubClass: public VirtualTableSize
        {
        public:
            virtual int method3(int param);
            virtual int method4(int param);
        };";
    $SOURCE2 .= "
        int VirtualTableSize::method1(int param) { return param; }
        int VirtualTableSize::method2(int param) { return param; }
        int VirtualTableSize::addedMethod(int param) { return param; }
        int VirtualTableSize_SubClass::method3(int param) { return param; }
        int VirtualTableSize_SubClass::method4(int param) { return param; }";
    
    # Virtual_Method_Position
    $HEADER1 .= "
        class $DECL_SPEC VirtualMethodPosition
        {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };";
    $SOURCE1 .= "
        int VirtualMethodPosition::method1(int param) { return param; }";
    $SOURCE1 .= "
        int VirtualMethodPosition::method2(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC VirtualMethodPosition
        {
        public:
            virtual int method2(int param);
            virtual int method1(int param);
        };";
    $SOURCE2 .= "
        int VirtualMethodPosition::method1(int param) { return param; }";
    $SOURCE2 .= "
        int VirtualMethodPosition::method2(int param) { return param; }";

    # Pure_Virtual_Method_Position
    $HEADER1 .= "
        class $DECL_SPEC PureVirtualFunctionPosition {
        public:
            virtual int method1(int param)=0;
            virtual int method2(int param)=0;
            int method3(int param);
        };";
    $SOURCE1 .= "
        int PureVirtualFunctionPosition::method3(int param) { return method1(7)+method2(7); }";
    
    $HEADER2 .= "
        class $DECL_SPEC PureVirtualFunctionPosition {
        public:
            virtual int method2(int param)=0;
            virtual int method1(int param)=0;
            int method3(int param);
        };";
    $SOURCE2 .= "
        int PureVirtualFunctionPosition::method3(int param) { return method1(7)+method2(7); }";

    # Virtual_Method_Position
    $HEADER1 .= "
        class $DECL_SPEC VirtualFunctionPosition {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };";
    $SOURCE1 .= "
        int VirtualFunctionPosition::method1(int param) { return 1; }
        int VirtualFunctionPosition::method2(int param) { return 2; }";
    
    $HEADER2 .= "
        class $DECL_SPEC VirtualFunctionPosition {
        public:
            virtual int method2(int param);
            virtual int method1(int param);
        };";
    $SOURCE2 .= "
        int VirtualFunctionPosition::method1(int param) { return 1; }
        int VirtualFunctionPosition::method2(int param) { return 2; }";
    
    # Virtual_Method_Position (safe)
    $HEADER1 .= "
        class $DECL_SPEC VirtualFunctionPositionSafe_Base {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };
        class $DECL_SPEC VirtualFunctionPositionSafe: public VirtualFunctionPositionSafe_Base {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };";
    $SOURCE1 .= "
        int VirtualFunctionPositionSafe_Base::method1(int param) { return param; }
        int VirtualFunctionPositionSafe_Base::method2(int param) { return param; }
        int VirtualFunctionPositionSafe::method1(int param) { return param; }
        int VirtualFunctionPositionSafe::method2(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC VirtualFunctionPositionSafe_Base {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };
        class $DECL_SPEC VirtualFunctionPositionSafe: public VirtualFunctionPositionSafe_Base {
        public:
            virtual int method2(int param);
            virtual int method1(int param);
        };";
    $SOURCE2 .= "
        int VirtualFunctionPositionSafe_Base::method1(int param) { return param; }
        int VirtualFunctionPositionSafe_Base::method2(int param) { return param; }
        int VirtualFunctionPositionSafe::method1(int param) { return param; }
        int VirtualFunctionPositionSafe::method2(int param) { return param; }";
    
    # Overridden_Virtual_Method
    $HEADER1 .= "
        class $DECL_SPEC OverriddenVirtualMethod_Base {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };
        class $DECL_SPEC OverriddenVirtualMethod: public OverriddenVirtualMethod_Base {
        public:
            OverriddenVirtualMethod();
            virtual int method3(int param);
        };";
    $SOURCE1 .= "
        int OverriddenVirtualMethod_Base::method1(int param) { return param; }
        int OverriddenVirtualMethod_Base::method2(int param) { return param; }
        OverriddenVirtualMethod::OverriddenVirtualMethod() {}
        int OverriddenVirtualMethod::method3(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC OverriddenVirtualMethod_Base {
        public:
            virtual int method1(int param);
            virtual int method2(int param);
        };
        class $DECL_SPEC OverriddenVirtualMethod:public OverriddenVirtualMethod_Base {
            OverriddenVirtualMethod();
            virtual int method2(int param);
            virtual int method3(int param);
        };";
    $SOURCE2 .= "
        int OverriddenVirtualMethod_Base::method1(int param) { return param; }
        int OverriddenVirtualMethod_Base::method2(int param) { return param; }
        OverriddenVirtualMethod::OverriddenVirtualMethod() {}
        int OverriddenVirtualMethod::method2(int param) { return param; }
        int OverriddenVirtualMethod::method3(int param) { return param; }";

    # Overridden_Virtual_Method_B (+ removed)
    $HEADER1 .= "
        
    class $DECL_SPEC OverriddenVirtualMethodB: public OverriddenVirtualMethod_Base {
        public:
            OverriddenVirtualMethodB();
            virtual int method2(int param);
            virtual int method3(int param);
    };";
    $SOURCE1 .= "
        OverriddenVirtualMethodB::OverriddenVirtualMethodB() {}
        int OverriddenVirtualMethodB::method2(int param) { return param; }
        int OverriddenVirtualMethodB::method3(int param) { return param; }";
    
    $HEADER2 .= "
        
    class $DECL_SPEC OverriddenVirtualMethodB:public OverriddenVirtualMethod_Base {
        public:
            OverriddenVirtualMethodB();
            virtual int method3(int param);
    };";
    $SOURCE2 .= "
        OverriddenVirtualMethodB::OverriddenVirtualMethodB() {}
        int OverriddenVirtualMethodB::method3(int param) { return param; }";
    
    # Size
    $HEADER1 .= "
        struct $DECL_SPEC TypeSize
        {
        public:
            TypeSize method(TypeSize param);
            int i[5];
            long j;
            double k;
            TypeSize* p;
        };";
    $SOURCE1 .= "
        TypeSize TypeSize::method(TypeSize param) { return param; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC TypeSize
        {
        public:
            TypeSize method(TypeSize param);
            int i[15];
            long j;
            double k;
            TypeSize* p;
            int added_member;
        };";
    $SOURCE2 .= "
        TypeSize TypeSize::method(TypeSize param) { return param; }";

    # Size_Of_Allocable_Class_Increased
    $HEADER1 .= "
        class $DECL_SPEC AllocableClassSize
        {
        public:
            AllocableClassSize();
            int method();
            double p[5];
        };";
    $SOURCE1 .= "
        AllocableClassSize::AllocableClassSize() { }";
    $SOURCE1 .= "
        int AllocableClassSize::method() { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AllocableClassSize
        {
        public:
            AllocableClassSize();
            int method();
            double p[15];
        };";
    $SOURCE2 .= "
        AllocableClassSize::AllocableClassSize() { }";
    $SOURCE2 .= "
        int AllocableClassSize::method() { return 0; }";
    
    # Size_Of_Allocable_Class_Decreased (decreased size, has derived class, has public members)
    $HEADER1 .= "
        class $DECL_SPEC DecreasedClassSize
        {
        public:
            DecreasedClassSize();
            int method();
            double p[15];
        };";
    $SOURCE1 .= "
        DecreasedClassSize::DecreasedClassSize() { }";
    $SOURCE1 .= "
        int DecreasedClassSize::method() { return 0; }";
    $HEADER1 .= "
        class $DECL_SPEC DecreasedClassSize_SubClass: public DecreasedClassSize
        {
        public:
            DecreasedClassSize_SubClass();
            int method();
            int f;
        };";
    $SOURCE1 .= "
        DecreasedClassSize_SubClass::DecreasedClassSize_SubClass() { f=7; }";
    $SOURCE1 .= "
        int DecreasedClassSize_SubClass::method() { return f; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC DecreasedClassSize
        {
        public:
            DecreasedClassSize();
            int method();
            double p[5];
        };";
    $SOURCE2 .= "
        DecreasedClassSize::DecreasedClassSize() { }";
    $SOURCE2 .= "
        int DecreasedClassSize::method() { return 0; }";
    $HEADER2 .= "
        class $DECL_SPEC DecreasedClassSize_SubClass: public DecreasedClassSize
        {
        public:
            DecreasedClassSize_SubClass();
            int method();
            int f;
        };";
    $SOURCE2 .= "
        DecreasedClassSize_SubClass::DecreasedClassSize_SubClass() { f=7; }";
    $SOURCE2 .= "
        int DecreasedClassSize_SubClass::method() { return f; }";

    # Size_Of_Copying_Class
    $HEADER1 .= "
        class $DECL_SPEC CopyingClassSize
        {
        public:
            int method();
            int p[5];
        };";
    $SOURCE1 .= "
        int CopyingClassSize::method() { return p[4]; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC CopyingClassSize
        {
        public:
            int method();
            int p[15];
        };";
    $SOURCE2 .= "
        int CopyingClassSize::method() { return p[10]; }";

    # Base_Class_Became_Virtually_Inherited
    $HEADER1 .= "
        class $DECL_SPEC BecameVirtualBase
        {
        public:
            BecameVirtualBase();
            int method();
            double p[5];
        };";
    $SOURCE1 .= "
        BecameVirtualBase::BecameVirtualBase() { }";
    $SOURCE1 .= "
        int BecameVirtualBase::method() { return 0; }";
    $HEADER1 .= "
        class $DECL_SPEC AddedVirtualBase1:public BecameVirtualBase
        {
        public:
            AddedVirtualBase1();
            int method();
        };";
    $SOURCE1 .= "
        AddedVirtualBase1::AddedVirtualBase1() { }";
    $SOURCE1 .= "
        int AddedVirtualBase1::method() { return 0; }";
    $HEADER1 .= "
        class $DECL_SPEC AddedVirtualBase2: public BecameVirtualBase
        {
        public:
            AddedVirtualBase2();
            int method();
        };";
    $SOURCE1 .= "
        AddedVirtualBase2::AddedVirtualBase2() { }";
    $SOURCE1 .= "
        int AddedVirtualBase2::method() { return 0; }";
    $HEADER1 .= "
        class $DECL_SPEC BaseClassBecameVirtuallyInherited:public AddedVirtualBase1, public AddedVirtualBase2
        {
        public:
            BaseClassBecameVirtuallyInherited();
        };";
    $SOURCE1 .= "
        BaseClassBecameVirtuallyInherited::BaseClassBecameVirtuallyInherited() { }";
    
    $HEADER2 .= "
        class $DECL_SPEC BecameVirtualBase
        {
        public:
            BecameVirtualBase();
            int method();
            double p[5];
        };";
    $SOURCE2 .= "
        BecameVirtualBase::BecameVirtualBase() { }";
    $SOURCE2 .= "
        int BecameVirtualBase::method() { return 0; }";
    $HEADER2 .= "
        class $DECL_SPEC AddedVirtualBase1:public virtual BecameVirtualBase
        {
        public:
            AddedVirtualBase1();
            int method();
        };";
    $SOURCE2 .= "
        AddedVirtualBase1::AddedVirtualBase1() { }";
    $SOURCE2 .= "
        int AddedVirtualBase1::method() { return 0; }";
    $HEADER2 .= "
        class $DECL_SPEC AddedVirtualBase2: public virtual BecameVirtualBase
        {
        public:
            AddedVirtualBase2();
            int method();
        };";
    $SOURCE2 .= "
        AddedVirtualBase2::AddedVirtualBase2() { }";
    $SOURCE2 .= "
        int AddedVirtualBase2::method() { return 0; }";
    $HEADER2 .= "
        class $DECL_SPEC BaseClassBecameVirtuallyInherited:public AddedVirtualBase1, public AddedVirtualBase2
        {
        public:
            BaseClassBecameVirtuallyInherited();
        };";
    $SOURCE2 .= "
        BaseClassBecameVirtuallyInherited::BaseClassBecameVirtuallyInherited() { }";

    # Added_Base_Class, Removed_Base_Class
    $HEADER1 .= "
        class $DECL_SPEC BaseClass
        {
        public:
            BaseClass();
            int method();
            double p[5];
        };
        class $DECL_SPEC RemovedBaseClass
        {
        public:
            RemovedBaseClass();
            int method();
        };
        class $DECL_SPEC ChangedBaseClass:public BaseClass, public RemovedBaseClass
        {
        public:
            ChangedBaseClass();
        };";
    $SOURCE1 .= "
        BaseClass::BaseClass() { }
        int BaseClass::method() { return 0; }
        RemovedBaseClass::RemovedBaseClass() { }
        int RemovedBaseClass::method() { return 0; }
        ChangedBaseClass::ChangedBaseClass() { }";
    
    $HEADER2 .= "
        class $DECL_SPEC BaseClass
        {
        public:
            BaseClass();
            int method();
            double p[5];
        };
        class $DECL_SPEC AddedBaseClass
        {
        public:
            AddedBaseClass();
            int method();
        };
        class $DECL_SPEC ChangedBaseClass:public BaseClass, public AddedBaseClass
        {
        public:
            ChangedBaseClass();
        };";
    $SOURCE2 .= "
        BaseClass::BaseClass() { }
        int BaseClass::method() { return 0; }
        AddedBaseClass::AddedBaseClass() { }
        int AddedBaseClass::method() { return 0; }
        ChangedBaseClass::ChangedBaseClass() { }";

    # Added_Base_Class_And_Shift, Removed_Base_Class_And_Shift
    $HEADER1 .= "
        struct $DECL_SPEC BaseClass2
        {
            BaseClass2();
            int method();
            double p[15];
        };
        class $DECL_SPEC ChangedBaseClassAndSize:public BaseClass
        {
        public:
            ChangedBaseClassAndSize();
        };";
    $SOURCE1 .= "
        BaseClass2::BaseClass2() { }
        int BaseClass2::method() { return 0; }
        ChangedBaseClassAndSize::ChangedBaseClassAndSize() { }";
    
    $HEADER2 .= "
        struct $DECL_SPEC BaseClass2
        {
            BaseClass2();
            int method();
            double p[15];
        };
        class $DECL_SPEC ChangedBaseClassAndSize:public BaseClass2
        {
        public:
            ChangedBaseClassAndSize();
        };";
    $SOURCE2 .= "
        BaseClass2::BaseClass2() { }
        int BaseClass2::method() { return 0; }
        ChangedBaseClassAndSize::ChangedBaseClassAndSize() { }";
    
    # Added_Field_And_Size
    $HEADER1 .= "
        struct $DECL_SPEC AddedFieldAndSize
        {
            int method(AddedFieldAndSize param);
            double i, j, k;
            AddedFieldAndSize* p;
        };";
    $SOURCE1 .= "
        int AddedFieldAndSize::method(AddedFieldAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AddedFieldAndSize
        {
            int method(AddedFieldAndSize param);
            double i, j, k;
            AddedFieldAndSize* p;
            int added_member1;
            long long added_member2;
        };";
    $SOURCE2 .= "
        int AddedFieldAndSize::method(AddedFieldAndSize param) { return 0; }";
    
    # Added_Field
    $HEADER1 .= "
        class $DECL_SPEC ObjectAddedMember
        {
        public:
            int method(int param);
            double i, j, k;
            AddedFieldAndSize* p;
        };";
    $SOURCE1 .= "
        int ObjectAddedMember::method(int param) { return param; }";
    
    $HEADER2 .= "
        class $DECL_SPEC ObjectAddedMember
        {
        public:
            int method(int param);
            double i, j, k;
            AddedFieldAndSize* p;
            int added_member1;
            long long added_member2;
        };";
    $SOURCE2 .= "
        int ObjectAddedMember::method(int param) { return param; }";
    
    # Added_Field (safe)
    $HEADER1 .= "
        struct $DECL_SPEC AddedBitfield
        {
            int method(AddedBitfield param);
            double i, j, k;
            int b1 : 32;
            int b2 : 31;
            AddedBitfield* p;
        };";
    $SOURCE1 .= "
        int AddedBitfield::method(AddedBitfield param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AddedBitfield
        {
            int method(AddedBitfield param);
            double i, j, k;
            int b1 : 32;
            int b2 : 31;
            int added_bitfield : 1;
            AddedBitfield* p;
        };";
    $SOURCE2 .= "
        int AddedBitfield::method(AddedBitfield param) { return 0; }";
    
    # BitField_Size
    $HEADER1 .= "
        struct $DECL_SPEC BitfieldSize
        {
            int method(BitfieldSize param);
            short changed_bitfield : 1;
        };";
    $SOURCE1 .= "
        int BitfieldSize::method(BitfieldSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC BitfieldSize
        {
            int method(BitfieldSize param);
            short changed_bitfield : 7;
        };";
    $SOURCE2 .= "
        int BitfieldSize::method(BitfieldSize param) { return 0; }";
    
    # Removed_Field
    $HEADER1 .= "
        struct $DECL_SPEC RemovedBitfield
        {
            int method(RemovedBitfield param);
            double i, j, k;
            int b1 : 32;
            int b2 : 31;
            int removed_bitfield : 1;
            RemovedBitfield* p;
        };";
    $SOURCE1 .= "
        int RemovedBitfield::method(RemovedBitfield param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC RemovedBitfield
        {
            int method(RemovedBitfield param);
            double i, j, k;
            int b1 : 32;
            int b2 : 31;
            RemovedBitfield* p;
        };";
    $SOURCE2 .= "
        int RemovedBitfield::method(RemovedBitfield param) { return 0; }";
    
    # Added_Middle_Field_And_Size
    $HEADER1 .= "
        struct $DECL_SPEC AddedMiddleFieldAndSize
        {
            int method(AddedMiddleFieldAndSize param);
            int i;
            long j;
            double k;
            AddedMiddleFieldAndSize* p;
        };";
    $SOURCE1 .= "
        int AddedMiddleFieldAndSize::method(AddedMiddleFieldAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AddedMiddleFieldAndSize
        {
            int method(AddedMiddleFieldAndSize param);
            int i;
            int added_middle_member;
            long j;
            double k;
            AddedMiddleFieldAndSize* p;
        };";
    $SOURCE2 .= "
        int AddedMiddleFieldAndSize::method(AddedMiddleFieldAndSize param) { return 0; }";
    
    # Renamed_Field
    $HEADER1 .= "
        struct $DECL_SPEC RenamedField
        {
            int method(RenamedField param);
            long i;
            long j;
            double k;
            RenamedField* p;
        };";
    $SOURCE1 .= "
        int RenamedField::method(RenamedField param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC RenamedField
        {
            int method(RenamedField param);
            long renamed_member;
            long j;
            double k;
            RenamedField* p;
        };";
    $SOURCE2 .= "
        int RenamedField::method(RenamedField param) { return 0; }";
    
    # Removed_Field_And_Size
    $HEADER1 .= "
        struct $DECL_SPEC RemovedFieldAndSize
        {
            int method(RemovedFieldAndSize param);
            double i, j, k;
            RemovedFieldAndSize* p;
            int removed_member1;
            long removed_member2;
        };";
    $SOURCE1 .= "
        int RemovedFieldAndSize::method(RemovedFieldAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC RemovedFieldAndSize
        {
            int method(RemovedFieldAndSize param);
            double i, j, k;
            RemovedFieldAndSize* p;
        };";
    $SOURCE2 .= "
        int RemovedFieldAndSize::method(RemovedFieldAndSize param) { return 0; }";
    
    # Removed_Middle_Field_And_Size
    $HEADER1 .= "
        struct $DECL_SPEC RemovedMiddleFieldAndSize
        {
            int method(RemovedMiddleFieldAndSize param);
            int i;
            int removed_middle_member;
            long j;
            double k;
            RemovedMiddleFieldAndSize* p;
        };";
    $SOURCE1 .= "
        int RemovedMiddleFieldAndSize::method(RemovedMiddleFieldAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC RemovedMiddleFieldAndSize
        {
            int method(RemovedMiddleFieldAndSize param);
            int i;
            long j;
            double k;
            RemovedMiddleFieldAndSize* p;
        };";
    $SOURCE2 .= "
        int RemovedMiddleFieldAndSize::method(RemovedMiddleFieldAndSize param) { return 0; }";
    
    # Enum_Member_Value
    $HEADER1 .= "
        enum EnumMemberValue
        {
            MEMBER_1=1,
            MEMBER_2=2
        };";
    $HEADER1 .= "
        $DECL_SPEC int enumMemberValueChange(enum EnumMemberValue param);";
    $SOURCE1 .= "
        int enumMemberValueChange(enum EnumMemberValue param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    $HEADER2 .= "
        enum EnumMemberValue
        {
            MEMBER_1=2,
            MEMBER_2=1
        };";
    $HEADER2 .= "
        $DECL_SPEC int enumMemberValueChange(enum EnumMemberValue param);";
    $SOURCE2 .= "
        int enumMemberValueChange(enum EnumMemberValue param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    # Enum_Member_Name
    $HEADER1 .= "
        enum EnumMemberRename
        {
            BRANCH_1=1,
            BRANCH_2=2
        };";
    $HEADER1 .= "
        $DECL_SPEC int enumMemberRename(enum EnumMemberRename param);";
    $SOURCE1 .= "
        int enumMemberRename(enum EnumMemberRename param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    $HEADER2 .= "
        enum EnumMemberRename
        {
            BRANCH_FIRST=1,
            BRANCH_SECOND=2
        };";
    $HEADER2 .= "
        $DECL_SPEC int enumMemberRename(enum EnumMemberRename param);";
    $SOURCE2 .= "
        int enumMemberRename(enum EnumMemberRename param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    # Field_Type_And_Size
    $HEADER1 .= "
        struct $DECL_SPEC FieldTypeAndSize
        {
            int method(FieldTypeAndSize param);
            int i;
            long j;
            double k;
            FieldTypeAndSize* p;
        };";
    $SOURCE1 .= "
        int FieldTypeAndSize::method(FieldTypeAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC FieldTypeAndSize
        {
            int method(FieldTypeAndSize param);
            long long i;
            long j;
            double k;
            FieldTypeAndSize* p;
        };";
    $SOURCE2 .= "
        int FieldTypeAndSize::method(FieldTypeAndSize param) { return 0; }";
    
    # Member_Type
    $HEADER1 .= "
        struct $DECL_SPEC MemberType
        {
            int method(MemberType param);
            int i;
            long j;
            double k;
            MemberType* p;
        };";
    $SOURCE1 .= "
        int MemberType::method(MemberType param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC MemberType
        {
            int method(MemberType param);
            float i;
            long j;
            double k;
            MemberType* p;
        };";
    $SOURCE2 .= "
        int MemberType::method(MemberType param) { return 0; }";
    
    # Field_BaseType
    $HEADER1 .= "
        struct $DECL_SPEC FieldBaseType
        {
            int method(FieldBaseType param);
            int *i;
            long j;
            double k;
            FieldBaseType* p;
        };";
    $SOURCE1 .= "
        int FieldBaseType::method(FieldBaseType param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC FieldBaseType
        {
            int method(FieldBaseType param);
            long long *i;
            long j;
            double k;
            FieldBaseType* p;
        };";
    $SOURCE2 .= "
        int FieldBaseType::method(FieldBaseType param) { return 0; }";
    
    # Field_PointerLevel_Increased (and size)
    $HEADER1 .= "
        struct $DECL_SPEC FieldPointerLevelAndSize
        {
            int method(FieldPointerLevelAndSize param);
            long long i;
            long j;
            double k;
            FieldPointerLevelAndSize* p;
        };";
    $SOURCE1 .= "
        int FieldPointerLevelAndSize::method(FieldPointerLevelAndSize param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC FieldPointerLevelAndSize
        {
            int method(FieldPointerLevelAndSize param);
            long long *i;
            long j;
            double k;
            FieldPointerLevelAndSize* p;
        };";
    $SOURCE2 .= "
        int FieldPointerLevelAndSize::method(FieldPointerLevelAndSize param) { return 0; }";
    
    # Field_PointerLevel
    $HEADER1 .= "
        struct $DECL_SPEC FieldPointerLevel
        {
            int method(FieldPointerLevel param);
            int **i;
            long j;
            double k;
            FieldPointerLevel* p;
        };";
    $SOURCE1 .= "
        int FieldPointerLevel::method(FieldPointerLevel param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC FieldPointerLevel
        {
            int method(FieldPointerLevel param);
            int *i;
            long j;
            double k;
            FieldPointerLevel* p;
        };";
    $SOURCE2 .= "
        int FieldPointerLevel::method(FieldPointerLevel param) { return 0; }";
    
    # Added_Interface (method)
    $HEADER1 .= "
        struct $DECL_SPEC AddedInterface
        {
            int method(AddedInterface param);
            int i;
            long j;
            double k;
            AddedInterface* p;
        };";
    $SOURCE1 .= "
        int AddedInterface::method(AddedInterface param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AddedInterface
        {
            int method(AddedInterface param);
            int added_func(AddedInterface param);
            int i;
            long j;
            double k;
            AddedInterface* p;
        };";
    $SOURCE2 .= "
        int AddedInterface::method(AddedInterface param) { return 0; }";
    $SOURCE2 .= "
        int AddedInterface::added_func(AddedInterface param) { return 0; }";
    
    # Added_Interface (function)
    $HEADER2 .= "
        $DECL_SPEC int addedFunc2(void *** param);";
    $SOURCE2 .= "
        int addedFunc2(void *** param) { return 0; }";
    
    # Added_Interface (global variable)
    $HEADER1 .= "
        struct $DECL_SPEC AddedVariable
        {
            int method(AddedVariable param);
            int i1, i2;
            long j;
            double k;
            AddedVariable* p;
        };";
    $SOURCE1 .= "
        int AddedVariable::method(AddedVariable param) {
            return i1;
        }";
    
    $HEADER2 .= "
        struct $DECL_SPEC AddedVariable
        {
            int method(AddedVariable param);
            static int i1;
            static int i2;
            long j;
            double k;
            AddedVariable* p;
        };";
    $SOURCE2 .= "
        int AddedVariable::method(AddedVariable param) { return AddedVariable::i1; }";
    $SOURCE2 .= "
        int AddedVariable::i1=0;";
    $SOURCE2 .= "
        int AddedVariable::i2=0;";
    
    # Removed_Interface (method)
    $HEADER1 .= "
        struct $DECL_SPEC RemovedInterface
        {
            int method(RemovedInterface param);
            int removed_func(RemovedInterface param);
            int i;
            long j;
            double k;
            RemovedInterface* p;
        };";
    $SOURCE1 .= "
        int RemovedInterface::method(RemovedInterface param) { return 0; }";
    $SOURCE1 .= "
        int RemovedInterface::removed_func(RemovedInterface param) { return 0; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC RemovedInterface
        {
            int method(RemovedInterface param);
            int i;
            long j;
            double k;
            RemovedInterface* p;
        };";
    $SOURCE2 .= "
        int RemovedInterface::method(RemovedInterface param) { return 0; }";
    
    # Removed_Interface (function)
    $HEADER1 .= "
        $DECL_SPEC int removedFunc2(void *** param);";
    $SOURCE1 .= "
        int removedFunc2(void *** param) { return 0; }";
    
    # Method_Became_Static
    $HEADER1 .= "
        struct $DECL_SPEC MethodBecameStatic
        {
            MethodBecameStatic becameStatic(MethodBecameStatic param);
            int **i;
            long j;
            double k;
            MethodBecameStatic* p;
        };";
    $SOURCE1 .= "
        MethodBecameStatic MethodBecameStatic::becameStatic(MethodBecameStatic param) { return param; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC MethodBecameStatic
        {
            static MethodBecameStatic becameStatic(MethodBecameStatic param);
            int **i;
            long j;
            double k;
            MethodBecameStatic* p;
        };";
    $SOURCE2 .= "
        MethodBecameStatic MethodBecameStatic::becameStatic(MethodBecameStatic param) { return param; }";
    
    # Method_Became_NonStatic
    $HEADER1 .= "
        struct $DECL_SPEC MethodBecameNonStatic
        {
            static MethodBecameNonStatic becameNonStatic(MethodBecameNonStatic param);
            int **i;
            long j;
            double k;
            MethodBecameNonStatic* p;
        };";
    $SOURCE1 .= "
        MethodBecameNonStatic MethodBecameNonStatic::becameNonStatic(MethodBecameNonStatic param) { return param; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC MethodBecameNonStatic
        {
            MethodBecameNonStatic becameNonStatic(MethodBecameNonStatic param);
            int **i;
            long j;
            double k;
            MethodBecameNonStatic* p;
        };";
    $SOURCE2 .= "
        MethodBecameNonStatic MethodBecameNonStatic::becameNonStatic(MethodBecameNonStatic param) { return param; }";
    
    # Parameter_Type_And_Size
    $HEADER1 .= "
        $DECL_SPEC int funcParameterTypeAndSize(int param, int other_param);";
    $SOURCE1 .= "
        int funcParameterTypeAndSize(int param, int other_param) { return other_param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int funcParameterTypeAndSize(long long param, int other_param);";
    $SOURCE2 .= "
        int funcParameterTypeAndSize(long long param, int other_param) { return other_param; }";
    
    # Parameter_Type
    $HEADER1 .= "
        $DECL_SPEC int funcParameterType(int param, int other_param);";
    $SOURCE1 .= "
        int funcParameterType(int param, int other_param) { return other_param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int funcParameterType(float param, int other_param);";
    $SOURCE2 .= "
        int funcParameterType(float param, int other_param) { return other_param; }";
    
    # Parameter_BaseType
    $HEADER1 .= "
        $DECL_SPEC int funcParameterBaseType(int *param);";
    $SOURCE1 .= "
        int funcParameterBaseType(int *param) { return sizeof(*param); }";
    
    $HEADER2 .= "
        $DECL_SPEC int funcParameterBaseType(long long *param);";
    $SOURCE2 .= "
        int funcParameterBaseType(long long *param) { return sizeof(*param); }";
    
    # Parameter_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC long long funcParameterPointerLevelAndSize(long long param);";
    $SOURCE1 .= "
        long long funcParameterPointerLevelAndSize(long long param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long funcParameterPointerLevelAndSize(long long *param);";
    $SOURCE2 .= "
        long long funcParameterPointerLevelAndSize(long long *param) { return param[5]; }";
    
    # Parameter_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC int funcParameterPointerLevel(int *param);";
    $SOURCE1 .= "
        int funcParameterPointerLevel(int *param) { return param[5]; }";
    
    $HEADER2 .= "
        $DECL_SPEC int funcParameterPointerLevel(int **param);";
    $SOURCE2 .= "
        int funcParameterPointerLevel(int **param) { return param[5][5]; }";
    
    # Return_Type_And_Size
    $HEADER1 .= "
        $DECL_SPEC int funcReturnTypeAndSize(int param);";
    $SOURCE1 .= "
        int funcReturnTypeAndSize(int param) { return 2^(sizeof(int)*8-1)-1; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long funcReturnTypeAndSize(int param);";
    $SOURCE2 .= "
        long long funcReturnTypeAndSize(int param) { return 2^(sizeof(long long)*8-1)-1; }";
    
    # Return_Type
    $HEADER1 .= "
        $DECL_SPEC int funcReturnType(int param);";
    $SOURCE1 .= "
        int funcReturnType(int param) { return 2^(sizeof(int)*8-1)-1; }";
    
    $HEADER2 .= "
        $DECL_SPEC float funcReturnType(int param);";
    $SOURCE2 .= "
        float funcReturnType(int param) { return 0.7; }";

    # Return_Type_Became_Void
    $HEADER1 .= "
        $DECL_SPEC int funcReturnTypeBecameVoid(int param);";
    $SOURCE1 .= "
        int funcReturnTypeBecameVoid(int param) { return 7; }";
    
    $HEADER2 .= "
        $DECL_SPEC void funcReturnTypeBecameVoid(int param);";
    $SOURCE2 .= "
        void funcReturnTypeBecameVoid(int param) { return; }";
    
    # Return_BaseType
    $HEADER1 .= "
        $DECL_SPEC int* funcReturnBaseType(int param);";
    $SOURCE1 .= "
        int* funcReturnBaseType(int param) {
            int *x = new int[10];
            return x;
        }";
    
    $HEADER2 .= "
        $DECL_SPEC long long* funcReturnBaseType(int param);";
    $SOURCE2 .= "
        long long* funcReturnBaseType(int param) {
            long long *x = new long long[10];
            return x;
        }";
    
    # Return_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC long long funcReturnPointerLevelAndSize(int param);";
    $SOURCE1 .= "
        long long funcReturnPointerLevelAndSize(int param) { return 100; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long* funcReturnPointerLevelAndSize(int param);";
    $SOURCE2 .= "
        long long* funcReturnPointerLevelAndSize(int param) { return new long long[10]; }";
    
    # Return_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC int* funcReturnPointerLevel(int param);";
    $SOURCE1 .= "
        int* funcReturnPointerLevel(int param) { return new int[10]; }";
    
    $HEADER2 .= "
        $DECL_SPEC int** funcReturnPointerLevel(int param);";
    $SOURCE2 .= "
        int** funcReturnPointerLevel(int param) { return new int*[10]; }";
    
    # Size (anon type)
    $HEADER1 .= "
        typedef struct {
            int i;
            long j;
            double k;
        } AnonTypedef;
        $DECL_SPEC int funcAnonTypedef(AnonTypedef param);";
    $SOURCE1 .= "
        int funcAnonTypedef(AnonTypedef param) { return 0; }";
    
    $HEADER2 .= "
        typedef struct {
            int i;
            long j;
            double k;
            union {
                int dummy[256];
                struct {
                    char q_skiptable[256];
                    const char *p;
                    int l;
                } p;
            };
        } AnonTypedef;
        $DECL_SPEC int funcAnonTypedef(AnonTypedef param);";
    $SOURCE2 .= "
        int funcAnonTypedef(AnonTypedef param) { return 0; }";
    
    # Added_Field (safe: opaque)
    $HEADER1 .= "
        struct $DECL_SPEC OpaqueType
        {
        public:
            OpaqueType method(OpaqueType param);
            int i;
            long j;
            double k;
            OpaqueType* p;
        };";
    $SOURCE1 .= "
        OpaqueType OpaqueType::method(OpaqueType param) { return param; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC OpaqueType
        {
        public:
            OpaqueType method(OpaqueType param);
            int i;
            long j;
            double k;
            OpaqueType* p;
            int added_member;
        };";
    $SOURCE2 .= "
        OpaqueType OpaqueType::method(OpaqueType param) { return param; }";
    
    # Added_Field (safe: internal)
    $HEADER1 .= "
        struct $DECL_SPEC InternalType {
            InternalType method(InternalType param);
            int i;
            long j;
            double k;
            InternalType* p;
        };";
    $SOURCE1 .= "
        InternalType InternalType::method(InternalType param) { return param; }";
    
    $HEADER2 .= "
        struct $DECL_SPEC InternalType {
            InternalType method(InternalType param);
            int i;
            long j;
            double k;
            InternalType* p;
            int added_member;
        };";
    $SOURCE2 .= "
        InternalType InternalType::method(InternalType param) { return param; }";
    
    # Size (unnamed struct/union fields within structs/unions)
    $HEADER1 .= "
        typedef struct {
            int a;
            struct {
                int u1;
                float u2;
            };
            int d;
        } UnnamedTypeSize;
        $DECL_SPEC int unnamedTypeSize(UnnamedTypeSize param);";
    $SOURCE1 .= "
        int unnamedTypeSize(UnnamedTypeSize param) { return 0; }";
    
    $HEADER2 .= "
        typedef struct {
            int a;
            struct {
                long double u1;
                float u2;
            };
            int d;
        } UnnamedTypeSize;
        $DECL_SPEC int unnamedTypeSize(UnnamedTypeSize param);";
    $SOURCE2 .= "
        int unnamedTypeSize(UnnamedTypeSize param) { return 0; }";
    
    # constants
    $HEADER1 .= "
        #define PUBLIC_CONSTANT \"old_value\"";
    $HEADER2 .= "
        #define PUBLIC_CONSTANT \"new_value\"";
    
    $HEADER1 .= "
        #define PRIVATE_CONSTANT \"old_value\"
        #undef PRIVATE_CONSTANT";
    $HEADER2 .= "
        #define PRIVATE_CONSTANT \"new_value\"
        #undef PRIVATE_CONSTANT";
    
    # Added_Field (union)
    $HEADER1 .= "
        union UnionAddedField {
            int a;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionAddedField(UnionAddedField param);";
    $SOURCE1 .= "
        int unionAddedField(UnionAddedField param) { return 0; }";
    
    $HEADER2 .= "
        union UnionAddedField {
            int a;
            struct {
                long double x, y;
            } new_field;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionAddedField(UnionAddedField param);";
    $SOURCE2 .= "
        int unionAddedField(UnionAddedField param) { return 0; }";
    
    # Removed_Field (union)
    $HEADER1 .= "
        union UnionRemovedField {
            int a;
            struct {
                long double x, y;
            } removed_field;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionRemovedField(UnionRemovedField param);";
    $SOURCE1 .= "
        int unionRemovedField(UnionRemovedField param) { return 0; }";
    
    $HEADER2 .= "
        union UnionRemovedField {
            int a;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionRemovedField(UnionRemovedField param);";
    $SOURCE2 .= "
        int unionRemovedField(UnionRemovedField param) { return 0; }";

    # Added (typedef change)
    $HEADER1 .= "
        typedef float TYPEDEF_TYPE;
        $DECL_SPEC int parameterTypedefChange(TYPEDEF_TYPE param);";
    $SOURCE1 .= "
        int parameterTypedefChange(TYPEDEF_TYPE param) { return 1; }";
    
    $HEADER2 .= "
        typedef int TYPEDEF_TYPE;
        $DECL_SPEC int parameterTypedefChange(TYPEDEF_TYPE param);";
    $SOURCE2 .= "
        int parameterTypedefChange(TYPEDEF_TYPE param) { return 1; }";

    # Parameter_Default_Value_Changed
    $HEADER1 .= "
        $DECL_SPEC int parameterDefaultValueChanged(int param = 0xf00f); ";
    $SOURCE1 .= "
        int parameterDefaultValueChanged(int param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterDefaultValueChanged(int param = 0xf00b); ";
    $SOURCE2 .= "
        int parameterDefaultValueChanged(int param) { return param; }";

    # Parameter_Default_Value_Changed (char const *)
    $HEADER1 .= "
        $DECL_SPEC int parameterDefaultStringValueChanged(char const* param = \" str  1 \"); ";
    $SOURCE1 .= "
        int parameterDefaultStringValueChanged(char const* param) { return 0; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterDefaultStringValueChanged(char const* param = \" str  2 \"); ";
    $SOURCE2 .= "
        int parameterDefaultStringValueChanged(char const* param) { return 0; }";

    # Parameter_Default_Value_Removed
    $HEADER1 .= "
        $DECL_SPEC int parameterDefaultValueRemoved(int param = 0xf00f);
    ";
    $SOURCE1 .= "
        int parameterDefaultValueRemoved(int param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterDefaultValueRemoved(int param);";
    $SOURCE2 .= "
        int parameterDefaultValueRemoved(int param) { return param; }";
    
    # Field_Type (typedefs in member type)
    $HEADER1 .= "
        typedef float TYPEDEF_TYPE_2;
        struct $DECL_SPEC FieldTypedefChange{
        public:
            TYPEDEF_TYPE_2 m;
            TYPEDEF_TYPE_2 n;
        };
        $DECL_SPEC int fieldTypedefChange(FieldTypedefChange param);";
    $SOURCE1 .= "
        int fieldTypedefChange(FieldTypedefChange param) { return 1; }";
    
    $HEADER2 .= "
        typedef int TYPEDEF_TYPE_2;
        struct $DECL_SPEC FieldTypedefChange{
        public:
            TYPEDEF_TYPE_2 m;
            TYPEDEF_TYPE_2 n;
        };
        $DECL_SPEC int fieldTypedefChange(FieldTypedefChange param);";
    $SOURCE2 .= "
        int fieldTypedefChange(FieldTypedefChange param) { return 1; }";

    # Callback (testCallback symbol should be affected
    # instead of callback1 and callback2)
    $HEADER1 .= "
        class $DECL_SPEC Callback {
        public:
            virtual int callback1(int x, int y)=0;
            virtual int callback2(int x, int y)=0;
        };
        $DECL_SPEC int testCallback(Callback* p);";
    $SOURCE1 .= "
        int testCallback(Callback* p) {
            p->callback2(1, 2);
            return 0;
        }";

    $HEADER2 .= "
        class $DECL_SPEC Callback {
        public:
            virtual int callback1(int x, int y)=0;
            virtual int added_callback(int x, int y)=0;
            virtual int callback2(int x, int y)=0;
        };
        $DECL_SPEC int testCallback(Callback* p);";
    $SOURCE2 .= "
        int testCallback(Callback* p) {
            p->callback2(1, 2);
            return 0;
        }";
    
    # End namespace
    $HEADER1 .= "\n}\n";
    $HEADER2 .= "\n}\n";
    $SOURCE1 .= "\n}\n";
    $SOURCE2 .= "\n}\n";
    
    if($OSgroup eq "linux")
    {# Check versioning
        # Start versioning
        $HEADER1 .= "
            $DECL_SPEC int startVersioning(int param);";
        $SOURCE1 .= "
            int startVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }";
        
        $HEADER2 .= "
            $DECL_SPEC int startVersioning(int param);";
        $SOURCE2 .= "
            int startVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver _Z15startVersioningi,_Z15startVersioningi\@VERSION_2.0\");";
        
        # Return_Type (good versioning)
        $HEADER1 .= "
            $DECL_SPEC int goodVersioning(int param);";
        $SOURCE1 .= "
            int goodVersioning(int param) {
                return 2^(sizeof(int)*8-1)-1;
            }
            __asm__(\".symver _Z14goodVersioningi,_Z14goodVersioningi\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC int goodVersioningOld(int param);";
        $SOURCE2 .= "
            int goodVersioningOld(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver _Z17goodVersioningOldi,_Z14goodVersioningi\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float goodVersioning(int param);";
        $SOURCE2 .= "
            float goodVersioning(int param) { return 0.7; }
            __asm__(\".symver _Z14goodVersioningi,_Z14goodVersioningi\@VERSION_2.0\");";
        
        # Return_Type (bad versioning)
        $HEADER1 .= "
            $DECL_SPEC int badVersioning(int param);";
        $SOURCE1 .= "
            int badVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver _Z13badVersioningi,_Z13badVersioningi\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float badVersioningOld(int param);";
        $SOURCE2 .= "
            float badVersioningOld(int param) { return 0.7; }
            __asm__(\".symver _Z16badVersioningOldi,_Z13badVersioningi\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float badVersioning(int param);";
        $SOURCE2 .= "
            float badVersioning(int param) { return 0.7; }
            __asm__(\".symver _Z13badVersioningi,_Z13badVersioningi\@VERSION_2.0\");";
    }
    runTests("simple_lib_cpp", "C++", $HEADER1, $SOURCE1, $HEADER2, $SOURCE2, "TestNS::OpaqueType", "_ZN6TestNS12InternalType6methodES0_");
}

sub testSystemC()
{
    print "\nverifying detectable C library changes\n";
    my ($HEADER1, $SOURCE1, $HEADER2, $SOURCE2) = ();
    my $DECL_SPEC = ($OSgroup eq "windows")?"__declspec( dllexport )":"";
    my $EXTERN = ($OSgroup eq "windows")?"extern ":"";
    
    # Global_Data_Value (int)
    $HEADER1 .= "
        $EXTERN $DECL_SPEC const int globalDataValue = 10;";
    
    $HEADER2 .= "
        $EXTERN $DECL_SPEC const int globalDataValue = 15;";
    
    # Global_Data_Became_Non_Const (int)
    $HEADER1 .= "
        $EXTERN $DECL_SPEC const int globalDataBecameNonConst = 10;";
    
    $HEADER2 .= "
        extern $DECL_SPEC int globalDataBecameNonConst;";
    $SOURCE2 .= "
        int globalDataBecameNonConst = 15;";

    # Global_Data_Became_Const (safe)
    $HEADER1 .= "
        extern $DECL_SPEC int globalDataBecameConst;";
    $SOURCE1 .= "
        int globalDataBecameConst = 10;";
    
    $HEADER2 .= "
        $EXTERN $DECL_SPEC const int globalDataBecameConst=15;";
    
    # Global_Data_Became_Non_Const (struct)
    $HEADER1 .= "
        struct GlobalDataType{int a;int b;struct GlobalDataType* p;};";
    $HEADER1 .= "
        $EXTERN $DECL_SPEC const struct GlobalDataType globalStructDataBecameConst = {1, 2, (struct GlobalDataType*)0};";
    
    $HEADER2 .= "
        struct GlobalDataType{int a;int b;struct GlobalDataType* p;};";
    $HEADER2 .= "
        $EXTERN $DECL_SPEC struct GlobalDataType globalStructDataBecameConst = {1, 2, (struct GlobalDataType*)0};";
    
    # Removed_Parameter
    $HEADER1 .= "
        $DECL_SPEC int removedParameter(int param, int removed_param);";
    $SOURCE1 .= "
        int removedParameter(int param, int removed_param) { return 0; }";
    
    $HEADER2 .= "
        $DECL_SPEC int removedParameter(int param);";
    $SOURCE2 .= "
        int removedParameter(int param) { return 0; }";
    
    # Added_Parameter
    $HEADER1 .= "
        $DECL_SPEC int addedParameter(int param);";
    $SOURCE1 .= "
        int addedParameter(int param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int addedParameter(int param, int added_param, int added_param2);";
    $SOURCE2 .= "
        int addedParameter(int param, int added_param, int added_param2) { return added_param2; }";
    
    # Added (typedef to funcptr parameter)
    $HEADER2 .= "
        typedef int (*FUNCPTR_TYPE)(int a, int b);
        $DECL_SPEC int addedFunc(FUNCPTR_TYPE*const** f);";
    $SOURCE2 .= "
        int addedFunc(FUNCPTR_TYPE*const** f) { return 0; }";
    
    # Added (funcptr parameter)
    $HEADER2 .= "
        $DECL_SPEC int addedFunc2(int(*func)(int, int));";
    $SOURCE2 .= "
        int addedFunc2(int(*func)(int, int)) { return 0; }";
    
    # Added (no limited parameters)
    $HEADER2 .= "
        $DECL_SPEC int addedFunc3(float p1, ...);";
    $SOURCE2 .= "
        int addedFunc3(float p1, ...) { return 0; }";
    
    # Size
    $HEADER1 .= "
        struct TypeSize
        {
            long long i[5];
            long j;
            double k;
            struct TypeSize* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int testSize(struct TypeSize param, int param_2);";
    $SOURCE1 .= "
        int testSize(struct TypeSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct TypeSize
        {
            long long i[15];
            long long j;
            double k;
            struct TypeSize* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int testSize(struct TypeSize param, int param_2);";
    $SOURCE2 .= "
        int testSize(struct TypeSize param, int param_2) { return param_2; }";
    
    # Added_Field_And_Size
    $HEADER1 .= "
        struct AddedFieldAndSize
        {
            int i;
            long j;
            double k;
            struct AddedFieldAndSize* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int addedFieldAndSize(struct AddedFieldAndSize param, int param_2);";
    $SOURCE1 .= "
        int addedFieldAndSize(struct AddedFieldAndSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct AddedFieldAndSize
        {
            int i;
            long j;
            double k;
            struct AddedFieldAndSize* p;
            int added_member1;
            int added_member2;
        };";
    $HEADER2 .= "
        $DECL_SPEC int addedFieldAndSize(struct AddedFieldAndSize param, int param_2);";
    $SOURCE2 .= "
        int addedFieldAndSize(struct AddedFieldAndSize param, int param_2) { return param_2; }";
    
    # Added_Middle_Field_And_Size
    $HEADER1 .= "
        struct AddedMiddleFieldAndSize
        {
            int i;
            long j;
            double k;
            struct AddedMiddleFieldAndSize* p;
        };
        $DECL_SPEC int addedMiddleFieldAndSize(struct AddedMiddleFieldAndSize param, int param_2);";
    $SOURCE1 .= "
        int addedMiddleFieldAndSize(struct AddedMiddleFieldAndSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct AddedMiddleFieldAndSize
        {
            int i;
            int added_middle_member;
            long j;
            double k;
            struct AddedMiddleFieldAndSize* p;
        };
        $DECL_SPEC int addedMiddleFieldAndSize(struct AddedMiddleFieldAndSize param, int param_2);";
    $SOURCE2 .= "
        int addedMiddleFieldAndSize(struct AddedMiddleFieldAndSize param, int param_2) { return param_2; }";

    # Added_Middle_Field
    $HEADER1 .= "
        struct AddedMiddleField
        {
            unsigned char field1;
            unsigned short field2;
        };
        $DECL_SPEC int addedMiddleField(struct AddedMiddleField param, int param_2);";
    $SOURCE1 .= "
        int addedMiddleField(struct AddedMiddleField param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct AddedMiddleField
        {
            unsigned char field1;
            unsigned char added_field;
            unsigned short field2;
        };
        $DECL_SPEC int addedMiddleField(struct AddedMiddleField param, int param_2);";
    $SOURCE2 .= "
        int addedMiddleField(struct AddedMiddleField param, int param_2) { return param_2; }";
    
    # Renamed_Field
    $HEADER1 .= "
        struct RenamedField
        {
            long i;
            long j;
            double k;
            struct RenamedField* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int renamedField(struct RenamedField param, int param_2);";
    $SOURCE1 .= "
        int renamedField(struct RenamedField param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct RenamedField
        {
            long renamed_member;
            long j;
            double k;
            struct RenamedField* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int renamedField(struct RenamedField param, int param_2);";
    $SOURCE2 .= "
        int renamedField(struct RenamedField param, int param_2) { return param_2; }";
    
    # Removed_Field_And_Size
    $HEADER1 .= "
        struct RemovedFieldAndSize
        {
            int i;
            long j;
            double k;
            struct RemovedFieldAndSize* p;
            int removed_member1;
            int removed_member2;
        };";
    $HEADER1 .= "
        $DECL_SPEC int removedFieldAndSize(struct RemovedFieldAndSize param, int param_2);";
    $SOURCE1 .= "
        int removedFieldAndSize(struct RemovedFieldAndSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct RemovedFieldAndSize
        {
            int i;
            long j;
            double k;
            struct RemovedFieldAndSize* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int removedFieldAndSize(struct RemovedFieldAndSize param, int param_2);";
    $SOURCE2 .= "
        int removedFieldAndSize(struct RemovedFieldAndSize param, int param_2) { return param_2; }";
    
    # Removed_Middle_Field
    $HEADER1 .= "
        struct RemovedMiddleField
        {
            int i;
            int removed_middle_member;
            long j;
            double k;
            struct RemovedMiddleField* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int removedMiddleField(struct RemovedMiddleField param, int param_2);";
    $SOURCE1 .= "
        int removedMiddleField(struct RemovedMiddleField param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct RemovedMiddleField
        {
            int i;
            long j;
            double k;
            struct RemovedMiddleField* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int removedMiddleField(struct RemovedMiddleField param, int param_2);";
    $SOURCE2 .= "
        int removedMiddleField(struct RemovedMiddleField param, int param_2) { return param_2; }";
    
    # Enum_Member_Value
    $HEADER1 .= "
        enum EnumMemberValue
        {
            MEMBER_1=1,
            MEMBER_2=2
        };";
    $HEADER1 .= "
        $DECL_SPEC int enumMemberValue(enum EnumMemberValue param);";
    $SOURCE1 .= "
        int enumMemberValue(enum EnumMemberValue param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    $HEADER2 .= "
        enum EnumMemberValue
        {
            MEMBER_1=2,
            MEMBER_2=1
        };";
    $HEADER2 .= "
        $DECL_SPEC int enumMemberValue(enum EnumMemberValue param);";
    $SOURCE2 .= "
        int enumMemberValue(enum EnumMemberValue param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";

    # Enum_Member_Removed
    $HEADER1 .= "
        enum EnumMemberRemoved
        {
            MEMBER=1,
            MEMBER_REMOVED=2
        };";
    $HEADER1 .= "
        $DECL_SPEC int enumMemberRemoved(enum EnumMemberRemoved param);";
    $SOURCE1 .= "
        int enumMemberRemoved(enum EnumMemberRemoved param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    $HEADER2 .= "
        enum EnumMemberRemoved
        {
            MEMBER=1
        };";
    $HEADER2 .= "
        $DECL_SPEC int enumMemberRemoved(enum EnumMemberRemoved param);";
    $SOURCE2 .= "
        int enumMemberRemoved(enum EnumMemberRemoved param)
        {
            switch(param)
            {
                case 1:
                    return 1;
            }
            return 0;
        }";
    
    # Enum_Member_Name
    $HEADER1 .= "
        enum EnumMemberName
        {
            BRANCH_1=1,
            BRANCH_2=2
        };";
    $HEADER1 .= "
        $DECL_SPEC int enumMemberName(enum EnumMemberName param);";
    $SOURCE1 .= "
        int enumMemberName(enum EnumMemberName param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    $HEADER2 .= "
        enum EnumMemberName
        {
            BRANCH_FIRST=1,
            BRANCH_SECOND=2
        };";
    $HEADER2 .= "
        $DECL_SPEC int enumMemberName(enum EnumMemberName param);";
    $SOURCE2 .= "
        int enumMemberName(enum EnumMemberName param)
        {
            switch(param)
            {
                case 1:
                    return 1;
                case 2:
                    return 2;
            }
            return 0;
        }";
    
    # Field_Type_And_Size
    $HEADER1 .= "
        struct FieldTypeAndSize
        {
            int i;
            long j;
            double k;
            struct FieldTypeAndSize* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int fieldTypeAndSize(struct FieldTypeAndSize param, int param_2);";
    $SOURCE1 .= "
        int fieldTypeAndSize(struct FieldTypeAndSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct FieldTypeAndSize
        {
            int i;
            long long j;
            double k;
            struct FieldTypeAndSize* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int fieldTypeAndSize(struct FieldTypeAndSize param, int param_2);";
    $SOURCE2 .= "
        int fieldTypeAndSize(struct FieldTypeAndSize param, int param_2) { return param_2; }";
    
    # Field_Type
    $HEADER1 .= "
        struct FieldType
        {
            int i;
            long j;
            double k;
            struct FieldType* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int fieldType(struct FieldType param, int param_2);";
    $SOURCE1 .= "
        int fieldType(struct FieldType param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct FieldType
        {
            float i;
            long j;
            double k;
            struct FieldType* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int fieldType(struct FieldType param, int param_2);";
    $SOURCE2 .= "
        int fieldType(struct FieldType param, int param_2) { return param_2; }";
    
    # Field_BaseType
    $HEADER1 .= "
        struct FieldBaseType
        {
            int i;
            long *j;
            double k;
            struct FieldBaseType* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int fieldBaseType(struct FieldBaseType param, int param_2);";
    $SOURCE1 .= "
        int fieldBaseType(struct FieldBaseType param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct FieldBaseType
        {
            int i;
            long long *j;
            double k;
            struct FieldBaseType* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int fieldBaseType(struct FieldBaseType param, int param_2);";
    $SOURCE2 .= "
        int fieldBaseType(struct FieldBaseType param, int param_2) { return param_2; }";
    
    # Field_PointerLevel (and Size)
    $HEADER1 .= "
        struct FieldPointerLevelAndSize
        {
            int i;
            long long j;
            double k;
            struct FieldPointerLevelAndSize* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int fieldPointerLevelAndSize(struct FieldPointerLevelAndSize param, int param_2);";
    $SOURCE1 .= "
        int fieldPointerLevelAndSize(struct FieldPointerLevelAndSize param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct FieldPointerLevelAndSize
        {
            int i;
            long long *j;
            double k;
            struct FieldPointerLevelAndSize* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int fieldPointerLevelAndSize(struct FieldPointerLevelAndSize param, int param_2);";
    $SOURCE2 .= "
        int fieldPointerLevelAndSize(struct FieldPointerLevelAndSize param, int param_2) { return param_2; }";
    
    # Field_PointerLevel
    $HEADER1 .= "
        struct FieldPointerLevel
        {
            int i;
            long *j;
            double k;
            struct FieldPointerLevel* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int fieldPointerLevel(struct FieldPointerLevel param, int param_2);";
    $SOURCE1 .= "
        int fieldPointerLevel(struct FieldPointerLevel param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct FieldPointerLevel
        {
            int i;
            long **j;
            double k;
            struct FieldPointerLevel* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int fieldPointerLevel(struct FieldPointerLevel param, int param_2);";
    $SOURCE2 .= "
        int fieldPointerLevel(struct FieldPointerLevel param, int param_2) { return param_2; }";
    
    # Added_Interface
    $HEADER2 .= "
        $DECL_SPEC int addedFunc4(int param);";
    $SOURCE2 .= "
        int addedFunc4(int param) { return param; }";
    
    # Removed_Interface
    $HEADER1 .= "
        $DECL_SPEC int removedFunc(int param);";
    $SOURCE1 .= "
        int removedFunc(int param) { return param; }";
    
    # Parameter_Type_And_Size
    $HEADER1 .= "
        $DECL_SPEC int parameterTypeAndSize(int param, int other_param);";
    $SOURCE1 .= "
        int parameterTypeAndSize(int param, int other_param) { return other_param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterTypeAndSize(long long param, int other_param);";
    $SOURCE2 .= "
        int parameterTypeAndSize(long long param, int other_param) { return other_param; }";

    # Parameter_Type_And_Size (test calling conventions)
    $HEADER1 .= "\n
        $DECL_SPEC int parameterCallingConvention(int p1, int p2, int p3);";
    $SOURCE1 .= "
        int parameterCallingConvention(int p1, int p2, int p3) { return 0; }";
    
    $HEADER2 .= "\n
        $DECL_SPEC float parameterCallingConvention(char p1, int p2, int p3);";
    $SOURCE2 .= "
        float parameterCallingConvention(char p1, int p2, int p3) { return 7.0f; }";
    
    # Parameter_Type
    $HEADER1 .= "
        $DECL_SPEC int parameterType(int param, int other_param);";
    $SOURCE1 .= "
        int parameterType(int param, int other_param) { return other_param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterType(float param, int other_param);";
    $SOURCE2 .= "
        int parameterType(float param, int other_param) { return other_param; }";
    
    # Parameter_Became_Non_Const
    $HEADER1 .= "
        $DECL_SPEC int parameterBecameNonConst(int const* param);";
    $SOURCE1 .= "
        int parameterBecameNonConst(int const* param) { return *param; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterBecameNonConst(int* param);";
    $SOURCE2 .= "
        int parameterBecameNonConst(int* param) {
            *param=10;
            return *param;
        }";
    
    # Parameter_BaseType (Typedef)
    $HEADER1 .= "
        typedef int* PARAM_TYPEDEF;
        $DECL_SPEC int parameterBaseTypedefChange(PARAM_TYPEDEF param);";
    $SOURCE1 .= "
        int parameterBaseTypedefChange(PARAM_TYPEDEF param) { return 0; }";
    
    $HEADER2 .= "
        typedef const int* PARAM_TYPEDEF;
        $DECL_SPEC int parameterBaseTypedefChange(PARAM_TYPEDEF param);";
    $SOURCE2 .= "
        int parameterBaseTypedefChange(PARAM_TYPEDEF param) { return 0; }";
    
    # Parameter_BaseType
    $HEADER1 .= "
        $DECL_SPEC int parameterBaseTypeChange(int *param);";
    $SOURCE1 .= "
        int parameterBaseTypeChange(int *param) { return sizeof(*param); }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterBaseTypeChange(long long *param);";
    $SOURCE2 .= "
        int parameterBaseTypeChange(long long *param) { return sizeof(*param); }";
    
    # Parameter_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC long long parameterPointerLevelAndSize(long long param);";
    $SOURCE1 .= "
        long long parameterPointerLevelAndSize(long long param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long parameterPointerLevelAndSize(long long *param);";
    $SOURCE2 .= "
        long long parameterPointerLevelAndSize(long long *param) { return param[5]; }";
    
    # Parameter_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC int parameterPointerLevel(int *param);";
    $SOURCE1 .= "
        int parameterPointerLevel(int *param) { return param[5]; }";
    
    $HEADER2 .= "
        $DECL_SPEC int parameterPointerLevel(int **param);";
    $SOURCE2 .= "
        int parameterPointerLevel(int **param) { return param[5][5]; }";
    
    # Return_Type_And_Size
    $HEADER1 .= "
        $DECL_SPEC int returnTypeAndSize(int param);";
    $SOURCE1 .= "
        int returnTypeAndSize(int param) { return 2^(sizeof(int)*8-1)-1; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long returnTypeAndSize(int param);";
    $SOURCE2 .= "
        long long returnTypeAndSize(int param) { return 2^(sizeof(long long)*8-1)-1; }";
    
    # Return_Type
    $HEADER1 .= "
        $DECL_SPEC int returnType(int param);";
    $SOURCE1 .= "
        int returnType(int param) { return 2^(sizeof(int)*8-1)-1; }";
    
    $HEADER2 .= "
        $DECL_SPEC float returnType(int param);";
    $SOURCE2 .= "
        float returnType(int param) { return 0.7; }";

    # Return_Type (to void)
    $HEADER1 .= "
        $DECL_SPEC int returnTypeChangeToVoid(int param);";
    $SOURCE1 .= "
        int returnTypeChangeToVoid(int param) { return 2^(sizeof(int)*8-1)-1; }";
    
    $HEADER2 .= "
        $DECL_SPEC void returnTypeChangeToVoid(int param);";
    $SOURCE2 .= "
        void returnTypeChangeToVoid(int param) { return; }";

    # Return_Type (to void*)
    $HEADER1 .= "
        struct SomeStruct {int A;long B;};";
    $HEADER1 .= "
        $DECL_SPEC struct SomeStruct* returnTypeChangeToVoidPtr(int param);";
    $SOURCE1 .= "
        struct SomeStruct* returnTypeChangeToVoidPtr(int param) { return (struct SomeStruct*)0; }";
    
    $HEADER2 .= "
        struct SomeStruct {int A;int B;};";
    $HEADER2 .= "
        $DECL_SPEC void* returnTypeChangeToVoidPtr(int param);";
    $SOURCE2 .= "
        void* returnTypeChangeToVoidPtr(int param) { return (void*)0; }";

    # Return_Type_And_Stack_Layout (from void to struct)
    $HEADER1 .= "
        $DECL_SPEC void returnTypeChangeFromVoidToStruct(int param);";
    $SOURCE1 .= "
        void returnTypeChangeFromVoidToStruct(int param) { return; }";
    
    $HEADER2 .= "
        $DECL_SPEC struct SomeStruct returnTypeChangeFromVoidToStruct(int param);";
    $SOURCE2 .= "
        struct SomeStruct returnTypeChangeFromVoidToStruct(int param) {
            struct SomeStruct obj = {1,2};
            return obj;
        }";

    # Return_Type_And_Stack_Layout (from struct to void)
    $HEADER1 .= "
        $DECL_SPEC struct SomeStruct returnTypeChangeFromStructToVoid(int param);";
    $SOURCE1 .= "
        struct SomeStruct returnTypeChangeFromStructToVoid(int param) {
            struct SomeStruct obj = {1,2};
            return obj;
        }";
    
    $HEADER2 .= "
        $DECL_SPEC void returnTypeChangeFromStructToVoid(int param);";
    $SOURCE2 .= "
        void returnTypeChangeFromStructToVoid(int param) { return; }";
    
    # Return_Type_And_Stack_Layout (from void to long)
    $HEADER1 .= "
        $DECL_SPEC void returnTypeChangeFromVoidToLong(int param);";
    $SOURCE1 .= "
        void returnTypeChangeFromVoidToLong(int param) { return; }";
    
    $HEADER2 .= "
        $DECL_SPEC long returnTypeChangeFromVoidToLong(int param);";
    $SOURCE2 .= "
        long returnTypeChangeFromVoidToLong(int param) { return 0; }";
    
    # Return_Type_And_Stack_Layout (from int to struct)
    $HEADER1 .= "
        $DECL_SPEC int returnTypeChangeFromIntToStruct(int param);";
    $SOURCE1 .= "
        int returnTypeChangeFromIntToStruct(int param) { return param; }";
    
    $HEADER2 .= "
        $DECL_SPEC struct SomeStruct returnTypeChangeFromIntToStruct(int param);";
    $SOURCE2 .= "
        struct SomeStruct returnTypeChangeFromIntToStruct(int param) {
            struct SomeStruct obj = {1,2};
            return obj;
        }";
    
    # Return_Type_And_Stack_Layout (from struct to int)
    $HEADER1 .= "
        $DECL_SPEC struct SomeStruct returnTypeChangeFromStructToInt(int param);";
    $SOURCE1 .= "
        struct SomeStruct returnTypeChangeFromStructToInt(int param) {
            struct SomeStruct obj = {1,2};
            return obj;
        }";
    
    $HEADER2 .= "
        $DECL_SPEC int returnTypeChangeFromStructToInt(int param);";
    $SOURCE2 .= "
        int returnTypeChangeFromStructToInt(int param) { return param; }";

     # Return_Type_And_Return_Method (from struct to int without parameters)
    $HEADER1 .= "
        $DECL_SPEC struct SomeStruct returnTypeChangeFromStructToIntWithNoParams();";
    $SOURCE1 .= "
        struct SomeStruct returnTypeChangeFromStructToIntWithNoParams() {
            struct SomeStruct obj = {1,2};
            return obj;
        }";
    
    $HEADER2 .= "
        $DECL_SPEC int returnTypeChangeFromStructToIntWithNoParams();";
    $SOURCE2 .= "
        int returnTypeChangeFromStructToIntWithNoParams() { return 0; }";
    
    # Return_BaseType
    $HEADER1 .= "
        $DECL_SPEC int *returnBaseTypeChange(int param);";
    $SOURCE1 .= "
        int *returnBaseTypeChange(int param) { return (int*)0; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long *returnBaseTypeChange(int param);";
    $SOURCE2 .= "
        long long *returnBaseTypeChange(int param) { return (long long*)0; }";
    
    # Return_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC long long returnPointerLevelAndSize(int param);";
    $SOURCE1 .= "
        long long returnPointerLevelAndSize(int param) { return 100; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long *returnPointerLevelAndSize(int param);";
    $SOURCE2 .= "
        long long *returnPointerLevelAndSize(int param) { return (long long *)0; }";
    
    # Return_PointerLevel
    $HEADER1 .= "
        $DECL_SPEC long long *returnPointerLevel(int param);";
    $SOURCE1 .= "
        long long *returnPointerLevel(int param) { return (long long *)0; }";
    
    $HEADER2 .= "
        $DECL_SPEC long long **returnPointerLevel(int param);";
    $SOURCE2 .= "
        long long **returnPointerLevel(int param) { return (long long **)0; }";
    
    # Size (typedef to anon structure)
    $HEADER1 .= "
        typedef struct
        {
            int i;
            long j;
            double k;
        } AnonTypedef;
        $DECL_SPEC int anonTypedef(AnonTypedef param);";
    $SOURCE1 .= "
        int anonTypedef(AnonTypedef param) { return 0; }";
    
    $HEADER2 .= "
        typedef struct
        {
            int i;
            long j;
            double k;
            union {
                int dummy[256];
                struct {
                    char q_skiptable[256];
                    const char *p;
                    int l;
                } p;
            };
        } AnonTypedef;
        $DECL_SPEC int anonTypedef(AnonTypedef param);";
    $SOURCE2 .= "
        int anonTypedef(AnonTypedef param) { return 0; }";
    
    # Size (safe: opaque)
    $HEADER1 .= "
        struct OpaqueType
        {
            long long i[5];
            long j;
            double k;
            struct OpaqueType* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int opaqueTypeUse(struct OpaqueType param, int param_2);";
    $SOURCE1 .= "
        int opaqueTypeUse(struct OpaqueType param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct OpaqueType
        {
            long long i[5];
            long long j;
            double k;
            struct OpaqueType* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int opaqueTypeUse(struct OpaqueType param, int param_2);";
    $SOURCE2 .= "
        int opaqueTypeUse(struct OpaqueType param, int param_2) { return param_2; }";
    
    # Size (safe: internal)
    $HEADER1 .= "
        struct InternalType
        {
            long long i[5];
            long j;
            double k;
            struct InternalType* p;
        };";
    $HEADER1 .= "
        $DECL_SPEC int internalTypeUse(struct InternalType param, int param_2);";
    $SOURCE1 .= "
        int internalTypeUse(struct InternalType param, int param_2) { return param_2; }";
    
    $HEADER2 .= "
        struct InternalType
        {
            long long i[5];
            long long j;
            double k;
            struct InternalType* p;
        };";
    $HEADER2 .= "
        $DECL_SPEC int internalTypeUse(struct InternalType param, int param_2);";
    $SOURCE2 .= "
        int internalTypeUse(struct InternalType param, int param_2) { return param_2; }";
    
    if($OSgroup eq "linux")
    {# Check versioning
        # Start versioning
        $HEADER1 .= "
            $DECL_SPEC int startVersioning(int param);";
        $SOURCE1 .= "
            int startVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }";
        
        $HEADER2 .= "
            $DECL_SPEC int startVersioning(int param);";
        $SOURCE2 .= "
            int startVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver startVersioning,startVersioning\@VERSION_2.0\");";
        
        # Return_Type (good versioning)
        $HEADER1 .= "
            $DECL_SPEC int goodVersioning(int param);";
        $SOURCE1 .= "
            int goodVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver goodVersioning,goodVersioning\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC int goodVersioningOld(int param);";
        $SOURCE2 .= "
            int goodVersioningOld(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver goodVersioningOld,goodVersioning\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float goodVersioning(int param);";
        $SOURCE2 .= "
            float goodVersioning(int param) { return 0.7; }
            __asm__(\".symver goodVersioning,goodVersioning\@VERSION_2.0\");";
        
        # Return_Type (bad versioning)
        $HEADER1 .= "
            $DECL_SPEC int badVersioning(int param);";
        $SOURCE1 .= "
            int badVersioning(int param) { return 2^(sizeof(int)*8-1)-1; }
            __asm__(\".symver badVersioning,badVersioning\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float badVersioningOld(int param);";
        $SOURCE2 .= "
            float badVersioningOld(int param) { return 0.7; }
            __asm__(\".symver badVersioningOld,badVersioning\@VERSION_1.0\");";
        
        $HEADER2 .= "
            $DECL_SPEC float badVersioning(int param);";
        $SOURCE2 .= "
            float badVersioning(int param) { return 0.7; }
            __asm__(\".symver badVersioning,badVersioning\@VERSION_2.0\");";
    }
    # unnamed struct/union fields within structs/unions
    $HEADER1 .= "
        typedef struct
        {
            int a;
            union {
                int b;
                float c;
            };
            int d;
        } UnnamedTypeSize;
        $DECL_SPEC int unnamedTypeSize(UnnamedTypeSize param);";
    $SOURCE1 .= "
        int unnamedTypeSize(UnnamedTypeSize param) { return 0; }";
    
    $HEADER2 .= "
        typedef struct
        {
            int a;
            union {
                long double b;
                float c;
            };
            int d;
        } UnnamedTypeSize;
        $DECL_SPEC int unnamedTypeSize(UnnamedTypeSize param);";
    $SOURCE2 .= "
        int unnamedTypeSize(UnnamedTypeSize param) { return 0; }";
    
    # constants
    $HEADER1 .= "
        #define PUBLIC_CONSTANT \"old_value\"";
    $HEADER2 .= "
        #define PUBLIC_CONSTANT \"new_value\"";

    $HEADER1 .= "
        #define INTEGER_CONSTANT 0x01";
    $HEADER2 .= "
        #define INTEGER_CONSTANT 1";
    
    $HEADER1 .= "
        #define PRIVATE_CONSTANT \"old_value\"
        #undef PRIVATE_CONSTANT";
    $HEADER2 .= "
        #define PRIVATE_CONSTANT \"new_value\"
        #undef PRIVATE_CONSTANT";
    
    # Added_Field (union)
    $HEADER1 .= "
        union UnionTypeAddedField
        {
            int a;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionTypeAddedField(union UnionTypeAddedField param);";
    $SOURCE1 .= "
        int unionTypeAddedField(union UnionTypeAddedField param) { return 0; }";
    
    $HEADER2 .= "
        union UnionTypeAddedField
        {
            int a;
            struct {
                long double x, y;
            } new_field;
            struct {
                int b;
                float c;
            };
            int d;
        };
        $DECL_SPEC int unionTypeAddedField(union UnionTypeAddedField param);";
    $SOURCE2 .= "
        int unionTypeAddedField(union UnionTypeAddedField param) { return 0; }";
    
    # Prameter_BaseType (typedef)
    $HEADER1 .= "
        typedef float TYPEDEF_TYPE;
        $DECL_SPEC int parameterTypedefChange(TYPEDEF_TYPE param);";
    $SOURCE1 .= "
        int parameterTypedefChange(TYPEDEF_TYPE param) { return 1.0; }";
    
    $HEADER2 .= "
        typedef int TYPEDEF_TYPE;
        $DECL_SPEC int parameterTypedefChange(TYPEDEF_TYPE param);";
    $SOURCE2 .= "
        int parameterTypedefChange(TYPEDEF_TYPE param) { return 1; }";
    
    # Field_BaseType (typedef in member type)
    $HEADER1 .= "
        typedef float TYPEDEF_TYPE_2;
        struct FieldBaseTypedefChange{
        TYPEDEF_TYPE_2 m;};
        $DECL_SPEC int fieldBaseTypedefChange(struct FieldBaseTypedefChange param);";
    $SOURCE1 .= "
        int fieldBaseTypedefChange(struct FieldBaseTypedefChange param) { return 1; }";
    
    $HEADER2 .= "
        typedef int TYPEDEF_TYPE_2;
        struct FieldBaseTypedefChange{
        TYPEDEF_TYPE_2 m;};
        $DECL_SPEC int fieldBaseTypedefChange(struct FieldBaseTypedefChange param);";
    $SOURCE2 .= "
        int fieldBaseTypedefChange(struct FieldBaseTypedefChange param) { return 1; }";
    
    # "C++" keywords in "C" code
    $HEADER1 .= "
        $DECL_SPEC int explicit(int class, int virtual, int (*this)(int));";
    $SOURCE1 .= "
        $DECL_SPEC int explicit(int class, int virtual, int (*this)(int)) { return 0; }";
    
    $HEADER2 .= "
        $DECL_SPEC int explicit(int class, int virtual);";
    $SOURCE2 .= "
        $DECL_SPEC int explicit(int class, int virtual) { return 0; }";
    
    # Regression (compare ABI dumps using 1.21 and 1.23.2)
    $HEADER1 .= "
        $DECL_SPEC int* testRegression(int *pointer, char const *name, ...);";
    $SOURCE1 .= "
        int* testRegression(int *pointer, char const *name, ...) { return 0; }";

    $HEADER2 .= "
        $DECL_SPEC int* testRegression(int *pointer, char const *name, ...);";
    $SOURCE2 .= "
        int* testRegression(int *pointer, char const *name, ...) { return 0; }";
    
    runTests("simple_lib_c", "C", $HEADER1, $SOURCE1, $HEADER2, $SOURCE2, "OpaqueType", "internalTypeUse");
}

sub runTests($$$$$$$$)
{
    my ($LibName, $Lang, $HEADER1, $SOURCE1, $HEADER2, $SOURCE2, $Opaque, $Private) = @_;
    my $Ext = ($Lang eq "C++")?"cpp":"c";
    rmtree($LibName);
    # creating test suite
    my $Path_v1 = "$LibName/simple_lib.v1";
    my $Path_v2 = "$LibName/simple_lib.v2";
    mkpath($Path_v1);
    mkpath($Path_v2);
    writeFile("$Path_v1/simple_lib.h", $HEADER1."\n");
    writeFile("$Path_v1/simple_lib.$Ext", "#include \"simple_lib.h\"\n".$SOURCE1."\n");
    writeFile("$LibName/v1.xml", "<version>\n    1.0\n</version>\n\n<headers>\n    ".abs_path($Path_v1)."\n</headers>\n\n<libs>\n    ".abs_path($Path_v1)."\n</libs>\n\n<opaque_types>\n    $Opaque\n</opaque_types>\n\n<skip_symbols>\n    $Private\n</skip_symbols>\n\n<include_paths>\n    ".abs_path($Path_v1)."\n</include_paths>\n");
    
    writeFile("$Path_v2/simple_lib.h", $HEADER2."\n");
    writeFile("$Path_v2/simple_lib.$Ext", "#include \"simple_lib.h\"\n".$SOURCE2."\n");
    writeFile("$LibName/v2.xml", "<version>\n    2.0\n</version>\n\n<headers>\n    ".abs_path($Path_v2)."\n</headers>\n\n<libs>\n    ".abs_path($Path_v2)."\n</libs>\n\n<opaque_types>\n    $Opaque\n</opaque_types>\n\n<skip_symbols>\n    $Private\n</skip_symbols>\n\n<include_paths>\n    ".abs_path($Path_v2)."\n</include_paths>\n");
    my $BuildCmd = "";
    if($OSgroup eq "windows")
    {
        my $CL = get_CmdPath("cl");
        if(not $CL) {
            exitStatus("Not_Found", "can't find \"cl\" compiler");
        }
        $BuildCmd = "$CL /LD simple_lib.$Ext >build_log.txt 2>&1";
    }
    elsif($OSgroup eq "linux")
    {
        writeFile("$Path_v1/version", "VERSION_1.0 {\n};\nVERSION_2.0 {\n};\n");
        writeFile("$Path_v2/version", "VERSION_1.0 {\n};\nVERSION_2.0 {\n};\n");
        $BuildCmd = "$GCC_PATH -Wl,--version-script version -shared -fkeep-inline-functions".($Lang eq "C++"?" -x c++":"")." simple_lib.$Ext -o simple_lib.$LIB_EXT";
    }
    elsif($OSgroup eq "macos") {# using GCC -dynamiclib -x c++ -lstdc++
        $BuildCmd = "$GCC_PATH -dynamiclib -fkeep-inline-functions".($Lang eq "C++"?" -x c++ -lstdc++":"")." simple_lib.$Ext -o simple_lib.$LIB_EXT";
    }
    else {
        $BuildCmd = "$GCC_PATH -shared -fkeep-inline-functions".($Lang eq "C++"?" -x c++":"")." simple_lib.$Ext -o simple_lib.$LIB_EXT";
    }
    writeFile("$Path_v1/Makefile", "all:\n\t$BuildCmd\n");
    writeFile("$Path_v2/Makefile", "all:\n\t$BuildCmd\n");
    system("cd $Path_v1 && $BuildCmd");
    if($?) {
        print STDERR "ERROR: can't compile \'$Path_v1/simple_lib.$Ext\'\n";
        return;
    }
    system("cd $Path_v2 && $BuildCmd");
    if($?) {
        print STDERR "ERROR: can't compile \'$Path_v2/simple_lib.$Ext\'\n";
        return;
    }
    # running the tool
    my $Cmd = "perl $0 -l $LibName -d1 $LibName/v1.xml -d2 $LibName/v2.xml";
    if($TestDump) {
        $Cmd .= " -use-dumps";
    }
    elsif($OSgroup ne "windows") {
        #$Cmd .= " -check-implementation";
    }
    system($Cmd);
    my $CReport = readAttributes("compat_reports/$LibName/1.0_to_2.0/abi_compat_report.html");
    my $NProblems = $CReport->{"type_problems_high"}+$CReport->{"type_problems_medium"};
    $NProblems += $CReport->{"interface_problems_high"}+$CReport->{"interface_problems_medium"};
    $NProblems += $CReport->{"removed"};
    if(($LibName eq "simple_lib_c" and $NProblems>30)
    or ($LibName eq "simple_lib_cpp" and $NProblems>60)) {
        print "result: SUCCESS ($NProblems problems found)\n\n";
    }
    else {
        print STDERR "result: FAILED ($NProblems problems found)\n\n";
    }
}

sub appendFile($$)
{
    my ($Path, $Content) = @_;
    return if(not $Path);
    if(my $Dir = get_dirname($Path)) {
        mkpath($Dir);
    }
    open(FILE, ">>".$Path) || die ("can't open file \'$Path\': $!\n");
    print FILE $Content;
    close(FILE);
}

sub writeFile($$)
{
    my ($Path, $Content) = @_;
    return if(not $Path);
    if(my $Dir = get_dirname($Path)) {
        mkpath($Dir);
    }
    open (FILE, ">".$Path) || die ("can't open file \'$Path\': $!\n");
    print FILE $Content;
    close(FILE);
}

sub readFile($)
{
    my $Path = $_[0];
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    local $/ = undef;
    my $Content = <FILE>;
    close(FILE);
    $Content=~s/\r//g;
    return $Content;
}

sub getGccVersion($)
{
    my $LibVersion = $_[0];
    if($GCC_VERSION{$LibVersion})
    { # dump version
        return $GCC_VERSION{$LibVersion};
    }
    elsif($UsedDump{$LibVersion})
    { # old-version dumps
        return "unknown";
    }
    my $GccVersion = get_dumpversion($GCC_PATH); # host version
    if(not $GccVersion) {
        return "unknown";
    }
    return $GccVersion;
}

sub showArch($)
{
    my $Arch = $_[0];
    if($Arch eq "arm") {
        return "ARM";
    }
    return $Arch;
}

sub getArch($)
{
    my $LibVersion = $_[0];
    if($CPU_ARCH{$LibVersion})
    { # dump version
        return $CPU_ARCH{$LibVersion};
    }
    elsif($UsedDump{$LibVersion})
    { # old-version dumps
        return "unknown";
    }
    if(defined $Cache{"getArch"}{$LibVersion}) {
        return $Cache{"getArch"}{$LibVersion};
    }
    my $Arch = get_dumpmachine($GCC_PATH); # host version
    if(not $Arch) {
        return "unknown";
    }
    if($Arch=~/\A([\w]{3,})(-|\Z)/) {
        $Arch = $1;
    }
    $Arch = "x86" if($Arch=~/\Ai[3-7]86\Z/);
    if($OSgroup eq "windows") {
        $Arch = "x86" if($Arch=~/win32|mingw32/i);
        $Arch = "x86_64" if($Arch=~/win64|mingw64/i);
    }
    $Cache{"getArch"}{$LibVersion} = $Arch;
    return $Arch;
}

sub get_Report_Header()
{
    my $ArchInfo = " on <span style='color:Blue;'>".showArch(getArch(1))."</span>";
    if(getArch(1) ne getArch(2) or getArch(1) eq "unknown")
    { # don't show architecture in the header
        $ArchInfo="";
    }
    my $Report_Header = "<h1 class='title1'><span style='white-space:nowrap;'>Binary compatibility report for the <span style='color:Blue;'>$TargetLibraryFullName</span> library</span>";
    $Report_Header .= " <span style='white-space:nowrap;'>&nbsp;between <span style='color:Red;'>".$Descriptor{1}{"Version"}."</span> and <span style='color:Red;'>".$Descriptor{2}{"Version"}."</span> versions$ArchInfo</span>";
    if($AppPath) {
        $Report_Header .= " <span style='white-space:nowrap;'>&nbsp;(relating to the portability of application <span style='color:Blue;'>".get_filename($AppPath)."</span>)</span>";
    }
    $Report_Header .= "</h1>\n";
    return $Report_Header;
}

sub get_SourceInfo()
{
    my $CheckedHeaders = "<a name='Checked_Headers'></a><h2 class='title2'>Header Files (".keys(%{$Target_Headers{1}}).")</h2><hr/>\n";
    $CheckedHeaders .= "<div class='header_list'>\n";
    foreach my $Header_Path (sort {lc($Target_Headers{1}{$a}{"Identity"}) cmp lc($Target_Headers{1}{$b}{"Identity"})} keys(%{$Target_Headers{1}}))
    {
        my $Identity = $Target_Headers{1}{$Header_Path}{"Identity"};
        my $Header_Name = get_filename($Identity);
        my $Dest_Comment = ($Identity=~/[\/\\]/)?" ($Identity)":"";
        $CheckedHeaders .= "$Header_Name$Dest_Comment<br/>\n";
    }
    $CheckedHeaders .= "</div>\n";
    $CheckedHeaders .= "<br/><a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    my $CheckedLibs = "<a name='Checked_Libs'></a><h2 class='title2'>".($OSgroup=~/macos|windows/?"Dynamic Libraries":"Shared Libraries")." (".keys(%{$DyLibNames_All{1}}).")</h2><hr/>\n";
    $CheckedLibs .= "<div class='dylib_list'>\n";
    foreach my $Library (sort {lc($a) cmp lc($b)}  keys(%{$DyLibNames_All{1}}))
    {
        $Library.=" (.$LIB_EXT)" if($Library!~/\.\w+\Z/);
        $CheckedLibs .= "$Library<br/>\n";
    }
    $CheckedLibs .= "</div>\n";
    $CheckedLibs .= "<br/><a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    if($CheckObjectsOnly) {
        $CheckedHeaders = "";
    }
    if($CheckHeadersOnly) {
        $CheckedLibs = "";
    }
    return $CheckedHeaders.$CheckedLibs;
}

sub get_TypeProblems_Count($$)
{
    my ($TypeChanges, $TargetPriority) = @_;
    my $Type_Problems_Count = 0;
    foreach my $Type_Name (sort keys(%{$TypeChanges}))
    {
        my %Kinds_Target = ();
        foreach my $Kind (keys(%{$TypeChanges->{$Type_Name}}))
        {
            foreach my $Location (keys(%{$TypeChanges->{$Type_Name}{$Kind}}))
            {
                my $Priority = $TypeChanges->{$Type_Name}{$Kind}{$Location}{"Priority"};
                next if($Priority ne $TargetPriority);
                my $Target = $TypeChanges->{$Type_Name}{$Kind}{$Location}{"Target"};
                next if($Kinds_Target{$Kind}{$Target});
                next if(cmp_priority($Type_MaxPriority{$Type_Name}{$Kind}{$Target}, $Priority));# select a problem with the highest priority
                $Kinds_Target{$Kind}{$Target} = 1;
                $Type_Problems_Count += 1;
            }
        }
    }
    return $Type_Problems_Count;
}

sub get_Summary()
{
    my ($Added, $Removed, $I_Problems_High, $I_Problems_Medium, $I_Problems_Low, $T_Problems_High, $T_Problems_Medium, $T_Problems_Low) = (0,0,0,0,0,0,0,0);
    foreach my $Interface (sort keys(%CompatProblems))
    {
        foreach my $Kind (sort keys(%{$CompatProblems{$Interface}}))
        {
            if($InterfaceProblems_Kind{$Kind})
            {
                foreach my $Location (sort keys(%{$CompatProblems{$Interface}{$Kind}}))
                {
                    if($Kind eq "Added_Interface") {
                        $Added += 1;
                    }
                    elsif($Kind eq "Removed_Interface") {
                        $Removed += 1;
                    }
                    else {
                        if($CompatProblems{$Interface}{$Kind}{$Location}{"Priority"} eq "High") {
                            $I_Problems_High += 1;
                        }
                        elsif($CompatProblems{$Interface}{$Kind}{$Location}{"Priority"} eq "Medium") {
                            $I_Problems_Medium += 1;
                        }
                        elsif($CompatProblems{$Interface}{$Kind}{$Location}{"Priority"} eq "Low") {
                            $I_Problems_Low += 1;
                        }
                    }
                }
            }
        }
    }
    my %TypeChanges = ();
    foreach my $Interface (sort keys(%CompatProblems))
    {
        foreach my $Kind (keys(%{$CompatProblems{$Interface}}))
        {
            if($TypeProblems_Kind{$Kind})
            {
                foreach my $Location (sort {cmp_locations($b, $a)} sort keys(%{$CompatProblems{$Interface}{$Kind}}))
                {
                    my $Type_Name = $CompatProblems{$Interface}{$Kind}{$Location}{"Type_Name"};
                    my $Priority = $CompatProblems{$Interface}{$Kind}{$Location}{"Priority"};
                    my $Target = $CompatProblems{$Interface}{$Kind}{$Location}{"Target"};
                    next if(cmp_priority($Type_MaxPriority{$Type_Name}{$Kind}{$Target}, $Priority));# select a problem with the highest priority
                    %{$TypeChanges{$Type_Name}{$Kind}{$Location}} = %{$CompatProblems{$Interface}{$Kind}{$Location}};
                    $Type_MaxPriority{$Type_Name}{$Kind}{$Target} = max_priority($Type_MaxPriority{$Type_Name}{$Kind}{$Target}, $Priority);
                }
            }
        }
    }
    $T_Problems_High = get_TypeProblems_Count(\%TypeChanges, "High");
    $T_Problems_Medium = get_TypeProblems_Count(\%TypeChanges, "Medium");
    $T_Problems_Low = get_TypeProblems_Count(\%TypeChanges, "Low");

    # test info
    my $TestInfo = "<h2 class='title2'>Test Info</h2><hr/>\n";
    $TestInfo .= "<table cellpadding='3' cellspacing='0' class='summary'>\n";
    $TestInfo .= "<tr><th>Library Name</th><td>$TargetLibraryFullName</td></tr>\n";
    my ($Arch1, $Arch2) = (getArch(1), getArch(2));
    my ($GccV1, $GccV2) = (getGccVersion(1), getGccVersion(2));
    my (@VInf1, @VInf2, $AddTestInfo) = ();
    if($Arch1 ne "unknown"
    and $Arch2 ne "unknown")
    { # CPU arch
        if($Arch1 eq $Arch2)
        { # go to the separate section
            $AddTestInfo .= "<tr><th>CPU Architecture</th><td>".showArch($Arch1)."</td></tr>\n";
        }
        else
        { # go to the version number
            push(@VInf1, showArch($Arch1));
            push(@VInf2, showArch($Arch2));
        }
    }
    if($GccV1 ne "unknown"
    and $GccV2 ne "unknown")
    { # GCC version
        if($GccV1 eq $GccV2)
        { # go to the separate section
            $AddTestInfo .= "<tr><th>GCC Version</th><td>$GccV1</td></tr>\n";
        }
        else
        { # go to the version number
            push(@VInf1, "gcc ".$GccV1);
            push(@VInf2, "gcc ".$GccV2);
        }
    }
    # show long version names with GCC version and CPU architecture name (if different)
    $TestInfo .= "<tr><th>Version #1</th><td>".$Descriptor{1}{"Version"}.(@VInf1?" (".join(", ", reverse(@VInf1)).")":"")."</td></tr>\n";
    $TestInfo .= "<tr><th>Version #2</th><td>".$Descriptor{2}{"Version"}.(@VInf2?" (".join(", ", reverse(@VInf2)).")":"")."</td></tr>\n";
    $TestInfo .= $AddTestInfo;
    $TestInfo .= "</table>\n";
    
    # test results
    my $Summary = "<h2 class='title2'>Test Results</h2><hr/>\n";
    $Summary .= "<table cellpadding='3' cellspacing='0' class='summary'>";
    
    my $Checked_Headers_Link = "0";
    $Checked_Headers_Link = "<a href='#Checked_Headers' style='color:Blue;'>".keys(%{$Target_Headers{1}})."</a>" if(keys(%{$Target_Headers{1}})>0);
    $Summary .= "<tr><th>Total Header Files</th><td>".($CheckObjectsOnly?"0&nbsp;(not&nbsp;analyzed)":$Checked_Headers_Link)."</td></tr>\n";
    
    my $Checked_Libs_Link = "0";
    $Checked_Libs_Link = "<a href='#Checked_Libs' style='color:Blue;'>".keys(%{$DyLibNames_All{1}})."</a>" if(keys(%{$DyLibNames_All{1}})>0);
    $Summary .= "<tr><th>Total ".ucfirst($SHARED)." Libraries</th><td>".($CheckHeadersOnly?"0&nbsp;(not&nbsp;analyzed)":$Checked_Libs_Link)."</td></tr>\n";
    my $TotalTypes = keys(%CheckedTypes);
    if(not $TotalTypes) {# list all the types
        $TotalTypes = keys(%{$TName_Tid{1}});
    }
    $Summary .= "<tr><th>Total Symbols / Types</th><td>".keys(%CheckedInterfaces)." / ".$TotalTypes."</td></tr>\n";
    
    my $Verdict = "";
    $CHECKER_VERDICT = $Removed+$I_Problems_High+$T_Problems_High+$T_Problems_Medium+$I_Problems_Medium;
    if($StrictCompat) {
        $CHECKER_VERDICT+=$T_Problems_Low+$I_Problems_Low;
    }
    else {
        $CHECKER_WARNING+=$T_Problems_Low+$I_Problems_Low;
    }
    if($CHECKER_VERDICT) {
        $Verdict = "<span style='color:Red;'><b>Incompatible</b></span>";
    }
    else {
        $Verdict = "<span style='color:Green;'><b>Compatible</b></span>";
    }
    my $STAT_FIRST_LINE = $CHECKER_VERDICT?"verdict:incompatible;":"verdict:compatible;";
    $Summary .= "<tr><th>Verdict</th><td>$Verdict</td></tr>";
    
    $Summary .= "</table>\n";
    
    # problem summary
    my $Problem_Summary = "<h2 class='title2'>Problem Summary</h2><hr/>\n";
    $Problem_Summary .= "<table cellpadding='3' cellspacing='0' class='summary'>";
    $Problem_Summary .= "<tr><th></th><th style='text-align:center;'>Severity</th><th style='text-align:center;'>Count</th></tr>";
    
    my $Added_Link = "0";
    $Added_Link = "<a href='#Added' style='color:Blue;'>$Added</a>" if($Added>0);
    $Added_Link = "n/a" if($CheckHeadersOnly);
    $STAT_FIRST_LINE .= "added:$Added;";
    $Problem_Summary .= "<tr><th>Added Symbols</th><td>-</td><td>$Added_Link</td></tr>\n";
    
    my $Removed_Link = "0";
    $Removed_Link = "<a href='#Removed' style='color:Blue;'>$Removed</a>" if($Removed>0);
    $Removed_Link = "n/a" if($CheckHeadersOnly);
    $STAT_FIRST_LINE .= "removed:$Removed;";
    $Problem_Summary .= "<tr><th>Removed Symbols</th><td style='color:Red;'>High</td><td>$Removed_Link</td></tr>\n";
    
    my $TH_Link = "0";
    $TH_Link = "<a href='#Type_Problems_High' style='color:Blue;'>$T_Problems_High</a>" if($T_Problems_High>0);
    $TH_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "type_problems_high:$T_Problems_High;";
    $Problem_Summary .= "<tr><th rowspan='3'>Problems with<br/>Data Types</th><td style='color:Red;'>High</td><td>$TH_Link</td></tr>\n";
    
    my $TM_Link = "0";
    $TM_Link = "<a href='#Type_Problems_Medium' style='color:Blue;'>$T_Problems_Medium</a>" if($T_Problems_Medium>0);
    $TM_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "type_problems_medium:$T_Problems_Medium;";
    $Problem_Summary .= "<tr><td>Medium</td><td>$TM_Link</td></tr>\n";
    
    my $TL_Link = "0";
    $TL_Link = "<a href='#Type_Problems_Low' style='color:Blue;'>$T_Problems_Low</a>" if($T_Problems_Low>0);
    $TL_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "type_problems_low:$T_Problems_Low;";
    $Problem_Summary .= "<tr><td>Low</td><td>$TL_Link</td></tr>\n";
    
    my $IH_Link = "0";
    $IH_Link = "<a href='#Interface_Problems_High' style='color:Blue;'>$I_Problems_High</a>" if($I_Problems_High>0);
    $IH_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "interface_problems_high:$I_Problems_High;";
    $Problem_Summary .= "<tr><th rowspan='3'>Problems with<br/>Symbols</th><td style='color:Red;'>High</td><td>$IH_Link</td></tr>\n";
    
    my $IM_Link = "0";
    $IM_Link = "<a href='#Interface_Problems_Medium' style='color:Blue;'>$I_Problems_Medium</a>" if($I_Problems_Medium>0);
    $IM_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "interface_problems_medium:$I_Problems_Medium;";
    $Problem_Summary .= "<tr><td>Medium</td><td>$IM_Link</td></tr>\n";
    
    my $IL_Link = "0";
    $IL_Link = "<a href='#Interface_Problems_Low' style='color:Blue;'>$I_Problems_Low</a>" if($I_Problems_Low>0);
    $IL_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "interface_problems_low:$I_Problems_Low;";
    $Problem_Summary .= "<tr><td>Low</td><td>$IL_Link</td></tr>\n";
    
    my $ChangedConstants_Link = "0";
    $ChangedConstants_Link = "<a href='#Changed_Constants' style='color:Blue;'>".keys(%ProblemsInDefines)."</a>" if($ComparedABI and keys(%ProblemsInDefines)>0);
    $ChangedConstants_Link = "n/a" if($CheckObjectsOnly);
    $STAT_FIRST_LINE .= "changed_constants:".keys(%ProblemsInDefines).";";
    $Problem_Summary .= "<tr><th>Problems with<br/>Constants</th><td>Low</td><td>$ChangedConstants_Link</td></tr>\n";
    $CHECKER_WARNING+=keys(%ProblemsInDefines);
    
    if($CheckImplementation)
    {
        my $ChangedImplementation_Link = "0";
        $ChangedImplementation_Link = "<a href='#Changed_Implementation' style='color:Blue;'>".keys(%ImplProblems)."</a>" if(keys(%ImplProblems)>0);
        $ChangedImplementation_Link = "n/a" if($CheckHeadersOnly);
        $STAT_FIRST_LINE .= "changed_implementation:".keys(%ImplProblems).";";
        $Problem_Summary .= "<tr><th>Problems with<br/>Implementation</th><td>Low</td><td>$ChangedImplementation_Link</td></tr>\n";
        $CHECKER_WARNING+=keys(%ImplProblems);
    }
    $STAT_FIRST_LINE .= "tool_version:$ABI_COMPLIANCE_CHECKER_VERSION";
    $Problem_Summary .= "</table>\n";
    return ($TestInfo.$Summary.$Problem_Summary, $STAT_FIRST_LINE);
}

sub get_Report_ChangedConstants()
{
    my ($CHANGED_CONSTANTS, %HeaderConstant);
    foreach my $Constant (keys(%ProblemsInDefines))
    {
        $HeaderConstant{$Constants{1}{$Constant}{"Header"}}{$Constant} = 1;
    }
    my $Constants_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%HeaderConstant))
    {
        $CHANGED_CONSTANTS .= "<span class='header_name'>$HeaderName</span><br/>\n";
        foreach my $Name (sort {lc($a) cmp lc($b)} keys(%{$HeaderConstant{$HeaderName}}))
        {
            $Constants_Number += 1;
            my $Old_Value = htmlSpecChars($ProblemsInDefines{$Name}{"Old_Value"});
            my $New_Value = htmlSpecChars($ProblemsInDefines{$Name}{"New_Value"});
            my $Change = "The value of constant <b>$Name</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.";
            my $Effect = "Applications will pass an old value of this constant as the parameter to the new-version library functions, that expect a new one. This may result in crash of incorrect behavior of applications.";
            my $ProblemsInDefinesReport = "<tr><th>1</th><td align='left' valign='top'>".$Change."</td><td align='left' valign='top'>$Effect</td></tr>\n";
            $CHANGED_CONSTANTS .= $ContentSpanStart."<span class='extension'>[+]</span> ".$Name.$ContentSpanEnd."<br/>\n$ContentDivStart<table cellpadding='3' cellspacing='0' class='problems_table'><tr><th width='2%'></th><th width='47%'>Change</th><th>Effect</th></tr>$ProblemsInDefinesReport</table><br/>$ContentDivEnd\n";
            $CHANGED_CONSTANTS = insertIDs($CHANGED_CONSTANTS);
        }
        $CHANGED_CONSTANTS .= "<br/>\n";
    }
    if($CHANGED_CONSTANTS) {
        $CHANGED_CONSTANTS = "<a name='Changed_Constants'></a><h2 class='title2'>Problems with Constants ($Constants_Number)</h2><hr/>\n".$CHANGED_CONSTANTS."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $CHANGED_CONSTANTS;
}

sub get_Report_Implementation()
{
    my ($CHANGED_IMPLEMENTATION, %HeaderLibFunc);
    foreach my $Interface (sort keys(%ImplProblems))
    {
        my $HeaderName = $CompleteSignature{1}{$Interface}{"Header"};
        my $DyLib = $Symbol_Library{1}{$Interface};
        $HeaderLibFunc{$HeaderName}{$DyLib}{$Interface} = 1;
    }
    my $Changed_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%HeaderLibFunc))
    {
        foreach my $DyLib (sort {lc($a) cmp lc($b)} keys(%{$HeaderLibFunc{$HeaderName}}))
        {
            my $FDyLib=$DyLib.($DyLib!~/\.\w+\Z/?" (.$LIB_EXT)":"");
            if($HeaderName) {
                $CHANGED_IMPLEMENTATION .= "<span class='header_name'>$HeaderName</span>, <span class='dylib_name'>$FDyLib</span><br/>\n";
            }
            else {
                $CHANGED_IMPLEMENTATION .= "<span class='dylib_name'>$DyLib</span><br/>\n";
            }
            my %NameSpace_Interface = ();
            foreach my $Interface (keys(%{$HeaderLibFunc{$HeaderName}{$DyLib}})) {
                $NameSpace_Interface{get_IntNameSpace($Interface, 2)}{$Interface} = 1;
            }
            foreach my $NameSpace (sort keys(%NameSpace_Interface))
            {
                $CHANGED_IMPLEMENTATION .= ($NameSpace)?"<span class='namespace_title'>namespace</span> <span class='namespace'>$NameSpace</span>"."<br/>\n":"";
                my @SortedInterfaces = sort {lc(get_Signature($a, 1)) cmp lc(get_Signature($b, 1))} keys(%{$NameSpace_Interface{$NameSpace}});
                foreach my $Interface (@SortedInterfaces)
                {
                    $Changed_Number += 1;
                    my $Signature = get_Signature($Interface, 1);
                    if($NameSpace) {
                        $Signature=~s/(\W|\A)\Q$NameSpace\E\:\:(\w)/$1$2/g;
                    }
                    $CHANGED_IMPLEMENTATION .= insertIDs($ContentSpanStart.highLight_Signature_Italic_Color($Signature).$ContentSpanEnd."<br/>\n".$ContentDivStart."<span class='mangled'>[ symbol: <b>$Interface</b> ]</span>".$ImplProblems{$Interface}{"Diff"}."<br/><br/>".$ContentDivEnd."\n");
                }
            }
            $CHANGED_IMPLEMENTATION .= "<br/>\n";
        }
    }
    if($CHANGED_IMPLEMENTATION) {
        $CHANGED_IMPLEMENTATION = "<a name='Changed_Implementation'></a><h2 class='title2'>Problems with Implementation ($Changed_Number)</h2><hr/>\n".$CHANGED_IMPLEMENTATION."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $CHANGED_IMPLEMENTATION;
}

sub get_Report_Added()
{
    my ($ADDED_INTERFACES, %FuncAddedInHeaderLib);
    foreach my $Interface (sort keys(%CompatProblems))
    {
        foreach my $Kind (sort keys(%{$CompatProblems{$Interface}}))
        {
            if($Kind eq "Added_Interface") {
                my $HeaderName = $CompleteSignature{2}{$Interface}{"Header"};
                my $DyLib = $Symbol_Library{2}{$Interface};
                $FuncAddedInHeaderLib{$HeaderName}{$DyLib}{$Interface} = 1;
            }
        }
    }
    my $Added_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%FuncAddedInHeaderLib))
    {
        foreach my $DyLib (sort {lc($a) cmp lc($b)} keys(%{$FuncAddedInHeaderLib{$HeaderName}}))
        {
            my $FDyLib=$DyLib.($DyLib!~/\.\w+\Z/?" (.$LIB_EXT)":"");
            if($HeaderName and $DyLib) {
                $ADDED_INTERFACES .= "<span class='header_name'>$HeaderName</span>, <span class='dylib_name'>$FDyLib</span><br/>\n";
            }
            elsif($DyLib) {
                $ADDED_INTERFACES .= "<span class='dylib_name'>$FDyLib</span><br/>\n";
            }
            elsif($HeaderName) {
                $ADDED_INTERFACES .= "<span class='header_name'>$HeaderName</span><br/>\n";
            }
            my %NameSpace_Interface = ();
            foreach my $Interface (keys(%{$FuncAddedInHeaderLib{$HeaderName}{$DyLib}})) {
                $NameSpace_Interface{get_IntNameSpace($Interface, 2)}{$Interface} = 1;
            }
            foreach my $NameSpace (sort keys(%NameSpace_Interface))
            {
                $ADDED_INTERFACES .= ($NameSpace)?"<span class='namespace_title'>namespace</span> <span class='namespace'>$NameSpace</span>"."<br/>\n":"";
                my @SortedInterfaces = sort {lc(get_Signature($a, 2)) cmp lc(get_Signature($b, 2))} keys(%{$NameSpace_Interface{$NameSpace}});
                foreach my $Interface (@SortedInterfaces)
                {
                    $Added_Number += 1;
                    my $SubReport = "";
                    my $Signature = get_Signature($Interface, 2);
                    if($NameSpace) {
                        $Signature=~s/(\W|\A)\Q$NameSpace\E\:\:(\w)/$1$2/g;
                    }
                    if($Interface=~/\A(_Z|\?)/) {
                        if($Signature) {
                            $SubReport = insertIDs($ContentSpanStart.highLight_Signature_Italic_Color($Signature).$ContentSpanEnd."<br/>\n".$ContentDivStart."<span class='mangled'>[ symbol: <b>$Interface</b> ]</span><br/><br/>".$ContentDivEnd."\n");
                        }
                        else {
                            $SubReport = "<span class=\"iname\">".$Interface."</span>"."<br/>\n";
                        }
                    }
                    else {
                        if($Signature) {
                            $SubReport = "<span class=\"iname\">".highLight_Signature_Italic_Color($Signature)."</span>"."<br/>\n";
                        }
                        else {
                            $SubReport = "<span class=\"iname\">".$Interface."</span>"."<br/>\n";
                        }
                    }
                    $ADDED_INTERFACES .= $SubReport;
                }
            }
            $ADDED_INTERFACES .= "<br/>\n";
        }
    }
    if($ADDED_INTERFACES) {
        $ADDED_INTERFACES = "<a name='Added'></a><h2 class='title2'>Added Symbols ($Added_Number)</h2><hr/>\n".$ADDED_INTERFACES."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $ADDED_INTERFACES;
}

sub get_Report_Removed()
{
    my ($REMOVED_INTERFACES, %FuncRemovedFromHeaderLib);
    foreach my $Interface (sort keys(%CompatProblems))
    {
        foreach my $Kind (sort keys(%{$CompatProblems{$Interface}}))
        {
            if($Kind eq "Removed_Interface")
            {
                my $HeaderName = $CompleteSignature{1}{$Interface}{"Header"};
                my $DyLib = $Symbol_Library{1}{$Interface};
                $FuncRemovedFromHeaderLib{$HeaderName}{$DyLib}{$Interface} = 1;
            }
        }
    }
    my $Removed_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%FuncRemovedFromHeaderLib))
    {
        foreach my $DyLib (sort {lc($a) cmp lc($b)} keys(%{$FuncRemovedFromHeaderLib{$HeaderName}}))
        {
            my $FDyLib=$DyLib.($DyLib!~/\.\w+\Z/?" (.$LIB_EXT)":"");
            if($HeaderName and $DyLib) {
                $REMOVED_INTERFACES .= "<span class='header_name'>$HeaderName</span>, <span class='dylib_name'>$FDyLib</span><br/>\n";
            }
            elsif($DyLib) {
                $REMOVED_INTERFACES .= "<span class='dylib_name'>$FDyLib</span><br/>\n";
            }
            elsif($HeaderName) {
                $REMOVED_INTERFACES .= "<span class='header_name'>$HeaderName</span><br/>\n";
            }
            my %NameSpace_Interface = ();
            foreach my $Interface (keys(%{$FuncRemovedFromHeaderLib{$HeaderName}{$DyLib}}))
            {
                $NameSpace_Interface{get_IntNameSpace($Interface, 1)}{$Interface} = 1;
            }
            foreach my $NameSpace (sort keys(%NameSpace_Interface))
            {
                $REMOVED_INTERFACES .= ($NameSpace)?"<span class='namespace_title'>namespace</span> <span class='namespace'>$NameSpace</span>"."<br/>\n":"";
                my @SortedInterfaces = sort {lc(get_Signature($a, 1)) cmp lc(get_Signature($b, 1))} keys(%{$NameSpace_Interface{$NameSpace}});
                foreach my $Interface (@SortedInterfaces)
                {
                    $Removed_Number += 1;
                    my $SubReport = "";
                    my $Signature = get_Signature($Interface, 1);
                    if($NameSpace) {
                        $Signature=~s/(\W|\A)\Q$NameSpace\E\:\:(\w)/$1$2/g;
                    }
                    if($Interface=~/\A(_Z|\?)/) {
                        if($Signature) {
                            $SubReport = insertIDs($ContentSpanStart.highLight_Signature_Italic_Color($Signature).$ContentSpanEnd."<br/>\n".$ContentDivStart."<span class='mangled'>[ symbol: <b>$Interface</b> ]</span><br/><br/>".$ContentDivEnd."\n");
                        }
                        else {
                            $SubReport = "<span class=\"iname\">".$Interface."</span>"."<br/>\n";
                        }
                    }
                    else {
                        if($Signature) {
                            $SubReport = "<span class=\"iname\">".highLight_Signature_Italic_Color($Signature)."</span>"."<br/>\n";
                        }
                        else {
                            $SubReport = "<span class=\"iname\">".$Interface."</span>"."<br/>\n";
                        }
                    }
                    $REMOVED_INTERFACES .= $SubReport;
                }
            }
            $REMOVED_INTERFACES .= "<br/>\n";
        }
    }
    if($REMOVED_INTERFACES) {
        $REMOVED_INTERFACES = "<a name='Removed'></a><a name='Withdrawn'></a><h2 class='title2'>Removed Symbols ($Removed_Number)</h2><hr/>\n".$REMOVED_INTERFACES."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $REMOVED_INTERFACES;
}

sub get_Report_InterfaceProblems($)
{
    my $TargetPriority = $_[0];
    my ($INTERFACE_PROBLEMS, %FuncHeaderLib);
    foreach my $Interface (sort keys(%CompatProblems))
    {
        next if($Interface=~/\A([^\@\$\?]+)[\@\$]+/ and defined $CompatProblems{$1});
        foreach my $Kind (sort keys(%{$CompatProblems{$Interface}}))
        {
            if($InterfaceProblems_Kind{$Kind} and ($Kind ne "Added_Interface") and ($Kind ne "Removed_Interface"))
            {
                my $HeaderName = $CompleteSignature{1}{$Interface}{"Header"};
                my $DyLib = $Symbol_Library{1}{$Interface};
                $FuncHeaderLib{$HeaderName}{$DyLib}{$Interface} = 1;
            }
        }
    }
    my $Problems_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%FuncHeaderLib))
    {
        foreach my $DyLib (sort {lc($a) cmp lc($b)} keys(%{$FuncHeaderLib{$HeaderName}}))
        {
            my ($HEADER_LIB_REPORT, %NameSpace_Interface, %NewSignature) = ();
            foreach my $Interface (keys(%{$FuncHeaderLib{$HeaderName}{$DyLib}})) {
                $NameSpace_Interface{get_IntNameSpace($Interface, 1)}{$Interface} = 1;
            }
            foreach my $NameSpace (sort keys(%NameSpace_Interface))
            {
                my $NAMESPACE_REPORT = "";
                my @SortedInterfaces = sort {lc($tr_name{$a}) cmp lc($tr_name{$b})} keys(%{$NameSpace_Interface{$NameSpace}});
                foreach my $Interface (@SortedInterfaces)
                {
                    my $Signature = get_Signature($Interface, 1);
                    my $InterfaceProblemsReport = "";
                    my $ProblemNum = 1;
                    foreach my $Kind (keys(%{$CompatProblems{$Interface}}))
                    {
                        foreach my $Location (keys(%{$CompatProblems{$Interface}{$Kind}}))
                        {
                            my ($Change, $Effect) = ();
                            my %Problems = %{$CompatProblems{$Interface}{$Kind}{$Location}};
                            my $Old_Value = htmlSpecChars($Problems{"Old_Value"});
                            my $New_Value = htmlSpecChars($Problems{"New_Value"});
                            my $Old_Type = $Problems{"Old_Type"};
                            my $New_Type = $Problems{"New_Type"};
                            my $Old_Reg = $Problems{"Old_Reg"};
                            my $New_Reg = $Problems{"New_Reg"};
                            my $Priority = $Problems{"Priority"};
                            my $Target = $Problems{"Target"};
                            my $PTypeName = $Problems{"Parameter_TypeName"};
                            my $Old_Size = $Problems{"Old_Size"};
                            my $New_Size = $Problems{"New_Size"};
                            my $InitialType_Type = $Problems{"InitialType_Type"};
                            my $Parameter_Position = numToStr($Problems{"Parameter_Position"} + 1);
                            next if($Priority ne $TargetPriority);
                            if($Kind eq "Method_Became_Static")
                            {
                                $Change = "Method became <b>static</b>.\n";
                                $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                            }
                            elsif($Kind eq "Method_Became_NonStatic")
                            {
                                $Change = "Method became <b>non-static</b>.\n";
                                $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_Static")
                            {
                                $Change = "Method became <b>static</b>.\n";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_NonStatic")
                            {
                                $Change = "Method became <b>non-static</b>.\n";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_Virtual")
                            {
                                $Change = "Method became <b>virtual</b>.\n";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_NonVirtual")
                            {
                                $Change = "Method became <b>non-virtual</b>.\n";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_Return")
                            {
                                $Change = "Type of return value has been changed from <b>$Old_Type</b> to <b>$New_Type</b>.";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                            }
                            elsif($Kind eq "Symbol_Changed_Signature")
                            {
                                $Change = "Parameters list has been changed.\n";
                                $Effect = "The name of the appropriate symbol for this function on binary level has been changed from <b>$Old_Value</b> to <b>$New_Value</b>. This may cause \"udefined reference\" linker error in old client applications.";
                                $NewSignature{$Interface} = $Problems{"New_Signature"};
                            }
                            elsif($Kind eq "Parameter_Default_Value_Changed")
                            {
                                $Change = "The default argument of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                $Effect = "Applications will pass an old default (compile-time) argument that may not be properly handled yet. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Parameter_Default_Value_Removed")
                            {
                                $Change = "The default argument <span style='white-space:nowrap;'><b>$Old_Value</b></span> of $Parameter_Position parameter <b>$Target</b> has been removed.\n";
                                $Effect = "Applications will pass an old default argument (that is not default any more) that may not be properly handled yet. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Parameter_Type_And_Size"
                            or $Kind eq "Parameter_Type_And_Stack")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                            }
                            elsif($Kind eq "Parameter_Type_And_Register")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                $Effect = "The parameter became passed through the <b>$Old_Reg</b> register instead of <b>$New_Reg</b>. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Parameter_Type_And_Pass_Through_Register")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                $Effect = "The parameter became passed through the <b>$New_Reg</b> register instead of the stack. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Parameter_Type_And_Pass_Through_Stack")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                $Effect = "The parameter became passed through the stack instead of the <b>$Old_Reg</b> register. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Parameter_Type")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                $Effect = "Replacement of parameter data type may indicate a change in its semantic meaning.";
                            }
                            elsif($Kind eq "Parameter_Became_Non_Const")
                            {
                                $Change = "Type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                $Effect = "This function may change parameter <b>$Target</b>, but it will be treated as const by applications. This may result in crash or incorrect behavior of applications.";
                            }
                            elsif($Kind eq "Removed_Parameter")
                            {
                                if($Target=~/\Ap\d+\Z/i) {
                                    $Change = "Parameter <b>$Target</b> of type <b>$PTypeName</b> has been removed from the calling stack.\n";
                                }
                                else {
                                    $Change = "$Parameter_Position parameter <b>$Target</b> has been removed from the calling stack.\n";
                                }
                                $Effect = "This parameter will be ignored by the function.";
                                $NewSignature{$Interface} = $Problems{"New_Signature"};
                            }
                            elsif($Kind eq "Added_Parameter")
                            {
                                if($Target=~/\Ap\d+\Z/i) {
                                    $Change = "Parameter <b>$Target</b> of type <b>$PTypeName</b> has been added to the calling stack.\n";
                                }
                                else {
                                    $Change = "$Parameter_Position parameter <b>$Target</b> has been added to the calling stack.\n";
                                }
                                $Effect = "This parameter will not be initialized by old clients.";
                                $NewSignature{$Interface} = $Problems{"New_Signature"};
                            }
                            elsif($Kind eq "Removed_Middle_Parameter")
                            {
                                if($Target=~/\Ap\d+\Z/i) {
                                    $Change = "Middle parameter <b>$Target</b> of type <b>$PTypeName</b> has been removed from the calling stack.\n";
                                }
                                else {
                                    $Change = "$Parameter_Position middle parameter <b>$Target</b> has been removed from the calling stack.\n";
                                }
                                $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                                $NewSignature{$Interface} = $Problems{"New_Signature"};
                            }
                            elsif($Kind eq "Added_Middle_Parameter")
                            {
                                if($Target=~/\Ap\d+\Z/i) {
                                    $Change = "Parameter <b>$Target</b> of type <b>$PTypeName</b> has been added to the calling stack at the middle position.\n";
                                }
                                else {
                                    $Change = "Parameter <b>$Target</b> has been added to the calling stack at the middle position.\n";
                                }
                                $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                                $NewSignature{$Interface} = $Problems{"New_Signature"};
                            }
                            elsif($Kind eq "Parameter_Rename")
                            {
                                $Change = "$Parameter_Position parameter <b>$Target</b> has been renamed to <b>$New_Value</b>.\n";
                                $Effect = "Renaming of a parameter may indicate a change in its semantic meaning.";
                            }
                            elsif($Kind eq "Parameter_BaseType_And_Size")
                            {
                                if($InitialType_Type eq "Pointer") {
                                    $Change = "Base type of $Parameter_Position parameter <b>$Target</b> (pointer) has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                    $Effect = "Memory stored by pointer may be incorrectly initialized by applications.";
                                }
                                else {
                                    $Change = "Base type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                    $Effect = "Layout of parameter's stack has been changed and therefore parameters at higher positions in the stack may be incorrectly initialized by applications.";
                                }
                            }
                            elsif($Kind eq "Parameter_BaseType")
                            {
                                if($InitialType_Type eq "Pointer") {
                                    $Change = "Base type of $Parameter_Position parameter <b>$Target</b> (pointer) has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                }
                                else {
                                    $Change = "Base type of $Parameter_Position parameter <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                }
                                $Effect = "Replacement of parameter base type may indicate a change in its semantic meaning.";
                            }
                            elsif($Kind eq "Parameter_PointerLevel_Increased")
                            {
                                $Change = "The pointer level of $Parameter_Position parameter <b>$Target</b> has been increased from <b>$Old_Value</b> to <b>$New_Value</b>.\n";
                                $Effect = "The library function may try to access unallocated memory by the dereferencing of old parameter value and therefore cause a crash of applications.";
                            }
                            elsif($Kind eq "Parameter_PointerLevel_Decreased")
                            {
                                $Change = "The pointer level of $Parameter_Position parameter <b>$Target</b> has been decreased from <b>$Old_Value</b> to <b>$New_Value</b>.\n";
                                $Effect = "The library function will treat the parameter as the lower-dimention array and will not read all elements. This may change the behavior of applications.<br/><b>ATTENTION:</b> if this is <b>out</b>-parameter then this change may cause a crash of applications.";
                            }
                            elsif($Kind eq "Return_PointerLevel_Increased")
                            {
                                $Change = "The pointer level of return value has been increased from <b>$Old_Value</b> to <b>$New_Value</b>.\n";
                                $Effect = "Applications will treat the return value as the lower-dimention array and will not read all elements. This may change the behavior of applications.";
                            }
                            elsif($Kind eq "Return_PointerLevel_Decreased")
                            {
                                $Change = "The pointer level of return value has been decreased from <b>$Old_Value</b> to <b>$New_Value</b>.\n";
                                $Effect = "Applications may try to access unallocated memory by the dereferencing of new return value and therefore cause a crash.";
                            }
                            elsif($Kind eq "Global_Data_Became_Non_Const")
                            {
                                $Change = "This global data became <b>non-const</b>.\n";
                                $Effect = "This data will be copied to applications at compile time. Any attempts to change this global data by library functions may result in crash of applications.";
                            }
                            elsif($Kind eq "Return_Type_And_Return_Method")
                            {
                                $Change = "Type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b>".($Old_Size?" (<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b>".($New_Size?" (<b>$New_Size</b> bytes)":"")."</span>.\n";
                                if($New_Value eq "void") {
                                    $Effect = "Applications will not obtain a return value and execution may change.";
                                }
                                else {
                                    if($Old_Type eq "Intrinsic") {
                                        $Effect = "Return value became passed in memory as the hidden first parameter instead of the register. Applications will obtain a different return value and execution may change.";
                                    }
                                    else {
                                        $Effect = "Return value became passed in register instead of the hidden first parameter. Applications will obtain a different return value and execution may change.";
                                    }
                                }
                            }
                            elsif($Kind eq "Return_Type_And_Stack_Layout")
                            {
                                $Change = "Type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b>".($Old_Size?" (<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b>".($New_Size?" (<b>$New_Size</b> bytes)":"")."</span>.\n";
                                if($Old_Type eq "Intrinsic") {
                                    if($Old_Value eq "void") {
                                        $Effect = "Layout of parameter's stack has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes because the return value became passed in memory as the hidden first parameter. All the parameters will be incorrectly initialized by applications.";
                                    }
                                    else {
                                        $Effect = "Layout of parameter's stack has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes because the return value became passed in memory as the hidden first parameter instead of the register. All the parameters will be incorrectly initialized by applications.";
                                    }
                                }
                                else {
                                    if($New_Value eq "void") {
                                        $Effect = "Layout of parameter's stack has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes because the hidden first parameter (return value) has been removed. All the parameters will be incorrectly initialized by applications.";
                                    }
                                    else {
                                        $Effect = "Layout of parameter's stack has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes because the return value became passed in register instead of the hidden first parameter. All the parameters will be incorrectly initialized by applications.";
                                    }
                                }
                            }
                            elsif($Kind eq "Return_Type_Became_Void")
                            {
                                $Change = "Type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b>".($Old_Size?" (<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>void</b></span>.\n";
                                $Effect = "Applications will not obtain a return value and execution may change.";
                            }
                            elsif($Kind eq "Return_Type_And_Size")
                            {
                                if($CompleteSignature{1}{$Interface}{"Data"}) {
                                    $Change = "Type of this global data has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b>".($Old_Size?" (<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b>".($New_Size?" (<b>$New_Size</b> bytes)":"")."</span>.\n";
                                    $Effect = "Applications will obtain a different value and execution may change.";
                                }
                                else {
                                    $Change = "Type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b>".($Old_Size?" (<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b>".($New_Size?" (<b>$New_Size</b> bytes)":"")."</span>.\n";
                                    $Effect = "Applications will obtain a different return value and execution may change.";
                                }
                            }
                            elsif($Kind eq "Return_Type")
                            {
                                if($CompleteSignature{1}{$Interface}{"Data"}) {
                                    $Change = "Type of this global data has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                    $Effect = "Replacement of data type may indicate a change in semantic meaning.";
                                }
                                else {
                                    $Change = "Type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                    $Effect = "Replacement of return type may indicate a change in its semantic meaning.";
                                }
                            }
                            elsif($Kind eq "Return_BaseType_And_Size")
                            {
                                $Change = "Base type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>$New_Value</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.\n";
                                $Effect = "Applications will obtain a different return value and execution may change.";
                            }
                            elsif($Kind eq "Return_BaseType")
                            {
                                $Change = "Base type of return value has been changed from <span style='white-space:nowrap;'><b>$Old_Value</b></span> to <span style='white-space:nowrap;'><b>$New_Value</b></span>.\n";
                                $Effect = "Replacement of return base type may indicate a change in its semantic meaning.";
                            }
                            if($Change) {
                                $InterfaceProblemsReport .= "<tr><th>$ProblemNum</th><td align='left' valign='top'>".$Change."</td><td align='left' valign='top'>".$Effect."</td></tr>\n";
                                $ProblemNum += 1;
                                $Problems_Number += 1;
                            }
                        }
                    }
                    $ProblemNum -= 1;
                    if($InterfaceProblemsReport)
                    {
                        if($Signature) {
                            $NAMESPACE_REPORT .= $ContentSpanStart."<span class='extension'>[+]</span> ".highLight_Signature_Italic_Color($Signature)." ($ProblemNum)".$ContentSpanEnd."<br/>\n$ContentDivStart\n";
                            if($Interface=~/\A(_Z|\?)/
                            and not $NewSignature{$Interface}) {
                                $NAMESPACE_REPORT .= "<span class='mangled'>&nbsp;&nbsp;[ symbol: <b>$Interface</b> ]</span><br/>\n";
                            }
                        }
                        else {
                            $NAMESPACE_REPORT .= $ContentSpanStart."<span class='extension'>[+]</span> ".$Interface." ($ProblemNum)".$ContentSpanEnd."<br/>\n$ContentDivStart\n";
                        }
                        if($NewSignature{$Interface}) {# argument list changed to
                            $NAMESPACE_REPORT .= "\n<span class='new_signature_label'>changed to:</span><br/><span class='new_signature'>".highLight_Signature_Italic_Color($NewSignature{$Interface})."</span>\n";
                        }
                        $NAMESPACE_REPORT .= "<table cellpadding='3' cellspacing='0' class='problems_table'><tr><th width='2%'></th><th width='47%'>Change</th><th>Effect</th></tr>$InterfaceProblemsReport</table><br/>$ContentDivEnd\n";
                        $NAMESPACE_REPORT = insertIDs($NAMESPACE_REPORT);
                        if($NameSpace) {
                            $NAMESPACE_REPORT=~s/(\W|\A)\Q$NameSpace\E\:\:(\w)/$1$2/g;
                        }
                    }
                }
                if($NAMESPACE_REPORT) {
                    $HEADER_LIB_REPORT .= (($NameSpace)?"<span class='namespace_title'>namespace</span> <span class='namespace'>$NameSpace</span>"."<br/>\n":"").$NAMESPACE_REPORT;
                }
            }
            if($HEADER_LIB_REPORT)
            {
                my $FDyLib=$DyLib.($DyLib!~/\.\w+\Z/?" (.$LIB_EXT)":"");
                if($HeaderName and $DyLib) {
                    $INTERFACE_PROBLEMS .= "<span class='header_name'>$HeaderName</span>, <span class='dylib_name'>$FDyLib</span><br/>\n".$HEADER_LIB_REPORT."<br/>";
                }
                elsif($HeaderName) {
                    $INTERFACE_PROBLEMS .= "<span class='header_name'>$HeaderName</span><br/>\n".$HEADER_LIB_REPORT."<br/>";
                }
                elsif($DyLib) {
                    $INTERFACE_PROBLEMS .= "<span class='dylib_name'>$FDyLib</span><br/>\n".$HEADER_LIB_REPORT."<br/>";
                }
                else {
                    $INTERFACE_PROBLEMS .= $HEADER_LIB_REPORT."<br/>";
                }
            }
        }
    }
    if($INTERFACE_PROBLEMS) {
        $INTERFACE_PROBLEMS = "<a name=\'Interface_Problems_$TargetPriority\'></a>\n<h2 class='title2'>Problems with Symbols, $TargetPriority Severity ($Problems_Number)</h2><hr/>\n".$INTERFACE_PROBLEMS."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $INTERFACE_PROBLEMS;
}

sub black_name($)
{
    my $Name = $_[0];
    return "<span class='iname_b'>".highLight_Signature($Name)."</span>";
}

sub get_Report_TypeProblems($)
{
    my $TargetPriority = $_[0];
    my ($TYPE_PROBLEMS, %TypeHeader, %TypeChanges, %TypeType) = ();
    foreach my $Interface (sort keys(%CompatProblems))
    {
        foreach my $Kind (keys(%{$CompatProblems{$Interface}}))
        {
            if($TypeProblems_Kind{$Kind})
            {
                foreach my $Location (sort {cmp_locations($b, $a)} sort keys(%{$CompatProblems{$Interface}{$Kind}}))
                {
                    my $Type_Name = $CompatProblems{$Interface}{$Kind}{$Location}{"Type_Name"};
                    if(not $TypeType{$Type_Name}
                    or $TypeType{$Type_Name} eq "Struct")
                    {# register type of the type, select "class" if type has "class" and "struct"-type changes
                        $TypeType{$Type_Name} = $CompatProblems{$Interface}{$Kind}{$Location}{"Type_Type"};
                    }
                    my $Priority = $CompatProblems{$Interface}{$Kind}{$Location}{"Priority"};
                    my $Target = $CompatProblems{$Interface}{$Kind}{$Location}{"Target"};
                    next if(cmp_priority($Type_MaxPriority{$Type_Name}{$Kind}{$Target}, $Priority));# select a problem with the highest priority
                    %{$TypeChanges{$Type_Name}{$Kind}{$Location}} = %{$CompatProblems{$Interface}{$Kind}{$Location}};
                    my $HeaderName = get_TypeHeader($TName_Tid{1}{$Type_Name}, 1);
                    $TypeHeader{$HeaderName}{$Type_Name} = 1;
                }
            }
        }
    }
    my $Problems_Number = 0;
    foreach my $HeaderName (sort {lc($a) cmp lc($b)} keys(%TypeHeader))
    {
        my ($HEADER_REPORT, %NameSpace_Type) = ();
        foreach my $TypeName (keys(%{$TypeHeader{$HeaderName}})) {
            $NameSpace_Type{get_TypeNameSpace($TypeName, 1)}{$TypeName} = 1;
        }
        foreach my $NameSpace (sort keys(%NameSpace_Type))
        {
            my $NAMESPACE_REPORT = "";
            my @SortedTypes = sort {lc($TypeType{$a}." ".$a) cmp lc($TypeType{$b}." ".$b)} keys(%{$NameSpace_Type{$NameSpace}});
            foreach my $TypeName (@SortedTypes)
            {
                my $ProblemNum = 1;
                my ($TypeProblemsReport, %Kinds_Locations, %Kinds_Target) = ();
                foreach my $Kind (sort {$b=~/Size/ <=> $a=~/Size/} sort keys(%{$TypeChanges{$TypeName}}))
                {
                    foreach my $Location (sort {cmp_locations($b, $a)} sort keys(%{$TypeChanges{$TypeName}{$Kind}}))
                    {
                        my $Priority = $TypeChanges{$TypeName}{$Kind}{$Location}{"Priority"};
                        next if($Priority ne $TargetPriority);
                        $Kinds_Locations{$Kind}{$Location} = 1;
                        my ($Change, $Effect) = ("", "");
                        my %Problems = %{$TypeChanges{$TypeName}{$Kind}{$Location}};
                        my $Target = $Problems{"Target"};
                        my $Reserved = $Problems{"Reserved"};
                        next if($Kinds_Target{$Kind}{$Target});
                        $Kinds_Target{$Kind}{$Target} = 1;
                        my $Old_Value = $Problems{"Old_Value"};
                        my $New_Value = $Problems{"New_Value"};
                        my $Old_Size = $Problems{"Old_Size"};
                        my $New_Size = $Problems{"New_Size"};
                        my $InitialType_Type = $Problems{"InitialType_Type"};
                        my $Member_Position = $Problems{"Member_Position"};
                        my $Max_Position = $Problems{"Max_Position"};
                        my $Memory_Shift = $Problems{"Memory_Shift"};
                        if($Kind eq "Added_Virtual_Method")
                        {
                            $Change = "Virtual method ".black_name($Target)." has been added to this class.";
                            $Effect = "The layout of virtual table has been changed. Call of any virtual method at higher position in this class or its subclasses may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Pure_Virtual_Method")
                        {
                            $Change = "Pure virtual method ".black_name($Target)." has been added to this class.";
                            $Effect = "1) Applications will not provide the implementation for this pure virtual method and therefore cause a crash in the library trying to call this method.<br/>2) The layout of virtual table has been changed. Call of any virtual method at higher position in this class or its subclasses may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Virtual_Method_At_End")
                        {
                            $Change = "Virtual method ".black_name($Target)." has been added to this class.";
                            $Effect = "The layout of virtual table has been changed. This class has no exported constructors and therefore applications will copy an old v-table of the class that will not contain a pointer to added virtual method. Call of any method in this class may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_First_Virtual_Method")
                        {
                            $Change = "First virtual method ".black_name($Target)." has been added to this class.";
                            $Effect = "1) The layout of type structure has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes by the added v-table pointer.<br/> 2) Size of class has been increased by <b>".$POINTER_SIZE{2}."</b> bytes.";
                        }
                        elsif($Kind eq "Removed_Virtual_Method")
                        {
                            $Change = "Virtual method ".black_name($Target)." has been removed from this class.";
                            $Effect = "The layout of virtual table has been changed. Call of this virtual method or any virtual method at higher position in this class or its subclasses may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Removed_Last_Virtual_Method")
                        {
                            $Change = "Last virtual method ".black_name($Target)." has been removed from this class.";
                            $Effect = "1) The layout of type structure has been shifted by <b>".$POINTER_SIZE{2}."</b> bytes by the removed v-table pointer.<br/>2) Size of class has been decreased by <b>".$POINTER_SIZE{2}."</b> bytes.";
                        }
                        elsif($Kind eq "Virtual_Replacement")
                        {
                            $Change = "Virtual method ".black_name($Target)." has been added to this class instead of ".black_name($Old_Value);
                            $Effect = "Applications will pass parameters of older replaced method to newely added virtual method. It may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Pure_Virtual_Replacement")
                        {
                            $Change = "Pure virtual method ".black_name($Target)." has been added to this class instead of ".black_name($Old_Value);
                            $Effect = "Applications will provide an older method to the library instead of expected newely added virtual method. It may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Virtual_Table_Size")
                        {
                            $Change = "Size of virtual table in this class has been changed from <b>$Old_Value</b> to <b>$New_Value</b> bytes.";
                            $Effect = "The layout of virtual tables in its subclasses has been changed. Call of any virtual method in subclasses may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Virtual_Method_Position")
                        {
                            $Change = "The relative position of virtual method ".black_name($Target)." has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "The layout of virtual table has been changed. Call of this virtual method may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Pure_Virtual_Method_Position")
                        {
                            $Change = "The relative position of pure virtual method ".black_name($Target)." has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "The layout of virtual table has been changed. Call of this pure virtual method implementation may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Overridden_Virtual_Method"
                        or $Kind eq "Overridden_Virtual_Method_B")
                        {
                            $Change = "Virtual method ".black_name($Old_Value)." has been overridden by ".black_name($New_Value);
                            $Effect = "Method ".black_name($New_Value)." will be called instead of ".black_name($Old_Value);
                        }
                        elsif($Kind eq "Size")
                        {
                            $Change = "Size of this type has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b> bytes.";
                            $Effect = "Change of type size may cause different effects. $ContentSpanStart"."<span style='color:Black'>[+] ...</span>"."$ContentSpanEnd <label id=\"CONTENT_ID\" style=\"display:none;\"> In the context of function parameters, this change affects the parameter's stack layout and may cause incorrect initialization of parameters at higher positions in the stack. In the context of structure fields, this change affects the field's layout and may cause incorrect access to the fields at higher positions. Also this change may lead to partially uninitialized input parameters. Other effects are possible. See the list of affected symbols below for detailes.</label>";
                        }
                        elsif($Kind eq "Size_Of_Allocable_Class_Increased")
                        {
                            $Change = "Size of this class has been increased from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b> bytes.";
                            $Effect = "1) An object of this class can be allocated by the applications and old size will be hardcoded at the compile time. Call of any exported costructor will break the memory of neighbouring objects on the stack or heap.<br/>2) The memory layout and size of subclasses will be changed.";
                        }
                        elsif($Kind eq "Size_Of_Allocable_Class_Decreased")
                        {
                            $Change = "Size of this class has been decreased from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b> bytes.";
                            $Effect = "Previous accesses of applications to public fields of this class or its subclasses may be incorrect.";
                        }
                        elsif($Kind eq "Size_Of_Copying_Class")
                        {
                            $Change = "Size of this class has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b> bytes.";
                            $Effect = "1) The class has only inline or auto-generated constructors which will be copied to applications at compile time and will allocate an older memory layout. Call of any exported method of this class may access a memory outside the allocated objects or inside the older memory structure and result in crash or incorrect behavior of applications.<br/>2) The memory layout and size of subclasses will be changed.";
                        }
                        elsif($Kind eq "Base_Class_Position")
                        {
                            $Change = "The relative position of class <b>".htmlSpecChars($Target)."</b> has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b> in the list of base classes.";
                            $Effect = "Possible incorrect access of applications to the memory occupied by the base classes.";
                        }
                        elsif($Kind eq "Base_Class_Became_Non_Virtually_Inherited")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> became <b>non-virtually</b> inherited.";
                            $Effect = "Size, memory layout and v-table layout of this class and subclasses may change.";
                        }
                        elsif($Kind eq "Base_Class_Became_Virtually_Inherited")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> became <b>virtually</b> inherited.";
                            $Effect = "Size, memory layout and v-table layout of this class and subclasses may change.";
                        }
                        elsif($Kind eq "Added_Base_Class_And_Shift")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> has been added.";
                            if($Old_Size ne $New_Size) {
                                $Effect = "1) Size of the class has been changed from <b>$Old_Size</b> to <b>$New_Size</b> bytes.<br/>2) The memory layout in this class has been shifted by <b>$Memory_Shift</b> bytes.";
                            }
                            else {
                                $Effect = "The memory layout in this class has been shifted by <b>$Memory_Shift</b> bytes.";
                            }
                        }
                        elsif($Kind eq "Added_Base_Class")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> has been added.";
                            $Effect = "Possible incorrect access of applications to the memory occupied by the base classes.";
                        }
                        elsif($Kind eq "Removed_Base_Class_And_Shift")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> has been removed.";
                            if($Old_Size ne $New_Size) {
                                $Effect = "1) Size of the class has been changed from <b>$Old_Size</b> to <b>$New_Size</b> bytes.<br/>2) The memory layout in this class has been shifted by <b>$Memory_Shift</b> bytes.";
                            }
                            else {
                                $Effect = "The memory layout in this class has been shifted by <b>$Memory_Shift</b> bytes.";
                            }
                        }
                        elsif($Kind eq "Removed_Base_Class")
                        {
                            $Change = "Base class <b>".htmlSpecChars($Target)."</b> has been removed.";
                            $Effect = "Possible incorrect access of applications to the memory occupied by the base classes.";
                        }
                        elsif($Kind eq "Typedef_BaseType")
                        {
                            $Change = "Base type has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "Possible incorrect initialization of parameters by applications.";
                        }
                        elsif($Kind eq "Added_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been added to this type.";
                            $Effect = "Size of the inclusive type has been changed.<br/><b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Union_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been added to this type.";
                            $Effect = "Size of the union has been changed.<br/><b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Union_Field")
                        {
                            $Change = "Field <b>$Target</b> has been added to this type.";
                            $Effect = "<b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Field"
                        or $Kind eq "Added_Middle_Field")
                        {
                            $Change = "Field <b>$Target</b> has been added to this type.";
                            $Effect = "<b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Added_Middle_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been added at the middle position of this structural type.";
                            $Effect = "Layout of structure fields has been changed and therefore fields at higher positions of the structure definition may be incorrectly accessed by applications.";
                        }
                        elsif($Kind eq "Added_Reserved_Field")
                        {
                            $Change = "Field <b>$Target</b> has been added to this structural type at the reserved middle position (before <b>$Reserved</b>).";
                            $Effect = "This field will not be initialized by old clients.<br/><b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Used_Reserved_Field")
                        {
                            $Change = "Field <b>$Target</b> has been added to this structural type at the reserved middle position (instead of <b>$Reserved</b>).";
                            $Effect = "This field will not be initialized by old clients.<br/><b>ATTENTION:</b> this field should be accessed only from the new library symbols, otherwise it may result in crash or incorrect behavior of applications.";
                        }
                        elsif($Kind eq "Renamed_Field")
                        {
                            $Change = "Field <b>$Target</b> has been renamed to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "Renaming of a field in a structural data type may indicate a change in the semantic meaning of the field.";
                        }
                        elsif($Kind eq "Removed_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been removed from this type.";
                            $Effect = "1) Applications will access incorrect memory when attempting to access this field.<br/>2) Size of the inclusive type has been changed.";
                        }
                        elsif($Kind eq "Removed_Union_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been removed from this union.";
                            $Effect = "1) Applications may access incorrect memory when attempting to access this field.<br/>2) Size of the union has been changed.";
                        }
                        elsif($Kind eq "Removed_Field")
                        {
                            $Change = "Field <b>$Target</b> has been removed from this type.";
                            $Effect = "Applications will access incorrect memory when attempting to access this field.";
                        }
                        elsif($Kind eq "Removed_Union_Field")
                        {
                            $Change = "Field <b>$Target</b> has been removed from this union.";
                            $Effect = "Applications may access incorrect memory when attempting to access this field.";
                        }
                        elsif($Kind eq "Removed_Middle_Field_And_Size")
                        {
                            $Change = "Field <b>$Target</b> has been removed from the middle position of this structural type.";
                            $Effect = "1) Previous accesses of applications to the removed field will be incorrect.<br/>2) Layout of structure fields has been changed and therefore fields at higher positions of the structure definition may be incorrectly accessed by applications.";
                        }
                        elsif($Kind eq "Removed_Middle_Field")
                        {
                            $Change = "Field <b>$Target</b> has been removed from the middle position of this structural type.";
                            $Effect = "1) Applications will access incorrect memory when attempting to access this field.<br/>2) Layout of structure fields has been changed and therefore fields at higher positions of the structure definition may be incorrectly accessed by applications.";
                        }
                        elsif($Kind eq "Enum_Member_Value"
                        or $Kind eq "Enum_Last_Member_Value")
                        {
                            $Change = "Value of member <b>$Target</b> has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "Applications may execute a wrong branch of code in the library and therefore change the behavior.";
                        }
                        elsif($Kind eq "Enum_Member_Name")
                        {
                            $Change = "Name of member with value <b>$Target</b> has been changed from <b>".htmlSpecChars($Old_Value)."</b> to <b>".htmlSpecChars($New_Value)."</b>.";
                            $Effect = "Applications may execute a wrong branch of code in the library and therefore change the behavior.";
                        }
                        elsif($Kind eq "Enum_Member_Removed")
                        {
                            $Change = "The member <b>$Target</b> has been removed.";
                            $Effect = "This may result in crash or incorrect behavior of applications because the library may not handle removed member yet.";
                        }
                        elsif($Kind eq "Field_Type_And_Size")
                        {
                            $Change = "Type of field <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.";
                            if($Member_Position==$Max_Position) {
                                $Effect = "This field may be incorrectly initialized or accessed by applications.";
                            }
                            else {
                                $Effect = "Layout of structure fields has been changed and therefore fields at higher positions of the structure definition may be incorrectly accessed by applications.";
                            }
                        }
                        elsif($Kind eq "Field_Size"
                        or $Kind eq "Private_Field_Size")
                        {
                            $Change = "Size of field <b>$Target</b> has been changed from <b>$Old_Size</b> bytes to <b>$New_Size</b> bytes.";
                            $Effect = "1) Previous accesses of applications and library functions to this field and fields at higher positions of the structure definition may be broken.<br/>2) Size of the inclusive type may change.";
                        }
                        elsif($Kind eq "BitField_Size")
                        {
                            $Change = "Size of bitfield <b>$Target</b> has been changed from <b>$Old_Size</b> bits to <b>$New_Size</b> bits.";
                            $Effect = "Previous accesses of applications and library functions to this field and fields at higher positions of the structure definition may be broken.";
                        }
                        elsif($Kind eq "Reserved_Field_Size")
                        {
                            $Change = "Size of reserved field <b>$Target</b> has been changed from <b>$Old_Size</b> bytes to <b>$New_Size</b> bytes.";
                            $Effect = "Previous accesses of applications and library functions to this field may be broken.";
                        }
                        elsif($Kind eq "Field_Type")
                        {
                            $Change = "Type of field <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b></span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b></span>.";
                            $Effect = "Replacement of the field data type may indicate a change in the semantic meaning of the field.";
                        }
                        elsif($Kind eq "Field_BaseType_And_Size")
                        {
                            if($InitialType_Type eq "Pointer") {
                                $Change = "Base type of field <b>$Target</b> (pointer) has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.";
                                $Effect = "Possible access of applications to incorrect memory through the pointer.";
                            }
                            else {
                                $Change = "Base type of field <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b> ".($Old_Size?"(<b>$Old_Size</b> bytes)":"")."</span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b> ".($New_Size?"(<b>$New_Size</b> bytes)":"")."</span>.";
                                $Effect = "Layout of structure fields has been changed and therefore fields at higher positions of structure definition may be incorrectly accessed by applications.";
                            }
                        }
                        elsif($Kind eq "Field_BaseType")
                        {
                            if($InitialType_Type eq "Pointer") {
                                $Change = "Base type of field <b>$Target</b> (pointer) has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b></span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b></span>.";
                                $Effect = "Replacement of field base type may indicate a change in the semantic meaning of the field.";
                            }
                            else {
                                $Change = "Base type of field <b>$Target</b> has been changed from <span style='white-space:nowrap;'><b>".htmlSpecChars($Old_Value)."</b></span> to <span style='white-space:nowrap;'><b>".htmlSpecChars($New_Value)."</b></span>.";
                                $Effect = "Replacement of field base type may indicate a change in the semantic meaning of the field.";
                            }
                        }
                        elsif($Kind eq "Field_PointerLevel_Increased")
                        {
                            $Change = "The pointer level of field <b>$Target</b> has been increased from <b>$Old_Value</b> to <b>$New_Value</b>.";
                            $Effect = "The library functions may try to access unallocated memory by the dereferencing of old field value and therefore cause a crash of applications.";
                        }
                        elsif($Kind eq "Field_PointerLevel_Decreased")
                        {
                            $Change = "The pointer level of field <b>$Target</b> has been decreased from <b>$Old_Value</b> to <b>$New_Value</b>.";
                            $Effect = "The library functions will treat the value of this field as the lower-dimention array and will not read all elements. This may change the behavior of applications.";
                        }
                        if($Change) {
                            $TypeProblemsReport .= "<tr><th>$ProblemNum</th><td align='left' valign='top'>".$Change."</td><td align='left' valign='top'>$Effect</td></tr>\n";
                            $ProblemNum += 1;
                            $Problems_Number += 1;
                            $Kinds_Locations{$Kind}{$Location} = 1;
                        }
                    }
                }
                $ProblemNum -= 1;
                if($TypeProblemsReport)
                {
                    my $Affected_Interfaces = getAffectedInterfaces($TypeName, \%Kinds_Locations);
                    my $ShowVTables = "";
                    if(grep {$_=~/Virtual|Base_Class/} keys(%Kinds_Locations)) {
                        $ShowVTables = showVTables($TypeName);
                    }
                    $NAMESPACE_REPORT .= $ContentSpanStart."<span class='extension'>[+]</span> <span class='ttype'>".lc($TypeType{$TypeName})."</span> ".htmlSpecChars($TypeName)." ($ProblemNum)".$ContentSpanEnd;
                    $NAMESPACE_REPORT .= "<br/>\n$ContentDivStart<table cellpadding='3' cellspacing='0' class='problems_table'><tr>\n";
                    $NAMESPACE_REPORT .= "<th width='2%'></th><th width='47%'>Change</th>\n";
                    $NAMESPACE_REPORT .= "<th>Effect</th></tr>$TypeProblemsReport</table>\n";
                    $NAMESPACE_REPORT .= "$ShowVTables$Affected_Interfaces<br/><br/>$ContentDivEnd\n";
                    $NAMESPACE_REPORT = insertIDs($NAMESPACE_REPORT);
                    if($NameSpace) {
                        $NAMESPACE_REPORT=~s/(\W|\A)\Q$NameSpace\E\:\:(\w)/$1$2/g;
                    }
                }
            }
            if($NAMESPACE_REPORT) {
                $HEADER_REPORT .= (($NameSpace)?"<span class='namespace_title'>namespace</span> <span class='namespace'>$NameSpace</span>"."<br/>\n":"").$NAMESPACE_REPORT;
            }
        }
        if($HEADER_REPORT) {
            $TYPE_PROBLEMS .= "<span class='header_name'>$HeaderName</span><br/>\n".$HEADER_REPORT."<br/>";
        }
    }
    if($TYPE_PROBLEMS) {
        $TYPE_PROBLEMS = "<a name=\'Type_Problems_$TargetPriority\'></a>\n<h2 class='title2'>Problems with Data Types, $TargetPriority Severity ($Problems_Number)</h2><hr/>\n".$TYPE_PROBLEMS."<a style='font-size:11px;' href='#Top'>to the top</a><br/>\n";
    }
    return $TYPE_PROBLEMS;
}

sub showVTables($)
{
    my $TypeName = $_[0];
    my $TypeId1 = $TName_Tid{1}{$TypeName};
    my %Type1 = get_Type($Tid_TDid{1}{$TypeId1}, $TypeId1, 1);
    if(defined $Type1{"VTable"}
    and keys(%{$Type1{"VTable"}}))
    {
        my $TypeId2 = $TName_Tid{2}{$TypeName};
        my %Type2 = get_Type($Tid_TDid{2}{$TypeId2}, $TypeId2, 2);
        if(defined $Type2{"VTable"}
        and keys(%{$Type2{"VTable"}}))
        {
            my $VTABLES = "<table cellpadding='3' cellspacing='0' class='virtual_table'>";
            $VTABLES .= "<tr><th width='2%'>Offset</th>";
            $VTABLES .= "<th width='45%'>Virtual Table (Old) - ".(keys(%{$Type1{"VTable"}}))." entries</th>";
            $VTABLES .= "<th>Virtual Table (New) - ".(keys(%{$Type2{"VTable"}}))." entries</th></tr>";
            my %Indexes = map {$_=>1} (keys(%{$Type1{"VTable"}}), keys(%{$Type2{"VTable"}}));
            foreach (sort {int($a)<=>int($b)} (keys(%Indexes)))
            {
                my ($VL1, $VL2) = ($Type1{"VTable"}{$_}, $Type2{"VTable"}{$_});
                my ($SetColor1, $SetColor2) = ("", "");
                if($VL1 ne $VL2) {
                    if($VL1) {
                        $SetColor1 = " class='vtable_red'";
                        $SetColor2 = " class='vtable_red'";
                    }
                    else {
                        $SetColor2 = " class='vtable_yellow'";
                    }
                }
                else {
                    $VL1 = showVEntry($VL1);
                    $VL2 = showVEntry($VL2);
                }
                $VL1 = htmlSpecChars($VL1);
                $VL2 = htmlSpecChars($VL2);
                $VTABLES .= "<tr><th>$_</th>";
                $VTABLES .= "<td$SetColor1>$VL1</td>";
                $VTABLES .= "<td$SetColor2>$VL2</td></tr>\n";
            }
            $VTABLES .= "</table><br/>";
            my $SHOW_VTABLES = "<span style='padding-left:15px;'>".$ContentSpanStart_Info."[+] show v-table (old and new)".$ContentSpanEnd."</span><br/>\n";
            $SHOW_VTABLES .= $ContentDivStart.$VTABLES.$ContentDivEnd;
            return $SHOW_VTABLES;
        }
    }
    return "";
}

sub showVEntry($)
{
    my $VEntry = $_[0];
    $VEntry=~s/_ZTI\w+/typeinfo/g;
    if($VEntry=~s/ \[with (\w+) = (.+?)(, \w+ = .+|])\Z//g)
    {# std::basic_streambuf<_CharT, _Traits>::imbue [with _CharT = char, _Traits = std::char_traits<char>]
        # become std::basic_streambuf<char, ...>::imbue
        my ($Pname, $Pval) = ($1, $2);
        if($Pname eq "_CharT" and $VEntry=~/\Astd::/)
        {# stdc++ typedefs
            $VEntry=~s/<$Pname(, [^<>]+|)>/<$Pval>/g;
            # FIXME: simplify names using stdcxx typedefs (StdCxxTypedef)
            # The typedef info should be added to ABI dumps
        }
        else
        {
            $VEntry=~s/<$Pname>/<$Pval>/g;
            $VEntry=~s/<$Pname, [^<>]+>/<$Pval, ...>/g;
        }
    }
    if($VEntry=~/\A(.+)::(_ZThn.+)\Z/) {
        $VEntry = "non-virtual thunk";
    }
    if($VEntry=~/\A(.+)::\~(.+)\Z/ and $1 eq $2) {
        $VEntry = "~".$1;
    }
    return $VEntry;
}

sub getAffectedInterfaces($$)
{
    my ($Target_TypeName, $Kinds_Locations) = @_;
    my ($Affected_Interfaces_Header, $Affected_Interfaces, %FunctionNumber, %FunctionProblems) = ();
    foreach my $Interface (sort {lc($tr_name{$a}?$tr_name{$a}:$a) cmp lc($tr_name{$b}?$tr_name{$b}:$b)} keys(%CompatProblems))
    {
        last if(keys(%FunctionNumber)>1000);
        if(($Interface=~/C2E|D2E|D0E/)) {
            # duplicated problems for C2 constructors, D2 and D0 destructors
            next;
        }
        my ($MinPath_Length, $ProblemLocation_Last) = ();
        my $MaxPriority = 0;
        my $Signature = get_Signature($Interface, 1);
        foreach my $Kind (keys(%{$CompatProblems{$Interface}}))
        {
            foreach my $Location (keys(%{$CompatProblems{$Interface}{$Kind}}))
            {
                if(not defined $Kinds_Locations->{$Kind}
                or not $Kinds_Locations->{$Kind}{$Location}) {
                    next;
                }
                if($Interface=~/\A([^\@\$\?]+)[\@\$]+/ and defined $CompatProblems{$1}
                and defined $CompatProblems{$1}{$Kind}{$Location})
                {# duplicated problems for versioned symbols
                    next;
                }
                my $Type_Name = $CompatProblems{$Interface}{$Kind}{$Location}{"Type_Name"};
                next if($Type_Name ne $Target_TypeName);
                my $ProblemLocation = clear_location($Location, $Type_Name);
                my $Parameter_Position = $CompatProblems{$Interface}{$Kind}{$Location}{"Parameter_Position"};
                my $Priority = $CompatProblems{$Interface}{$Kind}{$Location}{"Priority"};
                $FunctionNumber{$Interface} = 1;
                my $Path_Length = 0;
                while($ProblemLocation=~/\-\>/g){$Path_Length += 1;}
                if($MinPath_Length eq "" or ($Path_Length<=$MinPath_Length and $Priority_Value{$Priority}>$MaxPriority)
                or (cmp_locations($ProblemLocation, $ProblemLocation_Last) and $Priority_Value{$Priority}==$MaxPriority))
                {
                    $MinPath_Length = $Path_Length;
                    $MaxPriority = $Priority_Value{$Priority};
                    $ProblemLocation_Last = $ProblemLocation;
                    my $Description = getAffectDescription($Interface, $Kind, $Location);
                    $FunctionProblems{$Interface}{"Issue"} = "<span class='iname_b'>".highLight_Signature_PPos_Italic($Signature, $Parameter_Position, 1, 0, 0)."</span>:<br/>"."<div class='affect'>$Description</div>\n";
                    $FunctionProblems{$Interface}{"Priority"} = $MaxPriority;
                }
            }
        }
    }
    foreach my $Interface (sort {$FunctionProblems{$b}{"Priority"}<=>$FunctionProblems{$a}{"Priority"}}
    sort {lc($tr_name{$a}?$tr_name{$a}:$a) cmp lc($tr_name{$b}?$tr_name{$b}:$b)} keys(%FunctionProblems)) {
        $Affected_Interfaces .= $FunctionProblems{$Interface}{"Issue"};
    }
    $Affected_Interfaces = "<div class='affected'>".$Affected_Interfaces."</div>";
    if(keys(%FunctionNumber)>1000) {
        $Affected_Interfaces .= "and others ...<br/>";
    }
    if($Affected_Interfaces) {
        $Affected_Interfaces_Header = $ContentSpanStart_Affected."[+] affected symbols (".(keys(%FunctionNumber)>1000?"more than 1000":keys(%FunctionNumber)).")".$ContentSpanEnd;
        $Affected_Interfaces =  $ContentDivStart.$Affected_Interfaces.$ContentDivEnd;
    }
    return "<span style='padding-left:15px'>".$Affected_Interfaces_Header."</span>".$Affected_Interfaces;
}

sub cmp_locations($$)
{
    my ($Location1, $Location2) = @_;
    if($Location2=~/(\A|\W)(RetVal|this)(\W|\Z)/
    and $Location1!~/(\A|\W)(RetVal|this)(\W|\Z)/ and $Location1!~/\-\>/) {
        return 1;
    }
    if($Location2=~/(\A|\W)(RetVal|this)(\W|\Z)/ and $Location2=~/\-\>/
    and $Location1!~/(\A|\W)(RetVal|this)(\W|\Z)/ and $Location1=~/\-\>/) {
        return 1;
    }
    return 0;
}

sub clear_location($$)
{
    my ($Location, $TypeName) = @_;
    $Location=~s/->\Q$TypeName\E\Z//g if($TypeName);
    return $Location;
}

sub getAffectDescription($$$)
{
    my ($Interface, $Kind, $Location) = @_;
    my %Affect = %{$CompatProblems{$Interface}{$Kind}{$Location}};
    my $Signature = get_Signature($Interface, 1);
    my $Old_Value = $Affect{"Old_Value"};
    my $New_Value = $Affect{"New_Value"};
    my $Type_Name = $Affect{"Type_Name"};
    my $Parameter_Name = $Affect{"Parameter_Name"};
    my $Start_Type_Name = $Affect{"Start_Type_Name"};
    my $InitialType_Type = $Affect{"InitialType_Type"};
    my $Parameter_Position = numToStr($Affect{"Parameter_Position"} + 1);
    my @Sentence_Parts = ();
    my $Location_To_Type = $Location;
    $Location_To_Type=~s/\A(.*)\-\>.+?\Z/$1/;
    if($Kind=~/(\A|_)Virtual(_|\Z)/)
    {
        if($Kind eq "Overridden_Virtual_Method"
        or $Kind eq "Overridden_Virtual_Method_B") {
            push(@Sentence_Parts, "The method '".highLight_Signature($New_Value)."' will be called instead of this method.");
        }
        elsif($Kind eq "Added_Virtual_Method_At_End") {
            push(@Sentence_Parts, "Call of this method may result in crash or incorrect behavior of applications.");
        }
        elsif($Kind eq "Pure_Virtual_Replacement")
        {
            if($Old_Value eq $Signature) {
                push(@Sentence_Parts, "Call of this method may result in crash or incorrect behavior of applications because it has been replaced by other virtual method.");
            }
            else {
                push(@Sentence_Parts, "Call of this method may result in crash or incorrect behavior of applications because some pure virtual method in v-table has been replaced by other.");
            }
        }
        elsif($Kind eq "Virtual_Replacement") {
            push(@Sentence_Parts, "Call of this method may result in crash or incorrect behavior of applications because it has been replaced by other virtual method.");
        }
        elsif($Kind eq "Virtual_Method_Position"
        or $Kind eq "Pure_Virtual_Method_Position") {
            push(@Sentence_Parts, "Call of this pure virtual method implementation may result in crash or incorrect behavior of applications because the layout of virtual table has been changed.");
        }
        else {
            push(@Sentence_Parts, "Call of this virtual method may result in crash or incorrect behavior of applications because the layout of virtual table has been changed.");
        }
    }
    elsif($TypeProblems_Kind{$Kind})
    {
        if($Location_To_Type eq "this")
        {
            my $METHOD_TYPE = $CompleteSignature{1}{$Interface}{"Constructor"}?"constructor":"method";
            my $ClassName = get_TypeName($CompleteSignature{1}{$Interface}{"Class"}, 1);
            if($ClassName eq $Type_Name) {
                return "This $METHOD_TYPE is from \'".htmlSpecChars($Type_Name)."\' class.";
            }
            else {
                return "This $METHOD_TYPE is from derived class \'".htmlSpecChars($ClassName)."\'.";
            }
        }
        if($Location_To_Type=~/RetVal/)
        {# return value
            if($Location_To_Type=~/\-\>/) {
                push(@Sentence_Parts, "Field \'".htmlSpecChars($Location_To_Type)."\' in return value");
            }
            else {
                push(@Sentence_Parts, "Return value");
            }
            if($InitialType_Type eq "Pointer") {
                push(@Sentence_Parts, "(pointer)");
            }
            elsif($InitialType_Type eq "Ref") {
                push(@Sentence_Parts, "(reference)");
            }
        }
        elsif($Location_To_Type=~/this/)
        {# "this" pointer
            if($Location_To_Type=~/\-\>/) {
                push(@Sentence_Parts, "Field \'".htmlSpecChars($Location_To_Type)."\' in the object of this method");
            }
            else {
                push(@Sentence_Parts, "\'this\' pointer");
            }
        }
        else
        {# parameters
            if($Location_To_Type=~/\-\>/) {
                push(@Sentence_Parts, "Field \'".htmlSpecChars($Location_To_Type)."\' in $Parameter_Position parameter");
            }
            else {
                push(@Sentence_Parts, "$Parameter_Position parameter");
            }
            if($Parameter_Name) {
                push(@Sentence_Parts, "\'$Parameter_Name\'");
            }
            if($InitialType_Type eq "Pointer") {
                push(@Sentence_Parts, "(pointer)");
            }
            elsif($InitialType_Type eq "Ref") {
                push(@Sentence_Parts, "(reference)");
            }
        }
        if($Location_To_Type eq "this") {
            push(@Sentence_Parts, "has base type \'".htmlSpecChars($Type_Name)."\'.");
        }
        elsif($Start_Type_Name eq $Type_Name) {
            push(@Sentence_Parts, "has type \'".htmlSpecChars($Type_Name)."\'.");
        }
        else {
            push(@Sentence_Parts, "has base type \'".htmlSpecChars($Type_Name)."\'.");
        }
    }
    return join(" ", @Sentence_Parts);
}

sub create_HtmlReport()
{
    my $CssStyles = "
    <style type=\"text/css\">
    body {
        font-family:Arial;
        color:Black;
        font-size:14px;
    }
    hr {
        color:Black;
        background-color:Black;
        height:1px;
        border:0;
    }
    h1.title1 {
        margin-bottom:0px;
        padding-bottom:0px;
        font-size:26px;
    }
    h2.title2 {
        margin-bottom:0px;
        padding-bottom:0px;
        font-size:20px;
    }
    span.section {
        font-weight:bold;
        cursor:pointer;
        margin-left:7px;
        font-size:16px;
        color:#003E69;
    }
    span.new_signature {
        font-weight:bold;
        margin-left:28px;
        font-size:16px;
        color:#003E69;
    }
    span.new_signature_label {
        margin-left:28px;
        font-size:14px;
        color:Black;
    }
    span:hover.section {
        color:#336699;
    }
    span.section_affected {
        cursor:pointer;
        margin-left:7px;
        font-size:14px;
        color:#cc3300;
    }
    span.section_info {
        cursor:pointer;
        margin-left:7px;
        font-size:14px;
        color:Black;
    }
    span.extension {
        font-weight:100;
        font-size:16px;
    }
    span.header_name {
        color:#cc3300;
        font-size:14px;
        font-weight:bold;
    }
    div.header_list {
        padding-left:10px;
        color:#333333;
        font-size:15px;
    }
    span.namespace_title {
        margin-left:2px;
        color:#408080;
        font-size:13px;
    }
    span.namespace {
        color:#408080;
        font-size:13px;
        font-weight:bold;
    }
    div.dylib_list {
        padding-left:10px;
        color:#333333;
        font-size:15px;
    }
    span.dylib_name {
        color:Green;
        font-size:14px;
        font-weight:bold;
    }
    span.iname {
        font-weight:bold;
        font-size:16px;
        color:#003E69;
        margin-left:7px;
    }
    span.iname_b {
        font-weight:bold;
        font-size:15px;
        color:#333333;
    }
    span.int_p {
        font-weight:normal;
    }
    div.affect {
        padding-left:15px;
        padding-bottom:4px;
        font-size:14px;
        font-style:italic;
        line-height:13px;
    }
    div.affected {
        padding-left:30px;
        padding-top:5px;
    }
    table.problems_table {
        border-collapse:collapse;
        border:1px outset black;
        line-height:16px;
        margin-left:15px;
        margin-top:3px;
        margin-bottom:3px;
        width:900px;
    }
    table.problems_table td {
        border:1px solid Gray;
    }
    table.problems_table th {
        background-color:#eeeeee;
        color:#333333;
        font-weight:bold;
        font-size:13px;
        font-family:Verdana;
        border:1px solid Gray;
        text-align:center;
        vertical-align:top;
        white-space:nowrap;
    }
    table.virtual_table {
        border-collapse:collapse;
        border:1px outset black;
        line-height:16px;
        margin-left:30px;
        margin-top:3px;
        width:100px;
    }
    table.virtual_table td {
        border:1px solid Gray;
        white-space:nowrap;
        border:1px solid Gray;
    }
    table.virtual_table th {
        background-color:#eeeeee;
        color:#333333;
        font-weight:bold;
        font-size:13px;
        font-family:Verdana;
        border:1px solid Gray;
        text-align:center;
        vertical-align:top;
        white-space:nowrap;
    }
    td.vtable_red {
        background-color:#FFCCCC;
    }
    td.vtable_yellow {
        background-color:#FFFFCC;
    }
    table.summary {
        border-collapse:collapse;
        border:1px outset black;
    }
    table.summary th {
        background-color:#eeeeee;
        font-weight:100;
        text-align:left;
        font-size:15px;
        border:1px inset gray;
    }
    table.summary td {
        padding-left:10px;
        padding-right:5px;
        text-align:right;
        font-size:16px;
        white-space:nowrap;
        border:1px inset gray;
    }
    table.code_view {
        cursor:text;
        margin-top:7px;
        width:50%;
        margin-left:20px;
        font-family:Monaco, \"Courier New\", Courier;
        font-size:14px;
        padding:10px;
        border:1px solid #e0e8e5;
        color:#444444;
        background-color:#eff3f2;
        overflow:auto;
    }
    table.code_view td {
        padding-left:15px;
        text-align:left;
        white-space:nowrap;
    }
    span.mangled {
        padding-left:15px;
        font-size:13px;
        cursor:text;
        color:#444444;
    }
    span.symver {
        color:#333333;
        font-size:14px;
        white-space:nowrap;
    }
    span.color_p {
        font-style:italic;
        color:Brown;
    }
    span.focus_p {
        font-style:italic;
        color:Red;
    }
    span.ttype {
        font-weight:100;
    }
    </style>";
    my $JScripts = "
    <script type=\"text/javascript\" language=\"JavaScript\">
    <!--
    function showContent(header, id)   {
        e = document.getElementById(id);
        if(e.style.display == 'none')
        {
            e.style.display = 'block';
            e.style.visibility = 'visible';
            header.innerHTML = header.innerHTML.replace(/\\\[[^0-9 ]\\\]/gi,\"[&minus;]\");
        }
        else
        {
            e.style.display = 'none';
            e.style.visibility = 'hidden';
            header.innerHTML = header.innerHTML.replace(/\\\[[^0-9 ]\\\]/gi,\"[+]\");
        }
    }
    -->
    </script>";
    my @AllLibs = keys(%{$DyLibNames_All{1}});
    foreach my $LibName1 (sort {length{$a} <=> length($b)} @AllLibs)
    {# remove duplicated shared objects from the report
        foreach my $LibName2 (@AllLibs)
        {
            if($LibName2=~/\A\Q$LibName1\E[\d-_.]+\Z/) {
                delete($DyLibNames_All{1}{$LibName1});
            }
        }
    }
    my ($Summary, $StatLine) = get_Summary();
    my $Title = "$TargetLibraryFullName: ".$Descriptor{1}{"Version"}." to ".$Descriptor{2}{"Version"}." binary compatibility report";
    my $Keywords = "$TargetLibraryFullName, binary compatibility, API, report";
    my $Description = "Backward binary compatibility report for the $TargetLibraryFullName library from version ".$Descriptor{1}{"Version"}." to ".$Descriptor{2}{"Version"};
    if(getArch(1) eq getArch(2)
    and getArch(1) ne "unknown") {
        $Description .= " on ".showArch(getArch(1));
    }
    writeFile($REPORT_PATH, "<!-\- $StatLine -\->\n".composeHTML_Head($Title, $Keywords, $Description, $CssStyles."\n".$JScripts)."\n<body>
    <div><a name='Top'></a>\n".get_Report_Header()."\n$Summary\n".get_Report_Added().get_Report_Removed().get_Report_Problems("High").get_Report_Problems("Medium").get_Report_Problems("Low").get_SourceInfo()."</div>
    <br/><br/><br/><hr/>
    ".getReportFooter($TargetLibraryFullName)."
    <div style='height:999px;'></div>\n</body></html>");
}

sub getReportFooter($)
{
    my $LibName = $_[0];
    my $Footer = "<div style='width:100%;font-size:11px;' align='right'><i>Generated on ".(localtime time); # report date
    $Footer .= " for <span style='font-weight:bold'>$LibName</span>"; # tested library/system name
    $Footer .= " by <a href='http://ispras.linux-foundation.org/index.php/ABI_compliance_checker'>ABI Compliance Checker</a>"; # tool name
    my $ToolSummary = "<br/>A tool for checking backward binary compatibility of a shared C/C++ library API&nbsp;&nbsp;";
    $Footer .= " $ABI_COMPLIANCE_CHECKER_VERSION &nbsp;$ToolSummary</i></div>"; # tool version
    return $Footer;
}

sub get_Report_Problems($)
{
    my $Priority = $_[0];
    my $Problems = get_Report_TypeProblems($Priority).get_Report_InterfaceProblems($Priority);
    if($Priority eq "Low") {
        $Problems .= get_Report_ChangedConstants();
        if($CheckImplementation) {
            $Problems .= get_Report_Implementation();
        }
    }
    if($Problems) {
        $Problems = "<a name=\'".$Priority."_Risk_Problems\'></a>".$Problems;
    }
    return $Problems;
}

sub composeHTML_Head($$$$)
{
    my ($Title, $Keywords, $Description, $OtherInHead) = @_;
    return "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n<head>
    <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />
    <meta name=\"keywords\" content=\"$Keywords\" />
    <meta name=\"description\" content=\"$Description\" />
    <title>\n        $Title\n    </title>\n$OtherInHead\n</head>";
}

sub insertIDs($)
{
    my $Text = $_[0];
    while($Text=~/CONTENT_ID/)
    {
        if(int($Content_Counter)%2) {
            $ContentID -= 1;
        }
        $Text=~s/CONTENT_ID/c_$ContentID/;
        $ContentID += 1;
        $Content_Counter += 1;
    }
    return $Text;
}

sub checkPreprocessedUnit($)
{
    my $Path = $_[0];
    my $CurHeader = "";
    open(PREPROC, "$Path") || die ("can't open file \'$Path\': $!\n");
    while(<PREPROC>)
    {# detecting public and private constants
        next if(not /\A#/);
        chomp($_);
        if(/#[ \t]+\d+[ \t]+\"(.+)\"/) {
            $CurHeader=$1;
        }
        next if(not $DependencyHeaders_All_FullPath{$Version}{get_filename($CurHeader)});
        if(/\#[ \t]*define[ \t]+([_A-Z0-9]+)[ \t]+(.+)[ \t]*\Z/)
        {
            my ($Name, $Value) = ($1, $2);
            if(not $Constants{$Version}{$Name}{"Access"})
            {
                $Constants{$Version}{$Name}{"Access"} = "public";
                $Constants{$Version}{$Name}{"Value"} = $Value;
                $Constants{$Version}{$Name}{"Header"} = get_filename($CurHeader);
            }
        }
        elsif(/\#[ \t]*undef[ \t]+([_A-Z]+)[ \t]*/) {
            $Constants{$Version}{$1}{"Access"} = "private";
        }
    }
    close(PREPROC);
    foreach my $Constant (keys(%{$Constants{$Version}}))
    {
        if($Constants{$Version}{$Constant}{"Access"} eq "private" or $Constant=~/_h\Z/i
        or $Constants{$Version}{$Constant}{"Header"}=~/\<built\-in\>|\<internal\>|\A\./)
        { # skip private constants
            delete($Constants{$Version}{$Constant});
        }
        else {
            delete($Constants{$Version}{$Constant}{"Access"});
        }
    }
}

my %IgnoreConstant=(
    "VERSION"=>1,
    "VERSIONCODE"=>1,
    "VERNUM"=>1,
    "VERS_INFO"=>1,
    "PATCHLEVEL"=>1,
    "INSTALLPREFIX"=>1,
    "VBUILD"=>1,
    "VPATCH"=>1,
    "VMINOR"=>1,
    "BUILD_STRING"=>1,
    "BUILD_TIME"=>1,
    "PACKAGE_STRING"=>1,
    "PRODUCTION"=>1,
    "CONFIGURE_COMMAND"=>1,
    "INSTALLDIR"=>1,
    "BINDIR"=>1,
    "CONFIG_FILE_PATH"=>1,
    "DATADIR"=>1,
    "EXTENSION_DIR"=>1,
    "INCLUDE_PATH"=>1,
    "LIBDIR"=>1,
    "LOCALSTATEDIR"=>1,
    "SBINDIR"=>1,
    "SYSCONFDIR"=>1,
    "RELEASE"=>1,
    "SOURCE_ID"=>1,
    "SUBMINOR"=>1,
    "MINOR"=>1,
    "MINNOR"=>1,
    "MINORVERSION"=>1,
    "MAJOR"=>1,
    "MAJORVERSION"=>1,
    "MICRO"=>1,
    "MICROVERSION"=>1,
    "BINARY_AGE"=>1,
    "INTERFACE_AGE"=>1,
    "CORE_ABI"=>1,
    "PATCH"=>1,
    "COPYRIGHT"=>1,
    "TIMESTAMP"=>1,
    "REVISION"=>1,
    "PACKAGE_TAG"=>1,
    "PACKAGEDATE"=>1,
    "NUMVERSION"=>1
);

sub mergeConstants()
{
    foreach my $Constant (keys(%{$Constants{1}}))
    {
        next if($SkipConstants{1}{$Constant});
        next if($Constants{2}{$Constant}{"Value"} eq "");
        my ($Old_Value, $New_Value, $Old_Value_Pure, $New_Value_Pure);
        $Old_Value = $Old_Value_Pure = uncover_constant(1, $Constant);
        $New_Value = $New_Value_Pure = uncover_constant(2, $Constant);
        $Old_Value_Pure=~s/(\W)\s+/$1/g;
        $Old_Value_Pure=~s/\s+(\W)/$1/g;
        $New_Value_Pure=~s/(\W)\s+/$1/g;
        $New_Value_Pure=~s/\s+(\W)/$1/g;
        next if($New_Value_Pure eq "" or $Old_Value_Pure eq "");
        if($New_Value_Pure ne $Old_Value_Pure)
        {# different values
            if(grep {$Constant=~/(\A|_)$_(_|\Z)/} keys(%IgnoreConstant))
            {# ignore library version
                next;
            }
            if($Constant=~/(\A|_)(lib|open|)$TargetLibraryShortName(_|)(VERSION|VER|DATE|API|PREFIX)(_|\Z)/i)
            {# ignore library version
                next;
            }
            if($Old_Value=~/\A('|"|)[\/\\]\w+([\/\\]|:|('|"|)\Z)/ or $Old_Value=~/[\/\\]\w+[\/\\]\w+/)
            {# ignoring path defines: /lib64:/usr/lib64:/lib:/usr/lib:/usr/X11R6/lib/Xaw3d ...
                next;
            }
            if(convert_integer($Old_Value) eq convert_integer($New_Value))
            {# 0x0001 and 0x1, 0x1 and 1 equal constants
                next;
            }
            %{$ProblemsInDefines{$Constant}} = (
                "Old_Value"=>$Old_Value,
                "New_Value"=>$New_Value  );
        }
    }
}

sub convert_integer($)
{
    my $Value = $_[0];
    if($Value=~/\A0x[a-f0-9]+\Z/)
    {# hexadecimal
        return hex($Value);
    }
    elsif($Value=~/\A0[0-7]+\Z/)
    {# octal
        return oct($Value);
    }
    elsif($Value=~/\A0b[0-1]+\Z/)
    {# binary
        return oct($Value);
    }
    else {
        return $Value;
    }
}

sub uncover_constant($$)
{
    my ($LibVersion, $Constant) = @_;
    return "" if(not $LibVersion or not $Constant);
    return $Constant if(isCyclical(\@RecurConstant, $Constant));
    if($Cache{"uncover_constant"}{$LibVersion}{$Constant} ne "") {
        return $Cache{"uncover_constant"}{$LibVersion}{$Constant};
    }
    my $Value = $Constants{$LibVersion}{$Constant}{"Value"};
    if($Value=~/\A[A-Z0-9_]+\Z/ and $Value=~/[A-Z]/)
    {
        push(@RecurConstant, $Constant);
        if((my $Uncovered = uncover_constant($LibVersion, $Value)) ne "") {
            $Value = $Uncovered;
        }
        pop(@RecurConstant);
    }
    # FIXME: uncover $Value using all the enum constants
    # USECASE: change of define NC_LONG from NC_INT (enum value) to NC_INT (define)
    $Cache{"uncover_constant"}{$LibVersion}{$Constant} = $Value;
    return $Value;
}

sub getSymbols($)
{
    my $LibVersion = $_[0];
    my @DyLibPaths = getSoPaths($LibVersion);
    if($#DyLibPaths==-1 and not $CheckHeadersOnly)
    {
        if($LibVersion==1) {
            $CheckHeadersOnly = 1;
        }
        else {
            exitStatus("Error", "$SHARED libraries are not found in the ".$Descriptor{$LibVersion}{"Version"});
        }
    }
    my %GroupNames = map {parse_libname(get_filename($_), "name+ext")=>1} @DyLibPaths;
    foreach my $DyLibPath (sort {length($a)<=>length($b)} @DyLibPaths) {
        getSymbols_Lib($LibVersion, $DyLibPath, 0, \%GroupNames);
    }
}

sub get_VTableSymbolSize($$)
{
    my ($ClassName, $LibVersion) = @_;
    return 0 if(not $ClassName);
    my $Symbol = $ClassVTable{$ClassName};
    if(defined $Symbol_Library{$LibVersion}{$Symbol}
    and my $DyLib = $Symbol_Library{$LibVersion}{$Symbol})
    {# bind class name and v-table size
        if(defined $Library_Symbol{$LibVersion}{$DyLib}{$Symbol}
        and my $Size = -$Library_Symbol{$LibVersion}{$DyLib}{$Symbol})
        {# size from the shared library
            if($Size>=12) {
#               0     (int (*)(...))0
#               4     (int (*)(...))(& _ZTIN7mysqlpp8DateTimeE)
#               8     mysqlpp::DateTime::~DateTime
                return $Size;
            }
            else {
                return 0;
            }
        }
    }
}

sub canonifyName($)
{# make TIFFStreamOpen(char const*, std::basic_ostream<char, std::char_traits<char> >*)
# to be TIFFStreamOpen(char const*, std::basic_ostream<char>*)
    my $Name = $_[0];
    $Name=~s/,\s*std::(allocator|less|char_traits|regex_traits)<\w+> //g;
    return $Name;
}

sub translateSymbols(@_$)
{
    my $LibVersion = pop(@_);
    my (@MnglNames1, @MnglNames2, @UnMnglNames) = ();
    foreach my $Interface (sort @_)
    {
        if($Interface=~/\A_Z/)
        {
            next if($tr_name{$Interface});
            $Interface=~s/[\@\$]+(.*)\Z//;
            push(@MnglNames1, $Interface);
        }
        elsif($Interface=~/\A\?/)
        {
            push(@MnglNames2, $Interface);
        }
        else
        {# not mangled
            $tr_name{$Interface} = $Interface;
            $mangled_name_gcc{$Interface} = $Interface;
            $mangled_name{$LibVersion}{$Interface} = $Interface;
        }
    }
    if($#MnglNames1 > -1)
    {# gcc names
        @UnMnglNames = reverse(unmangleArray(@MnglNames1));
        foreach my $MnglName (@MnglNames1)
        {
            my $Unmangled = $tr_name{$MnglName} = correctName(canonifyName(pop(@UnMnglNames)));
            if(not $mangled_name_gcc{$Unmangled}) {
                $mangled_name_gcc{$Unmangled} = $MnglName;
            }
            if($MnglName=~/\A_ZTV/ and $Unmangled=~/vtable for (.+)/)
            {# bind class name and v-table symbol
                $ClassVTable{$1} = $MnglName;
                $VTableClass{$MnglName} = $1;
            }
        }
    }
    if($#MnglNames2 > -1)
    {# cl names
        @UnMnglNames = reverse(unmangleArray(@MnglNames2));
        foreach my $MnglName (@MnglNames2)
        {
            $tr_name{$MnglName} = correctName(pop(@UnMnglNames));
            $mangled_name{$LibVersion}{$tr_name{$MnglName}} = $MnglName;
        }
    }
}

sub detectAdded()
{
    foreach my $Interface (keys(%{$Symbol_Library{2}}))
    {
        if(not $Symbol_Library{1}{$Interface}) {
            $AddedInt{$Interface} = 1;
        }
        if($CheckObjectsOnly) {
            $CheckedInterfaces{$Interface} = 1;
        }
    }
}

sub detectRemoved()
{
    foreach my $Interface (keys(%{$Symbol_Library{1}}))
    {
        if(not $Symbol_Library{2}{$Interface}
        and not $Symbol_Library{2}{$SymVer{2}{$Interface}})
        {
            if($DepSymbols{2}{$Interface}
            or $DepSymbols_Versioned{2}{$Interface}) {
                next;
            }
            my ($Sym, $Del, $Ver) = get_symbol_name_version($Interface);
            if($Sym and $Ver)
            { # search for the symbol with the same version or without version
                if($DepSymbols{2}{$Sym."\@\@".$Ver}
                or $DepSymbols{2}{$Sym."\@".$Ver} or $DepSymbols{2}{$Sym})
                { # search in the dependencies
                    next;
                }
                if($Symbol_Library{2}{$Sym."\@\@".$Ver}
                or $Symbol_Library{2}{$Sym."\@".$Ver} or $Symbol_Library{2}{$Sym})
                { # search in the library
                    next;
                }
            }
            $RemovedInt{$Interface} = 1;
        }
        if($CheckObjectsOnly) {
            $CheckedInterfaces{$Interface} = 1;
        }
    }
}

sub getSymbols_App($)
{
    my $Path = $_[0];
    return () if(not $Path or not -f $Path);
    my @Imported = ();
    if($OSgroup eq "macos")
    {
        my $OtoolCmd = get_CmdPath("otool");
        if(not $OtoolCmd) {
            exitStatus("Not_Found", "can't find \"otool\"");
        }
        open(APP, "$OtoolCmd -IV $Path 2>$TMP_DIR/null |");
        while(<APP>) {
            if(/[^_]+\s+_?([\w\$]+)\s*\Z/) {
                push(@Imported, $1);
            }
        }
        close(APP);
    }
    elsif($OSgroup eq "windows")
    {
        my $DumpBinCmd = get_CmdPath("dumpbin");
        if(not $DumpBinCmd) {
            exitStatus("Not_Found", "can't find \"dumpbin.exe\"");
        }
        open(APP, "$DumpBinCmd /IMPORTS $Path 2>$TMP_DIR/null |");
        while(<APP>) {
            if(/\s*\w+\s+\w+\s+\w+\s+([\w\?\@]+)\s*/) {
                push(@Imported, $1);
            }
        }
        close(APP);
    }
    else
    {
        my $ReadelfCmd = get_CmdPath("readelf");
        if(not $ReadelfCmd) {
            exitStatus("Not_Found", "can't find \"readelf\"");
        }
        open(APP, "$ReadelfCmd -WhlSsdA $Path |");
        my $symtab=0; # indicates that we are processing 'symtab' section of 'readelf' output
        while(<APP>)
        {
            if( /'.dynsym'/ ) {
                $symtab=0;
            }
            elsif($symtab == 1) {
                # do nothing with symtab (but there are some plans for the future)
                next;
            }
            elsif( /'.symtab'/ ) {
                $symtab=1;
            }
            elsif(my ($fullname, $idx, $Ndx, $type, $size) = readlile_ELF($_)) {
                if( $Ndx eq "UND" ) {
                    #only imported symbols
                    push(@Imported, $fullname);
                }
            }
        }
        close(APP);
    }
    return @Imported;
}

sub readlile_ELF($)
{
    if($_[0]=~/\s*\d+:\s+(\w*)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/)
    { # the line of 'readelf' output corresponding to the interface
        my ($value, $size, $type, $bind,
        $vis, $Ndx, $fullname)=($1, $2, $3, $4, $5, $6, $7);
        if($bind!~/\A(WEAK|GLOBAL)\Z/) {
            return ();
        }
        if($type!~/\A(FUNC|IFUNC|OBJECT|COMMON)\Z/) {
            return ();
        }
        if($vis!~/\A(DEFAULT|PROTECTED)\Z/) {
            return ();
        }
        if($Ndx eq "ABS" and $value!~/\D|1|2|3|4|5|6|7|8|9/) {
            return ();
        }
        return ($fullname, $value, $Ndx, $type, $size);
    }
    else {
        return ();
    }
}

sub getSymbols_Lib($$$$)
{
    my ($LibVersion, $Lib_Path, $IsNeededLib, $GroupNames) = @_;
    return if(not $Lib_Path or not -f $Lib_Path);
    my ($Lib_Dir, $Lib_Name) = separate_path(resolve_symlink($Lib_Path));
    return if($CheckedDyLib{$LibVersion}{$Lib_Name} and $IsNeededLib);
    return if(isCyclical(\@RecurLib, $Lib_Name) or $#RecurLib>=1);
    $CheckedDyLib{$LibVersion}{$Lib_Name} = 1;
    getImplementations($LibVersion, $Lib_Path) if($CheckImplementation and not $IsNeededLib);
    push(@RecurLib, $Lib_Name);
    my (%Value_Interface, %Interface_Value, %NeededLib) = ();
    if(not $IsNeededLib) {
        $DyLibNames_All{$LibVersion}{$Lib_Name} = 1;
    }
    if(not $IsNeededLib)
    { # libstdc++ and libc are always used by other libs
      # if you test one of these libs then you not need
      # to find them in the system for reusing
        if(parse_libname($Lib_Name, "short") eq "libstdc++")
        { # libstdc++.so.6
            $STDCXX_TESTING = 1;
        }
        if(parse_libname($Lib_Name, "short") eq "libc")
        { # libc-2.11.3.so
            $GLIBC_TESTING = 1 if();
        }
    }
    if($OSgroup eq "macos")
    {
        my $OtoolCmd = get_CmdPath("otool");
        if(not $OtoolCmd) {
            exitStatus("Not_Found", "can't find \"otool\"");
        }
        open(DYLIB, "$OtoolCmd -TV $Lib_Path 2>$TMP_DIR/null |");
        while(<DYLIB>)
        {
            if(/[^_]+\s+_([\w\$]+)\s*\Z/)
            {
                my $realname = $1;
                next if($InterfacesListPath and not $InterfacesList{$realname});
                next if($AppPath and not $InterfacesList_App{$realname});
                if($IsNeededLib and $GroupNames
                and not $GroupNames->{parse_libname($Lib_Name, "name+ext")}) {
                    $DepSymbols{$LibVersion}{$realname} = 1;
                }
                if(not $IsNeededLib or ($InterfacesListPath and $InterfacesList{$realname})
                or ($AppPath and $InterfacesList_App{$realname}))
                {
                    $Symbol_Library{$LibVersion}{$realname} = $Lib_Name;
                    $Library_Symbol{$LibVersion}{$Lib_Name}{$realname} = 1;
                    if(not $Language{$LibVersion}{$Lib_Name}) {
                        if($realname =~/\A(_Z|\?)/) {
                            $Language{$LibVersion}{$Lib_Name} = "C++";
                            if(not $IsNeededLib) {
                                $COMMON_LANGUAGE{$LibVersion} = "C++";
                            }
                        }
                    }
                    if($CheckObjectsOnly) {
                        $CheckedInterfaces{$realname} = 1;
                    }
                }
            }
        }
        close(DYLIB);
        open(DYLIB, "$OtoolCmd -L $Lib_Path 2>$TMP_DIR/null |");
        while(<DYLIB>) {
            if(/\s*([\/\\].+\.$LIB_EXT)\s*/
            and $1 ne $Lib_Path) {
                $NeededLib{$1} = 1;
            }
        }
        close(DYLIB);
    }
    elsif($OSgroup eq "windows")
    {
        my $DumpBinCmd = get_CmdPath("dumpbin");
        if(not $DumpBinCmd) {
            exitStatus("Not_Found", "can't find \"dumpbin\"");
        }
        open(DLL, "$DumpBinCmd /EXPORTS \"$Lib_Path\" 2>$TMP_DIR/null |");
        while(<DLL>)
        {
            if(/\A\s*\w+\s+\w+\s+\w+\s+([\w\?\@]+)\s*\Z/
            or /\A\s+_([\w\?\@]+)\s*\Z/
            or /\A\s*\w+\s+\w+\s+([\w\?\@]+)\s*\(\s*forwarded/)
            {# dynamic, static and forwarded symbols
                my $realname = $1;
                $realname=~s/\A_//g;
                next if($InterfacesListPath and not $InterfacesList{$realname});
                next if($AppPath and not $InterfacesList_App{$realname});
                if($IsNeededLib and not $GroupNames->{parse_libname($Lib_Name, "name+ext")}) {
                    $DepSymbols{$LibVersion}{$realname} = 1;
                }
                if(not $IsNeededLib or ($InterfacesListPath and $InterfacesList{$realname})
                or ($AppPath and $InterfacesList_App{$realname}))
                {
                    $Symbol_Library{$LibVersion}{$realname} = $Lib_Name;
                    $Library_Symbol{$LibVersion}{$Lib_Name}{$realname} = 1;
                    if(not $Language{$LibVersion}{$Lib_Name})
                    {
                        if($realname =~/\A(_Z|\?)/) {
                            $Language{$LibVersion}{$Lib_Name} = "C++";
                            if(not $IsNeededLib) {
                                $COMMON_LANGUAGE{$LibVersion} = "C++";
                            }
                        }
                    }
                    if($CheckObjectsOnly) {
                        $CheckedInterfaces{$realname} = 1;
                    }
                }
            }
        }
        close(DLL);
        open(DLL, "$DumpBinCmd /DEPENDENTS $Lib_Path 2>$TMP_DIR/null |");
        while(<DLL>) {
            if(/\s*(.+?\.$LIB_EXT)\s*/i
            and $1 ne $Lib_Path) {
                $NeededLib{$1} = 1;
            }
        }
        close(DLL);
    }
    else
    {# UNIX, ELF
        my $ReadelfCmd = get_CmdPath("readelf");
        if(not $ReadelfCmd) {
            exitStatus("Not_Found", "can't find \"readelf\"");
        }
        open(SOLIB, "$ReadelfCmd -WhlSsdA $Lib_Path 2>$TMP_DIR/null |");
        my $symtab=0; # indicates that we are processing 'symtab' section of 'readelf' output
        while(<SOLIB>)
        {
            if(/NEEDED.+\[([^\[\]]+)\]/) {
                $NeededLib{$1} = 1;
            }
            elsif(/'.dynsym'/) {
                $symtab=0;
            }
            elsif($symtab == 1) {
                # do nothing with symtab (but there are some plans for the future)
                next;
            }
            elsif(/'.symtab'/) {
                $symtab=1;
            }
            elsif(my ($fullname, $idx, $Ndx, $type, $size) = readlile_ELF($_)) {
                if( $Ndx eq "UND" ) {
                    # ignore interfaces that are imported from somewhere else
                    next;
                }
                my ($realname, $version_spec, $version) = get_symbol_name_version($fullname);
                next if($InterfacesListPath and not $InterfacesList{$realname});
                next if($AppPath and not $InterfacesList_App{$realname});
                if($type eq "OBJECT") {
                    $CompleteSignature{$LibVersion}{$fullname}{"Object"} = 1;
                    $CompleteSignature{$LibVersion}{$realname}{"Object"} = 1;
                }
                if($IsNeededLib and not $GroupNames->{parse_libname($Lib_Name, "name+ext")}) {
                    $DepSymbols{$LibVersion}{$fullname} = 1;
                    if($realname ne $fullname) {
                        $DepSymbols_Versioned{$LibVersion}{$realname} = 1;
                    }
                }
                if(not $IsNeededLib or ($InterfacesListPath and $InterfacesList{$realname})
                or ($AppPath and $InterfacesList_App{$realname}))
                {
                    $Symbol_Library{$LibVersion}{$fullname} = $Lib_Name;
                    $Library_Symbol{$LibVersion}{$Lib_Name}{$fullname} = ($type eq "OBJECT")?-$size:1;
                    $Interface_Value{$LibVersion}{$fullname} = $idx;
                    $Value_Interface{$LibVersion}{$idx}{$fullname} = 1;
                    if(not $Language{$LibVersion}{$Lib_Name})
                    {
                        if($fullname=~/\A(_Z|\?)/) {
                            $Language{$LibVersion}{$Lib_Name} = "C++";
                            if(not $IsNeededLib) {
                                $COMMON_LANGUAGE{$LibVersion} = "C++";
                            }
                        }
                    }
                    if($CheckObjectsOnly) {
                        $CheckedInterfaces{$fullname} = 1;
                    }
                }
            }
        }
        close(SOLIB);
    }
    if(not $IsNeededLib)
    {
        foreach my $Interface_Name (keys(%{$Symbol_Library{$LibVersion}}))
        {
            next if($Interface_Name!~/\@/);
            my $Interface_SymName = "";
            foreach my $InterfaceName_SameValue (keys(%{$Value_Interface{$LibVersion}{$Interface_Value{$LibVersion}{$Interface_Name}}}))
            {
                if($InterfaceName_SameValue ne $Interface_Name) {
                    $SymVer{$LibVersion}{$InterfaceName_SameValue} = $Interface_Name;
                    $Interface_SymName = $InterfaceName_SameValue;
                    last;
                }
            }
            if(not $Interface_SymName) {
                if($Interface_Name=~/\A([^\@\$\?]*)[\@\$]+([^\@\$]*)\Z/
                and not $SymVer{$LibVersion}{$1}) {
                    $SymVer{$LibVersion}{$1} = $Interface_Name;
                }
            }
        }
    }
    foreach my $DyLib (sort keys(%NeededLib))
    {
        my $DepPath = find_dylib_path($LibVersion, $DyLib);
        if($DepPath and -f $DepPath) {
            getSymbols_Lib($LibVersion, $DepPath, 1, $GroupNames);
        }
    }
    pop(@RecurLib);
}

sub detectSystemHeaders()
{
    foreach my $DevelPath (keys(%{$SystemPaths{"include"}}))
    {
        next if(not -d $DevelPath);
        foreach my $Path (cmd_find($DevelPath,"f","",""))
        { # search for all header files in the /usr/include
          # with or without extension (ncurses.h, QtCore, ...)
            $SystemHeaders{get_filename($Path)}{$Path}=1;
            if(get_depth($Path)>=3
            and my $Prefix = get_filename(get_dirname($Path))) {
                $SystemHeaders{joinPath($Prefix,get_filename($Path))}{$Path}=1;
            }
        }
    }
    foreach my $DevelPath (keys(%{$SystemPaths{"lib"}}))
    {
        next if(not -d $DevelPath);
        foreach my $Path (cmd_find($DevelPath,"f","*.h",""))
        { # search for config headers in the /usr/lib
            $SystemHeaders{get_filename($Path)}{$Path}=1;
            if(get_depth($Path)>=3
            and my $Prefix = get_filename(get_dirname($Path))) {
                $SystemHeaders{joinPath($Prefix,get_filename($Path))}{$Path}=1;
            }
        }
    }
}

sub detectSystemObjects()
{
    foreach my $DevelPath (keys(%{$SystemPaths{"lib"}}))
    {
        next if(not -d $DevelPath);
        foreach my $Path (find_libs($DevelPath,"",""))
        { # search for shared libraries in the /usr/lib
            $SystemObjects{parse_libname(get_filename($Path), "name+ext")}{$Path}=1;
        }
    }
}

sub find_dylib_path($$)
{
    my ($LibVersion, $DyLib) = @_;
    return "" if(not $DyLib or not $LibVersion);
    return $DyLib if($DyLib=~/\A(\/|\w+:[\/\\])/);
    return $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} if(defined $Cache{"find_dylib_path"}{$LibVersion}{$DyLib});
    if(my @Paths = sort keys(%{$SharedObject_Paths{$LibVersion}{$DyLib}}))
    {
        $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} = $Paths[0];
        return $Cache{"find_dylib_path"}{$LibVersion}{$DyLib};
    }
    elsif(my $DefaultPath = $DyLib_DefaultPath{$DyLib})
    {
        $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} = $DefaultPath;
        return $Cache{"find_dylib_path"}{$LibVersion}{$DyLib};
    }
    else
    {
        foreach my $Dir (sort keys(%DefaultLibPaths), sort keys(%{$SystemPaths{"lib"}}))
        {#search in default linker paths and then in the all system paths
            if(-f $Dir."/".$DyLib) {
                $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} = joinPath($Dir,$DyLib);
                return $Cache{"find_dylib_path"}{$LibVersion}{$DyLib};
            }
        }
        detectSystemObjects() if(not keys(%SystemObjects));
        if(my @AllObjects = keys(%{$SystemObjects{$DyLib}})) {
            $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} = $AllObjects[0];
            return $Cache{"find_dylib_path"}{$LibVersion}{$DyLib};
        }
        my $ShortName = parse_libname($DyLib, "name+ext");
        if($ShortName ne $DyLib
        and my $Path = find_dylib_path($ShortName)) {
            return $Path;
        }
        else {
            $Cache{"find_dylib_path"}{$LibVersion}{$DyLib} = "";
            return "";
        }
    }
}

sub symbols_Preparation($)
{# recreate environment using info from *.abi file
    my $LibVersion = $_[0];
    foreach my $Lib_Name (keys(%{$Library_Symbol{$LibVersion}}))
    {
        foreach my $Interface (keys(%{$Library_Symbol{$LibVersion}{$Lib_Name}}))
        {
            $Symbol_Library{$LibVersion}{$Interface} = $Lib_Name;
            $DyLibNames_All{$LibVersion}{$Lib_Name} = 1;
            if($Library_Symbol{$LibVersion}{$Lib_Name}{$Interface}<=-1)
            {# data marked as -size in the dump
                $CompleteSignature{$LibVersion}{$Interface}{"Object"} = 1;
            }
            if(not $Language{$LibVersion}{$Lib_Name}) {
                if($Interface=~/\A(_Z|\?)/) {
                    $Language{$LibVersion}{$Lib_Name} = "C++";
                }
            }
        }
    }
    my @VFunc = ();
    foreach my $FuncInfoId (keys(%{$SymbolInfo{$LibVersion}}))
    {
        my $MnglName = $SymbolInfo{$LibVersion}{$FuncInfoId}{"MnglName"};
        if(not $MnglName) {# C-functions
            next;
        }
        if(not $Symbol_Library{$LibVersion}{$MnglName}
        and not $DepSymbols{$LibVersion}{$MnglName}) {
            push(@VFunc, $MnglName);
        }
    }
    foreach my $Interface (keys(%{$DepSymbols{$LibVersion}}))
    {
        my ($MnglName, $VersionSpec, $SymbolVersion) = get_symbol_name_version($Interface);
        if($MnglName ne $Interface) {
            $DepSymbols_Versioned{$LibVersion}{$MnglName} = 1;
        }
    }
    translateSymbols(@VFunc, $LibVersion);
}

sub getSoPaths($)
{
    my $LibVersion = $_[0];
    my @SoPaths = ();
    foreach my $Dest (split(/\n/, $Descriptor{$LibVersion}{"Libs"}))
    {
        $Dest=~s/\A\s+|\s+\Z//g;
        next if(not $Dest);
        if(not -e $Dest) {
            print STDERR "ERROR: can't access \'$Dest\'\n";
            next;
        }
        my @SoPaths_Dest = getSOPaths_Dest($Dest, $LibVersion);
        foreach (@SoPaths_Dest) {
            push(@SoPaths, $_);
        }
    }
    return @SoPaths;
}

sub skip_lib($$)
{
    my ($Path, $LibVersion) = @_;
    return 1 if(not $Path or not $LibVersion);
    my $LibName = get_filename($Path);
    my $ShortName = parse_libname($LibName, "name+ext");
    return 1 if($SkipLibs{$LibVersion}{$LibName});
    return 1 if($SkipLibs{$LibVersion}{$ShortName});
    foreach my $Pattern (keys(%{$SkipLibs_Pattern{$LibVersion}}))
    {
        return 1 if($LibName=~/$Pattern/);
        return 1 if($Pattern=~/[\/\\]/ and $Path=~/$Pattern/);
    }
    return 0;
}

sub skip_header($$)
{
    my ($Path, $LibVersion) = @_;
    return 1 if(not $Path or not $LibVersion);
    my $HeaderName = get_filename($Path);
    return 1 if($SkipHeaders{$LibVersion}{$HeaderName});
    foreach my $Pattern (keys(%{$SkipHeaders_Pattern{$LibVersion}}))
    {
        return 1 if($HeaderName=~/$Pattern/);
        return 1 if($Pattern=~/[\/\\]/ and $Path=~/$Pattern/);
    }
    return 0;
}

sub getSOPaths_Dest($$)
{
    my ($Dest, $LibVersion) = @_;
    if(skip_lib($Dest, $LibVersion)) {
        return ();
    }
    if(-f $Dest)
    {
        $SharedObject_Paths{$LibVersion}{get_filename($Dest)}{$Dest} = 1;
        my $ContainingDir = get_dirname($Dest);
        if(not $SharedObject_Dirs{$LibVersion}{$ContainingDir}
        and not $SystemPaths{"lib"}{$ContainingDir}) {
            foreach my $Path (find_libs($ContainingDir,"",1))
            {
                next if(ignore_path($Path, $ContainingDir));
                next if(skip_lib($Path, $LibVersion));
                $SharedObject_Paths{$LibVersion}{get_filename($Path)}{$Path} = 1;
            }
            $SharedObject_Dirs{$LibVersion}{$ContainingDir} = 1;
        }
        return ($Dest);
    }
    elsif(-d $Dest)
    {
        $Dest=~s/[\/\\]+\Z//g;
        my @AllObjects = ();
        if($SystemPaths{"lib"}{$Dest})
        { # you have specified /usr/lib as the search directory (<libs>) in the XML descriptor
          # and the real name of the library by -l option (bz2, stdc++, Xaw, ...)
            foreach my $Path (cmd_find($Dest,"","*".esc($TargetLibraryName)."*\.$LIB_EXT*",2))
            {# all files and symlinks that match the name of a library
                if(get_filename($Path)=~/\A(|lib)\Q$TargetLibraryName\E[\d\-]*\.$LIB_EXT[\d\.]*\Z/i)
                {
                    $SharedObject_Paths{$LibVersion}{get_filename($Path)}{$Path} = 1;
                    push(@AllObjects, resolve_symlink($Path));
                }
            }
        }
        else
        {# search for all files and symlinks
            foreach my $Path (find_libs($Dest,"",""))
            {
                next if(ignore_path($Path, $Dest));
                next if(skip_lib($Path, $LibVersion));
                $SharedObject_Paths{$LibVersion}{get_filename($Path)}{$Path} = 1;
                push(@AllObjects, resolve_symlink($Path));
            }
            if($OSgroup eq "macos")
            {# shared libraries on MacOS X may have no extension
                foreach my $Path (cmd_find($Dest,"f","",""))
                {
                    next if(ignore_path($Path, $Dest));
                    next if(skip_lib($Path, $LibVersion));
                    if(get_filename($Path)!~/\./
                    and cmd_file($Path)=~/(shared|dynamic)\s+library/i) {
                        $SharedObject_Paths{$LibVersion}{get_filename($Path)}{$Path} = 1;
                        push(@AllObjects, resolve_symlink($Path));
                    }
                }
            }
        }
        return @AllObjects;
    }
    else {
        return ();
    }
}

sub isCyclical($$)
{
    return (grep {$_ eq $_[1]} @{$_[0]});
}

sub read_symlink($)
{
    my $Path = $_[0];
    return "" if(not $Path);
    return "" if(not -f $Path and not -l $Path);
    if(defined $Cache{"read_symlink"}{$Path}) {
        return $Cache{"read_symlink"}{$Path};
    }
    if(my $Res = readlink($Path)) {
        return ($Cache{"read_symlink"}{$Path} = $Res);
    }
    elsif(my $ReadlinkCmd = get_CmdPath("readlink")) {
        return ($Cache{"read_symlink"}{$Path} = `$ReadlinkCmd -n $Path`);
    }
    elsif(my $FileCmd = get_CmdPath("file"))
    {
        my $Info = `$FileCmd $Path`;
        if($Info=~/symbolic\s+link\s+to\s+['`"]*([\w\d\.\-\/\\]+)['`"]*/i) {
            return ($Cache{"read_symlink"}{$Path} = $1);
        }
    }
    return ($Cache{"read_symlink"}{$Path} = "");
}

sub resolve_symlink($)
{
    my $Path = $_[0];
    return "" if(not $Path);
    return "" if(not -f $Path and not -l $Path);
    if(defined $Cache{"resolve_symlink"}{$Path}) {
        return $Cache{"resolve_symlink"}{$Path};
    }
    return $Path if(isCyclical(\@RecurSymlink, $Path));
    push(@RecurSymlink, $Path);
    if(-l $Path and my $Redirect=read_symlink($Path))
    {
        if($Redirect=~/\A(\/|\w+:[\/\\])/)
        {# absolute path
            if($SystemRoot and $SystemRoot ne "/"
            and $Path=~/\A\Q$SystemRoot\E\//
            and (-f $SystemRoot.$Redirect or -l $SystemRoot.$Redirect))
            { # symbolic links from the sysroot
              # should be corrected to point to
              # the files inside sysroot
                $Redirect = $SystemRoot.$Redirect;
            }
            my $Res = resolve_symlink($Redirect);
            pop(@RecurSymlink);
            return ($Cache{"resolve_symlink"}{$Path} = $Res);
        }
        elsif($Redirect=~/\.\.[\/\\]/)
        {# relative path
            $Redirect = joinPath(get_dirname($Path),$Redirect);
            while($Redirect=~s&(/|\\)[^\/\\]+(\/|\\)\.\.(\/|\\)&$1&){};
            my $Res = resolve_symlink($Redirect);
            pop(@RecurSymlink);
            return ($Cache{"resolve_symlink"}{$Path} = $Res);
        }
        elsif(-f get_dirname($Path)."/".$Redirect)
        {# file name in the same directory
            my $Res = resolve_symlink(joinPath(get_dirname($Path),$Redirect));
            pop(@RecurSymlink);
            return ($Cache{"resolve_symlink"}{$Path} = $Res);
        }
    }
    pop(@RecurSymlink);
    return ($Cache{"resolve_symlink"}{$Path} = $Path);
}

sub genDescriptorTemplate()
{
    writeFile("VERSION.xml", $Descriptor_Template."\n");
    print "XML-descriptor template ./VERSION.xml has been generated\n";
}

sub detectPointerSize()
{
    return if(not $GCC_PATH);
    writeFile("$TMP_DIR/empty.h", "");
    my $Defines = `$GCC_PATH -E -dD $TMP_DIR/empty.h`;
    unlink("$TMP_DIR/empty.h");
    if($Defines=~/ __SIZEOF_POINTER__\s+(\d+)/)
    {# GCC 4
        $POINTER_SIZE{1} = $1;
        $POINTER_SIZE{2} = $1;
    }
    elsif($Defines=~/ __PTRDIFF_TYPE__\s+(\w+)/)
    {# GCC 3
        my $PTRDIFF = $1;
        if($PTRDIFF=~/long/) {
            $POINTER_SIZE{1} = 8;
            $POINTER_SIZE{2} = 8;
        }
        else {
            $POINTER_SIZE{1} = 4;
            $POINTER_SIZE{2} = 4;
        }
    }
    if(not int($POINTER_SIZE{1})) {
        exitStatus("Error", "can't check size of pointer");
    }
}

sub majorVersion($)
{
    my $Version = $_[0];
    return 0 if(not $Version);
    my @VParts = split(/\./, $Version);
    return $VParts[0];
}

sub cmpVersions($$)
{# compare two versions in dotted-numeric format
    my ($V1, $V2) = @_;
    return 0 if($V1 eq $V2);
    return undef if($V1!~/\A\d+[\.\d+]*\Z/);
    return undef if($V2!~/\A\d+[\.\d+]*\Z/);
    my @V1Parts = split(/\./, $V1);
    my @V2Parts = split(/\./, $V2);
    for (my $i = 0; $i <= $#V1Parts && $i <= $#V2Parts; $i++) {
        return -1 if(int($V1Parts[$i]) < int($V2Parts[$i]));
        return 1 if(int($V1Parts[$i]) > int($V2Parts[$i]));
    }
    return -1 if($#V1Parts < $#V2Parts);
    return 1 if($#V1Parts > $#V2Parts);
    return 0;
}

sub read_ABI_Dump($$)
{
    my ($LibVersion, $Path) = @_;
    return if(not $LibVersion or not -e $Path);
    my $FilePath = unpackDump($Path);
    return if($FilePath!~/\.abi\Z/);
    my $Content = readFile($FilePath);
    unlink($FilePath);
    if($Content!~/};\s*\Z/) {
        exitStatus("Invalid_Dump", "specified ABI dump \'$Path\' is not valid, try to recreate it");
    }
    my $LibraryABI = eval($Content);
    if(not $LibraryABI) {
        exitStatus("Error", "internal error - eval() procedure seem to not working correctly, try to remove 'use strict' and try again");
    }
    # new dumps (>=1.22) have a personal versioning
    my $DumpVersion = $LibraryABI->{"ABI_DUMP_VERSION"};
    my $ToolVersion = $LibraryABI->{"ABI_COMPLIANCE_CHECKER_VERSION"};
    if(not $DumpVersion)
    {# old dumps (<=1.21.6) have been marked by the tool version
        $DumpVersion = $ToolVersion;
    }
    $UsedDump{$LibVersion} = $DumpVersion;
    if(majorVersion($DumpVersion) ne $ABI_DUMP_MAJOR)
    {# should be compatible with dumps of the same major version
        if(cmpVersions($DumpVersion, $ABI_DUMP_VERSION)>0) {
            # Don't know how to parse future dump formats
            exitStatus("Dump_Version", "incompatible version $DumpVersion of specified ABI dump (newer than $ABI_DUMP_VERSION)");
        }
        elsif(cmpVersions($DumpVersion, $ABI_COMPLIANCE_CHECKER_VERSION)>0 and not $LibraryABI->{"ABI_DUMP_VERSION"}) {
            # Don't know how to parse future dump formats
            exitStatus("Dump_Version", "incompatible version $DumpVersion of specified ABI dump (newer than $ABI_COMPLIANCE_CHECKER_VERSION)");
        }
        if($UseOldDumps) {
            if(cmpVersions($DumpVersion, $OLDEST_SUPPORTED_VERSION)<0) {
                exitStatus("Dump_Version", "incompatible version $DumpVersion of specified ABI dump (older than $OLDEST_SUPPORTED_VERSION)");
            }
        }
        else
        {
            my $Msg = "incompatible version $DumpVersion of specified ABI dump (allowed only $ABI_DUMP_MAJOR.0<=V<=$ABI_DUMP_MAJOR.9)";
            if(cmpVersions($DumpVersion, $OLDEST_SUPPORTED_VERSION)>=0) {
                $Msg .= "\nUse -old-dumps option to use old-version dumps ($OLDEST_SUPPORTED_VERSION<=V<$ABI_DUMP_MAJOR.0)";
            }
            exitStatus("Dump_Version", $Msg);
        }
    }
    $TypeInfo{$LibVersion} = $LibraryABI->{"TypeInfo"};
    if(not $TypeInfo{$LibVersion})
    {# support for old dumps
        $TypeInfo{$LibVersion} = $LibraryABI->{"TypeDescr"};
    }
    if(cmpVersions($DumpVersion, "2.0")<0)
    {# support for old dumps: type names corrected in ACC 1.22 (dump 2.0 format)
        foreach my $TypeDeclId (keys(%{$TypeInfo{$LibVersion}})) {
            foreach my $TypeId (keys(%{$TypeInfo{$LibVersion}{$TypeDeclId}}))
            {
                my $TName = $TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId}{"Name"};
                if($TName=~/\A(\w+)::(\w+)/) {
                    my ($Part1, $Part2) = ($1, $2);
                    if($Part1 eq $Part2) {
                        $TName=~s/\A$Part1:\:$Part1(\W)/$Part1$1/;
                    }
                    else {
                        $TName=~s/\A(\w+:\:)$Part2:\:$Part2(\W)/$1$Part2$2/;
                    }
                }
                $TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId}{"Name"} = $TName;
            }
        }
    }
    foreach my $TypeDeclId (keys(%{$TypeInfo{$LibVersion}})) {
        foreach my $TypeId (keys(%{$TypeInfo{$LibVersion}{$TypeDeclId}}))
        {
            $TName_Tid{$LibVersion}{$TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId}{"Name"}} = $TypeId;
            if(defined $TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId}{"Base"})
            {
                foreach (keys(%{$TypeInfo{$LibVersion}{$TypeDeclId}{$TypeId}{"Base"}})) {
                    $Class_SubClasses{$LibVersion}{$_}{$TypeId}=1;
                }
            }
        }
    }
    $SymbolInfo{$LibVersion} = $LibraryABI->{"SymbolInfo"};
    if(not $SymbolInfo{$LibVersion})
    {# support for old dumps
        $SymbolInfo{$LibVersion} = $LibraryABI->{"FuncDescr"};
    }
    if(not keys(%{$SymbolInfo{$LibVersion}}))
    { # validation of old-version dumps
        exitStatus("Invalid_Dump", "the input dump d$LibVersion is invalid");
    }
    $Library_Symbol{$LibVersion} = $LibraryABI->{"Symbols"};
    if(not $Library_Symbol{$LibVersion})
    {# support for old dumps
        $Library_Symbol{$LibVersion} = $LibraryABI->{"Interfaces"};
    }
    $DepSymbols{$LibVersion} = $LibraryABI->{"DepSymbols"};
    if(not $DepSymbols{$LibVersion})
    {# support for old dumps
        $DepSymbols{$LibVersion} = $LibraryABI->{"DepInterfaces"};
    }
    if(not $DepSymbols{$LibVersion})
    {# support for old dumps
        # Cannot reconstruct DepSymbols. This may result in false
        # positives if the old dump is for library 2. Not a problem if
        # old dumps are only from old libraries.
        $DepSymbols{$LibVersion} = {};
    }
    $SymVer{$LibVersion} = $LibraryABI->{"SymbolVersion"};
    $Tid_TDid{$LibVersion} = $LibraryABI->{"Tid_TDid"};
    $Descriptor{$LibVersion}{"Version"} = $LibraryABI->{"LibraryVersion"};
    $OpaqueTypes{$LibVersion} = $LibraryABI->{"OpaqueTypes"};
    $SkipSymbols{$LibVersion} = $LibraryABI->{"SkipSymbols"};
    if(not $SkipSymbols{$LibVersion})
    {# support for old dumps
        $SkipSymbols{$LibVersion} = $LibraryABI->{"SkipInterfaces"};
    }
    if(not $SkipSymbols{$LibVersion})
    {# support for old dumps
        $SkipSymbols{$LibVersion} = $LibraryABI->{"InternalInterfaces"};
    }
    read_Headers_DumpInfo($LibraryABI, $LibVersion);
    read_Libs_DumpInfo($LibraryABI, $LibVersion);
    $Constants{$LibVersion} = $LibraryABI->{"Constants"};
    $NestedNameSpaces{$LibVersion} = $LibraryABI->{"NameSpaces"};
    if(not $NestedNameSpaces{$LibVersion})
    {# support for old dumps
        # Cannot reconstruct NameSpaces. This may affect design
        # of the compatibility report.
        $NestedNameSpaces{$LibVersion} = {};
    }
    read_Machine_DumpInfo($LibraryABI, $LibVersion);
    symbols_Preparation($LibVersion);
    $Descriptor{$LibVersion}{"Dump"} = 1;
}

sub read_Machine_DumpInfo($$)
{
    my ($LibraryABI, $LibVersion) = @_;
    if($LibraryABI->{"Arch"}) {
        $CPU_ARCH{$LibVersion} = $LibraryABI->{"Arch"};
    }
    if($LibraryABI->{"SizeOfPointer"}) {
        $POINTER_SIZE{$LibVersion} = $LibraryABI->{"SizeOfPointer"};
    }
    if($LibraryABI->{"GccVersion"}) {
        $GCC_VERSION{$LibVersion} = $LibraryABI->{"GccVersion"};
    }
}

sub read_Libs_DumpInfo($$)
{
    my ($LibraryABI, $LibVersion) = @_;
    $DyLibNames_All{$LibVersion} = $LibraryABI->{"DynamicLibs"};
    if(not $DyLibNames_All{$LibVersion})
    {# support for old dumps
        $DyLibNames_All{$LibVersion} = $LibraryABI->{"SharedObjects"};
    }
    if(keys(%{$DyLibNames_All{$LibVersion}})
    and not $DumpInfo_DPath) {
        $Descriptor{$LibVersion}{"Libs"} = "OK";
    }
}

sub read_Libs_DumpInfo($$)
{
    my ($LibraryABI, $LibVersion) = @_;
    $DyLibNames_All{$LibVersion} = $LibraryABI->{"DynamicLibs"};
    if(not $DyLibNames_All{$LibVersion})
    {# support for old dumps
        $DyLibNames_All{$LibVersion} = $LibraryABI->{"SharedObjects"};
    }
    if(keys(%{$DyLibNames_All{$LibVersion}})
    and not $DumpInfo_DPath) {
        $Descriptor{$LibVersion}{"Libs"} = "OK";
    }
}

sub read_Headers_DumpInfo($$)
{
    my ($LibraryABI, $LibVersion) = @_;
    if(keys(%{$LibraryABI->{"Headers"}})
    and not $DumpInfo_DPath) {
        $Descriptor{$LibVersion}{"Headers"} = "OK";
    }
    foreach my $Identity (keys(%{$LibraryABI->{"Headers"}}))
    {# headers info is stored in the old dumps in the different way
        if($UseOldDumps
        and my $Name = $LibraryABI->{"Headers"}{$Identity}{"Name"})
        {# support for old dumps: headers info corrected in 1.22
            $Identity = $Name;
        }
        $Target_Headers{$LibVersion}{$Identity}{"Identity"} = $Identity;
    }
}

sub find_libs($$$)
{
    my ($Path, $Type, $MaxDepth) = @_;
    return cmd_find($Path, $Type, ".*\\.$LIB_EXT\[0-9.]*", $MaxDepth);
}

sub createDescriptor_Rpm_Deb($$)
{
    my ($LibVersion, $VirtualPath) = @_;
    return if(not $LibVersion or not $VirtualPath or not -e $VirtualPath);
    my $IncDir = "none";
    $IncDir = $VirtualPath."/usr/include/" if(-d $VirtualPath."/usr/include/");
    $IncDir = $VirtualPath."/include/" if(-d $VirtualPath."/include/");
    my $LibDir = "none";
    $LibDir = $VirtualPath."/lib64/" if(-d $VirtualPath."/lib64/");
    $LibDir = $VirtualPath."/usr/lib64/" if(-d $VirtualPath."/usr/lib64/");
    $LibDir = $VirtualPath."/lib/" if(-d $VirtualPath."/lib/");
    $LibDir = $VirtualPath."/usr/lib/" if(-d $VirtualPath."/usr/lib/");
    if($IncDir ne "none")
    {
        my @Objects = ();
        foreach my $Path (find_libs($LibDir,"f","")) {
            push(@Objects, $Path);
        }
        $LibDir = "none" if($#Objects==-1);
    }
    if($IncDir eq "none" and $LibDir eq "none") {
        exitStatus("Error", "can't find headers and $SHARED libraries in the package d$LibVersion");
    }
    readDescriptor($LibVersion,"
      <version>
          ".$TargetVersion{$LibVersion}."
      </version>

      <headers>
          $IncDir
      </headers>

      <libs>
          $LibDir
      </libs>");
}

sub createDescriptor($$)
{
    my ($LibVersion, $Path) = @_;
    return if(not $LibVersion or not $Path or not -e $Path);
    if($Path=~/\Q.abi.tar.gz\E\Z/)
    {# read an ABI dump
        read_ABI_Dump($LibVersion, $Path);
    }
    else
    {
        if($Path=~/\.deb\Z/i)
        {# Deb "devel" package
            my $ArCmd = get_CmdPath("ar");
            if(not $ArCmd) {
                exitStatus("Not_Found", "can't find \"ar\"");
            }
            my $VirtualPath = "$TMP_DIR/package.v$LibVersion/";
            mkpath($VirtualPath);
            system("cd $VirtualPath && $ArCmd p ".abs_path($Path)." data.tar.gz | tar zx");
            if($?) {
                exitStatus("Error", "can't extract package d$LibVersion");
            }
            if(not $TargetVersion{$LibVersion})
            {# auto detect version of Deb package
                system("cd $VirtualPath && $ArCmd p ".abs_path($Path)." control.tar.gz | tar zx");
                if(-f "$VirtualPath/control") {
                    my $ControlInfo = `grep Version $VirtualPath/control`;
                    if($ControlInfo=~/Version\s*:\s*([\d\.\-\_]*\d)/) {
                        $TargetVersion{$LibVersion} = $1;
                    }
                }
            }
            if(not $TargetVersion{$LibVersion})
            {
                if($Path=~/_([\d\.\-\_]*\d)(.+?)\.deb\Z/i)
                {# auto detect version of Deb package
                    $TargetVersion{$LibVersion} = $1;
                }
            }
            if(not $TargetVersion{$LibVersion}) {
                exitStatus("Error", "specify ".($LibVersion==1?"1st":"2nd")." version number (-v$LibVersion <num> option)");
            }
            createDescriptor_Rpm_Deb($LibVersion, $VirtualPath);
        }
        elsif($Path=~/\.rpm\Z/i)
        {# RPM "devel" package
            my $Rpm2CpioCmd = get_CmdPath("rpm2cpio");
            if(not $Rpm2CpioCmd) {
                exitStatus("Not_Found", "can't find \"rpm2cpio\"");
            }
            my $CpioCmd = get_CmdPath("cpio");
            if(not $CpioCmd) {
                exitStatus("Not_Found", "can't find \"cpio\"");
            }
            my $VirtualPath = "$TMP_DIR/package.v$LibVersion/";
            mkpath($VirtualPath);
            system("cd $VirtualPath && $Rpm2CpioCmd ".abs_path($Path)." | cpio -id --quiet");
            if($?) {
                exitStatus("Error", "can't extract package d$LibVersion");
            }
            my $RpmCmd = get_CmdPath("rpm");
            if(not $RpmCmd) {
                exitStatus("Not_Found", "can't find \"rpm\"");
            }
            if(not $TargetVersion{$LibVersion}) {
                $TargetVersion{$LibVersion} = `$RpmCmd -qp --queryformat \%{version} $Path 2>$TMP_DIR/null`;
            }
            if(not $TargetVersion{$LibVersion})
            {
                if($Path=~/-([\d\.\-\_]*\d)(.+?)\.rpm\Z/i)
                {# auto detect version of Rpm package
                    $TargetVersion{$LibVersion} = $1;
                }
            }
            if(not $TargetVersion{$LibVersion}) {
                exitStatus("Error", "specify ".($LibVersion==1?"1st":"2nd")." version number (-v$LibVersion <num> option)");
            }
            createDescriptor_Rpm_Deb($LibVersion, $VirtualPath);
        }
        elsif(-d $Path)
        {# directory with headers files and shared objects
            readDescriptor($LibVersion,"
              <version>
                  ".$TargetVersion{$LibVersion}."
              </version>

              <headers>
                  $Path
              </headers>

              <libs>
                  $Path
              </libs>");
        }
        else
        {# files
            if(is_header($Path, 1, $LibVersion))
            {# header file
                readDescriptor($LibVersion,"
                  <version>
                      ".$TargetVersion{$LibVersion}."
                  </version>

                  <headers>
                      $Path
                  </headers>

                  <libs>
                      none
                  </libs>");
            }
            elsif(parse_libname($Path, "name"))
            {# shared object
                readDescriptor($LibVersion,"
                  <version>
                      ".$TargetVersion{$LibVersion}."
                  </version>

                  <headers>
                      none
                  </headers>

                  <libs>
                      $Path
                  </libs>");
            }
            else
            {# standard XML-descriptor
                readDescriptor($LibVersion, readFile($Path));
            }
        }
        # create log directory
        my ($LOG_DIR, $LOG_FILE) = ("logs/$TargetLibraryName/".$Descriptor{$LibVersion}{"Version"}, "log.txt");
        if($OutputLogPath{$LibVersion}) {# used defined by -log-path option
            ($LOG_DIR, $LOG_FILE) = separate_path($OutputLogPath{$LibVersion});
        }
        mkpath($LOG_DIR);
        $LOG_PATH{$LibVersion} = abs_path($LOG_DIR)."/".$LOG_FILE;
        unlink($LOG_PATH{$LibVersion});# remove an old log
    }
}

sub detect_lib_default_paths()
{
    my %LPaths = ();
    if($OSgroup eq "bsd")
    {
        if(my $LdConfig = get_CmdPath("ldconfig")) {
            foreach my $Line (split(/\n/, `$LdConfig -r 2>$TMP_DIR/null`)) {
                if($Line=~/\A[ \t]*\d+:\-l(.+) \=\> (.+)\Z/) {
                    $LPaths{"lib".$1} = $2;
                }
            }
        }
        else {
            print "WARNING: can't find ldconfig\n";
        }
    }
    else
    {
        if(my $LdConfig = get_CmdPath("ldconfig"))
        {
            if($SystemRoot and $Config{"osname"} eq "linux")
            {# use host (x86) ldconfig with the target (arm) ld.so.conf
                if(-e $SystemRoot."/etc/ld.so.conf") {
                    $LdConfig .= " -f ".$SystemRoot."/etc/ld.so.conf";
                }
            }
            foreach my $Line (split(/\n/, `$LdConfig -p 2>$TMP_DIR/null`)) {
                if($Line=~/\A[ \t]*([^ \t]+) .* \=\> (.+)\Z/) {
                    $LPaths{$1} = $2;
                }
            }
        }
        elsif($Config{"osname"}=~/linux/i) {
            print "WARNING: can't find ldconfig\n";
        }
    }
    return \%LPaths;
}

sub detect_bin_default_paths()
{
    my $EnvPaths = $ENV{"PATH"};
    if($OSgroup eq "beos") {
        $EnvPaths.=":".$ENV{"BETOOLS"};
    }
    my $Sep = ($OSgroup eq "windows")?";":":|;";
    foreach my $Path (sort {length($a)<=>length($b)} split(/$Sep/, $EnvPaths))
    {
        $Path=~s/[\/\\]+\Z//g;
        next if(not $Path);
        if($SystemRoot
        and $Path=~/\A\Q$SystemRoot\E\//)
        { # do NOT use binaries from target system
            next;
        }
        $DefaultBinPaths{$Path} = 1;
    }
}

sub detect_include_default_paths()
{
    return if(not $GCC_PATH);
    writeFile("$TMP_DIR/empty.h", "");
    foreach my $Line (split(/\n/, `$GCC_PATH -v -x c++ -E $TMP_DIR/empty.h 2>&1`))
    {# detecting gcc default include paths
        if($Line=~/\A[ \t]*((\/|\w+:\\)[^ ]+)[ \t]*\Z/)
        {
            my $Path = $1;
            while($Path=~s&(\/|\\)[^\/\\]+(\/|\\)\.\.(\/|\\)&$1&){};
            $Path=~s/[\/\\]+\Z//g;
            $Path=~s/\//\\/g if($OSgroup eq "windows");
            if($Path=~/c\+\+|g\+\+/) {
                $DefaultCppPaths{$Path}=1;
                if(not defined $MAIN_CPP_DIR
                or get_depth($MAIN_CPP_DIR)>get_depth($Path)) {
                    $MAIN_CPP_DIR = $Path;
                }
            }
            elsif($Path=~/gcc/) {
                $DefaultGccPaths{$Path}=1;
            }
            else {
                next if($Path=~/local(\/|\\)include/);
                if($SystemRoot
                and $Path!~/\A\Q$SystemRoot\E\//)
                { # The GCC include path for user headers is not a part of the system root
                  # The reason: you are not specified the --cross-gcc option or selected a wrong compiler
                  # or it is the internal cross-GCC path like arm-linux-gnueabi/include
                    next;
                }
                $DefaultIncPaths{$Path}=1;
            }
        }
    }
    unlink("$TMP_DIR/empty.h");
}

sub detect_default_paths()
{
    foreach my $Type (keys(%{$OS_AddPath{$OSgroup}}))
    {# additional search paths
        foreach my $Path (keys(%{$OS_AddPath{$OSgroup}{$Type}}))
        {
            next if(not -d $Path);
            $SystemPaths{$Type}{$Path} = $OS_AddPath{$OSgroup}{$Type}{$Path};
        }
    }
    if($OSgroup ne "windows")
    {# unix-like
        foreach my $Type ("include", "lib", "bin")
        {# automatic detection of system "devel" directories
            my ($UsrDir, $RootDir) = ("/usr", "/");
            if($SystemRoot and $Type ne "bin")
            {# 1. search for target headers and libraries
             # 2. use host commands: ldconfig, readelf, etc.
                ($UsrDir, $RootDir) = ("$SystemRoot/usr", $SystemRoot);
            }
            foreach my $Path (cmd_find($RootDir,"d","*$Type*",1)) {
                $SystemPaths{$Type}{$Path} = 1;
            }
            if(-d $RootDir."/".$Type)
            {# if "/lib" is symbolic link
                if($RootDir eq "/") {
                    $SystemPaths{$Type}{"/".$Type} = 1;
                }
                else {
                    $SystemPaths{$Type}{$RootDir."/".$Type} = 1;
                }
            }
            if(-d $UsrDir) {
                foreach my $Path (cmd_find($UsrDir,"d","*$Type*",1)) {
                    $SystemPaths{$Type}{$Path} = 1;
                }
                if(-d $UsrDir."/".$Type)
                {# if "/usr/lib" is symbolic link
                    $SystemPaths{$Type}{$UsrDir."/".$Type} = 1;
                }
            }
        }
    }
    detect_bin_default_paths();
    foreach my $Path (keys(%DefaultBinPaths)) {
        $SystemPaths{"bin"}{$Path} = $DefaultBinPaths{$Path};
    }
    # check environment variables
    if($OSgroup eq "beos")
    {
        foreach my $Path (split(/:|;/, $ENV{"BEINCLUDES"}))
        {
            if($Path=~/\A(\/|\w+:[\/\\])/) {
                $DefaultIncPaths{$Path} = 1;
            }
        }
        foreach my $Path (split(/:|;/, $ENV{"BELIBRARIES"}), split(/:|;/, $ENV{"LIBRARY_PATH"}))
        {
            if($Path=~/\A(\/|\w+:[\/\\])/) {
                $DefaultLibPaths{$Path} = 1;
            }
        }
    }
    else
    {
        my $Sep = ($OSgroup eq "windows")?";":":|;";
        foreach my $Var (keys(%ENV))
        {
            if($Var=~/INCLUDE/i) {
                foreach my $Path (split(/$Sep/, $ENV{$Var})) {
                    if($Path=~/\A(\/|\w+:[\/\\])/) {
                        $SystemPaths{"include"}{$Path} = 1;
                    }
                }
            }
            elsif($Var=~/(\ALIB)|LIBRAR(Y|IES)/i) {
                foreach my $Path (split(/$Sep/, $Var)) {
                    if($Path=~/\A(\/|\w+:[\/\\])/) {
                        $SystemPaths{"lib"}{$Path} = 1;
                    }
                }
            }
        }
    }
    if(my $LPaths = detect_lib_default_paths()) {
        foreach my $Path (keys(%{$LPaths}))
        {
            if($SystemRoot
            and $Path!~/\A\Q$SystemRoot\E\//)
            { # wrong ldconfig configuration
              # check your <sysroot>/etc/ld.so.conf
                next;
            }
            $DyLib_DefaultPath{$Path} = $LPaths->{$Path};
            $DefaultLibPaths{get_dirname($LPaths->{$Path})} = 1;
        }
    }
    foreach my $Path (keys(%DefaultLibPaths)) {
        $SystemPaths{"lib"}{$Path} = $DefaultLibPaths{$Path};
    }
    $GCC_PATH = get_CmdPath("gcc");
    if($CrossGcc)
    {
        if(-e $CrossGcc)
        {# absolute or relative path
            $GCC_PATH = abs_path($CrossGcc);
        }
        elsif($CrossGcc!~/\// and get_CmdPath($CrossGcc))
        {# -cross-gcc arm-linux-gcc
            $GCC_PATH = $CrossGcc;
        }
        else {
            exitStatus("Access_Error", "can't access \'$CrossGcc\'");
        }
    }
    if(not $GCC_PATH) {
        exitStatus("Not_Found", "can't find GCC>=3.0 in PATH");
    }
    $CPP_FILT = get_CmdPath("c++filt");
    if(not $CPP_FILT) {
        exitStatus("Not_Found", "can't find c++filt in PATH");
    }
    if(not $CheckObjectsOnly_Opt)
    {
        if(my $GCC_Ver = get_dumpversion($GCC_PATH)) {
            print "Using GCC $GCC_Ver (".get_dumpmachine($GCC_PATH).")\n";
        }
        else {
            exitStatus("Error", "something is going wrong with the GCC compiler");
        }
    }
    detect_include_default_paths();
    foreach my $Path (keys(%DefaultIncPaths)) {
        $SystemPaths{"include"}{$Path} = $DefaultIncPaths{$Path};
    }
    if(-d $SystemRoot."/usr/include") {
        $UserIncPath{$SystemRoot."/usr/include"}=1;
    }
}

sub get_dumpversion($)
{
    my $Cmd = $_[0];
    return "" if(not $Cmd);
    if($Cache{"get_dumpversion"}{$Cmd}) {
        return $Cache{"get_dumpversion"}{$Cmd};
    }
    my $Version = `$Cmd -dumpversion 2>$TMP_DIR/null`;
    chomp($Version);
    $Cache{"get_dumpversion"}{$Cmd} = $Version;
    return $Version;
}

sub get_dumpmachine($)
{
    my $Cmd = $_[0];
    return "" if(not $Cmd);
    if($Cache{"get_dumpmachine"}{$Cmd}) {
        return $Cache{"get_dumpmachine"}{$Cmd};
    }
    my $Version = `$Cmd -dumpmachine 2>$TMP_DIR/null`;
    chomp($Version);
    $Cache{"get_dumpmachine"}{$Cmd} = $Version;
    return $Version;
}

sub get_version($)
{
    my $Cmd = $_[0];
    return "" if(not $Cmd);
    my $Version = `$Cmd --version 2>$TMP_DIR/null`;
    return $Version;
}

sub check_gcc_version($$)
{
    my ($Cmd, $Req) = @_;
    return 0 if(not $Cmd or not $Req);
    my $GccVersion = get_dumpversion($Cmd);
    if(cmpVersions($GccVersion, $Req)>=0) {
        return $Cmd;
    }
    return "";
}

sub get_depth($)
{
    if(defined $Cache{"get_depth"}{$_[0]}) {
        return $Cache{"get_depth"}{$_[0]}
    }
    return ($Cache{"get_depth"}{$_[0]} = ($_[0]=~tr![/\]|::!!));
}

sub find_gcc_cxx_headers($)
{
    my $LibVersion = $_[0];
    return if($Cache{"find_gcc_cxx_headers"});# this function should be called once
    # detecting system header paths
    foreach my $Path (sort {get_depth($b) <=> get_depth($a)} keys(%DefaultGccPaths))
    {
        foreach my $HeaderPath (sort {get_depth($a) <=> get_depth($b)} cmd_find($Path,"f","",""))
        {
            my $FileName = get_filename($HeaderPath);
            next if($DefaultGccHeader{$FileName});
            $DefaultGccHeader{$FileName} = $HeaderPath;
        }
    }
    if(($COMMON_LANGUAGE{$LibVersion} eq "C++" or $CheckHeadersOnly) and not $STDCXX_TESTING)
    {
        foreach my $CppDir (sort {get_depth($b)<=>get_depth($a)} keys(%DefaultCppPaths))
        {
            my @AllCppHeaders = cmd_find($CppDir,"f","","");
            foreach my $Path (sort {get_depth($a)<=>get_depth($b)} @AllCppHeaders)
            {
                my $FileName = get_filename($Path);
                next if($DefaultCppHeader{$FileName});
                $DefaultCppHeader{$FileName} = $Path;
            }
        }
    }
    $Cache{"find_gcc_cxx_headers"} = 1;
}

sub parse_libname($$)
{# separate short name and version
    my ($Name, $Type) = @_;
    if($Name=~/((((lib|).+?)[\-\.\_]*([0-9\-\.\_]*))\.$LIB_EXT)[\.]*([0-9\-\.\_]*)\Z/)
    { # libSDL-1.2.so.0.7.1
        if($Type eq "name")
        { # libSDL-1.2
            return $2;
        }
        elsif($Type eq "name+ext")
        { # libSDL-1.2.so
            return $1;
        }
        elsif($Type eq "version")
        {
            if($6)
            { # 0.7.1
                return $6;
            }
            else
            { # for libc-2.5.so (=>2.5 version)
                return $5;
            }
        }
        elsif($Type eq "short")
        { # libSDL
            return $3;
        }
        elsif($Type eq "shortest")
        { # SDL
            return shortest_name($3);
        }
    }
    return "";# error
}

sub shortest_name($)
{
    my $Name = $_[0];
    # remove prefix
    $Name=~s/\A(lib|open)//;
    # remove suffix
    $Name=~s/[\W\d_]+\Z//i;
    $Name=~s/([a-z]{2,})(lib)\Z/$1/i;
    return $Name;
}

sub getPrefix($)
{
    my $Str = $_[0];
    if($Str=~/\A(Get|get|Set|set)([A-Z]|_)/)
    {# GetError
        return "";
    }
    if($Str=~/\A([_]*[A-Z][a-z]+)[A-Z]/)
    {# XmuValidArea: Xmu
        return $1;
    }
    elsif($Str=~/\A([_]*[a-z]+)[A-Z]/)
    {# snfReadFont: snf
        return $1;
    }
    elsif($Str=~/\A([_]*[A-Z]{2,})[A-Z][a-z]+([A-Z][a-z]+|\Z)/)
    {# XRRTimes: XRR
        return $1;
    }
    elsif($Str=~/\A([_]*[a-z0-9]{2,}_)[a-z]+/i)
    {# alarm_event_add: alarm_
        return $1;
    }
    elsif($Str=~/\A(([a-z])\2{1,})/i)
    {# ffopen
        return $1;
    }
    else {
        return "";
    }
}

sub problem_title($)
{
    if($_[0]==1)  {
        return "1 problem";
    }
    else  {
        return $_[0]." problems";
    }
}

sub warning_title($)
{
    if($_[0]==1)  {
        return "1 warning";
    }
    else  {
        return $_[0]." warnings";
    }
}

sub readLineNum($$)
{
    my ($Path, $Num) = @_;
    return "" if(not $Path or not -f $Path);
    open (FILE, $Path);
    foreach (1 ... $Num) {
        <FILE>;
    }
    my $Line = <FILE>;
    close(FILE);
    return $Line;
}

sub readAttributes($)
{
    my $Path = $_[0];
    return () if(not $Path or not -f $Path);
    my %Attributes = ();
    if(readLineNum($Path, 0)=~/<!--\s+(.+)\s+-->/) {
        foreach my $AttrVal (split(/;/, $1)) {
            if($AttrVal=~/(.+):(.+)/)
            {
                my ($Name, $Value) = ($1, $2);
                $Attributes{$Name} = $Value;
            }
        }
    }
    return \%Attributes;
}

sub cmpSystems()
{ # -cmp-systems option handler
  # should be used with -d1 and -d2 options
    if(not $Descriptor{1}{"Path"}) {
        exitStatus("Error", "the option -d1 should be specified");
    }
    elsif(not -d $Descriptor{1}{"Path"}) {
        exitStatus("Access_Error", "can't access directory \'".$Descriptor{1}{"Path"}."\'");
    }
    elsif(not -d $Descriptor{1}{"Path"}."/abi_dumps") {
        exitStatus("Access_Error", "can't access directory \'".$Descriptor{1}{"Path"}."/abi_dumps\'");
    }
    if(not $Descriptor{2}{"Path"}) {
        exitStatus("Error", "the option -d2 should be specified");
    }
    elsif(not -d $Descriptor{2}{"Path"}) {
        exitStatus("Access_Error", "can't access directory \'".$Descriptor{2}{"Path"}."\'");
    }
    elsif(not -d $Descriptor{2}{"Path"}."/abi_dumps") {
        exitStatus("Access_Error", "can't access directory \'".$Descriptor{2}{"Path"}."/abi_dumps\'");
    }
    # sys_dumps/<System>/<Arch>/...
    my $SystemName1 = get_filename(get_dirname($Descriptor{1}{"Path"}));
    my $SystemName2 = get_filename(get_dirname($Descriptor{2}{"Path"}));
    # sys_dumps/<System>/<Arch>/...
    my $ArchName = get_filename($Descriptor{1}{"Path"});
    if($ArchName ne get_filename($Descriptor{2}{"Path"})) {
        exitStatus("Error", "can't compare systems of different CPU architecture");
    }
    my $SYS_REPORT_PATH = "sys_compat_reports/".$SystemName1."_to_".$SystemName2."/$ArchName";
    rmtree($SYS_REPORT_PATH);
    my (%LibSoname1, %LibSoname2) = ();
    foreach (split(/\n/, readFile($Descriptor{1}{"Path"}."/sonames.txt"))) {
        if(my ($LFName, $Soname) = split(/;/, $_)) {
            $LibSoname1{$LFName} = $Soname;
        }
    }
    foreach (split(/\n/, readFile($Descriptor{2}{"Path"}."/sonames.txt"))) {
        if(my ($LFName, $Soname) = split(/;/, $_)) {
            $LibSoname2{$LFName} = $Soname;
        }
    }
    my @Dumps1 = cmd_find($Descriptor{1}{"Path"}."/abi_dumps","f","*.tar.gz",1);
    my @Dumps2 = cmd_find($Descriptor{2}{"Path"}."/abi_dumps","f","*.tar.gz",1);
    my (%LibVers1, %LibVers2) = ();
    foreach my $DPath (@Dumps1) {
        if(get_filename($DPath)=~/\A(.+)\Q.abi.tar.gz\E\Z/) {
            $LibVers1{$LibSoname1{$1}}{parse_libname($1, "version")} = $DPath;
        }
    }
    foreach my $DPath (@Dumps2) {
        if(get_filename($DPath)=~/\A(.+)\Q.abi.tar.gz\E\Z/) {
            $LibVers2{$LibSoname2{$1}}{parse_libname($1, "version")} = $DPath;
        }
    }
    my (%ChangedSoname_Safe, %ChangedSoname_Unsafe) = ();
    foreach my $LName (sort keys(%LibSoname2))
    {# libcurl.so.3 -> libcurl.so.4 (search for SONAME by the file name)
     # OS #1 => OS #2
        if(defined $LibVers2{$LName})
        { # already registered
            next;
        }
        my $Soname = $LibSoname2{$LName};
        if(defined $LibVers2{$Soname}
        and defined $LibVers1{$LName}) {
            $LibVers2{$LName} = $LibVers2{$Soname};
            $ChangedSoname_Safe{$Soname}=$LName;
        }
    }
    foreach my $LName (sort keys(%LibSoname1))
    {# libcurl.so.3 -> libcurl.so.4 (search for SONAME by the file name)
     # OS #1 <= OS #2
        if(defined $LibVers1{$LName})
        { # already registered
            next;
        }
        my $Soname = $LibSoname1{$LName};
        if(defined $LibVers1{$Soname}
        and defined $LibVers2{$LName}) {
            $LibVers1{$LName} = $LibVers1{$Soname};
        }
    }
    my (%TestResults, %Added, %Removed) = ();
    foreach my $LName (sort {lc($a) cmp lc($b)} keys(%LibVers1))
    {
        my @Versions1 = keys(%{$LibVers1{$LName}});
        if($#Versions1>=1)
        {# should be only one version
            next;
        }
        my @Versions2 = keys(%{$LibVers2{$LName}});
        if(not @Versions2)
        {# removed library
            if(not $LibSoname2{$LName}) {
                $Removed{$LName}=$Versions1[0];
            }
        }
        else
        {# the same or changed version of the library
            my ($LV1, $LV2) = ($Versions1[0], $Versions2[0]);
            my ($FV1, $FV2) = ($LV1."-".$SystemName1, $LV2."-".$SystemName2);
            my $DPath1 = $LibVers1{$LName}{$LV1};
            my $DPath2 = $LibVers2{$LName}{$LV2};
            my $ACC_compare = "perl $0 -l $LName -d1 \"$DPath1\" -d2 \"$DPath2\"";
            my $LReportPath = "compat_reports/$LName/abi_compat_report.html";
            my $LReportPath_Full = $SYS_REPORT_PATH."/".$LReportPath;
            $ACC_compare .= " -report-path \"$LReportPath_Full\"";
            print "Checking $LName: ";
            system($ACC_compare." 1>$TMP_DIR/null 2>$TMP_DIR/$LName.stderr");
            if(-s "$TMP_DIR/$LName.stderr")
            {
                my $ErrorLog = readFile("$TMP_DIR/$LName.stderr");
                chomp($ErrorLog);
                print "Failed ($ErrorLog)\n";
            }
            else
            {
                print "Ok\n";
                $TestResults{$LName} = readAttributes($LReportPath_Full);
                $TestResults{$LName}{"v1"} = $LV1;
                $TestResults{$LName}{"v2"} = $LV2;
                $TestResults{$LName}{"path"} = $LReportPath;
            }
        }
    }
    foreach my $LName (sort {lc($a) cmp lc($b)} keys(%LibVers2))
    {# added libs
        my @Versions2 = keys(%{$LibVers2{$LName}});
        if($#Versions2>=1)
        {# should be only one version
            next;
        }
        if($ChangedSoname_Safe{$LName})
        {# changed soname but added the symbolic link for old-version library
            next;
        }
        if(not keys(%{$LibVers1{$LName}})
        and not $LibSoname1{$LName}) {
            $Added{$LName}=$Versions2[0];
        }
    }
    my $SYS_REPORT_STYLES = "
    <style type=\"text/css\">
    body {
        font-family:Arial;
        font-size:14px;
        background:#ffffff;
        color:Black;
        padding-left:15px;
    }
    hr {
        color:Black;
        background-color:Black;
        height:1px;
        border:0;
    }
    h1.title1 {
        margin-bottom:3px;
        padding-bottom:3px;
        font-size:26px;
        white-space:nowrap;
    }
    table.wikitable {
        border-collapse:collapse;
        margin:1em 1em 1em 0;
        background:#f9f9f9;
        border:1px #aaaaaa solid;
    }
    table.wikitable td {
        border:1px #aaaaaa solid;
        text-align:center;
        padding:0.2em;
        padding-left:5px;
        padding-right:5px;
        white-space:nowrap;
    }
    table.wikitable th {
        background:#f2f2f2;
        border:1px #aaaaaa solid;
        padding:0.2em;
        padding-left:0.4em;
        padding-right:0.4em;
        text-align:center;
        white-space:nowrap;
    }
    table.wikitable td.left {
        text-align:left;
        background-color:#f2f2f2;
    }
    td.passed {
        background-color:#CCFFCC;
    }
    td.warning {
        background-color:#F4F4AF;
    }
    td.failed {
        background-color:#FFC3CE;
    }
    td.new {
        background-color:#C6DEFF;
    }
    a.default {
        color:#336699;
    }
    a.disabled {
        text-decoration:none;
        color:Black;
    }
    th.severity {
        width:85px;
        color:Black;
    }
    sup {
        font-size:10px;
    }
    </style>";
    my $SYS_REPORT = "<h1 class='title1'>Binary compatibility between <span style='color:Blue;'>$SystemName1</span> and <span style='color:Blue;'>$SystemName2</span> on <span style='color:Blue;'>".showArch($ArchName)."</span></h1>\n";
    # legend
    $SYS_REPORT .= "<table>
    <tr><td class='new' width='80px'>Added</td><td class='passed' width='80px'>Compatible</td></tr>
    <tr><td class='warning'>Warning</td><td class='failed'>Incompatible</td></tr>
    </table>";
    $SYS_REPORT .= "<table class='wikitable'>
    <tr><th rowspan='2'>SONAME<sup>".(keys(%TestResults) + keys(%Added) + keys(%Removed))."</sup></th>
    <th colspan='2'>VERSION</th>
    <th rowspan='2'>Test<br/>Result</th>
    <th rowspan='2'>Added<br/>Symbols</th>
    <th rowspan='2'>Removed<br/>Symbols</th>
    <th colspan='3' style='white-space:nowrap;'>Backward Binary Compatibility Problems</th></tr>
    <tr><th>$SystemName1</th><th>$SystemName2</th>
    <th class='severity'>High</th><th class='severity'>Medium</th><th class='severity'>Low</th></tr>\n";
    foreach my $LName (sort {lc($a) cmp lc($b)} (keys(%TestResults), keys(%Added), keys(%Removed)))
    {
        my $DefaultLink = "<a href='".$TestResults{$LName}{"path"}."' class='disabled'>0</a>";
        $SYS_REPORT .= "<tr>\n<td class='left'>$LName</td>\n";
        if(defined $Removed{$LName}) {
            $SYS_REPORT .= "<td class='failed'>".$Removed{$LName}."</td>\n";
        }
        elsif(defined $Added{$LName}) {
            $SYS_REPORT .= "<td class='new'>added</td>\n";
        }
        else {
            $SYS_REPORT .= "<td>".$TestResults{$LName}{"v1"}."</td>\n";
        }
        if(defined $Added{$LName})
        {
            $SYS_REPORT .= "<td class='new'>".$Added{$LName}."</td>\n";
            foreach (1 .. 6) {
                #$SYS_REPORT .= "<td></td>\n";
            }
            $SYS_REPORT .= "<td colspan='6'>n/a</td>\n";
            $SYS_REPORT .= "</tr>\n";
            next;
        }
        elsif(defined $Removed{$LName})
        {
            $SYS_REPORT .= "<td class='failed'>removed</td>\n";
            foreach (1 .. 6) {
                #$SYS_REPORT .= "<td></td>\n";
            }
            $SYS_REPORT .= "<td colspan='6'>n/a</td>\n";
            $SYS_REPORT .= "</tr>\n";
            next;
        }
        else {
            $SYS_REPORT .= "<td>".$TestResults{$LName}{"v2"}."</td>\n";
        }
        if($TestResults{$LName}{"verdict"} eq "compatible") {
            $SYS_REPORT .= "<td class='passed'>compatible</td>\n";
        }
        else {
            $SYS_REPORT .= "<td class='failed'>incompatible</td>\n";
        }
        my $AddedSym="";
        if(my $Count = $TestResults{$LName}{"added"}) {
            $AddedSym="<a href='".$TestResults{$LName}{"path"}."#Added'>$Count new</a>";
        }
        if($AddedSym) {
            $SYS_REPORT.="<td class='new'>$AddedSym</td>";
        }
        else {
            $SYS_REPORT.="<td class='passed'>$DefaultLink</td>";
        }
        my $RemovedSym="";
        if(my $Count = $TestResults{$LName}{"removed"}) {
            $RemovedSym="<a href='".$TestResults{$LName}{"path"}."#Removed'>$Count removed</a>";
        }
        if($RemovedSym) {
            $SYS_REPORT.="<td class='failed'>$RemovedSym</td>";
        }
        else {
            $SYS_REPORT.="<td class='passed'>$DefaultLink</td>";
        }
        my $High="";
        if(my $Count = $TestResults{$LName}{"type_problems_high"}+$TestResults{$LName}{"interface_problems_high"}) {
            $High="<a href='".$TestResults{$LName}{"path"}."#High_Risk_Problems'>".problem_title($Count)."</a>";
        }
        if($High) {
            $SYS_REPORT.="<td class='failed'>$High</td>";
        }
        else {
            $SYS_REPORT.="<td class='passed'>$DefaultLink</td>";
        }
        my $Medium="";
        if(my $Count = $TestResults{$LName}{"type_problems_medium"}+$TestResults{$LName}{"interface_problems_medium"}) {
            $Medium="<a href='".$TestResults{$LName}{"path"}."#Medium_Risk_Problems'>".problem_title($Count)."</a>";
        }
        if($Medium) {
            $SYS_REPORT.="<td class='failed'>$Medium</td>";
        }
        else {
            $SYS_REPORT.="<td class='passed'>$DefaultLink</td>";
        }
        my $Low="";
        if(my $Count = $TestResults{$LName}{"type_problems_low"}+$TestResults{$LName}{"interface_problems_low"}+$TestResults{$LName}{"changed_constants"}) {
            $Low="<a href='".$TestResults{$LName}{"path"}."#Low_Risk_Problems'>".warning_title($Count)."</a>";
        }
        if($Low) {
            $SYS_REPORT.="<td class='warning'>$Low</td>";
        }
        else {
            $SYS_REPORT.="<td class='passed'>$DefaultLink</td>";
        }
        $SYS_REPORT .= "</tr>\n";
    }
    # bottom header
    $SYS_REPORT .= "<tr><th rowspan='2'>SONAME</th>
    <th>$SystemName1</th><th>$SystemName2</th>
    <th rowspan='2'>Test<br/>Result</th>
    <th rowspan='2'>Added<br/>Symbols</th>
    <th rowspan='2'>Removed<br/>Symbols</th>
    <th class='severity'>High</th><th class='severity'>Medium</th><th class='severity'>Low</th></tr>
    <tr><th colspan='2'>VERSION</th>
    <th colspan='3' style='white-space:nowrap;'>Backward Binary Compatibility Problems</th></tr>\n";
    $SYS_REPORT .= "</table>";
    my $Title = "$SystemName1 to $SystemName2 binary compatibility report";
    my $Keywords = "compatibility, $SystemName1, $SystemName2, API";
    my $Description = "Binary compatibility between $SystemName1 and $SystemName2 on ".showArch($ArchName);
    writeFile($SYS_REPORT_PATH."/abi_compat_report.html", composeHTML_Head($Title, $Keywords, $Description, $SYS_REPORT_STYLES)."\n<body>
    <div>$SYS_REPORT</div>
    <br/><br/><br/><hr/>
    ".getReportFooter($SystemName2)."
    <div style='height:999px;'></div>\n</body></html>");
    print "see detailed report in the file:\n  $SYS_REPORT_PATH/abi_compat_report.html\n";
}

sub dumpSystem()
{ # -dump-system option handler
  # should be used with -sysroot and -cross-gcc options
  # tested on:
  # - Maemo SDK (5.0, 4.1.2, 4.1, 4.0, 3.1)
  # - Alpine SDK (2.2, 2.0, 1.10)
    my $SYS_DUMP_PATH = "sys_dumps/$DumpSystem/".getArch(1);
    rmtree($SYS_DUMP_PATH);
    my $SearchPrefix = $SystemRoot;
    my ($SearchLibs, $SearchHeaders) = ($SearchPrefix."/usr/lib", $SearchPrefix."/usr/include");
    if(not -e $SearchLibs) {
        exitStatus("Access_Error", "can't access \'$SearchLibs\'");
    }
    if(not -e $SearchHeaders) {
        exitStatus("Access_Error", "can't access \'$SearchHeaders\'");
    }
    my @SysHeaders = cmd_find($SearchHeaders,"f","","");# include subdirs
    my (%SysLib_Symbols, %Symbol_SysHeaders, %Plugins, %SysLib_SysHeaders) = ();
    my (%Skipped, %Failed, %Success) = ();
    my (%SysHeaderDir_SysLibs, %SysHeaderDir_SysHeaders) = ();
    my (%LibPrefixes, %SymbolCounter) = ();
    my $Total = 0;
    # This is the list of symbols that should NOT be used to match headers with libraries
    # - _init, _fini
    # - symbols with too common names
    # - other symbols to fix the algorithm errors
    my %IgnoreSymbols = map {$_=>1} ("_init", "_fini", "key_free", "create_module", "program_name",
    "get_version", "open64", "__open64", "__h_errno_location", "GetProperty", "SetProperty", "stat64",
    "DeleteProperty", "ReadBuffer", "WriteBuffer", "MD5_CTX", "get_volume", "client_name", "HMAC",
    "asn1_free", "MD5Final", "MD5Init", "MD5Update", "next_token", "get_interface", "gconf_client",
    "multiline_error", "multiline_warning", "MD5", "pthread_create", "disassembler");
    my %PluginSymbols = map {$_=>1} ("gst_plugin_desc", "fs_init_plugin", "null_driver_open",
    "gstreamer_driver_open", "pulse_driver_open", "gconf_backend_get_vtable");
    my %SysLibs = ();
    foreach my $LPath (find_libs($SearchLibs,"",1)) # ignore subdirs
    {
        my $LSName = parse_libname(get_filename($LPath), "short");
        if(-l $LPath)
        { # symlinks
            if(my $Path = resolve_symlink($LPath)) {
                $SysLibs{$Path} = 1;
            }
        }
        elsif(-f $LPath)
        {
            if($LSName=~/\A(libc|libpthread)\Z/
            and cmd_file($LPath)=~/ASCII/)
            {# GNU ld scripts (libc.so, libpthread.so)
                my @Candidates = cmd_find($SystemRoot."/lib","",$LSName.".".$LIB_EXT."*","1");
                if(@Candidates)
                {
                    my $Candidate = $Candidates[0];
                    if(-l $Candidate
                    and my $Path = resolve_symlink($Candidate)) {
                        $Candidate = $Path;
                    }
                    $SysLibs{$Candidate} = 1;
                }
            }
            else {
                $SysLibs{$LPath} = 1;
            }
        }
    }
    print "Reading symbols from libraries ...\n";
    my $ObjdumpCmd = get_CmdPath("objdump");
    if(not $ObjdumpCmd) {
        exitStatus("Not_Found", "can't find \"objdump\"");
    }
    foreach my $LPath (sort keys(%SysLibs))
    {
        my $LRelPath = cut_path_prefix($LPath, $SearchPrefix);
        my $LName = get_filename($LPath);
        if($LName!~/\Alib/
        or $LRelPath=~/\/(modules|loaders|filesystems|engines|gconv|extensions)\//
        or $LName=~/_module_/)
        { # modules and plugins
          # libasound_module_pcm_bluetooth.so
          # lib/gio/modules/libgiogconf.so
            $Plugins{$LPath} = 1;
            next;
        }
        getSymbols_Lib(1, $LPath, 0, ());
        if(`$ObjdumpCmd -T $LPath 2>&1 | grep __gxx_personality`)
        { # C++
            translateSymbols(keys(%{$Library_Symbol{1}{$LName}}), 1);
            foreach my $Symbol (keys(%{$Library_Symbol{1}{$LName}}))
            {
                if(skipGlobalData($Symbol)) {
                    next;
                }
                $Symbol=~s/[\@\$]+(.*)\Z//g;
                foreach my $Symbol (split(/\W+/, $tr_name{$Symbol}))
                {
                    $SysLib_Symbols{$LPath}{$Symbol}=1;
                    if(my $Prefix = getPrefix($Symbol)) {
                        $LibPrefixes{$Prefix}{$LName}+=1;
                    }
                    $SymbolCounter{$Symbol}{$LName}=1;
                }
            }
        }
        else
        { # C
            foreach my $Symbol (keys(%{$Library_Symbol{1}{$LName}}))
            {
                $Symbol=~s/[\@\$]+(.*)\Z//g;
                $SysLib_Symbols{$LPath}{$Symbol}=1;
                if(my $Prefix = getPrefix($Symbol)) {
                    $LibPrefixes{$Prefix}{$LName}+=1;
                }
                $SymbolCounter{$Symbol}{$LName}=1;
            }
        }
        foreach my $PSymbol (keys(%PluginSymbols))
        {
            if(defined $SysLib_Symbols{$LPath}{$PSymbol})
            {# gstreamer plugins
                $Plugins{$LPath} = 1;
                last;
            }
        }
    }
    my %DupLibs = ();
    foreach my $LPath1 (keys(%SysLib_Symbols))
    { # match duplicated libs
      # libmenu contains all symbols from libmenuw
        foreach my $LPath2 (keys(%SysLib_Symbols))
        {
            next if($LPath1 eq $LPath2);
            my $Duplicate=1;
            foreach (keys(%{$SysLib_Symbols{$LPath1}}))
            {
                if(not $SysLib_Symbols{$LPath2}{$_}) {
                    $Duplicate=0;
                    last;
                }
            }
            if($Duplicate) {
                $DupLibs{$LPath1}{$LPath2}=1;
            }
        }
    }
    foreach my $Prefix (keys(%LibPrefixes))
    {
        my @Libs = sort {$LibPrefixes{$Prefix}{$b}<=>$LibPrefixes{$Prefix}{$a}}
        keys(%{$LibPrefixes{$Prefix}});
        $LibPrefixes{$Prefix}=$Libs[0];
    }
    print "Reading symbols from headers ...\n";
    foreach my $HPath (@SysHeaders)
    {
        my $HRelPath = cut_path_prefix($HPath, $SearchPrefix);
        my ($HDir, $HName) = separate_path($HRelPath);
        if($HName=~/\.\w+\Z/
        and not is_header_file($HName))
        { # have a wrong extension: .gch, .in
            next;
        }
        if($HName=~/~\Z/)
        { # reserved copy
            next;
        }
        if($HRelPath=~/\/_gen/)
        { # telepathy-1.0/telepathy-glib/_gen
          # telepathy-1.0/libtelepathy/_gen-tp-constants-deprecated.h
            next;
        }
        if($HRelPath=~/include\/linux\//)
        {# kernel-space headers
            next;
        }
        if($HRelPath=~/include\/asm\//)
        {# asm headers
            next;
        }
        if($HRelPath=~/\/microb-engine\//)
        {# MicroB engine (Maemo 4)
            next;
        }
        my $Content = readFile($HPath);
        $Content=~s/\/\*(.|\n)+?\*\///g;
        $Content=~s/\/\/.+?\n/\n/g;# remove comments
        foreach my $Symbol (split(/\W+/, $Content)) {
            $Symbol_SysHeaders{$Symbol}{$HRelPath} = 1;
        }
        $SysHeaderDir_SysHeaders{$HDir}{$HName} = 1;
    }
    my %SkipLibHeaders = (
    # do NOT search for all symbols in some directories
        # Maemo 5.0
        "libgettextlib" => ["libxml2"],
        "libgvfsdaemon" => ["/dbus/"],
        "libts" => ["/linux/netfilter/"],
        "libxkbfile" => ["/xorg/", "XKBsrv.h"],
        "libXfont" => ["/xorg/"],
        # Maemo 4.1.2
        "libgdk-x" => ["gdkalias.h"],
        # Alpine
        "libstdc++" => ["/java/", "/javax/", "/gcj/", "/gnu/", "/sun/",
                        "/org/", "/parallel/", "/bits/", "/tr1_impl/",
                        "atomic", "stdatomic.h", "c++0xwarning.h"]
    );
    my %SearchLibHeaders = (
    # search for all symbols only in the directories
        "libstdc++" => ["/c++/"]
    );
    my %NonSelfCompiled = (
    # non self-compiled headers
        # Maemo 5.0
        "libclinkc" => ["csubscription.h"],
        "libgstbase" => ["gstbytereader.h"],
        "libjpeg" => ["jmorecfg.h", "jpegint.h"],
        "libasound" => ["mixer_abst.h"],
        "libbluetooth" => ["sdp_lib.h", "hci_lib.h"],
        "libnice" => ["win32_common.h"],
        "libdbus" => ["dbus-glib-error-enum.h"],
        "libedataserver" => ["e-component-listener.h", "e-sexp.h"],
        "libfreetype" => ["ftmac.h", "ftmodule.h", "fterrdef.h"],
        "libc" => ["regexp.h", "/bits/", "/scsi/",
                   "/arpa/", "/rpcsvc/", "net/if_packet.h",
                   "net/if.h", "asoundlib.h", "elf.h"],
        "libpthread" => ["/bits/"],
        "libgnomevfs" => ["gnome-vfs-cancellable-ops.h",
                          "gnome-vfs-module-shared.h"],
        "libgstreamer" => ["gstcheck.h", "gstbytereader.h",
                           "gstbufferstraw.h", "gstbytewriter.h",
                           "gstconsistencychecker.h"],
        "libgtk-x" => ["gtktextdisplay.h", "gtkitemfactory.h",
                       "gtkprivate.h"],
        "libgdk-x" => ["gdkalias.h", "gdkprivate.h"],
        "libglib" => ["gi18n-lib.h"],
        "libmcclient" => ["/libmcclient/_gen/"],
        "libmissioncontrol-server" => ["/mission-control/_gen/"],
        "libpulsecore" => ["poll.h", "/pulsecore/modules/"],
        "libpulsecommon" => ["poll.h", "/pulsecore/modules/"],
        "libtelepathy-glib" => ["telepathy-glib/_gen"],
        "libtelepathy" => ["telepathy/_gen", "*-gen.h"],
        "libsofia-sip-ua" => ["string0.h", "sip_parser.h",
                              "sip_util.h", "sl_utils.h",
                              "tstdef.h", "sres_cache.h"],
        "libsofia-tools" => ["sres_record.h"],
        "libstdc++" => ["/backward/", "/bits/", "/debug/",
                        "/tr1/", "/ext/", "c++config.h"],
        "libX" => ["/extensions/", "CoreP.h", "ResourceI.h", "/xorg/"],
        "libXaw" => ["[a-z]P", "VendorEP.h", "SmeBSBP.h"],
        "libXt" => ["[a-z]P", "[a-z]I"],
        "libjinglebase" => ["stl_decl.h"],
        "libnsl" => ["nislib.h", "nfs_prot.h", "yp_prot.h"],
        # Maemo 3.1
        "libhildonwidgets" => ["hildon-time-editor.h"],
        "libossohelp" => ["osso-helplib-private.h"],
        "libebook" => ["e-data-book-bindings.h",
                       "e-data-book-factory-bindings.h",
                       "e-data-book-view-bindings.h"],
        "libdbus" => ["dbus-glib-bindings.h", "dbus-glib-error-enum.h"],
        "libdbus-glib" => ["dbus-glib-bindings.h"]
    );
    if($DumpSystem=~/maemo/i) {# GL/gl.h: No such file
        $NonSelfCompiled{"libSDL"}=["SDL_opengl.h"];
    }
    print "Matching symbols ...\n";
    foreach my $LPath (sort keys(%SysLibs))
    {# matching
        my $LName = get_filename($LPath);
        my $LRelPath = cut_path_prefix($LPath, $SearchPrefix);
        my $LSName = parse_libname($LName, "short");
        my $SName = parse_libname($LName, "shortest");
        if($Plugins{$LPath})
        {# modules and plugins
            $Skipped{$LRelPath}=1;
            next;
        }
        if($LRelPath=~/\/debug\//)
        {# debug libs
            $Skipped{$LRelPath}=1;
            next;
        }
        if($LSName eq "libapt-pkg-libc")
        {# libapt-pkg-libc6.5-6.so.4.6.0
            $Skipped{$LRelPath}=1;
            next;
        }
        $Total+=1;
        $SysLib_SysHeaders{$LRelPath} = ();
        my %SymbolDirs = ();
        foreach my $Symbol (sort {length($b) cmp length($a)}
        sort keys(%{$SysLib_Symbols{$LPath}}))
        {
            if($IgnoreSymbols{$Symbol}
            or $PluginSymbols{$Symbol}) {
                next;
            }
            if(not $DupLibs{$LPath}
            and keys(%{$SymbolCounter{$Symbol}})>=2
            and my $Prefix = getPrefix($Symbol)) {
                if($LibPrefixes{$Prefix}
                and $LibPrefixes{$Prefix} ne $LName) {
                    next;
                }
            }
            if(length($Symbol)<=2) {
                next;
            }
            if($Symbol!~/[A-Z_0-9]/
            and length($Symbol)<10) {
                # FIXME: improve this filter
                # signalfd (libc)
                next;
            }
            if($Symbol=~/\A(_M_|_Rb_|_S_)/)
            {# _M_insert, _Rb_tree, _S_destroy_c_locale
                next;
            }
            if($Symbol=~/\A[A-Z][a-z]+\Z/)
            {# Clone, Initialize, Skip, Unlock, Terminate, Chunk
                next;
            }
            if($Symbol=~/\A[A-Z][A-Z]\Z/)
            {#  BC, PC, UP, SP
                next;
            }
            if($Symbol=~/_t\Z/)
            {# pthread_mutex_t, wchar_t
                next;
            }
            my @SymHeaders = sort {lc($a) cmp lc($b)} keys(%{$Symbol_SysHeaders{$Symbol}});# sort by name
            @SymHeaders = sort {length(get_dirname($a))<=>length(get_dirname($b))} @SymHeaders;# sort by length
            if(length($SName)>=3) {# sort candidate headers by name
                @SymHeaders = sort {$b=~/\Q$SName\E/i<=>$a=~/\Q$SName\E/i} @SymHeaders;
            }
            else {# libz, libX11
                @SymHeaders = sort {$b=~/lib\Q$SName\E/i<=>$a=~/lib\Q$SName\E/i} @SymHeaders;
                @SymHeaders = sort {$b=~/\Q$SName\Elib/i<=>$a=~/\Q$SName\Elib/i} @SymHeaders;
            }
            @SymHeaders = sort {$b=~/\Q$LSName\E/i<=>$a=~/\Q$LSName\E/i} @SymHeaders;
            @SymHeaders = sort {$SymbolDirs{get_dirname($b)}<=>$SymbolDirs{get_dirname($a)}} @SymHeaders;
            foreach my $HRelPath (@SymHeaders)
            {
                if(my $Search = $SearchLibHeaders{$LSName})
                {# search for headers only in the directory
                    if(not grep {$HRelPath=~/\Q$_\E/} @{$Search}) {
                        next;
                    }
                }
                if(my $Skip = $SkipLibHeaders{$LSName})
                {# do NOT search for headers in the directory
                    if(grep {$HRelPath=~/\Q$_\E/} @{$Skip}) {
                        next;
                    }
                }
                if(my $Skip = $NonSelfCompiled{$LSName})
                {# do NOT search for headers in the directory
                    if(grep {$HRelPath=~/\Q$_\E/} @{$Skip}) {
                        $SymbolDirs{get_dirname($HRelPath)}+=1;
                        last;
                    }
                }
                if($LSName ne "libstdc++" and $HRelPath=~/\/c\+\+\//)
                {# do NOT search in the include/c++ directory
                    next;
                }
                $SysLib_SysHeaders{$LRelPath}{$HRelPath}=$Symbol;
                $SysHeaderDir_SysLibs{get_dirname($HRelPath)}{$LRelPath}=1;
                $SysHeaderDir_SysLibs{get_dirname(get_dirname($HRelPath))}{$LRelPath}=1;
                $SymbolDirs{get_dirname($HRelPath)}+=1;
                last;# select one header for one symbol
            }
        }
        if(keys(%{$SysLib_SysHeaders{$LRelPath}})) {
            $Success{$LRelPath}=1;
        }
        else {
            $Failed{$LRelPath}=1;
        }
    }
    print "Indexing sonames ...\n";
    my %LibSoname = ();
    #foreach my $DevelPath (sort keys(%DefaultLibPaths))
    foreach my $DevelPath (sort keys(%{$SystemPaths{"lib"}}))
    {# register all libraries from default paths
        foreach my $LPath (find_libs($DevelPath,"",1))
        {
            my $LName = get_filename($LPath);
            if($LName!~/\Alib/) {
                next;
            }
            if(my $Soname = get_soname($LPath)) {
                if(not defined $LibSoname{$LName}) {
                    $LibSoname{$LName}=$Soname;
                }
            }
        }
    }
    my $SONAMES = "";
    foreach (sort {lc($a) cmp lc($b)} keys(%LibSoname)) {
        $SONAMES .= $_.";".$LibSoname{$_}."\n";
    }
    writeFile($SYS_DUMP_PATH."/sonames.txt", $SONAMES);
    writeFile($SYS_DUMP_PATH."/match.txt", Dumper(\%SysLib_SysHeaders));
    writeFile($SYS_DUMP_PATH."/symbols.txt", Dumper(\%SysLib_Symbols));
    writeFile($SYS_DUMP_PATH."/skipped.txt", join("\n", sort keys(%Skipped)));
    writeFile($SYS_DUMP_PATH."/failed.txt", join("\n", sort keys(%Failed)));
    my %IncPreamble = (
    # include preamble for header files
        # Maemo 5.0
        "libjpeg" => ["stdio.h"],
        "libebook" => ["glib/gtypes.h"],# obsolete
        "libfreetype" => ["ft2build.h"],
        "libhildon" => ["gtkenums.h"],
        "libhildon-im-ui" => ["hildon-im-ui.h"],
        "libpulsecore" => ["pulseaudio.h", "pulsecore-config.h"],
        "libpulsecommon" => ["pulseaudio.h", "pulsecore-config.h"],
        "libXRes" => ["Xmd.h"],
        "libXfont" => ["ft2build.h", "freetype.h", "fntfilst.h",
                       "fontenc.h", "bitmap.h"],
        "libxkbfile" => ["Xlib.h", "XKBstr.h", "stdio.h"],
        "libXaw" => ["X11/Intrinsic.h"],
        "libXext" => ["X11/Xproto.h"],
        "libgtkfilesystemmemorypublic" => ["gtk/gtktreemodel.h"],
        # Maemo 4.0
        "libossoemailinterface" => ["libosso.h"],
        # Maemo 3.1
        "libshadowappd" => ["glib.h", "gconf-client.h",
                            "libosso.h", "unistd.h"]
    );
    my %GccOpts = (
    # options for compiling headers
        # Maemo 5.0
        "liboil" => ["-DOIL_ENABLE_UNSTABLE_API"],
        "libjinglebase" => ["-DPOSIX"],# for _stricmp
        "libjinglep2pbase" => ["-DPOSIX"],
        "libjinglep2pclient" => ["-DPOSIX"],
        # Maemo 3.1
        "libtelepathy" => ["-DDBUS_GLIB_CLIENT_WRAPPERS_org_freedesktop_Telepathy_Channel",
            "-DDBUS_GLIB_CLIENT_WRAPPERS_org_freedesktop_Telepathy_Connection",
            "-DDBUS_GLIB_CLIENT_WRAPPERS_org_freedesktop_Telepathy_Connection_Interface_Forwarding",
            "-DDBUS_GLIB_CLIENT_WRAPPERS_org_freedesktop_Telepathy_ConnectionManager",
            "-DDBUS_GLIB_CLIENT_WRAPPERS_org_freedesktop_Telepathy_Properties"],
        "libdbus" => ["-DDBUS_API_SUBJECT_TO_CHANGE"],
        "libdbus-glib" => ["-DDBUS_API_SUBJECT_TO_CHANGE"],
        "libosso-ic" => ["-DDBUS_API_SUBJECT_TO_CHANGE"],
    );
    if(check_gcc_version($GCC_PATH, "4.4")) {
        $GccOpts{"libstdc++"} = ["-std=c++0x"];
    }
    print "Generating XML descriptors ...\n";
    foreach my $LRelPath (keys(%SysLib_SysHeaders))
    {
        if(my @LibHeaders = keys(%{$SysLib_SysHeaders{$LRelPath}}))
        {
            my (%DirsHeaders, %Includes) = ();
            foreach my $HRelPath (@LibHeaders) {
                $DirsHeaders{get_dirname($HRelPath)}{$HRelPath}=1;
            }
            foreach my $Dir (keys(%DirsHeaders))
            {
                my $TotalPart = 0;
                if(my $TotalHeaders = keys(%{$SysHeaderDir_SysHeaders{$Dir}})) {
                    $TotalPart = (keys(%{$DirsHeaders{$Dir}})*100)/$TotalHeaders;
                }
                if(get_depth($Dir)>=1
                and ((keys(%{$SysHeaderDir_SysLibs{$Dir}})==1 and $TotalPart>=10) or $TotalPart>=50))
                {# complete directory
                    $Includes{$Dir} = 1;
                }
                elsif((keys(%{$DirsHeaders{$Dir}})*100)/($#LibHeaders+1)>5)
                {# list of headers
                    @Includes{keys(%{$DirsHeaders{$Dir}})}=values(%{$DirsHeaders{$Dir}});
                }
            }
            my $LName = get_filename($LRelPath);
            my $LSName = parse_libname($LName, "short");
            my $LVersion = parse_libname($LName, "version");
            if($LVersion)
            {# append by system name
                $LVersion .= "-".$DumpSystem;
            }
            else {
                $LVersion = $DumpSystem;
            }
            my @Content = ("<version>\n    $LVersion\n</version>");
            my @IncHeaders = sort {natural_sorting($a, $b)} keys(%Includes);
            sort_by_word(\@IncHeaders, parse_libname($LName, "shortest"));
            if($IncHeaders[0]=~/\A\//) {
                push(@Content, "<headers>\n    ".join("\n    ", @IncHeaders)."\n</headers>");
            }
            else {
                push(@Content, "<headers>\n    {RELPATH}/".join("\n    {RELPATH}/", @IncHeaders)."\n</headers>");
            }
            if($LRelPath=~/\A\//) {
                push(@Content, "<libs>\n    $LRelPath\n</libs>");
            }
            else {
                push(@Content, "<libs>\n    {RELPATH}/$LRelPath\n</libs>");
            }
            my @Skip = ();
            if($SkipLibHeaders{$LSName}) {
                @Skip = (@Skip, @{$SkipLibHeaders{$LSName}});
            }
            if($NonSelfCompiled{$LSName}) {
                @Skip = (@Skip, @{$NonSelfCompiled{$LSName}});
            }
            if(@Skip) {
                push(@Content, "<skip_headers>\n    ".join("\n    ", @Skip)."\n</skip_headers>");
            }
            if(not $IncPreamble{$LSName}) {
                $IncPreamble{$LSName} = [];
            }
            if($LSName=~/\AlibX\w+\Z/)
            { # add Xlib.h for libXt, libXaw, libXext and others
                @{$IncPreamble{$LSName}} = ("Xlib.h", "X11/Intrinsic.h", @{$IncPreamble{$LSName}});
            }
            # add sys/types.h, stddef.h, stdint.h and stdbool.h (_Bool) preamble for all headers
            @{$IncPreamble{$LSName}} = ("sys/types.h", "stddef.h", "stdint.h", "stdbool.h", @{$IncPreamble{$LSName}});
            push(@Content, "<include_preamble>\n    ".join("\n    ", @{$IncPreamble{$LSName}})."\n</include_preamble>");
            if($DumpSystem=~/maemo/i) {
                push(@{$GccOpts{$LSName}}, "-DMAEMO_CHANGES");
            }
            if($GccOpts{$LSName}) {
                push(@Content, "<gcc_options>\n    ".join("\n    ", @{$GccOpts{$LSName}})."\n</gcc_options>");
            }
            writeFile($SYS_DUMP_PATH."/descriptors/$LName.xml", join("\n\n", @Content));
        }
    }
    print "Dumping libraries ABI:\n";
    my %DumpSuccess = ();
    my @Descriptors = cmd_find($SYS_DUMP_PATH."/descriptors","f","*.xml","1");
    foreach my $DPath (sort {lc($a) cmp lc($b)} @Descriptors)
    {
        my $DName = get_filename($DPath);
        my $LName = "";
        if($DName=~/\A(.+).xml\Z/) {
            $LName = $1;
        }
        else {
            next;
        }
        $DPath = cut_path_prefix($DPath, $ORIG_DIR);
        my $ACC_dump = "perl $0 -l ".parse_libname($LName, "name")." -dump \"$DPath\"";
        if($SearchPrefix) {
            $ACC_dump .= " -relpath \"$SearchPrefix\"";
        }
        if($SystemRoot) {
            $ACC_dump .= " -sysroot \"$SystemRoot\"";
        }
        my $DumpPath = "$SYS_DUMP_PATH/abi_dumps/$LName.abi.tar.gz";
        $ACC_dump .= " -dump-path \"$DumpPath\"";
        $ACC_dump .= " -log-path \"$SYS_DUMP_PATH/logs/$LName.txt\"";
        if($CrossGcc) {
            $ACC_dump .= " -cross-gcc \"$CrossGcc\"";
        }
        print "Dumping $LName: ";
        system($ACC_dump." 1>$TMP_DIR/null 2>$TMP_DIR/$LName.stderr");
        appendFile("$SYS_DUMP_PATH/logs/$LName.txt", "The ACC parameters:\n  $ACC_dump\n");
        if(-s "$TMP_DIR/$LName.stderr")
        {
            appendFile("$SYS_DUMP_PATH/logs/$LName.txt", readFile("$TMP_DIR/$LName.stderr"));
            print "Failed ($SYS_DUMP_PATH/logs/$LName.txt)\n";
        }
        elsif(not -f $DumpPath) {
            print "Failed ($SYS_DUMP_PATH/logs/$LName.txt)\n";
        }
        else
        {
            $DumpSuccess{$LName} = 1;
            print "Ok\n";
        }
    }
    print "Total libraries:         ".keys(%SysLibs)."\n";
    print "Skipped libraries:       ".keys(%Skipped)." ($SYS_DUMP_PATH/skipped.txt)\n";
    print "Failed to find headers:  ".keys(%Failed)."  ($SYS_DUMP_PATH/failed.txt)\n";
    print "Created descriptors:     ".keys(%Success)." ($SYS_DUMP_PATH/descriptors/)\n";
    print "Dumped libraries:        ".keys(%DumpSuccess)." ($SYS_DUMP_PATH/abi_dumps/)\n";
    print "The $DumpSystem system ABI has been dumped to:\n  $SYS_DUMP_PATH\n";
}

sub get_soname($)
{
    my $Path = $_[0];
    return if(not $Path or not -e $Path);
    if(defined $Cache{"get_soname"}{$Path}) {
        return $Cache{"get_soname"}{$Path};
    }
    my $ObjdumpCmd = get_CmdPath("objdump");
    if(not $ObjdumpCmd) {
        exitStatus("Not_Found", "can't find \"objdump\"");
    }
    if(my $SonameInfo = `$ObjdumpCmd -x $Path 2>&1 | grep SONAME`)
    {
        if($SonameInfo=~/SONAME\s+([^\s]+)/) {
            return ($Cache{"get_soname"}{$Path} = $1);
        }
    }
    return ($Cache{"get_soname"}{$Path}="");
}

sub checkVersionNum($$)
{
    my ($LibVersion, $Path) = @_;
    if(my $VerNum = $TargetVersion{$LibVersion}) {
        return $VerNum;
    }
    my $UsedAltDescr = 0;
    foreach my $Part (split(/\s*,\s*/, $Path))
    {# try to get version string from file path
        if(parse_libname($Part, "version")
        or is_header($Part, 1, $LibVersion) or -d $Part) {
            $UsedAltDescr = 1;
            if(my $VerNum = readStringVersion($Part)) {
                $TargetVersion{$LibVersion} = $VerNum;
                if($DumpInfo_DPath) {
                    print STDERR "WARNING: setting version number to $VerNum (use -vnum <num> option to change it)\n";
                }
                else {
                    print STDERR "WARNING: setting ".($LibVersion==1?"1st":"2nd")." version number to \"$VerNum\" (use -v$LibVersion <num> option to change it)\n";
                }
                return $TargetVersion{$LibVersion};
            }
        }
    }
    if($UsedAltDescr)
    {
        if($DumpInfo_DPath) {
            exitStatus("Error", "version number is not set (use -vnum <num> option)");
        }
        else {
            exitStatus("Error", ($LibVersion==1?"1st":"2nd")." version number is not set (use -v$LibVersion <num> option)");
        }
    }
}

sub readStringVersion($)
{
    my $Str = $_[0];
    return "" if(not $Str);
    $Str=~s/\Q$TargetLibraryName\E//g;
    if($Str=~/(\/|\\|\w|\A)[\-\_]*(\d+[\d\.\-]+\d+|\d+)/)
    {# .../libssh-0.4.0/...
        return $2;
    }
    elsif(my $V = parse_libname($Str, "version")) {
        return $V;
    }
    return "";
}

sub scan_libs_and_headers($)
{
    my $LibVersion = $_[0];
    if(not $CheckHeadersOnly) {
        getSymbols($LibVersion);
    }
    if(not $CheckObjectsOnly) {
        searchForHeaders($LibVersion);
    }
}

sub get_OSgroup()
{
    if($Config{"osname"}=~/macos|darwin|rhapsody/i) {
        return "macos";
    }
    elsif($Config{"osname"}=~/freebsd|openbsd|netbsd/i) {
        return "bsd";
    }
    elsif($Config{"osname"}=~/haiku|beos/i) {
        return "beos";
    }
    elsif($Config{"osname"}=~/symbian|epoc/i) {
        return "symbian";
    }
    elsif($Config{"osname"}=~/win/i) {
        return "windows";
    }
    else {
        return $Config{"osname"};
    }
}

sub dump_sorting($)
{
    my $hash = $_[0];
    return [] if(not $hash or not keys(%{$hash}));
    if((keys(%{$hash}))[0]=~/\A\d+\Z/) {
        return [sort {int($a) <=> int($b)} keys(%{$hash})];
    }
    else {
        return [sort {$a cmp $b} keys(%{$hash})];
    }
}

sub exitStatus($$)
{
    my ($Code, $Msg) = @_;
    print STDERR "ERROR: ". $Msg."\n";
    exit($ERROR_CODE{$Code});
}

sub exitNormal()
{# the tool has run without any errors
    print_Report();
    if($COMPILE_ERRORS)
    {# errors in headers may add false positives/negatives
        exit($ERROR_CODE{"Compile_Error"});
    }
    if($CHECKER_VERDICT) {
        exit($ERROR_CODE{"Incompatible"});
    }
    else {
        exit($ERROR_CODE{"Compatible"});
    }
}

sub print_Report()
{
    print "creating compatibility report ...\n";
    create_HtmlReport();
    if($CHECKER_VERDICT) {
        print "result: INCOMPATIBLE (total problems: $CHECKER_VERDICT, warnings: $CHECKER_WARNING)\n";
    }
    else {
        print "result: COMPATIBLE (total problems: $CHECKER_VERDICT, warnings: $CHECKER_WARNING)\n";
    }
    print "see detailed report in the file:\n  $REPORT_PATH\n";
}

sub scenario()
{
    if(defined $Help) {
        HELP_MESSAGE();
        exit(0);
    }
    if(defined $InfoMsg) {
        INFO_MESSAGE();
        exit(0);
    }
    if(defined $ShowVersion) {
        print "ABI Compliance Checker (ACC) $ABI_COMPLIANCE_CHECKER_VERSION\nCopyright (C) 2011 Institute for System Programming, RAS\nLicense: LGPLv2 or GPLv2 <http://www.gnu.org/licenses/>\nThis program is free software: you can redistribute it and/or modify it.\n\nWritten by Andrey Ponomarenko.\n";
        exit(0);
    }
    if(defined $DumpVersion) {
        print "$ABI_COMPLIANCE_CHECKER_VERSION\n";
        exit(0);
    }
    if($SystemRoot)
    {# user defined root
        if(not -e $SystemRoot) {
            exitStatus("Access_Error", "can't access \'$SystemRoot\'");
        }
        $SystemRoot=~s/[\/]+\Z//g;
        if($SystemRoot) {
            $SystemRoot = abs_path($SystemRoot);
        }
    }
    elsif($DumpSystem) {
        exitStatus("Error", "-sysroot <dirpath> option should be specified");
    }
    # FIXME: can't pass \&dump_sorting - cause segfault sometimes
    # $Data::Dumper::Sortkeys = \&dump_sorting;
    $Data::Dumper::Sortkeys = 1;
    detect_default_paths();
    if($OSgroup eq "windows")
    {
        if(not $ENV{"DevEnvDir"}
        or not $ENV{"LIB"}) {
            exitStatus("Error", "can't start without VS environment (vsvars32.bat)");
        }
    }
    if($DumpSystem) {
        dumpSystem();
        exit(0);
    }
    if($CmpSystems) {
        cmpSystems();
        exit(0);
    }
    if($TestTool
    or $TestDump)
    {
        testSystemC();
        testSystemCpp();
        exit(0);
    }
    if($GenDescriptor) {
        genDescriptorTemplate();
        exit(0);
    }
    if(not $TargetLibraryName) {
        exitStatus("Error", "library name is not selected (option -l <name>)");
    }
    else {
        # validate library name
        if($TargetLibraryName=~/[\*\/\\]/) {
            exitStatus("Error", "\"\\\", \"\/\" and \"*\" symbols are not allowed in the library name");
        }
    }
    if(not $TargetLibraryFullName) {
        $TargetLibraryFullName = $TargetLibraryName;
    }
    if($CheckHeadersOnly_Opt and $CheckObjectsOnly_Opt) {
        exitStatus("Error", "you can't specify both -headers-only and -objects-only options at the same time");
    }
    if($InterfacesListPath) {
        if(not -f $InterfacesListPath) {
            exitStatus("Access_Error", "can't access file \'$InterfacesListPath\'");
        }
        foreach my $Interface (split(/\n/, readFile($InterfacesListPath))) {
            $InterfacesList{$Interface} = 1;
        }
    }
    if($ParamNamesFilePath) {
        if(-f $ParamNamesFilePath) {
            foreach my $Line (split(/\n/, readFile($ParamNamesFilePath)))
            {
                if($Line=~s/\A(\w+)\;//) {
                    my $Interface = $1;
                    if($Line=~/;(\d+);/) {
                        while($Line=~s/(\d+);(\w+)//) {
                            $AddIntParams{$Interface}{$1}=$2;
                        }
                    }
                    else {
                        my $Num = 0;
                        foreach my $Name (split(/;/, $Line)) {
                            $AddIntParams{$Interface}{$Num++}=$Name;
                        }
                    }
                }
            }
        }
        else {
            exitStatus("Access_Error", "can't access file \'$ParamNamesFilePath\'");
        }
    }
    if($AppPath) {
        if(-f $AppPath) {
            foreach my $Interface (getSymbols_App($AppPath)) {
                $InterfacesList_App{$Interface} = 1;
            }
        }
        else {
            exitStatus("Access_Error", "can't access file \'$AppPath\'");
        }
    }
    if($DumpInfo_DPath)
    {# make an ABI dump
        if(not -e $DumpInfo_DPath) {
            exitStatus("Access_Error", "can't access \'$DumpInfo_DPath\'");
        }
        my $TarCmd = get_CmdPath("tar");
        if(not $TarCmd) {
            exitStatus("Not_Found", "can't find \"tar\"");
        }
        my $GzipCmd = get_CmdPath("gzip");
        if(not $GzipCmd) {
            exitStatus("Not_Found", "can't find \"gzip\"");
        }
        my @DParts = split(/\s*,\s*/, $DumpInfo_DPath);
        foreach my $Part (@DParts)
        {
            if(not -e $Part) {
                exitStatus("Access_Error", "can't access \'$Part\'");
            }
        }
        checkVersionNum(1, $DumpInfo_DPath);
        foreach my $Part (@DParts) {
            createDescriptor(1, $Part);
        }
        my $DumpPackage = "abi_dumps/$TargetLibraryName/".$TargetLibraryName."_".$Descriptor{1}{"Version"}.".abi.tar.gz";
        if($OutputDumpPath)
        { # user defined path
            $DumpPackage = $OutputDumpPath;
        }
        my $DumpPath = $DumpPackage;
        if(not $DumpPath=~s/\.(tar\.gz)\Z//g) {
            exitStatus("Error", "the dump path (-dump-path option) should be the path to a \"tar.gz\" file");
        }
        scan_libs_and_headers(1);
        detectPointerSize();
        translateSymbols(keys(%{$Symbol_Library{1}}), 1);
        translateSymbols(keys(%{$DepSymbols{1}}), 1);
        if($Descriptor{1}{"Headers"}
        and not $Descriptor{1}{"Dump"}) {
            readHeaders(1);
        }
        cleanData(1);
        if(not keys(%{$SymbolInfo{1}}) and not $CheckHeadersOnly and not $CheckObjectsOnly) {
            exitStatus("Error", "the symbol sets from headers and $SHARED libraries have empty intersection");
        }
        my %HeadersInfo = ();
        foreach my $HPath (keys(%{$Target_Headers{1}}))
        {# headers info stored without paths in the dump
            $HeadersInfo{$Target_Headers{1}{$HPath}{"Identity"}} = $Target_Headers{1}{$HPath}{"Position"};
        }
        print "creating library ABI dump ...\n";
        my %LibraryABI = (
            "TypeInfo" => $TypeInfo{1},
            "SymbolInfo" => $SymbolInfo{1},
            "Symbols" => $Library_Symbol{1},
            "DepSymbols" => $DepSymbols{1},
            "SymbolVersion" => $SymVer{1},
            "LibraryVersion" => $Descriptor{1}{"Version"},
            "LibraryName" => $TargetLibraryName,
            "DynamicLibs" => $DyLibNames_All{1},
            "Tid_TDid" => $Tid_TDid{1},
            "OpaqueTypes" => $OpaqueTypes{1},
            "SkipSymbols" => $SkipSymbols{1},
            "Headers" => \%HeadersInfo,
            "Constants" => $Constants{1},
            "NameSpaces" => $NestedNameSpaces{1},
            "Arch" => getArch(1),
            "SizeOfPointer" => $POINTER_SIZE{1},
            "GccVersion" => get_dumpversion($GCC_PATH),
            "ABI_DUMP_VERSION" => $ABI_DUMP_VERSION,
            "ABI_COMPLIANCE_CHECKER_VERSION" => $ABI_COMPLIANCE_CHECKER_VERSION,
        );
        my ($DumpDir, $DumpName) = separate_path($DumpPath);
        mkpath($DumpDir);
        unlink($DumpPackage);
        writeFile($DumpPath, Dumper(\%LibraryABI));
        if(not -s $DumpPath) {
            unlink($DumpPath);
            exitStatus("Error", "can't create ABI dump because something is going wrong with the Data::Dumper module");
        }
        chdir($DumpDir);
        system($TarCmd, "-cf", "$DumpName.tar", $DumpName);
        if($?)
        {# cannot allocate memory (or other problems with "tar")
            unlink($DumpPath);
            exitStatus("Error", "can't pack the ABI dump: ".$!);
        }
        system($GzipCmd, "$DumpName.tar");
        if($?)
        {# cannot allocate memory (or other problems with "gzip")
            unlink($DumpPath);
            exitStatus("Error", "can't pack the ABI dump: ".$!);
        }
        chdir($ORIG_DIR);
        unlink($DumpPath);
        print "library ABI has been dumped to:\n  ./$DumpPath\.tar\.gz\n";
        print "you can transfer this dump everywhere and use instead of the ".$Descriptor{1}{"Version"}." version descriptor\n";
        exit($COMPILE_ERRORS);
    }
    if(not $Descriptor{1}{"Path"}) {
        exitStatus("Error", "1st library descriptor is not selected (option -d1 <path>)");
    }
    my @DParts1 = split(/\s*,\s*/, $Descriptor{1}{"Path"});
    foreach my $Part (@DParts1)
    {
        if(not -e $Part) {
            exitStatus("Access_Error", "can't access \'$Part\'");
        }
    }
    if(not $Descriptor{2}{"Path"}) {
        exitStatus("Error", "2nd library descriptor is not selected (option -d2 <path>)");
    }
    my @DParts2 = split(/\s*,\s*/, $Descriptor{2}{"Path"});
    foreach my $Part (@DParts2)
    {
        if(not -e $Part) {
            exitStatus("Access_Error", "can't access \'$Part\'");
        }
    }
    if($#DParts1==0 and $#DParts2==0)
    {# Quick "empty" reports
     # 4 times faster than merging equal dumps
     # NOTE: the dump contains the "LibraryVersion" attribute
     # if you change the version, then your dump will be different
     # OVERCOME: use -v1 and v2 options for comparing dumps
     # and don't change version in the XML descriptor (and dumps)
     # OVERCOME 2: separate meta info from the dumps in ACC 2.0
        if($Descriptor{1}{"Path"}=~/\Q.abi.tar.gz\E\Z/
        and $Descriptor{2}{"Path"}=~/\Q.abi.tar.gz\E\Z/
        and (-s $Descriptor{1}{"Path"} == -s $Descriptor{2}{"Path"}))
        {
            my $FilePath1 = unpackDump($Descriptor{1}{"Path"});
            my $FilePath2 = unpackDump($Descriptor{2}{"Path"});
            if($FilePath1 and $FilePath2)
            {
                my $Content = readFile($FilePath1);
                if($Content eq readFile($FilePath2))
                {
                    # read a number of headers, libs, symbols and types
                    my $ABIdump = eval($Content);
                    if(not $ABIdump) {
                        exitStatus("Error", "internal error");
                    }
                    if(not $ABIdump->{"TypeInfo"})
                    {# support for old dumps
                        $ABIdump->{"TypeInfo"} = $ABIdump->{"TypeDescr"};
                    }
                    if(not $ABIdump->{"SymbolInfo"})
                    {# support for old dumps
                        $ABIdump->{"SymbolInfo"} = $ABIdump->{"FuncDescr"};
                    }
                    read_Headers_DumpInfo($ABIdump, 1);
                    read_Libs_DumpInfo($ABIdump, 1);
                    read_Machine_DumpInfo($ABIdump, 1);
                    %CheckedTypes = %{$ABIdump->{"TypeInfo"}};
                    %CheckedInterfaces = %{$ABIdump->{"SymbolInfo"}};
                    $Descriptor{1}{"Version"} = $TargetVersion{1}?$TargetVersion{1}:$ABIdump->{"LibraryVersion"};
                    $Descriptor{2}{"Version"} = $TargetVersion{2}?$TargetVersion{2}:$ABIdump->{"LibraryVersion"};
                    if(not $OutputReportPath) {
                        $REPORT_PATH = "compat_reports/$TargetLibraryName/".$Descriptor{1}{"Version"}."_to_".$Descriptor{2}{"Version"}."/abi_compat_report.html";
                    }
                    exitNormal();
                }
            }
        }
    }
    checkVersionNum(1, $Descriptor{1}{"Path"});
    checkVersionNum(2, $Descriptor{2}{"Path"});
    print "preparation, please wait ...\n";
    foreach my $Part (@DParts1) {
        createDescriptor(1, $Part);
    }
    foreach my $Part (@DParts2) {
        createDescriptor(2, $Part);
    }
    if($UseDumps)
    {# parallel processing
        if(not $Descriptor{1}{"Path"}) {
            exitStatus("Error", "-d1 option is not specified");
        }
        if(not $Descriptor{2}{"Path"}) {
            exitStatus("Error", "-d2 option is not specified");
        }
        my $pid = fork();
        if($pid) {# dump on two CPU cores
            my @PARAMS = ("-dump", $Descriptor{1}{"Path"}, "-l", $TargetLibraryName);
            if($RelativeDirectory{1}) {
                @PARAMS = (@PARAMS, "-relpath", $RelativeDirectory{1});
            }
            if($OutputLogPath{1}) {
                @PARAMS = (@PARAMS, "-log-path", $OutputLogPath{1});
            }
            system("perl", "$0", @PARAMS);
            if($?) {
                exit(1);
            }
        }
        else
        {# child
            my @PARAMS = ("-dump", $Descriptor{2}{"Path"}, "-l", $TargetLibraryName);
            if($RelativeDirectory{2}) {
                @PARAMS = (@PARAMS, "-relpath", $RelativeDirectory{2});
            }
            if($OutputLogPath{2}) {
                @PARAMS = (@PARAMS, "-log-path", $OutputLogPath{2});
            }
            system("perl", "$0", @PARAMS);
            if($?) {
                exit(1);
            }
            else {
                exit(0);
            }
        }
        waitpid($pid, 0);
        my @CMP_PARAMS = ("-l", $TargetLibraryName);
        @CMP_PARAMS = (@CMP_PARAMS, "-d1", "abi_dumps/$TargetLibraryName/".$TargetLibraryName."_".$Descriptor{1}{"Version"}.".abi.tar.gz");
        @CMP_PARAMS = (@CMP_PARAMS, "-d2", "abi_dumps/$TargetLibraryName/".$TargetLibraryName."_".$Descriptor{2}{"Version"}.".abi.tar.gz");
        if($TargetLibraryFullName ne $TargetLibraryName) {
            @CMP_PARAMS = (@CMP_PARAMS, "-l-full", $TargetLibraryFullName);
        }
        if($ShowRetVal) {
            @CMP_PARAMS = (@CMP_PARAMS, "-show-retval");
        }
        system("perl", "$0", @CMP_PARAMS);
        exit($?>>8);
    }
    if(not $Descriptor{1}{"Headers"}
    and not $Descriptor{1}{"Libs"}) {
        exitStatus("Error", "descriptor d1 does not contain both header files and libraries info");
    }
    if(not $Descriptor{2}{"Headers"}
    and not $Descriptor{2}{"Libs"}) {
        exitStatus("Error", "descriptor d2 does not contain both header files and libraries info");
    }
    if($Descriptor{1}{"Headers"} and not $Descriptor{1}{"Libs"}
    and not $Descriptor{2}{"Headers"} and $Descriptor{2}{"Libs"}) {
        exitStatus("Error", "can't compare headers with $SHARED libraries");
    }
    elsif(not $Descriptor{1}{"Headers"} and $Descriptor{1}{"Libs"}
    and $Descriptor{2}{"Headers"} and not $Descriptor{2}{"Libs"}) {
        exitStatus("Error", "can't compare $SHARED libraries with headers");
    }
    if(not $Descriptor{1}{"Headers"}) {
        if($CheckHeadersOnly_Opt) {
            exitStatus("Error", "can't find header files info in descriptor d1");
        }
    }
    if(not $Descriptor{2}{"Headers"}) {
        if($CheckHeadersOnly_Opt) {
            exitStatus("Error", "can't find header files info in descriptor d2");
        }
    }
    if(not $Descriptor{1}{"Headers"}
    or not $Descriptor{2}{"Headers"}) {
        if(not $CheckObjectsOnly_Opt) {
            print "WARNING: comparing $SHARED libraries only\n";
            $CheckObjectsOnly = 1;
        }
    }
    if(not $Descriptor{1}{"Libs"}) {
        if($CheckObjectsOnly_Opt) {
            exitStatus("Error", "can't find $SHARED libraries info in descriptor d1");
        }
    }
    if(not $Descriptor{2}{"Libs"}) {
        if($CheckObjectsOnly_Opt) {
            exitStatus("Error", "can't find $SHARED libraries info in descriptor d2");
        }
    }
    if(not $Descriptor{1}{"Libs"}
    or not $Descriptor{2}{"Libs"}) {
        if(not $CheckHeadersOnly_Opt) {
            print "WARNING: comparing headers only\n";
            $CheckHeadersOnly = 1;
        }
    }
    if(($Descriptor{1}{"Libs"} or $Descriptor{1}{"Headers"})
    and not $Descriptor{1}{"Dump"}) {
        scan_libs_and_headers(1);
    }
    if(($Descriptor{2}{"Libs"} or $Descriptor{2}{"Headers"})
    and not $Descriptor{2}{"Dump"}) {
        scan_libs_and_headers(2);
    }
    if($AppPath and not keys(%{$Symbol_Library{1}})) {
        print "WARNING: the application ".get_filename($AppPath)." has no symbols imported from the $SHARED libraries\n";
    }
    if(not $OutputReportPath) {
        $REPORT_PATH = "compat_reports/$TargetLibraryName/".$Descriptor{1}{"Version"}."_to_".$Descriptor{2}{"Version"}."/abi_compat_report.html";
    }
    mkpath(get_dirname($REPORT_PATH));
    unlink($REPORT_PATH);
    detectPointerSize();
    translateSymbols(keys(%{$Symbol_Library{1}}), 1);
    translateSymbols(keys(%{$DepSymbols{1}}), 1);
    translateSymbols(keys(%{$Symbol_Library{2}}), 2);
    translateSymbols(keys(%{$DepSymbols{2}}), 2);
    if(not $CheckHeadersOnly) {
        detectAdded();
        detectRemoved();
    }
    if(not $CheckObjectsOnly) {
        if($Descriptor{1}{"Headers"}
        and not $Descriptor{1}{"Dump"}) {
            readHeaders(1);
        }
        if($Descriptor{2}{"Headers"}
        and not $Descriptor{2}{"Dump"}) {
            readHeaders(2);
        }
        print "comparing headers ...\n";
        mergeSignatures();
        if($ComparedABI) {
            mergeConstants();
        }
    }
    if(not $CheckHeadersOnly) {
        print "comparing libraries ...\n";
        mergeLibs();
        if($CheckImplementation) {
            mergeImplementations();
        }
    }
    exitNormal();
}

scenario();
